/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrFFApi
// @module  CECFFTestResultsFlusher Implementation | This module contains the implementation
//          of the CECFFTestResultsFlusher class
//
///////////////////////////////////////////////////////////////////////////////
#include "CECFFTestResultsFlusher.h"
#include "CyrffApi\CECFFTestRun.h"
#include <CyrffApi/CECFFFile.h>
#include <Repository/CyrStHelper.inl>
#include <iostream>
void* CECFFTestResultsFlusher_Run(void* args)
{
    CECFFTestResultsFlusher* TRF=reinterpret_cast<CECFFTestResultsFlusher*>(args);
    TRF->Run();
    return NULL;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   useIfIdle
//
// This function determines whether or not we are flushing a list.
// If we are then reserve the thread for use with this list
bool CECFFTestResultsFlusher::useIfIdle()
{
    bool bUse = false;

    // We simply have to lock this part as all lists will hit this function at once
    static omni_mutex mutex;

    mutex.acquire();

    // See if the Thread is running...
    if (!m_bRunning)
    {
        // If not then tell all the other lists that it is in action
        m_bRunning = true;

        // And inform the caller that it is reserved for use
        bUse = true;
    }
    else
    {
        // The thread must still be busy so try again later when another string
        // is added to the list
        bUse = false;
    }

    mutex.release();

    return bUse;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   setIdle
//
// This function releases the thread so that other lists can now use it
void CECFFTestResultsFlusher::setIdle()
{
    // Must acquire a mutex before doing this as other threads may be accessing the
    // running flag
    static omni_mutex mutex;

    mutex.acquire();

    // Set it to false to indicate that we are doing nothing
    m_bRunning=false;

    mutex.release();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Standard Constructor
// @parm   The current Test Run Folder.
CECFFTestResultsFlusher::CECFFTestResultsFlusher()
{
    // Put the thread in an idle state
    setIdle();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   OutputList
// @parm   Output Table relating to this test run.
// @parm   Output List relating to this test run.
void CECFFTestResultsFlusher::kickoff(string strFullpath, string strFilename, OutputList outList)
{
    // Ensure something has been passed in
    if (outList.empty())
    {
        setIdle();
        return;
    }

    // Set up the internal Output list with the data to purge
    m_OutList = outList;

    // And set up the test run folder - must do this to keep support for multiple
    // test runs using this thread
    m_strFullpath = strFullpath;

    // And put an underscore before the filename to write to - this indicates that the file
    // is a temporary file so that the gui does not get fooled into believing the test results
    // are actually complete
    m_strFilename = "_" + strFilename;

    try
    {
        // Create Flusher thread and start it running.

        // This automatically create us a thread and calls the (virtual)
        // run member function
        flusher = new omni_thread(&CECFFTestResultsFlusher_Run ,this);
        flusher->start();

    }
    catch( ... )
    {}

}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Destructor - Perform any cleanup here
CECFFTestResultsFlusher::~CECFFTestResultsFlusher()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This virtual function is the standard start method for threads.
//          Simply flush the output for each list in turn.

void CECFFTestResultsFlusher::Run()
{

    try
    {
        // Simply flush the internal list
        flushList( );
    }
    catch( ... )
    {
        //just end!!;
    }

    // And put back into an idle state once more

    setIdle();

}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method saves the entire contents of the output list to
//          several files within the Test Output folder.  Each string is saved
//          one per line in a standard ASCII file.
//          Currently this purge procedure does not affect the OTA headers
//          file which remains in memory. The reason for this is to ensure
//          that TestCommander (or any similar monitoring tool may dynamically
//          know what stats are being collected.
// @parm    The filename of the current set of output.
// @parm    The list of output strings.
void CECFFTestResultsFlusher::flushList()
{

    // As a sanity check we ensure that we actually do have something in the list
    if ( m_OutList.empty() )
    {
        return;
    }

    // Get correct filename for file
    m_strFullpath += m_strFilename;

    // Open file for output and write contents to file
    std::ofstream* pFile = new std::ofstream( m_strFullpath.c_str(), std::ios::binary|std::ios::app|std::ios::out);
    if( pFile->good() )
    {
        // And loop round for the defined chunk of data we wish to flush
        while (!m_OutList.empty())
        {
            // Get the data - and remove from our internal (and temporary list)

            // Note: this does not actually remove the data from the main output list
            // as we are in a seperate thread here. The main list is still in use at this point
            // and it is NOT our responsibility to manage that list too. The caller (CECComTestRun)
            // has to do that.
            string outList = *(m_OutList.begin());
            m_OutList.pop_front();

            // And write to the file
            pFile->write( outList.c_str(), outList.length() );
            pFile->write( "\r\n", 2 );
        }

        // Delete file - flushing contents to disk
        delete pFile;
    }
}

