/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrStRepository
// @module  Link Implementation | This module contains the implementation of
//          the CyrStLink CORBA object.  This CORBA object is intended for use
//          as a base class for any item that can be linked in the OpenSTARep
//          interface.
// @devnote As these items will be implemented as CORBA interfaces this will
//          mean use of multiple inheritance.  Bleurgh! If I could think of a
//          better way I'd use it.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Repository/CyrStLink_i.h>

#ifdef __VDK002__
#include <Logging/CECLog.h>
extern CECLog* v2logger;
#endif


///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor simply assigns the member variables.
// @parm    Pointer to the parent Provider object.
CyrStLinkImpl::CyrStLinkImpl( CyrStProvider* pProvider )
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 50, "Constructing a CyrStLinkImpl." );
#endif
    // Assign member variables
    m_provider = pProvider;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor.
CyrStLinkImpl::~CyrStLinkImpl()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 50, "Destroying a CyrStLinkImpl." );
#endif

    // Destroy underlying server reference
    if (m_provider == NULL)
    {
#ifdef __VDK002__
        v2logger->AddTrace( CYR_VDK002, 50, "m_provider is null." );
#endif
    }
    else
    {
        try
        {
            m_provider->RemovePtr( m_regName );
#ifdef __VDK002__
            string msg = "Removed registered item " + m_regName;
            v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif
        }
        catch (...)
        {
#ifdef __VDK002__
            v2logger->AddTrace( CYR_VDK002, 50, "m_provider is corrupt." );
#endif
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Within the Reaper mechanism of the Factory, each object must be
//          given a unique name.  This method returns that name for this
//          particular object.  The name itself is generated by the Provider
//          when the underlying COM object is stored.  This name can be used
//          to increase the lifetime of the CORBA object.  The name itself is
//          generated from the name of the server, plus the name of the current
//          user.  This should guarantee uniqueness unless the same user wishes
//          to create the same object again - in which case they are
//          responsible for producing a different name.
// @rdesc   The registered name.
char* CyrStLinkImpl::GetRegisteredName()
{
    // And add some tracing
#ifdef __VDK002__
    string msg = "Obtaining registered Name: " + m_regName;
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif
    return CORBA::string_dup( m_regName.c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   To maintain a list of unique names for each item in the OpenSTARep
//          database, the complete hierarchy can be used as the name.  This
//          method constructs the full list of Folders into a single string by
//          walking backwards up the hierarchy.
// @rdesc   The full name of this OpenSTARep item.
// @parm    A pointer to the item.
string CyrStLinkImpl::BuildItemPath( CECFFLink* link )
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Building Item Path." );
#endif
    // Get parent folder of item
    CECFFFolder* Parent = link->GetParentFolder();

    // Walk backwards up hierarchy until folder is null
    string name( "" );
    name+=Parent->GetFolderHierarchy();

#ifdef __VDK002__
    string msg = "Item Path Calculated as:" + name;
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

    return name;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the full path to the item represented by this
//          class.  Effectively this is a list of folders that will lead to
//          the object in question when traversed down the hierarchy.  The list
//          is '\|' separated as usual.  The first value is the root folder,
//          and also the name of the Project.  The last name is the name of the
//          item.  Any entries in between represent Folder names.
// @rdesc   The separated path to the item.
char* CyrStLinkImpl::GetLinkPath()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Obtaining Link Path." );
#endif
    // Get underlying COM references
    CECFFLink* pLink = (CECFFLink*)m_provider->GetPtr( m_regName );
    return CORBA::string_dup( pLink->GetLinkPath().c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a string representing the complete path to an
//          object that has been linked to this one.  Calling this function
//          will invalidate any current enumeration.
// @rdesc   The first link path.
// @xref    <mf CyrStLinkImpl::GetLinkPath>,
//          <mf CyrStLinkImpl::GetNextLinkPath>
char* CyrStLinkImpl::GetFirstLinkPath()
{
    // And add some tracing
#ifdef __VDK002__
    string msg = "Obtaining First Link Path for:" + m_regName;
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif
    // Get underlying COM references
    CECFFLink* pLink = (CECFFLink*)m_provider->GetPtr( m_regName );

    // Return next link path in sequence
    return CORBA::string_dup( pLink->GetFirstLinkPath().c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the next link path in the sequence or an empty
//          string if there are no more linked items.
// @rdesc   The next link path.
// @xref    <mf CyrStlinkImpl::GetFirstLinkPath>,
//          <mf CyrStlinkImpl::GetLinkPath>
char* CyrStLinkImpl::GetNextLinkPath()
{
    // And add some tracing
#ifdef __VDK002__
    string msg = "Obtaining First Link Path for:" + m_regName;
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif
    // Get underlying COM references
    CECFFLink* pLink = (CECFFLink*)m_provider->GetPtr( m_regName );

    // Build link path for item
    return CORBA::string_dup( pLink->GetNextLinkPath().c_str() );
}
