/* Copyright (C) 2000-2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*
 *  Module: chk_dataname.c
 *  ======================
 *
 *  This module contains two routines - CHK_DATANAME is the standard
 *  routine for checking an opensTA dataname (maximum length 16
 *  characters). CHK_LONG_DATANAME is a 'temporary' routine for
 *  checking long OpenSTA dataname (up to 64 characters long).
 *  The intention is that the former should be changed to allow
 *  longer names. However, this change needs to be done particularly
 *  carefully, as there are likely to be a significant number of
 *  local variables - in SCL in particular - that have fixed length
 *  buffers that rely on datanames being 16 characters long. In
 *  order to allow longer Test, Script and Profile names NOW, a
 *  new 'long' dataname type has been introduced (internally) to
 *  OpenSTA.
 */

#include  <psl_macros.h>        /* PSL Standard Macros */

/*   Define useful Include structures   */

#include  <ctype.h>         /* Character functions */

/*   Define Function Prototypes for all external routines */
/*   used by the module (and for ourselves) */

#include  psl_common_file(psl_rtndef.h) /* Define ourselves */


/************************************************************************/
/*                                                                      */
/*  CHK_DATANAME - Check OpenSTA Dataname                               */
/*                                                                      */
/*  This routine checks a supplied string to see if it is a             */
/*  valid OpenSTA Dataname.                                             */
/*                                                                      */
/*      Calling Convention:                                             */
/*                                                                      */
/*  CHK_DATANAME (buf_ptr, buflen_ptr, status_ptr)                      */
/*                                                                      */
/*  Where:                                                              */
/*                                                                      */
/*  buf_ptr    = Pointer to buffer containing dataname                  */
/*  buflen_ptr = Pointer to Integer*4 length of buffer to scan          */
/*               for dataname.                                          */
/*  status_ptr = Pointer to Integer*4 for returned status:              */
/*             = -1 if no valid dataname was found                      */
/*             = number of characters in buffer after dataname          */
/*               otherwise (=0 if all SIZE characters in BUFFER         */
/*               constituted a valid OpenSTA dataname)                  */
/*                                                                      */
/************************************************************************/

void CHK_DATANAME (const char *buf_ptr,
                   const long *buflen_ptr,
                   long *status_ptr)
{
    /*
     *  Parameterize Maximum Valid Dataname Length
     */
#define DAT_MAXLEN  16          /* Currently 16 characters */

    /* Local Data */
    long    buflen;             /* Buffer length */
    long    count;              /* Characters left to process */
    char    ch;                 /* Current character */
    Boolean hyphen = PFALSE;    /* True if got hyphen or underscore */

    /*
     *   First get buffer size
     */
    buflen = *buflen_ptr;       /* Get buffer length */

    /*
     *   We now want to check for a valid dataname.  The rules are currently
     *   as follows:
     *
     *  - 1 - 16 alphanumeric hyphen or underscore characters
     *  - First character must be alphabetic
     *  - Two adjacent hyphens or underscores are not allowed
     *  - A trailing hyphen or underscore is not allowed
     *
     *    We check the first character first as it has special rules, and we
     *    return a different status (-1) if it is invalid.
     */
    ch = *buf_ptr;          /* Get first character */
    if (!isalpha (ch))          /* If not alphabetic ... */
    {
        *status_ptr = -1;       /* Return error */
        goto end;
    }
    /*
     *    We now know we have a valid dataname, and need merely sort out
     *    where it ends.
     */
    buflen -= 1;            /* Count first character checked */
    if (buflen == 0) goto end_of_string; /* If EQL, exit with Success */
    buf_ptr += 1;           /* Increment buffer pointer */
    for (count = DAT_MAXLEN-1; count > 0; count--)
        /* Another 15 characters allowed */
    {
        /*
         *    First see if next character is alphanumeric - always valid
         */
        ch = *buf_ptr;          /* Get next character */
        if (isalnum (ch))       /* If character is alphanumeric */
        {
            hyphen = PFALSE;        /* Reset hyphen flag */
        }
        /*
         *    The only thing left that the character can validly be is a hyphen or 
         *    an underscore and this is only valid if it doesn't follow another. 
         *    In the latter case we have reached the end of the name AND the 
         *    previous hyphen was invalid (trailing hyphens not allowed) so we need 
         *    to back-track over it.
         */
        else if (hyphen)        /* If previous char was hyphen/und */
        {
            buflen += 1;        /* It was illegal so ignore it */
            goto end_of_string;     /* And exit */
        }
        else if ((ch == '-') || (ch == '_'))
            /* If current char is '-' or '_' */
        {
            hyphen = PTRUE;     /* Set hyphen flag */
        }
        else
        {
            goto end_of_string;     /* End of valid string so exit */
        }
        /*
         *    The last character was valid.  See if there are any more and check
         *    the next unless we have already exceeded the valid dataname range.
         */
        buflen -= 1;            /* Another character checked */
        if (buflen == 0) goto check_trailing_hyphen;
        /* If EQL, exit loop and check */
        /* for a trailing hyphen */
        buf_ptr += 1;           /* Increment buffer pointer */
    }
    /*
     *    We've run out of characters, either in the input string or in
     *    what constitutes a valid dataname (16) so we just need to check for
     *    (and possibly discount) any trailing hyphen.
     */
check_trailing_hyphen:
    if (hyphen)             /* If previous char was hyphen/und */
    {
        buflen += 1;            /* It was illegal so ignore it */
    }
    /*
     *    We found the end of the string.  Return the number of characters
     *    remaining as the status.
     */
end_of_string:
    *status_ptr = buflen;       /* Set number of chars left as status */
    /*
     *   End of routine
     */
end:
    return;
}

/************************************************************************/
/*                                                                      */
/*  CHK_LONG_DATANAME - Check Long OpenSTA Dataname                     */
/*                                                                      */
/*  This routine checks a supplied string to see if it is a             */
/*  valid OpenSTA Dataname.                                             */
/*                                                                      */
/*      Calling Convention:                                             */
/*                                                                      */
/*  CHK_LONG_DATANAME (buf_ptr, buflen_ptr, status_ptr)                 */
/*                                                                      */
/*  Where:                                                              */
/*                                                                      */
/*  buf_ptr    = Pointer to buffer containing dataname                  */
/*  buflen_ptr = Pointer to Integer*4 length of buffer to scan          */
/*               for dataname.                                          */
/*  status_ptr = Pointer to Integer*4 for returned status:              */
/*             = -1 if no valid dataname was found                      */
/*             = number of characters in buffer after dataname          */
/*               otherwise (=0 if all SIZE characters in BUFFER         */
/*               constituted a valid dataname)                          */
/*                                                                      */
/************************************************************************/

void CHK_LONG_DATANAME (const char *buf_ptr,
                        const long *buflen_ptr,
                        long *status_ptr)
{
    /*
     *  Parameterize Maximum Valid Dataname Length
     */
#define LONG_DAT_MAXLEN  64     /* Currently 64 characters */

    /* Local Data */
    long    buflen;             /* Buffer length */
    long    count;              /* Characters left to process */
    char    ch;                 /* Current character */
    Boolean hyphen = PFALSE;    /* True if got hyphen or underscore */

    /*
     *   First get buffer size
     */
    buflen = *buflen_ptr;       /* Get buffer length */

    /*
     *   We now want to check for a valid dataname.  The rules are currently
     *   as follows:
     *
     *  - 1 - 16 alphanumeric hyphen or underscore characters
     *  - First character must be alphabetic
     *  - Two adjacent hyphens or underscores are not allowed
     *  - A trailing hyphen or underscore is not allowed
     *
     *    We check the first character first as it has special rules, and we
     *    return a different status (-1) if it is invalid.
     */
    ch = *buf_ptr;              /* Get first character */
    if (!isalpha (ch))          /* If not alphabetic ... */
    {
        *status_ptr = -1;       /* Return error */
        goto end;
    }
    /*
     *    We now know we have a valid dataname, and need merely sort out
     *    where it ends.
     */
    buflen -= 1;                /* Count first character checked */
    if (buflen == 0) goto end_of_string; /* If EQL, exit with Success */
    buf_ptr += 1;               /* Increment buffer pointer */
    for (count = LONG_DAT_MAXLEN-1; count > 0; count--)
        /* Another 15 characters allowed */
    {
        /*
         *    First see if next character is alphanumeric - always valid
         */
        ch = *buf_ptr;          /* Get next character */
        if (isalnum (ch))       /* If character is alphanumeric */
        {
            hyphen = PFALSE;        /* Reset hyphen flag */
        }
        /*
         *    The only thing left that the character can validly be is a hyphen or 
         *    an underscore and this is only valid if it doesn't follow another. 
         *    In the latter case we have reached the end of the name AND the 
         *    previous hyphen was invalid (trailing hyphens not allowed) so we need 
         *    to back-track over it.
         */
        else if (hyphen)        /* If previous char was hyphen/und */
        {
            buflen += 1;        /* It was illegal so ignore it */
            goto end_of_string;     /* And exit */
        }
        else if ((ch == '-') || (ch == '_'))
            /* If current char is '-' or '_' */
        {
            hyphen = PTRUE;     /* Set hyphen flag */
        }
        else
        {
            goto end_of_string;     /* End of valid string so exit */
        }
        /*
         *    The last character was valid.  See if there are any more and check
         *    the next unless we have already exceeded the valid dataname range.
         */
        buflen -= 1;            /* Another character checked */
        if (buflen == 0) goto check_trailing_hyphen;
        /* If EQL, exit loop and check */
        /* for a trailing hyphen */
        buf_ptr += 1;           /* Increment buffer pointer */
    }
    /*
     *    We've run out of characters, either in the input string or in
     *    what constitutes a valid dataname (16) so we just need to check for
     *    (and possibly discount) any trailing hyphen.
     */
check_trailing_hyphen:
    if (hyphen)             /* If previous char was hyphen/und */
    {
        buflen += 1;            /* It was illegal so ignore it */
    }
    /*
     *    We found the end of the string.  Return the number of characters
     *    remaining as the status.
     */
end_of_string:
    *status_ptr = buflen;       /* Set number of chars left as status */
    /*
     *   End of routine
     */
end:
    return;
}