/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
/*===========================================================================*/
/*									     */
/*			   ****  C O M M O N  ****   			     */
/*									     */
/*===========================================================================*/
/*									     */
/*			Module Name: PSL_REALPATH_UNIX			     */
/*			============					     */
/*									     */
/* Description:								     */
/* ============								     */
/*									     */
/*	This routine is provided to give us a realpath() type function,	     */
/*	available on SunOS but not HP-UX, but with the extension that the    */
/*      final filename does not have to exist, however the directory	     */
/*	specification must be correct. The final pathname may be a file,     */
/*	directory or yet to be created file.				     */
/*	Links are resolved at the callers discretion.			     */
/*									     */
/*===========================================================================*/

/*
 *   PSL Macros
 */
#include <psl_macros.h>

#ifdef UNIX
/*
 *   Standard ANSI header files
 */
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef HP
#include <symlink.h>
#endif /* HP */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>

/*
 *  Function Prototypes.
 */
#include psl_common_file (psl_rtndef.h)

/*
 *	Local function prototypes
 */
static unsigned long
CHECK_DIR(unsigned long count,
	  unsigned long max_path_len,
	  char *new_path,
	  unsigned long *npid_ptr,
	  Boolean do_resolve_links);

static unsigned long
RESOLVE_LINK(char *new_path,
	     unsigned long *npid_ptr,
	     unsigned long max_path_len);

/*
 *    Structure definitions.
 */
#include psl_common_file (shrmsg.h)	/* PSL Messages */

/*
 *  Define S_ISLNK for NCR machine.
 */
#ifdef NCR
#define S_ISLNK(mode)	((mode&S_IFMT) == S_IFLNK)
#endif

/*======================================================================*/
/*									*/
/*    Calling Convention:						*/
/*									*/
/*	iret = PSL_REALPATH(file_ptr, ret_file_ptr, max_ret_len,	*/
/*			    do_resolve_links)				*/
/*									*/
/*    Where:								*/
/*									*/
/*	iret		= unsigned long routine status			*/
/*			= PSL_NORMAL - If success			*/
/*			= PSL_INSVIRMEM - If error from MALLOC		*/
/*			= PSL_BADPARAM - If not enough space in buffer	*/
/*			= PSL_FILNOTFOU - If some part pathname invalid	*/
/*									*/
/*	file_ptr        = Pointer to the file specification string	*/
/*			  (Null terminated)				*/
/*								        */
/*	ret_file_ptr   	= Pointer to the Returned file specification.   */
/*			  (Null terminated)				*/
/*								        */
/*	max_ret_len	= unsigned long length of the return file name  */
/*			  buffer. This must not be greater than		*/
/*			  MAXPATHLEN					*/
/*									*/
/*	do_resolve_links= Boolean, PTRUE if to resolve links, otherwise	*/
/*			  their specication is left alone.		*/
/*									*/
/*======================================================================*/
unsigned long
PSL_REALPATH(char *old_file_ptr,
	     char *ret_file_ptr,
	     unsigned long max_ret_len,
	     Boolean do_resolve_links)
{
    unsigned long   iret = PSL_NORMAL;		/* Return status */
    char	    file_ptr[MAXPATHLEN];	/* Buffer for old file */
    unsigned long   fp_id;			/* file array subscript */
    unsigned long   rfp_id;			/* ret_file array subscript */
    unsigned long   count;			/* count for num of .'s */
    char	    *chr_ptr;

/*
 *  Copy the users old file into our buffer, this allows the user to specify
 *  the same buffer for the old_file_ptr and the ret_file_ptr.
 */
    (void)strcpy(file_ptr, old_file_ptr);

/*
 *  Initialise the old and new strings and their positional markers.
 */
    fp_id = 0;
    rfp_id = 0;

/*
 *  If the user has not specified the file from root then we need to start
 *  from the current working directory.
 */
    if (file_ptr[0] != '/')		/* If not specified from root */
    {
	chr_ptr = getcwd(ret_file_ptr,
			 max_ret_len);
	if (chr_ptr == NULL)		/* If error */
	{
	    if (errno == EACCES)	/* No privilege */
	    {
		iret = PSL_FILNOTFOU;
	    }
	    else			/* or Buffer not big enough */
	    {
		iret = PSL_BADPARAM;
	    }
	    return iret;		/* Bomb out here with error */
	}

/*
 *	Set the array index to be at end of current directory, then if we
 *	are going to add to this, also place on a trailing '/' to go before
 *	the first filename in the user specified bit.
 */
	rfp_id = strlen(ret_file_ptr);
	if (file_ptr[0] != '\0')		/* If caller specified path */
	{
	    ret_file_ptr[rfp_id] = '/';		/* slash ready for next dir */
	    if (rfp_id < max_ret_len)
	    {
		rfp_id += 1;			/* INcrement array index */
	    }
	    else
	    {
		return PSL_BADPARAM;		/* Bomb out with error */
	    }
	}
    }						/* End if not fully specified */

/*
 *  Now loop through the string doing the realpath business of checking
 *  for unnescessary directories.
 */
    for (count = 0;
	 file_ptr[fp_id] != '\0';
	 rfp_id++, fp_id++)
    {
/*
 *	Check to see we still have room in the given buffer
 */
	if (rfp_id >= max_ret_len)
	{
	    return PSL_BADPARAM;		/* Bomb out with error status */
	}

/*
 *	Copy the character from the old string to the expanded one we are
 *	creating. 
 */
	ret_file_ptr[rfp_id] = file_ptr[fp_id];

/*
 *	Check the character we hve just copied.
 */
	switch (file_ptr[fp_id]) 
	{
	    case '.':				/* Rememember to spot . & .. */
	    {
		count += 1;
		break;
	    }
	    case '/':				/* Signifies dir complete */
	    {
		iret = CHECK_DIR(count,
				 max_ret_len,
				 ret_file_ptr,
				 &rfp_id,
				 do_resolve_links);
		if (!psl_test_status(iret))	/* If error */
		{
		    return iret;		/* Bomb out with error status */
		}
		count = 0;			/* Reset the count */
		break;
	    }
	    default:				/* Any other character */
	    {
		count = 3;			/* Shows not found . or .. */
		break;
	    }
	}					/* End switch */
    }						/* End for */

/*
 *  Check the final path specification.
 */
    ret_file_ptr[rfp_id] = '\0';		/* NULL terminate pathname */
    iret = CHECK_DIR(count,
		     max_ret_len,
		     ret_file_ptr,
		     &rfp_id,
		     do_resolve_links);

/*
 *  Make sure final pathname is NULL terminated, with special case when pathname
 *  ends up to be "/".
 */
    if (rfp_id == 0)				/* Must be root spec. */
    {
	ret_file_ptr[1] = '\0';			/* NULL terminate '/' */
    }
    else					/* Reasonable pathname */
    {
	ret_file_ptr[rfp_id] = '\0';		/* NULL terminate pathname */
    }

/*
 *  Exit here with status.
 */
    return iret;
}

/*
 *  CHECK_DIR - check out the pathname supplied.
 *  =========
 */
static unsigned long
CHECK_DIR(unsigned long count,
	  unsigned long max_path_len,
	  char *new_path,
	  unsigned long *npid_ptr,
	  Boolean do_resolve_links)
{
/*
 *  Local variables
 */
    unsigned long   iret = PSL_NORMAL;	/* Return status */
    int	    	    rtnsts;		/* Return status */
    struct stat	    stat_buf;		/* File status buffer */

/*
 *  Perform different action depending on count of dots
 */
    switch (count)
    {
/*
 *	Found a '.' type directory, just needs removing.
 */
	case 1:				/* '.' type directory */
	{
	    if (*npid_ptr >= 2)		/* If enough space to */
	    {
		*npid_ptr -= 2;		/* Step back to last '/' */
	    }
	    else			/* Trying to step back over root */
	    {
		iret = PSL_BADPARAM;	/* Set up error status */
	    }
	    break;
	}

/*
 *	Found a '..' type directory step back over previous directory.
 */
	case 2:
	{
	    while (count != 0)		/* until removed .. and prev dir */
	    {
		if (*npid_ptr > 0)	/* if not reached beginning of path */
		{
		    *npid_ptr -= 1;
		    if (new_path[*npid_ptr] == '/') count -= 1;
		}
		else			/* If trying go back further than root*/
		{
		    iret = PSL_BADPARAM;/* Set up error status */
		}
	    }
	    break;
	}

/*
 *	Else we have found a complete filename of the path.
 */
	default:
	{
/*
 *	    ODD CASE:
 *	    If we have a directory spec with no name ie. // somewhere in
 *	    the pathname then ignore it. It's valid, but useless.
 */
	    if ((*npid_ptr != 0) &&
		(new_path[*npid_ptr -1] == '/'))
	    {
		*npid_ptr -= 1;
	    }

/*
 *	    MORE NORMAL:
 *	    Else if we have a directory file and it is not the root directory
 *	    then we check to make sure it exists, and if it is a link and the
 *	    caller requested links resolving, resolve it.
 */
	    else if ((new_path[*npid_ptr] == '/') &&
		     (*npid_ptr != 0))
	    {
		new_path[*npid_ptr] = '\0';
		rtnsts = lstat(new_path,
			       &stat_buf);
		if (rtnsts == -1)		/* Directory is not valid */
		{
		    iret = PSL_FILNOTFOU;
		}
		else if ((do_resolve_links) &&
			 (S_ISLNK(stat_buf.st_mode)))
		{
		    iret = RESOLVE_LINK(new_path,
					npid_ptr,
					max_path_len);
		}
		new_path[*npid_ptr] = '/';
	    }
/*
 *	    END CASE:
 *	    Else if we the final filename of the path. We don't care
 *	    if it exists, but we may want to resolve if it is a link.
 */
	    else if ((do_resolve_links) &&
		     (new_path[*npid_ptr] == '\0'))
	    {
		rtnsts = lstat(new_path,
			       &stat_buf);
		if ((rtnsts != -1) &&		/* File exists */
		    (S_ISLNK(stat_buf.st_mode)))/* and it is a link */
		{
		    iret = RESOLVE_LINK(new_path,
					npid_ptr,
					max_path_len);
		}		
	    }
	    break;
	}					/* End of default case */
    }						/* End of switch */
/*
 *  Exit here
 */
    return iret;
}

/*
 *	RESOLVE_LINK - takes a link file pathname and resolves.
 *	============
 */
static unsigned long
RESOLVE_LINK(char *new_path,
	     unsigned long *npid_ptr,
	     unsigned long max_path_len)
{
/*
 *  Local variables
 */
    unsigned long   iret = PSL_NORMAL;		/* Return status */
    char	    *resolved_path;		/* Pointer to buffer */
    int 	    rtnsts;			/* return value */
    char	    *dir_ptr;			/* position ptr */

/*
 *  Get a buffer for our old pathname and copy it into there.
 */
    resolved_path = (char *)malloc(max_path_len);
    if (resolved_path == NULL)			/* If some error */
    {
	return PSL_INSVIRMEM;			/* Insufficient vir mem */
    }

/*
 *  Now put the pathname of the file that this links to in our buffer.
 */
    rtnsts = readlink(new_path,
		      resolved_path,
		      max_path_len);
    if (rtnsts == -1)				/* If some problem */
    {
	iret = PSL_FILNOTFOU;			/* Set up error */
    }
    else					/* Found link OK */
    {
	if (resolved_path[0] == '/')		/* Link specified from root */
	{
	    (void)strcpy(new_path,		/* Straight copy */
			 resolved_path);
	    new_path[rtnsts] = '\0';		/* NULL terminate path */
        }
	else
	{
	    dir_ptr = strrchr(new_path, '/');	/* Find last directory */
	    dir_ptr += 1;
	    (void)strcpy(dir_ptr,		/* Copy link at this point */
			 resolved_path);
	    dir_ptr[rtnsts] = '\0';		/* NULL terminate path */
        }
/*
 *	Call PSL_REALPATH recursively to make sure that the link specified
 *	is fully resolved
 */
	iret = PSL_REALPATH(new_path,
			    new_path,
			    max_path_len,
			    PTRUE);
	*npid_ptr = strlen(new_path);		/* Set index in path 40X146 */
    }

/*
 *  Clean up, then exit here with status
 */
    free(resolved_path);
    return iret;
}
#endif	/* UNIX */
