/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
/*===========================================================================*/
/*									     */
/*			     **** C O M M O N ****   			     */
/*									     */
/*===========================================================================*/
/*									     */
/*			Module Name: SCR_RTNS_UNIX			     */
/*			============					     */
/*									     */
/* Description:								     */
/* ============								     */
/*									     */
/*	These routines allow the Screen Image routines to be used within     */
/*	the UNIX environment. They include:				     */
/*	 								     */
/*	SCR_PARSE_TRMCHR - Parse Terminal Characteristics Fields	     */
/*									     */
/*									     */
/*===========================================================================*/

#include <psl_macros.h>			/* Standard PSL macros */
/*
 *	Only valid for UNIX
 */
#ifdef UNIX

/*
 *	Define useful Include structures
 */
#include <stdio.h>			/* Standard I/O functions */
#include <curses.h>			/* Curses definitions */     
#include <term.h>			/* Terminal definitions */   	
#include <string.h>			/* String and mem rtns */    
#include <errno.h>                      /* Error codes */      
/*
 *  Function prototypes
 */
#include psl_common_file(psl_rtndef.h)  /* PSL functions */
#include psl_common_file(scr_rtndef.h)	/* SCRIMG routine prototypes */
#include "scr_intrtndef.h"		/* Internal SCR function prototypes */

/*
 *  Paramter and message files
 */
#include psl_common_file(shrmsg.h)	/* PSL message codes */
#include psl_common_file(trmtyp.h)	/* PSL terminal type codes */
#include psl_common_file(scr_prmdef.h)	/* SCR general parameter values */


/*
 *	SCR_PARSE_TRMCHR - Parse Terminal Characteristics Fields
 *	========================================================
 *
 *	Calling Format:
 *
 *	   status = SCR_PARSE_TRMCHR (trmnam_ptr, trmtyp, depth, width)
 *
 *	Where:
 *
 *	   status	Unsigned Long PSL Routine Return Status.
 *
 *			  PSL_NORMAL if the routine completed successfully.
 *			  PSL_BADPARAM if one of the parameters is invalid.
 *
 *	   trmnam_ptr	Pointer to terminal name string.
 *			Passed by Value.
 *
 *	   trmtyp	Unsigned short variable to hold the PSl terminal type
 *			corresponding to that defined by the terminal name.
 *			Passed by reference.
 *
 *	   depth	Unsigned short variable to hold the maximum number of
 *			lines currently available on the terminal. Passed by
 *			Reference.
 *
 *	   width	Unsigned short variable to hold the maximum number of
 *			columns currently available on the terminal. Passed by
 *			reference.
 */

unsigned long SCR_PARSE_TRMCHR (char *trmnam_ptr,
				unsigned short *trmtyp,
				unsigned short *depth,
				unsigned short *width)

{

/*
 *  Local Data
 */
    char                trmnam[512];    /* Local terminal name */     
    unsigned long       iret;           /* PSL return status */
    int                 sts;		/* setupterm() status */
#if defined(IBM) || defined(OSF)				      
    struct term         tmp_term;	/* Temporary terminal */
    struct term         *curses_term;   /* Curses terminal */
#else
    TERMINAL            tmp_term;	/* Temporary terminal */
    TERMINAL            *curses_term;   /* Curses terminal */
#endif
/*
 *  Check the parameters
 */
    if (trmnam_ptr == NULL) return PSL_BADPARAM;

    if (*trmnam_ptr == '\0') return PSL_BADPARAM;

    if (trmtyp == NULL) return PSL_BADPARAM;

    if (depth == NULL) return PSL_BADPARAM;

    if (width == NULL) return PSL_BADPARAM;

/*
 *  Check the various terminal alias files first.
 *  If we do not find an alias then we use it as it was passed to us.
 */
    iret = PSL_GET_TRMALIAS(trmnam_ptr, trmnam, sizeof(trmnam), NULL);
    if (!(psl_test_status(iret)))
    {
        (void) strcpy(trmnam, trmnam_ptr);
    }
    
/*
 *  Now look at the supplied terminal name and see if there is a PSL type
 *  for it.
 */
    if (memcmp(trmnam, "vt52", 4) == 0)
    {
	*trmtyp = TRTY_VT52;
    }
    else if (memcmp(trmnam, "vt1", 3) == 0)
    {
	*trmtyp = TRTY_VT100;
    }
    else if (memcmp(trmnam, "vt2", 3) == 0)
    {
	*trmtyp = TRTY_VT220;
    }
    else if (memcmp(trmnam, "vt3", 3) == 0)
    {
	*trmtyp = TRTY_VT220;
    }
    else if (memcmp(trmnam, "vt4", 3) == 0)
    {
	*trmtyp = TRTY_VT420;
    }
    else if (memcmp(trmnam, "wyse60", 5) == 0)
    {
	*trmtyp = TRTY_WYSE60;
    }
    else if (memcmp(trmnam, "hp", 2) == 0)			
    {
	*trmtyp = TRTY_HP700; 
    }                                                           
    else
    {
	return PSL_TRMUNSUP;
    }

/*
 *  Now see if we can work out the size of this terminal.
 *
 *  This is more complex than you would imagine because:
 *  [1] if this is an interactive utility then TRM_INIT may have already
 *	been called. If so calls to curses routines to retrieve information
 *	about a particular terminal will:
 *	  [a] replace curses current idea of the terminal with a new one
 *	  [b] send the reset escapes for the new terminal we want information
 *	    on to the terminal.
 *	[a] and [b] are OK so long as the terminals are the same.
 *
 *  [2] if this is not an interactive utility then escapes may be sent to
 *	the terminal as a result of this enquiry. This screws up listings
 *	to stdout (e.g. ldv listing a vrf in screen/linear format).
 *
 *  So we must fool curses by first saving a reference to the curses structures
 *  and then changing it to our own. We then call setupterm to initialize our
 *  TERMINAL structure (note setupterm() does not output the init escapes as
 *  newterm() and initscr() do). Lastly we give curses its TERMINAL struture
 *  back.
 */
    curses_term = cur_term;			/* Save curses TERMINAL */
#if defined(IBM) || defined(OSF)				      
    cur_term = &tmp_term;
#else
    set_curterm(&tmp_term);			/* Use our own TERMINAL */
#endif
    sts = setupterm(trmnam, 1, 0);		/* Initialize our TERMINAL */

/*
 *  If we have failed to initialize our terminal then just assume 24,80
 */
    if (sts != OK)
    {
        *depth = 24;
        *width = 80;
    }
/*
 *  else use the values obtained by setupterm().
 */
    else
    {
        *depth = lines;
        *width = columns;
    }
#if defined(IBM) || defined(OSF)				    
    cur_term = curses_term;
#else
    set_curterm(curses_term);		/* Restore the curses TERMINAL */
#endif

/*
 *  All done - return (success) status to caller
 */
    return PSL_NORMAL;			/* Exit with success */
}
#endif
