/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*
 * scl_ini_routines.c:
 * ===================
 *
 * This module contains all SCL's data initialization routines.
 */

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 *  PSL Macros.
 */
#include <psl_macros.h>
#include <vtest_macros.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL routines */
#include <trc_rtndef.h>         /* Tracing routines */
#include <vtest_rtndef.h>       /* Vtest routines */
#include "scl_rtndef.h"         /* SCL Routines */
#include "scl_def_tbl_rtndef.h"     /* SCL definitions routines */
#include "scl_thr_tbl_rtndef.h"     /* SCL thread script routines */
#include "scl_tst_tbl_rtndef.h"     /* SCL test script routines */
#include "scl_env_tbl_rtndef.h"     /* SCL environment routines */
#include "scl_cod_tbl_rtndef.h"     /* SCL code  routines */
#include "scl_sub_tbl_rtndef.h"     /* SCL subexpression routines */

/*
 *  Structure definitions.
 */
#include <shrmsg.h>         /* PSL Messages         */
#include <psldiag_prmdef.h>     /* Diagnostic Routine Symbols   */
#include <psl_trc_prmdef.h>     /* Tracing parameters       */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block    */
#include "scl_tbldef.h"         /* SCL Table Definitions    */

/*
 *    Data
 */
#include <psl_vm_prmdef.h>      /* Linked list parameters */
#include "scl_limits.h"         /* SCL limits */
#include <vtest_prmdef.h>       /* VTEST Parameters */
#include "scl_cmd.h"            /* Command Line Data Structure */
#include "scl_comdef.h"         /* Common Data */
#include "scl_tpacom.h"         /* Tparse common Data */
#include <scpdef.h>         /* Script object file */
#include <tofprmdef.h>          /* TOF parameters */
#include "scl_tlodef.h"         /* Test data */

/************************************************************************/
/*                                  */
/*    INI_ACQUIRE_MUTEX - Init routine for the ACQUIRE MUTEX command    */
/*    =================                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      ONTIMEOUT_LABEL = Cleared to show No on Timeout     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*      PIPE_FLAGSL = Cleared to show no flags yet      */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of an ACQUIRE MUTEX command.                      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_ACQUIRE_MUTEX (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the command options
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Initialize the Mutex Flags from the value passed to us.
     */
    (*tpa_ptr).pipe_flagsl = (*argblk_ptr).paraml;
    /* Store specified mutex flag */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    /*
     *    Clear the ON TIMEOUT label to show not specified.
     */
    (void) memset ((*tpa_ptr).ontimeout_labelc20 ,'\0'
                   ,sizeof ((*tpa_ptr).ontimeout_labelc20));
    /* Clear ON TIMEOUT Label buffer */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_ARRAY_INDICES - Init. routine for array indeces       */
/*    =================                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  Variable being parsed added to LL of array variables    */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of an array reference. It does the following:         */
/*                                  */
/*    1. Increments the array nesting level (may have nested reference) */
/*    2. Saves any current array reference data (may have netsed    */
/*   reference).                            */
/*    3. Initialises new level level data.              */
/*                                  */
/*    The information relating to the parsing of an array is held in    */
/*    an array block. If we have used the array block previously it is  */
/*    stored on a LL and the base block re-used for the next array  */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long INI_ARRAY_INDICES (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com  *com_ptr;     /* Common Data          */
    struct   scl_tpa  *tpa_ptr;     /* TPARSE Data          */
    struct   array_blk *array_blk_ptr;  /* Pointer to array block.  */
    unsigned long     iret;         /* Status return variable.  */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Increment array expression nesting level.
     */
    (*tpa_ptr).array_levell += 1;   /* Increment nesting level */

    /*
     *    If the array block pointer in the SCL_TPA structure is NULL then
     *    allocate a new block and save its pointer in this variable.
     */
    if ((*tpa_ptr).array_blk_ptrps == NULL)
    {
        (*tpa_ptr).array_blk_ptrps =
            (struct array_blk *) CALLOC (sizeof (struct array_blk) ,1);
        if ((*tpa_ptr).array_blk_ptrps == NULL)
        {
            PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"MALLOC (allocating space for an ARRAY_BLK)", 0);
        }
#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER5))
        {
            PSL_OUTPUT_TRACE ("MALLOC: array - '%d/%p' - INI_ARRAY\n",
                              sizeof (struct array_blk),
                              (*tpa_ptr).array_blk_ptrps);
        }
#endif  /* PSL_TRACING */
    }

    array_blk_ptr = (*tpa_ptr).array_blk_ptrps;
    /* Pointer to array block.  */

    /*
     *    If the current array block has been used, then save it on the LL of
     *    array blocks.
     */
    if ((*array_blk_ptr).var_ptrps != NULL)
    {
        iret = PSL_VM_INSERT ((*tpa_ptr).parsed_array_llps
                              ,(char *) array_blk_ptr ,NULL, NULL
                              ,PSL_VM_POS_END);
        /* ...insert into LL of parsed  */
        if  (!(psl_test_status (iret))) /* ...report any errors     */
        {
            PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"PSL_VM_INSERT (in INI_ARRAY_INDICES)", iret);
        }
    }

    /*
     *    ...and fill in the array block with:
     *
     *  The current variable symbol table pointer.
     *  The current parenthesis level.
     *      Clear the parsed index count.
     */
    (*array_blk_ptr).var_ptrps = (*tpa_ptr).var_ptrps;
    (*array_blk_ptr).nesting_levell = (*tpa_ptr).expr_levell;
    (*array_blk_ptr).array_indicesl = 0;

    /*
     *    ...and return success
     */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                      */
/*    INI_BREAK -  Initialise break count within case/switch statement  */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:                                                       */
/*      Variables symbol table                  */
/*    Outputs:  None                            */
/*                                  */
/*     This routine is called on parsing a BREAK statement.         */
/*     If case_count is > 0 at current switch level then a break is     */
/*     valid otherwise it is an error.                  */
/*     The SWITCH error flag has to be set if there is an error.    */
/*     A break must correspond to a case statement - although not every */
/*     case statement has to have a break                               */
/*                                  */
/*======================================================================*/

unsigned long   INI_BREAK (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   swstack *sw_ptr;               /* Switch stack                 */
    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    sw_ptr  = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl -1;
    /* Point to current entry      */


    /*
     *   There is an error to report if there are more
     *   breaks than there are "open" case statements. (ones
     *   not ending with a break.)
     *   Set the error flag so nothing is encoded relating to break.
     */

    if(((*sw_ptr).case_countl - (*sw_ptr).break_countl) <= 0)
    {
        REPORT_ERROR (argblk_ptr,SCL_INVBREAK);
        (*tpa_ptr).sw_error_fgl = 1;
    }

    /*
     * Set break_fgl as break parsed
     */
    if ((*tpa_ptr).sw_depthl != 0)                      /* Only set if switch */
    {
        (*tpa_ptr).break_fgl = 1;
    }
    /*
     *   Output tracing information if required
     */

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("1.CHK_BREAK*Break checked\n");
    }
#endif

    return PSL_NORMAL;           /* ...and return success    */

}

/************************************************************************/
/*                                  */
/*    INI_CALL_SCRIPT - Init. routine for the CALL SCRIPT command   */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  Variables used in parsing a call script command reset.  */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a CALL SCRIPT command.                                         */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long INI_CALL_SCRIPT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;   /* TPARSE data pointer */

    /*
     *  Clear flags and counts used in a CALL SCRIPT command.
     */
    tpa_ptr->script_flagl = 1;      /* Set script called indicator */
    tpa_ptr->call_prm_cntl = 0;     /* Clear call script parameter count */
    tpa_ptr->cmnd_opt_specifiedl = 0;   /* Show no command options yet */
    tpa_ptr->cmnd_optl = 0;     /* Clear options bitmask */

    /*
     *  Clear the ONERROR label to show not specified.
     */
    (void) memset (tpa_ptr->onerror_labelc16, '\0'
                   ,sizeof (tpa_ptr->onerror_labelc16));

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                  */
/*    INI_CALL_SUB - Init. routine for the CALL (subroutine) command    */
/*    ============                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  Variables used in parsing a call (subroutine) command   */
/*      reset.                          */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a CALL (subroutine) command.                                   */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long INI_CALL_SUB (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;   /* TPARSE data pointer */

    /*
     *  Clear flags and counts used in a CALL (subroutine) command.
     */
    tpa_ptr->subprm_cntl = 0;       /* Clear call subroutine parameter count */
    tpa_ptr->when_maskl = 0;        /* Show no WHEN options yet */
    tpa_ptr->cmnd_opt_specifiedl = 0;   /* Show no command options yet */
    tpa_ptr->cmnd_optl = 0;     /* Clear options bitmask */

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                      */
/*    INI_CASE -  Initial case                      */
/*    ===============                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                                                */
/*                                      */
/*    Outputs:  CASE_COUNTL at current switch level incremented     */
/*                                  */
/*     This routine is called on parsing an CASE  statement.        */
/*     It increments CASE_COUNTL to indicate that we have parsed        */
/*     a CASE  at the current switch depth.                 */
/*                                  */
/*======================================================================*/

unsigned long   INI_CASE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   swstack *sw_ptr;               /* Pointer to switch stack      */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    sw_ptr = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl - 1;
    /* Point to current entry */

    /*
     * Increment case count at current switch depth to indicate just parsed
     * another case command
     */
    if ((*tpa_ptr).sw_depthl <= 0 )
    {
        REPORT_ERROR (argblk_ptr, SCL_INVCASE);
    }
    else
    {
        (*sw_ptr).case_countl += 1;           /* Increment case count */
    }

    return PSL_NORMAL;                /* ...and return success */

}

/************************************************************************/
/*                                  */
/*    INI_CHAR_LIST - Init. Routine for Character List          */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  tpa-ptr->char_list_cnt = Zero               */
/*                                  */
/*    This routine initialises the variables that need to be set    */
/*    prior to parsing a character list.                */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_CHAR_LIST (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    /* Common data pointer */
    tpa_ptr = com_ptr->tpa_ptrps;   /* TPARSE data pointer */

    /*
     *    Clear character list value count.
     */
    tpa_ptr->char_list_val_cntl = 0;

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                                                      */
/*    INI_CHR_CMND - Init. Routine for Character Command Parsing        */
/*    ==============                                                    */
/*                                                                      */
/*    Type:     Action routine                                          */
/*    Status:   Always succeeds (PSL_NORMAL)                            */
/*    Pointers: Unchanged                                               */
/*    Inputs:   Tparse argument block.                                  */
/*    Outputs:  tpa-ptr->string_chrc8[0] = '\0'                         */
/*                                                                      */
/*    This routine initialises the variables that need to be set        */
/*    prior to parsing a character command.                             */
/*                                                                      */
/*----------------------------------------------------------------------*/
unsigned long INI_CHR_CMND (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;   /* TPARSE data pointer */

    /*
     *    Clear character command value buffer.
     */
    tpa_ptr->string_chrc8[0] = '\0';

    return PSL_NORMAL;              /* Return success */
}

/************************************************************************/
/*                                  */
/*    INI_CLEAR_SEMAPH - Init routine for the CLEAR SEMAPHORE command   */
/*    =================                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*      PIPE_FLAGSL = Cleared to show no flags yet      */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of an CLEAR SEMAPHORE command.                        */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_CLEAR_SEMAPH (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the command options
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Initialize the Semaphore Flags from the value passed to us.
     */
    (*tpa_ptr).pipe_flagsl = (*argblk_ptr).paraml;
    /* Store specified semaphore flag */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_CLOSE_PIPE - Init routine for the CLOSE PIPE command      */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of an CLOSE PIPE command.                         */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_CLOSE_PIPE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the Pipe Id and command options
     */
    (*tpa_ptr).open_fididxl = 0;    /* Clear Pipe Id        */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_COD_START - Init. Routine for Code Section            */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  sclsec_flagl = Set to SCLSEC_COD            */
/*                                  */
/*    This routine initialises the variables that need to be set    */
/*    prior to parsing the Code section of a script.            */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_COD_START (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *    Set the SCL script section flag to show that we are parsing the
     *    Code section.
     */
    (*tpa_ptr).sclsec_flagl = SCLSEC_COD;

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                  */
/*    INI_COND_EXPRESSION - Init. for Conditional Expressions       */
/*    ===================                       */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  (*tpa_ptr).cond_levell - Current level initialized. */
/*                                  */
/*    This routine initialises variables required during parsing    */
/*    of a conditional expression. It is called at the start        */
/*    of parsing of such an expression.                 */
/*                                  */
/************************************************************************/
unsigned long INI_COND_EXPRESSION (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct   scl_com *com_ptr;      /* SCL common data */
    struct   scl_tpa *tpa_ptr;      /* SCL Tparse data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;   /* SCL Tparse data */

    /*
     *  Initialise condiotion level. 
     */
    tpa_ptr->cond_levell = 0;

    return PSL_NORMAL;          /* ...before returning success  */
}

/************************************************************************/
/*                                  */
/*    INI_CONVERT - Init. routine for the CONVERT command       */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  ONERROR_LABEL = Cleared to show No On error     */
/*      CNV_OPT_SPECIFIED = Cleared to show no options yet. */
/*      CNV_OPTL = Cleared to default options           */
/*                                  */
/*    This routine initialises the variablles required during parsing   */
/*    of a CONVERT command.                                             */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_CONVERT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the ONERROR flags to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ONERROR Label buffer   */

    /*
     *    ...and the convert options bitmask.
     *    ...and set default options to SIGNED, LEADING ZEROS and LEFT JUSTIFIED.
     *
     */
    (*tpa_ptr).cnv_opt_specifiedl = 0;  /* No convert options yet   */
    (*tpa_ptr).cnv_optl = 0;        /* Set default convert options  */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/
/*                                  */
/*    INI_CREATE_PIPE - Init routine for the CREATE PIPE command    */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      ONTIMEOUT_LABEL = Cleared to show No on Timeout     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*      PIPE_FLAGSL = Cleared to show no flags yet      */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a CREATE PIPE command.                         */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_CREATE_PIPE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the command options
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Initialize the Pipe Flags from the value passed to us.
     */
    (*tpa_ptr).pipe_flagsl = (*argblk_ptr).paraml;
    /* Store specified pipe flag */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    /*
     *    Clear the ON TIMEOUT label to show not specified.
     */
    (void) memset ((*tpa_ptr).ontimeout_labelc20 ,'\0'
                   ,sizeof ((*tpa_ptr).ontimeout_labelc20));
    /* Clear ON TIMEOUT Label buffer */

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/
/*                                  */
/*    INI_DATAFILE - Init. routine for external datafile processing */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      OPT_OPT_SPECIFIED = Cleared to show no options yet. */
/*      OPT_OPTL = Cleared to default options           */
/*      OPEN_MASK - Cleared (will be defaulted later)       */
/*                                  */
/*    This routine initialises the variablles required during parsing   */
/*    of an external datafile accessing command.                */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_DATAFILE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the ONERROR flags to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ONERROR Label buffer   */

    /*
     *    ...and clear the standard label processinmg field
     */
    (void) memset ((*tpa_ptr).label_namec16 ,'\0'
                   ,sizeof ((*tpa_ptr).label_namec16));
    /* Clear Label buffer       */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */
    (*tpa_ptr).opt_opt_specifiedl = 0;  /* No options yet       */
    (*tpa_ptr).opt_optl = 0;        /* Set default options      */
    (*tpa_ptr).open_fididxl = 0;    /* Clear Fileid         */
    (*tpa_ptr).open_maskl = 0;      /* No open access specified     */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_DEF - Initialize definition section entry parsing data    */
/*    =======                               */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  INI_DEF (com_ptr);                      */
/*                                  */
/*    Where:                                */
/*                                  */
/*  com_ptr       = Pointer to Common Data Structure        */
/*                                  */
/*    This routine should be called before parsing each entry in the    */
/*    Definition section of a script. It (re)initializes the fields     */
/*    used in parsing a definition ((*tpa_ptr).def_*).          */
/*                                      */
/*----------------------------------------------------------------------*/
void INI_DEF (struct scl_com *com_ptr)
{
    /*
     *    Local data
     */
    struct   scl_tpa *tpa_ptr;      /* Pointer to TPARSE data */
    unsigned long     isub;         /* Loop control */

    /*
     *    Copy pointers for ease of reference.
     */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* Pointer to TPARSE data */

    /*
     *    Now initialize the variable parsing data
     */
    (void) memset ((*tpa_ptr).def_namec64 ,'\0'
                   ,sizeof ((*tpa_ptr).def_namec64));
    /* Clear variable definition name */

    (*tpa_ptr).def_nlengthl = 0;    /* Clear variable name length     */
    (*tpa_ptr).in_rec_headl = 0;        /* Reset - no longer specifying   */
    /* the record scope in header     */


    /*
     * Setting default values
     */
    if ((*tpa_ptr).def_recidxl != 0)     /* Initialisation if in  record */
    {
        (*tpa_ptr).def_addrl = (*tpa_ptr).rec_addrl; /*  Default address     */
        /*  is initial specified*/
    }                                                /*  record scope from   */
    /*  header REC_ADDRL    */
    else
    {
        (*tpa_ptr).rec_addrl = ADDR_LOCAL;
        (*tpa_ptr).def_addrl = ADDR_LOCAL;   /* Default addressing = local*/
    }


    (*tpa_ptr).def_typl = 0;        /* Clear definition type    */
    (*tpa_ptr).def_dtypl = 0;       /* Clear definition data type   */
    (*tpa_ptr).def_sizel = 1;       /* Default to size = 1      */
    (*tpa_ptr).def_values_offl = 0; /* Clear value list offset  */
    (*tpa_ptr).def_values_cntl = 0; /* Clear value count        */
    (*tpa_ptr).def_optl = 0;        /* Clear variable option mask   */
    (*tpa_ptr).def_opt_specifiedl = 0;  /* ...and no options specified  */
    (*tpa_ptr).array_dimcntl = 0;   /* No array dimensions yet */
    (*tpa_ptr).def_low_rangel = 0;  /* Reset Low range value    */
    (*tpa_ptr).def_high_rangel = 0; /* Reset high range value   */

    (void) memset ((*tpa_ptr).def_promptc50
                   ,'\0' ,sizeof ((*tpa_ptr).def_promptc50));

    /* Clear field used to save filename of file variables */
    (*tpa_ptr).def_filnamc256[0] = '\0';

    /*
     * 
     *    Initialize the array boundary LOW values to be 1. This ensures that
     *    arrays have boundaries starting at 1 unless the low value is explicitely
     *    specified.
     */
    for (isub = 0; isub < MAX_ARRAY_DIM; isub++)
    {
        (*tpa_ptr).def_array_bounds1l10 [isub] = 1;
    }


    return;
}

/************************************************************************/

/*    INI_DEF_START - Init. Routine for Definition Section      */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  sclsec_flagl = Set to SCLSEC_DEF            */
/*                                  */
/*    This routine initialises the variables that need to be set    */
/*    prior to parsing the Definition section of a script.      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_DEF_START (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *    Set the SCL script section flag to show that we are parsing the
     *    Definition section.
     */
    (*tpa_ptr).sclsec_flagl = SCLSEC_DEF;

    return PSL_NORMAL;          /* Return success */
}



/************************************************************************/
/*                                  */
/*    INI_DELETE_PIPE - Init routine for the DELETE PIPE command    */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      ONTIMEOUT_LABEL = Cleared to show No on Timeout     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*      PIPE_FLAGSL = Cleared to show no flags yet      */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a DELETE PIPE command.                         */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_DELETE_PIPE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the command options
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Initialize the Pipe Flags from the value passed to us.
     */
    (*tpa_ptr).pipe_flagsl = (*argblk_ptr).paraml;
    /* Store specified pipe flag */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    /*
     *    Clear the ON TIMEOUT label to show not specified.
     */
    (void) memset ((*tpa_ptr).ontimeout_labelc20 ,'\0'
                   ,sizeof ((*tpa_ptr).ontimeout_labelc20));
    /* Clear ON TIMEOUT Label buffer */

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/
/*                                  */
/*    INI_DISABLE - Init. routine for the DISABLE command       */
/*    ============                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  Script options bitmask cleared.             */
/*                                  */
/*    This routine initialises the variablles required during parsing   */
/*    of an DISABLE command.                                            */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_DISABLE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the script options bitmask.
     */
    (*tpa_ptr).scl_optl = 0;        /* Script options bitmask clear */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    INI_DISCONNECT - Init. routine for the DISCONNECT command parsing */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  scl_tpa.disconnect_maskl = 0.               */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a DISCONNECT command.                      */
/*                                  */
/************************************************************************/
unsigned long INI_DISCONNECT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  Clear the command options mask.
     */
    (*tpa_ptr).disconnect_maskl = 0;    /* Clear options mask */
    return PSL_NORMAL;          /* ...and return success */
}


/************************************************************************/
/*                                  */
/*    INI_DO        Init. Routine for the DO command        */
/*    ==================================================        */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   DO_STACK = Structured Command Stack         */
/*          DO_CNT = Count of stack entries             */
/*    Outputs:  DO_CNTL - Incremented to show next level        */
/*                                  */
/*    This routine pushes an entry for a DO command on the Structured   */
/*    Command Stack, reporting an error if the stack is full.       */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_DO (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct dostack  *do_ptr;        /* DO stack pointer     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;/* Common data pointer     */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Check to see if we have reached the maximum nesting depth and report
     *    an error if we have.
     */
    if ((*tpa_ptr).do_cntl >= MAX_DO)   /* If DO command stack full     */
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXCMNDNEST);
        /* Report nesting depth exceeded */
    }

    else                /* ...else if not exceeded  */
    {
        do_ptr = (*tpa_ptr).dostack_begps + (*tpa_ptr).do_cntl;
        /* Point to next FREE entry */

        /*
         *   Set this entry to be a DO loop and clear all of the variables on the
         *   structured command stack to show not specified.
         */
        (*do_ptr).start_addrpc = NULL;  /* Clear start address      */
        (*do_ptr).end_addrpc = NULL;    /* Clear start address      */
        (*tpa_ptr).do_cntl += 1;    /* ...and increment DO level    */
    }

    /*
     *    Clear the command options bitmask so that we can determine whether or
     *    not the STEP clause has been specified.
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    INI_ELSEIF  Init. Routine for the ELSEIF command          */
/*    ==========                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  None                            */
/*                                  */
/*    This routine initialises the 'TOF' line buffer for the        */
/*    Else record expr sub-record.                  */
/*                                  */
/************************************************************************/
unsigned long INI_ELSEIF (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* SCL Common data */
    struct scl_tpa  *tpa_ptr;       /* SCL Tparse data */
    char   *lin_ptr;       /* Pointer to line buffer */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;
    lin_ptr = com_ptr->toflinc1024; /* 'tof' line buffer pointer */

    /*
     *  Initialise the default 'TOF' line buffer for the Else record expr sub-record.
     */
    (void) strcpy (lin_ptr, "expr ");

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/
/*                                  */
/*    INI_ENABLE - Init. routine for the ENABLE command         */
/*    ===========                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  Script options bitmask cleared.             */
/*                                  */
/*    This routine initialises the variablles required during parsing   */
/*    of an ENABLE command.                                             */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_ENABLE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the script options bitmask.
     */
    (*tpa_ptr).scl_optl = 0;        /* Script options bitmask clear */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    INI_ENV_START - Init. Routine for Environment Section     */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  sclsec_flagl = Set to SCLSEC_ENV            */
/*                                  */
/*    This routine initialises the variables that need to be set    */
/*    prior to parsing the Environment section of a script.     */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_ENV_START (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *    Set the SCL script section flag to show that we are parsing the
     *    Environment section.
     */
    (*tpa_ptr).sclsec_flagl = SCLSEC_ENV;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/
/*                                  */
/*    INI_EXIT - Init. routine for the EXIT command parsing     */
/*    =========                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  DEF_VALUES_OFFL = Value list offset set         */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of an EXIT command.                                           */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_EXIT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the exit options mask and the status return operand descriptor
     */
    (*tpa_ptr).exit_maskl = 0;      /* Clear options mask       */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    INI_EXTRACT   Init. Routine for the EXTRACT function      */
/*    ========================================================      */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  tpa_ptr->opt_optl - Cleared - no options yet        */
/*      tpa_ptr->opt_opt_specifiedl - Cleared no options    */
/*          specified yet                   */
/*                                  */
/*    This routine initializes the operands required for parsing of the */
/*    EXTRACT function.                                         */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_EXTRACT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    (*tpa_ptr).opt_opt_specifiedl = 0;  /* No options yet       */
    (*tpa_ptr).opt_optl = 0;        /* Set default options      */

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/

/*    INI_FORMAT - Init. routine for the FORMAT command         */
/*    =========                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  onerror_labelc16 - cleared no ON ERROR label specified  */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a FORMAT command.                      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_FORMAT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the label used for error branching.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear Label buffer       */
    /*
     *    Initialize the command options
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     * Set format flag - indicate in format command
     */
    (*tpa_ptr).format_flagl = 1;
    (*tpa_ptr).format_arguml = 0;
    (*tpa_ptr).format_char_setl = 0;
    return PSL_NORMAL;          /* ...before returning success  */
}

/************************************************************************/
/*                                                                      */
/*    INI_GET - Init. routine for GET command parsing                   */
/*    =======                                                           */
/*                                                                      */
/*    Type: Action routine                                              */
/*    Status:   Always succeeds (PSL_NORMAL)                            */
/*    Pointers: Unchanged                                               */
/*    Inputs:   Tparse argument block.                                  */
/*    Outputs:  scl_tpa.http_reqtypl = TOF_REQTYP_GET.                  */
/*              scl_tpa.http_flagsl = 0.                                */
/*                                                                      */
/*    This routine initialises the variables required during parsing    */
/*    of a GET command.                                                 */
/*                                                                      */
/************************************************************************/
unsigned long INI_GET (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  Set the HTTP request type.
     */
    (*tpa_ptr).http_reqtypl = TOF_REQTYP_GET;
    (*tpa_ptr).http_flagsl = 0;

    return PSL_NORMAL;
}

/************************************************************************/
/*                                                                      */
/*    INI_HEAD - Init. routine for HEAD command parsing                 */
/*    ========                                                          */
/*                                                                      */
/*    Type: Action routine                                              */
/*    Status:   Always succeeds (PSL_NORMAL)                            */
/*    Pointers: Unchanged                                               */
/*    Inputs:   Tparse argument block.                                  */
/*    Outputs:  scl_tpa.http_reqtypl = TOF_REQTYP_HEAD.                 */
/*              scl_tpa.http_flagsl = 0.                                */
/*                                                                      */
/*    This routine initialises the variables required during parsing    */
/*    of a HEAD command.                                                */
/*                                                                      */
/************************************************************************/
unsigned long INI_HEAD (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  Set the HTTP request type.
     */
    (*tpa_ptr).http_reqtypl = TOF_REQTYP_HEAD;
    (*tpa_ptr).http_flagsl = 0;

    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    INI_HTTP_REQUEST - Init. routine for HTTP request parsing     */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  Initialise fields required during HTTP resquest     */
/*      parsing.                        */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of an HTTP request command.                   */
/*                                  */
/************************************************************************/
unsigned long INI_HTTP_REQUEST (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;   /* TPARSE data pointer */

    /*
     *  Initialise the HTTP request fields.
     */
    tpa_ptr->http_hdrval_cntl = 0;  /* No HEADER clause values */
    tpa_ptr->http_withval_cntl = 0; /* No WITH clause values */
    tpa_ptr->http_woutval_cntl = 0; /* No WITHOUT clause values */

    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    INI_IF    Init. Routine for the IF command            */
/*    ======                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   IFSTACK = IF Command Stack              */
/*          IFSTACK_LEVEL = Count of stack entries          */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks to determine whether or not there is space    */
/*    on the IF Stack for another IF command. If there is not we report */
/*    an error. If there is space then we clear the data at this level  */
/*    and initialise the 'TOF' line buffer for the IfThen/Conditional   */
/*    record expr sub-record.                       */
/*                                  */
/*    NOTE:                             */
/*  We do not actually update the IF Stack level because this may   */
/*  be an IF ... GOTO which does not actually need one. The stack   */
/*  level is incremented within the routine:            */
/*                                  */
/*  ENC_IF_THEN                         */
/*                                  */
/************************************************************************/
unsigned long INI_IF (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* SCL Common data */
    struct scl_tpa  *tpa_ptr;       /* SCL Tparse data */
    struct ifstack  *if_ptr;        /* IF stack pointer */
    char   *lin_ptr;       /* Pointer to line buffer */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;
    lin_ptr = com_ptr->toflinc1024; /* 'tof' line buffer pointer*/

    /*
     *  Check to see if we have reached the maximum nesting depth and report
     *  an error if we have, otherwise initialise the IF stack and the default
     *  'TOF' line buffer for the IfThen/Conditional record expr sub-record.
     */
    if (tpa_ptr->ifstack_levell >= MAX_IF)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXCMNDNEST);
    }

    else
    {
        if_ptr = tpa_ptr->ifstack_begps + tpa_ptr->ifstack_levell;
        /* Point to next FREE entry */
        if_ptr->else_parsedn = PFALSE;  /* Show ELSE not processed yet */
        (void) strcpy (lin_ptr, "expr ");
    }

    return PSL_NORMAL;          /* ...and return success */
}

/************************************************************************/
/*                                  */
/*    INI_INPUT - Initialization routine for the INPUT command.     */
/*    ==========                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                            */
/*    Outputs:  CMND_OPT_SPECIFIED = Command options specified bitmask. */
/*      OPT_OPTL = Input options mask initialised       */
/*      OPR = Operator initialized              */
/*                                  */
/*    This routine initializes the variables used in parsing an     */
/*    INPUT command OR an input stream.                 */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_INPUT (struct tpaarg *argblk_ptr)


{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the command options specified mask.
     */
    (*tpa_ptr).opt_opt_specifiedl = 0;  /* No input command options specified */
    (*tpa_ptr).opt_optl = 0;        /* Clear options bitmask.   */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_INPUT_RATE - Initialization routine for input rate parsing    */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                            */
/*    Outputs:  DEC_INTEGER = Zero                  */
/*          DEC_FRACTION = Zero                 */
/*          DEFINE_RATE_OPD cleared                 */
/*          SCPERR_OFFSET = Set to be start of this instruction */
/*                                      */
/*    This routine initializes the variables used in parsing the    */
/*    Environment section INPUT RATE statement and the SET INPUT    */
/*    RATE command, i.e. the variables used to hold the integer     */
/*    and fractional parts of the specified input rate. If we specify   */
/*    an integer variable then we will need to clear the rate operand   */
/*                                      */
/*----------------------------------------------------------------------*/

unsigned long   INI_INPUT_RATE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the data required during parsing of an input rate command
     */
    (*tpa_ptr).dec_integerl = 0;    /* Clear integer value      */
    (*tpa_ptr).dec_fractionl = 0;   /* Clear fractional value   */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_INT_VALUE  - Init. routine for integer value parsing      */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                            */
/*    Outputs:  SIGN_CHR = Zero                     */
/*                                  */
/*    This routine initializes the SIGN_CHR variable used in parsing    */
/*    an integer value.                         */
/*                                      */
/*----------------------------------------------------------------------*/

unsigned long   INI_INT_VALUE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the data required during parsing of an integer value
     */
    (*tpa_ptr).sign_chrb = 0;       /* Clear sign character     */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    INI_LOAD - Init. routine for the LOAD command parsing     */
/*    ========                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*          (*argblk_ptr).paraml = Load data (TOF_LODDAT_*)     */
/*    Outputs:  scl_tpa.loddatl = Set from argument block parameter.    */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a generic LOAD command.                            */
/*                                  */
/************************************************************************/
unsigned long INI_LOAD (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  Set the load data flag to the value passed in the Tparse argument block
     *  parameter field.
     */
    (*tpa_ptr).loddatl = (*argblk_ptr).paraml;
    return PSL_NORMAL;          /* ...and return success */
}


/************************************************************************/

/*    INI_LOAD_HTTP - Init. routine for the HTTP LOAD command parsing   */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*          (*argblk_ptr).paraml = Load data (TOF_LLDDAT_*)     */
/*    Outputs:  scl_tpa.loddatl = Set from argument block parameter.    */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of an HTTP-specific LOAD command.                         */
/*                                  */
/************************************************************************/
unsigned long INI_LOAD_HTTP (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  Set the load data flag to the value passed in the Tparse argument block
     *  parameter field.
     */
    (*tpa_ptr).loddatl = (*argblk_ptr).paraml;
    return PSL_NORMAL;          /* ...and return success */
}


/************************************************************************/
/*                                  */
/*    INI_LOCATE    Init. Routine for the LOCATE function       */
/*    =======================================================       */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  OPT_OPTL - Cleared - no options yet         */
/*      OPT_OPT_SPECIFIEDL - Cleared no options specified yet   */
/*                                  */
/*    This routine initializes the operands required for parsing of the */
/*    LOCATE function.                                          */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_LOCATE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    (*tpa_ptr).opt_opt_specifiedl = 0;  /* No options yet       */
    (*tpa_ptr).opt_optl = 0;        /* Set default options      */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    INI_LOG - Init. routine for SCL logging command parsing       */
/*    =======                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*          (*argblk_ptr).paraml = Log destination (LOG_*)      */
/*    Outputs:  scl_tpa.logdstl = Set from argument block parameter.    */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of an SCL logging command, i.e. a LOG, HISTORY, NOTE, REPORT  */
/*    or TRACE command.                         */
/*                                  */
/************************************************************************/
unsigned long INI_LOG (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  Set the log destination flag to the value passed in the Tparse argument block
     *  parameter field.
     */
    (*tpa_ptr).logdstl = (*argblk_ptr).paraml;
    return PSL_NORMAL;          /* ...and return success */
}

/************************************************************************/
/*                                  */
/*    INI_NEG_EXP_DIST    Init. Routine for the NEG_EXP_DIST function       */
/*    =======================================================       */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  OPT_OPTL - Cleared - no options yet         */
/*      OPT_OPT_SPECIFIEDL - Cleared no options specified yet   */
/*                                  */
/*    This routine initializes the operands required for parsing of the */
/*    NEG_EXP_DIST function.                                          */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_NEG_EXP_DIST (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    (*tpa_ptr).opt_opt_specifiedl = 0;  /* No options yet       */
    (*tpa_ptr).opt_optl = 0;        /* Set default options      */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_ON - Init. routine for the ON ERROR command           */
/*    =======                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  onerror_labelc16 - Cleared as not specified yet     */
/*                                  */
/*    This routine initialises the variablles required during parsing   */
/*    of an ON ERROR Command.                       */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_ON (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the ONERROR flags to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ONERROR Label buffer   */

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    INI_OPEN_PIPE - Initialization routine for the OPEN PIPE command  */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      ONTIMEOUT_LABEL = Cleared to show No on Timeout     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*      OPEN_MASKL = Cleared (will be defaulted later)      */
/*      PIPE_FLAGSL = Cleared to show no flags yet      */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of an OPEN PIPE command.                          */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_OPEN_PIPE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the Pipe Id, Access Mask and command options
     */
    (*tpa_ptr).open_fididxl = 0;    /* Clear Pipe Id        */
    (*tpa_ptr).open_maskl = 0;      /* No open access specified     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Initialize the Pipe Flags from the value passed to us.
     */
    (*tpa_ptr).pipe_flagsl = (*argblk_ptr).paraml;
    /* Store specified pipe flag */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    /*
     *    Clear the ON TIMEOUT label to show not specified.
     */
    (void) memset ((*tpa_ptr).ontimeout_labelc20 ,'\0'
                   ,sizeof ((*tpa_ptr).ontimeout_labelc20));
    /* Clear ON TIMEOUT Label buffer */

    return PSL_NORMAL;          /* ...and return success    */
}



/************************************************************************/
/*                                  */
/*    INI_OPTS  - Reset Options masks                   */
/*    =========                                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Operand                */
/*          OPT_OPTL = Current options bitmask          */
/*    Outputs:  Operand descriptor set up with options mask     */
/*          OPT_OPTL = Cleared                  */
/*          OPT_OPT_SPECIFIED = Cleared to show no options yet. */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_OPTS (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;


    (*tpa_ptr).opt_optl = 0;        /* Clear options bit mask.  */
    (*tpa_ptr).opt_opt_specifiedl = 0;  /* Clear specified options bit mask.*/

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/
/*                                                                      */
/*    INI_POST - Init. routine for POST command parsing                 */
/*    ========                                                          */
/*                                                                      */
/*    Type: Action routine                                              */
/*    Status:   Always succeeds (PSL_NORMAL)                            */
/*    Pointers: Unchanged                                               */
/*    Inputs:   Tparse argument block.                                  */
/*    Outputs:  scl_tpa.http_reqtypl = TOF_REQTYP_POST.                 */
/*              scl_tpa.http_flagsl = 0.                                */
/*                                                                      */
/*    This routine initialises the variables required during parsing    */
/*    of a POST command.                                                */
/*                                                                      */
/************************************************************************/
unsigned long INI_POST (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  Set the HTTP request type.
     */
    (*tpa_ptr).http_reqtypl = TOF_REQTYP_POST;
    (*tpa_ptr).http_flagsl = 0;

    return PSL_NORMAL;
}

/************************************************************************/

/*    INI_PRM_HEADER - Init. routine for Command Parameter HEADER   */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  scl_tpa.this_cmnd_optl = CMND_OPT_HEADER        */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of the command parameter HEADER - available on HTTP request   */
/*    commands.                             */
/*                                  */
/************************************************************************/
unsigned long INI_PRM_HEADER (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  Set the HEADER command option ID.
     *
     *  The HEADER parameter is not strictly an option, but this field is
     *  used to indicate that the following operand(s) are associated with
     *  the HEADER parameter.
     */
    (*tpa_ptr).this_cmnd_optl = CMND_OPT_HEADER;
    return PSL_NORMAL;
}



/************************************************************************/
/*                                  */
/*    INI_READ - Initialization routine for the READ command        */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a READ PIPE command and calls INI_DATAFILE to initialize   */
/*    those needed by a READ FILE command.                  */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_READ (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the Pipe Id and command options
     */
    (*tpa_ptr).open_fididxl = 0;    /* Clear Pipe Id        */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    /*
     *    Clear the ON TIMEOUT label to show not specified.
     */
    (void) memset ((*tpa_ptr).ontimeout_labelc20 ,'\0'
                   ,sizeof ((*tpa_ptr).ontimeout_labelc20));
    /* Clear ON TIMEOUT Label buffer */
    /*
     *    Call INI_DATAFILE and return its status.
     */
    return INI_DATAFILE (argblk_ptr);   /* ...and return success    */
}



/************************************************************************/
/*                                  */
/*    INI_RELEASE_MUTEX - Init routine for the RELEASE MUTEX command    */
/*    =================                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*      PIPE_FLAGSL = Cleared to show no flags yet      */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of an RELEASE MUTEX command.                      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_RELEASE_MUTEX (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the command options
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Initialize the Mutex Flags from the value passed to us.
     */
    (*tpa_ptr).pipe_flagsl = (*argblk_ptr).paraml;
    /* Store specified mutex flag */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/
/*                                  */
/*    INI_RESUME - Init. routine for the RESUME command         */
/*    ===========                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  ONERROR_LABEL = Cleared to show No On error     */
/*                                  */
/*      cmnd_optl -  Cleared to default options           */
/*                                  */
/*    This routine initialises the variablles required during parsing   */
/*    of a RESUME command.                      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_RESUME (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the ONERROR flags to show not specified.
     */

    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */

    (*tpa_ptr).cmnd_optl  = 0;          /* Clear options mask           */


    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ONERROR Label buffer   */

    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    INI_RESET - Initialization routine for the RESET command.     */
/*    =========                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                            */
/*    Outputs:  CMND_OPT_SPECIFIED = Command options specified bitmask. */
/*      CMND_OPTL = Input options mask initialised      */
/*                                  */
/*    This routine initializes the variables used in parsing a      */
/*    RESET command.                            */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_RESET   (struct tpaarg *argblk_ptr)


{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the command options specified mask.
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */
    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/
/*                                  */
/*    INI_REPLY_PIPE - Init routine for the REPLY PIPE command      */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a REPLY PIPE command.                          */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_REPLY_PIPE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the Pipe Id and command options
     */
    (*tpa_ptr).open_fididxl = 0;    /* Clear Pipe Id        */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/

/*    INI_SWITCH -   On  parsing a switch statement         */
/*    ===========                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:                                                       */
/*                                      */
/*    Outputs:  SW_STARTL set to 1 to indicate switch parsed        */
/*      SW_DEPTHL incremented if not exceeded nesting level     */
/*      Initialises all switch stack data                       */
/*                                      */
/*          sw_dtypl         Data type switch var           */
/*          end_switch_offpc     End switch offset              */
/*          last_case_offpc      Last case offset               */
/*          case_countl      Case count in switch           */
/*          break_countl         Break count in switch          */
/*          break_addrpl         End switch (break)  offset     */
/*              def_case_fgl             Default case flag              */
/*      sw_error_fgl         Error in switch        */
/*     This routine is called on parsing an SWITCH statement.       */
/*     It checks the nesting  level of switch statements                */
/*     If within allowed nesting level :                                */
/*     It sets a variable to indicate that we have parsed a SWITCH      */
/*                                  */
/*======================================================================*/

unsigned long   INI_SWITCH (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   swstack *sw_ptr;               /* Pointer to switch stack      */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;


    if((*tpa_ptr).sw_depthl >= 10)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXSWNEST);
    }
    else
    {                            /* ...else if not exceeded  */

        (*tpa_ptr).sw_depthl += 1 ;            /* Increment switch counter  */


        sw_ptr = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl - 1;
        /* Point to next FREE entry */

        /*
         * Clear variables on structured command switch stack.
         */

        (*sw_ptr).sw_dtypl = 0;               /* Clear dtype of switch var */
        (*sw_ptr).end_switch_offpc = NULL;        /* Clear end switch offset   */
        (*sw_ptr).last_case_offpc = NULL;         /* Clear last case offset    */
        (*sw_ptr).case_countl = 0;                /* Case count in switch init */
        (*sw_ptr).break_countl = 0;           /* Break count in switch     */
        (*sw_ptr).break_addrpl = NULL;            /* Clear break address offset*/
        (*sw_ptr).def_case_fgl = 0;           /* Clear default_case flag   */
        /*
         * Set sw_start = 1 as we have just parsed a switch statement
         */

        (*tpa_ptr).sw_startl = 1  ;            /* Set as parsed a switch     */

    }
    return PSL_NORMAL;               /* ...and return success */

}
/*
 * 
 */

/************************************************************************/
/*                                  */
/*    INI_SUSPEND - Init. routine for the SUSPEND command       */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  ONERROR_LABEL = Cleared to show No On error     */
/*      cmnd_optl -  Cleared to default options           */
/*                                  */
/*    This routine initialises the variablles required during parsing   */
/*    of a SUSPEND command.                         */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_SUSPEND (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */


    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */

    (*tpa_ptr).cmnd_optl  = 0;          /* Clear options mask           */


    /*
     *    Clear the ONERROR flags to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ONERROR Label buffer   */

    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    INI_SNAP - Init. routine for the SNAP command         */
/*    =========                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  onerror_labelc16 - cleared no ON ERROR label specified  */
/*      snap_maskl - Cleared as nop SNAP options yet        */
/*      cmnd_optl -  Cleared to default options           */
/*    This routine initialises the variablles required during parsing   */
/*    of a SNAP command.                        */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_SNAP (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    ...and clear the SNAP options mask.
     */
    (*tpa_ptr).snap_maskl = 0;      /* Clear snap options mask. */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */

    (*tpa_ptr).cmnd_optl  = 0;          /* Clear options mask           */


    /*
     *    ...and clear the label used for error branching.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear Label buffer       */
    return PSL_NORMAL;          /* ...before returning success  */
}

/************************************************************************/
/*                                  */
/*    INI_SET - Init. routine for the SET command           */
/*    ========                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*                                  */
/*    This routine initialises the variablles required during parsing   */
/*    of a SET command.                         */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_SET (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the ONERROR flags to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ONERROR Label buffer   */

    /*
     *    ...and the convert options bitmask.
     *    ...and set default options to SIGNED, LEADING ZEROS and LEFT JUSTIFIED.
     */
    (*tpa_ptr).cnv_opt_specifiedl = 0;  /* No convert options yet   */
    (*tpa_ptr).cnv_optl = 0;        /* Set default convert options  */
    /*
     * 
     * Must initialise all options.
     */
    (*tpa_ptr).opt_optl = 0;        /* Clear options bit mask.  */
    (*tpa_ptr).opt_opt_specifiedl = 0;  /* Clear specified options bit mask.*/

    /*
     *  Reset expression nesting levels.
     */
    (*tpa_ptr).expr_levell = 0;     /* Reset expression nesting level */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_SET_OPD - Init. routine for the SET command operands.     */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  CNV_OPT_SPECIFIED = Cleared to show no options yet. */
/*      CNV_OPTL = Cleared to default options           */
/*                                  */
/*    This routine initialises the operands required during parsing */
/*    of a SET command.                         */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_SET_OPD (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    ...and the convert options bitmask.
     *    ...and set default options to SIGNED, LEADING ZEROS and LEFT JUSTIFIED.
     */
    (*tpa_ptr).cnv_opt_specifiedl = 0;  /* No convert options yet   */
    (*tpa_ptr).cnv_optl = 0;        /* Set default convert options  */

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/
/*                                  */
/*    INI_SET_SEMAPH - Init routine for the SET SEMAPHORE command   */
/*    =================                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*      PIPE_FLAGSL = Cleared to show no flags yet      */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of an SET SEMAPHORE command.                      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_SET_SEMAPH (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the command options
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Initialize the Semaphore Flags from the value passed to us.
     */
    (*tpa_ptr).pipe_flagsl = (*argblk_ptr).paraml;
    /* Store specified semaphore flag */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/
/*                                  */
/*    INI_SIMPLE_COND - Init. routine for the Simple condition parsing  */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  OPT_OPT_SPECIFIED = Cleared to show no options yet. */
/*      OPT_OPTL = Cleared to default options           */
/*                                  */
/*    This routine initialises the operands required during parsing */
/*    of a Simple condition.                        */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_SIMPLE_COND (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    ...and the options bitmask.
     */
    (*tpa_ptr).opt_opt_specifiedl = 0;  /* No options yet       */
    (*tpa_ptr).opt_optl = 0;        /* Set default options      */
    (*tpa_ptr).attr_opt_specifiedl = 0; /* No character attributes yet  */
    (*tpa_ptr).attr_optl = 0;       /* ...and clear character attributes */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* Show no command options yet. */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_STR_PARSING - Init. routine for string parsing        */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                            */
/*    Outputs:  CHR_STRING_LENGTH = Zero                */
/*                                  */
/*    This routine initializes the variables used to parse a character  */
/*    string. It sets the character string length variable      */
/*    CHR_STRING_LENGTH to zero and enables blank recognition by    */
/*    setting blanks bit in the options field of the TPARSE argument    */
/*    block.                                */
/*                                      */
/*----------------------------------------------------------------------*/

unsigned long   INI_STR_PARSING (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the data required during parsing of a string value
     */
    (void) BLANKS_ON (argblk_ptr);  /* Switch blank recognition on  */
    (*tpa_ptr).chr_strlenl = 0;     /* Initialize string length */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    INI_THR - Initialize thread definition work structure     */
/*    =======                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine initializes the thread definition work structure.    */
/*                                  */
/************************************************************************/
unsigned long INI_THR (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Return status */
    struct   scl_com    *com_ptr;       /* Pointer to common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl     *thr_ptr;       /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to test definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */
    thr_ptr = (*tpa_ptr).def_thr_wrkps; /* Thread definition work data */

    /*
     *  Clear the work structure.
     */
    (void) memset ((void *) thr_ptr, '\0', sizeof (struct thrtbl));

    /*
     *  Allocate space for TDF structure.
     */
    tdf_ptr = (struct tdf *) calloc (1, sizeof (struct tdf));

    if (tdf_ptr == NULL)        /* If an error */
    {
        iret = PSL_INSVIRMEM;       /* Set up error */
        PSL_SIGDIAG (           /* Signal Diagnostic */
            (*com_ptr).diag_ptrps,  /* Use this diagnostic stack */
            2,              /* Number of message segments */
            PSL_SYSERVERR,      /* First message */
            "CALLOC for TDF Structure", /* Parameter */
            iret);          /* Secondary message */
    }

    (*thr_ptr).thr_tdf_ptrps = tdf_ptr; /* Save pointer */

    /*
     *  Initialize TDF structure - ensure that all fields are 'defaulted'.
     */
    SCL_DEFAULT_TDF ((*com_ptr).diag_ptrps, tdf_ptr);
    (void) SCL_UNDEFAULT_TDF
    ((*com_ptr).diag_ptrps, NULL, NULL, tdf_ptr, NULL, tdf_ptr);
    SCL_FREE_TDF (tdf_ptr);

    return PSL_NORMAL;          /* Always return success */
}


/************************************************************************/

/*    INI_THR_DEF - Initialize thread script definition parsing data    */
/*    ===========                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  INI_THR_DEF (com_ptr);                      */
/*                                  */
/*    Where:                                */
/*                                  */
/*  com_ptr       = Pointer to Common Data Structure        */
/*                                  */
/*    This routine should be called before parsing each entry in the    */
/*    Definition section of a Thread script. It (re)initializes the */
/*    fields used in parsing a thread definition ((*tpa_ptr).def_*).    */
/*                                      */
/************************************************************************/
void INI_THR_DEF (struct scl_com *com_ptr)
{
    /*
     *    Local data
     */
    unsigned long     jsub;         /* Loop control */
    struct thrtbl    *thr_ptr;      /* Pointer to thread definition */
    struct scl_tpa   *tpa_ptr;      /* Pointer to TPARSE data */

    /*
     *    Copy pointers for ease of reference.
     */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* Pointer to TPARSE data */
    thr_ptr = (*tpa_ptr).def_thr_wrkps; /* Thread definition work data */

    /*
     *    Initialize the variable parsing data.
     */
    (void) memset ((*tpa_ptr).def_namec64,
                   '\0',
                   sizeof ((*tpa_ptr).def_namec64));
    /* Clear definition name */
    (*tpa_ptr).def_nlengthl = 0;    /* Clear definition name length */
    (*tpa_ptr).def_typl = 0;        /* Clear definition type */
    (*tpa_ptr).def_dtypl = 0;       /* Clear definition data type */
    (*tpa_ptr).def_sizel = 1;       /* Default item size */

    /*
     *  Clear specified options flags and option source line numbers.
     */
    for (jsub = 0; jsub < SCL_OPT_MAX; jsub++)
    {
        (*tpa_ptr).opt_flagn64 [jsub] = PFALSE;
        (*tpa_ptr).opt_linel64 [jsub] = 0;
    }

    /*
     *  If the thread definition work structure has a TDF structure
     *  allocated there must have been a compilation error while parsing
     *  a thread definition. If so, free the space allocated for the TDF
     *  structure now.
     */
    if ((*thr_ptr).thr_tdf_ptrps != NULL)
    {
        free ((void *) (*thr_ptr).thr_tdf_ptrps);
    }

    /*
     *    Set the SCL script section flag to show that we are parsing the
     *    Definition section.
     */
    (*tpa_ptr).sclsec_flagl = SCLSEC_DEF;

    return;
}


/************************************************************************/

/*    INI_THR_RECOPT - Initialize thread recording options      */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine initializes the recording options fields in the  */
/*    thread definition work structure. It should be called immediately */
/*    before parsing the options in a thread recording options clause.  */
/*    Note that this routine is required to clear the default options   */
/*    set up in the thread definition work structure initialization */
/*    routine.                              */
/*                                  */
/************************************************************************/
unsigned long INI_THR_RECOPT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Pointer to common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl     *thr_ptr;       /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to test definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Since the recording type is initialized to Input+Output we must
     *  explicitly clear it here if the user is to specify what he wants.
     */
    (*tdf_ptr).tdf_record_typel = 0;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_RECORD_TYPE;

    return PSL_NORMAL;          /* Always return success */
}


/************************************************************************/

/*    INI_TST - Initialize test definition work structure       */
/*    =======                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   (*argblk_ptr).paraml = Test type (DEF_TEST or       */
/*                      DEF_SUBTEST).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine initializes the test definition work structure   */
/*    for an external TEST declaration. External declarations cannot    */
/*    specify
/*                                  */
/************************************************************************/
unsigned long INI_TST (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Return status */
    unsigned long       jsub;       /* Loop control */
    struct   scl_com    *com_ptr;       /* Pointer to common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   tlh        *tlh_ptr;       /* Pointer to test header data */
    struct   prm        *prm_ptr;       /* Pointer to test parameter data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */
    tst_ptr = (*tpa_ptr).def_tst_wrkps; /* Test definition work data */

    /*
     *  If the test definition work structure has any structures 'hanging off'
     *  it there must have been a compilation error while parsing
     *  a test definition. If so, free the space allocated for the structures.
     */
    if ((*tst_ptr).tst_prm_ptrps != NULL)
    {
        free ((void *) (*tst_ptr).tst_prm_ptrps);
    }

    if ((*tst_ptr).tst_tlh_ptrps != NULL)
    {
        free ((void *) (*tst_ptr).tst_tlh_ptrps);
    }

    /*
     *  Clear the work structure.
     */
    (void) memset ((void *) tst_ptr, '\0', sizeof (struct tsttbl));

    /*
     *  No test references yet.
     */
    (*tst_ptr).tst_exe_refn = PFALSE;
    (*tst_ptr).tst_wait_refn = PFALSE;

    /*
     *  Set the sub-test flag.
     */
    if ((*argblk_ptr).paraml == DEF_SUBTEST)
    {
        (*tst_ptr).tst_subtstn = PTRUE;     /* Test is a SUB-TEST */
        (*tst_ptr).tst_definedn = PFALSE;   /* SUB-TEST not defined yet */
    }
    else
    {
        (*tst_ptr).tst_subtstn = PFALSE;
    }

    /*
     *  Allocate space for TLH structure.
     */
    tlh_ptr = (struct tlh *) calloc (1, sizeof (struct tlh));

    if (tlh_ptr == NULL)         /* If an error */
    {
        iret = PSL_INSVIRMEM;       /* Set up error */
        PSL_SIGDIAG (           /* Signal Diagnostic */
            (*com_ptr).diag_ptrps,  /* Use this diagnostic stack */
            2,          /* Number of message segments */
            PSL_SYSERVERR,      /* First message */
            "CALLOC for TLH Structure",
            /* Parameter */
            iret);          /* Secondary message */
    }

    (*tst_ptr).tst_tlh_ptrps = tlh_ptr; /* Save pointer */

    /*
     *  Initialize TLH structure.
     */
    (void) memcpy ((void *) (*tlh_ptr).tlh_ftypc4, (void *) "TLO ", 4);
    (*tlh_ptr).tlh_fvrnl = TLH_SCP_FVRN;
    (*tlh_ptr).tlh_subverl = TLH_TST_SVRN;
    (*tlh_ptr).tlh_testtypeb = TLH_TESTTYPE_TERMINALS;

    /*
     *  Set the default protection mask - NO ACCESS!
     */
    (*tlh_ptr).tlh_sprotl = 0xFFFFFFFF;

    PSL_GET_DATTIM (&(*tlh_ptr).tlh_ctimq [0], &(*tlh_ptr).tlh_ctimq [1]);

    /*
     *  Set up fields to indicate which platform and OS the test
     *  is being compiled under.
     *  Flags will be set in the TLH header to indicate if there is
     *  platform or OS specific code in each test.
     *
     *  These flags and fields will be checked when test is validated in LGR.
     *  If there is platform or OS specific code in the test the fields
     *  indicating the platform the test was compiled on will be checked
     *  to ensure compatibility.
     *  Errors will be reported if there are any mis-matches.
     */

    /*
     *  Set up field indicating OPERATING SYSTEM the test is being
     *  compiled under
     */

#ifdef UNIX
    (*tlh_ptr).tlh_compile_opsysl = TLH_OS_UNIX;
#elif WIN32
(*tlh_ptr).tlh_compile_opsysl = TLH_OS_WIN32;
#else

#error operating system not supported at line __LINE__

#endif

    /*
     *  Set up field indicating PLATFORM the test is being
     *  compiled under
     */

#ifdef VAX
    (*tlh_ptr).tlh_compile_platl = TLH_PLAT_VAX;
#elif ALPHA
    (*tlh_ptr).tlh_compile_platl = TLH_PLAT_ALPHA;
#elif HP
    (*tlh_ptr).tlh_compile_platl = TLH_PLAT_HP;
#elif IBM
    (*tlh_ptr).tlh_compile_platl = TLH_PLAT_IBM;
#elif NCR
    (*tlh_ptr).tlh_compile_platl = TLH_PLAT_NCR;
#elif NTI
    (*tlh_ptr).tlh_compile_platl = TLH_PLAT_NTI;
#elif OSF
    (*tlh_ptr).tlh_compile_platl = TLH_PLAT_OSF;
#elif SUN
    (*tlh_ptr).tlh_compile_platl = TLH_PLAT_SUN;
#else
#error platform not supported at line __LINE__
#endif

    /*
     *  Initialise Booleans that indicate if there has been any platform
     *  specific code.
     *  Initialised to FALSE.
     */
    (*tlh_ptr).tlh_plat_spec_coden = PFALSE;
    (*tlh_ptr).tlh_opsys_spec_coden = PFALSE;

    /*
     *  Allocate space for PRM structure.
     */
    prm_ptr = (struct prm *) calloc (1, sizeof (struct prm));

    if (prm_ptr == NULL)         /* If an error */
    {
        iret = PSL_INSVIRMEM;       /* Set up error */
        PSL_SIGDIAG (           /* Signal Diagnostic */
            (*com_ptr).diag_ptrps,  /* Use this diagnostic stack */
            2,          /* Number of message segments */
            PSL_SYSERVERR,      /* First message */
            "CALLOC for PRM Structure",
            /* Parameter */
            iret);          /* Secondary message */
    }

    (*tst_ptr).tst_prm_ptrps = prm_ptr; /* Save pointer */

    /*
     *  Initialize PRM structure - ensure that all fields marked as 'defaulted' and
     *  have 'real' values.
     */
    SCL_DEFAULT_PRM ((*com_ptr).diag_ptrps, prm_ptr);
    (void) SCL_UNDEFAULT_PRM (
        (*com_ptr).diag_ptrps, NULL, prm_ptr, NULL, prm_ptr);
    SCL_FREE_PRM (prm_ptr);

    /*
     *  Set the default protection masks - NO ACCESS!
     */
    (*prm_ptr).prm_scp_protl = 0xFFFFFFFF;
    (*prm_ptr).prm_tst_protl = 0xFFFFFFFF;

    /*
     *  Clear specified options flags and option line numbers.
     */
    for (jsub = 0; jsub < SCL_OPT_MAX; jsub++)
    {
        (*tpa_ptr).opt_flagn64 [jsub] = PFALSE;
        (*tpa_ptr).opt_linel64 [jsub] = 0;
    }

    return PSL_NORMAL;          /* Always return success */
}


/************************************************************************/

/*    INI_TSTTHR - Initialize test thread definition work structure */
/*    ==========                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine initializes the test thread definition work      */
/*    structure.                            */
/*                                  */
/************************************************************************/
unsigned long INI_TSTTHR (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       jsub;       /* Loop control */
    unsigned long       iret;       /* Routine return status */
    struct scl_com      *com_ptr;       /* Pointer to common data */
    struct scl_tpa      *tpa_ptr;       /* Pointer to TPARSE data */
    struct tstthrtbl    *tstthr_ptr;    /* Pointer to test thread definition */
    struct tthr     *tthr_ptr;      /* Pointer to test thread data */
    struct tdf      *tdf_ptr;       /* Pointer to thread data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */
    tstthr_ptr = (*tpa_ptr).def_tstthr_wrkps;
    /* Test Thread definition work data */
    /*
     *  If the test thread definition work structure has any structures 'hanging off'
     *  it there must have been a compilation error while parsing a test thread
     *  definition. If so, free the space allocated for the structures.
     */
    if ((*tstthr_ptr).tstthr_tthr_ptrps != NULL)
    {
        free ((void *) (*tstthr_ptr).tstthr_tthr_ptrps);
    }

    if ((*tstthr_ptr).tstthr_tdf_ptrps != NULL)
    {
        free ((void *) (*tstthr_ptr).tstthr_tdf_ptrps);
    }

    /*
     *  Clear the work structure.
     */
    (void) memset ((void *) tstthr_ptr, '\0', sizeof (struct tstthrtbl));

    /*
     *  No test thread references yet.
     */
    (*tstthr_ptr).tstthr_exe_refn = PFALSE;

    /*
     *  Assume test thread defined externally.
     */
    (*tstthr_ptr).tstthr_defn = PTRUE;

    /*
     *  Allocate space for test thread data structure.
     */
    tthr_ptr = (struct tthr *) calloc (1, sizeof (struct tthr));

    if (tthr_ptr == NULL)        /* If an error */
    {
        iret = PSL_INSVIRMEM;       /* Set up error */
        PSL_SIGDIAG (           /* Signal Diagnostic */
            (*com_ptr).diag_ptrps,  /* Use this diagnostic stack */
            2,          /* Number of message segments */
            PSL_SYSERVERR,      /* First message */
            "CALLOC for TTHR Structure",
            /* Parameter */
            iret);          /* Secondary message */
    }

    (*tstthr_ptr).tstthr_tthr_ptrps = tthr_ptr;
    /* Save pointer */
    /*
     *  Allocate space for TDF structure.
     */
    tdf_ptr = (struct tdf *) calloc (1, sizeof (struct tdf));

    if (tdf_ptr == NULL)        /* If an error */
    {
        iret = PSL_INSVIRMEM;       /* Set up error */
        PSL_SIGDIAG (           /* Signal Diagnostic */
            (*com_ptr).diag_ptrps,  /* Use this diagnostic stack */
            2,              /* Number of message segments */
            PSL_SYSERVERR,      /* First message */
            "CALLOC for TDF Structure (in INI_TSTTHR)",
            /* Parameter */
            iret);          /* Secondary message */
    }

    (*tstthr_ptr).tstthr_tdf_ptrps = tdf_ptr;
    /* Save pointer */
    /*
     *  Initialize TDF structure - ensure that all fields are 'defaulted'.
     */
    SCL_DEFAULT_TDF ((*com_ptr).diag_ptrps, tdf_ptr);
    (void) SCL_UNDEFAULT_TDF
    ((*com_ptr).diag_ptrps, NULL, NULL, tdf_ptr, NULL, tdf_ptr);
    SCL_FREE_TDF (tdf_ptr);

    /*
     *  Clear specified options flags and option line numbers.
     */
    for (jsub = 0; jsub < SCL_OPT_MAX; jsub++)
    {
        (*tpa_ptr).opt_flagn64 [jsub] = PFALSE;
        (*tpa_ptr).opt_linel64 [jsub] = 0;
    }

    return PSL_NORMAL;          /* Always return success */
}


/************************************************************************/

/*    INI_UPDATE - Init. routine for the UPDATE command         */
/*    ==========                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  ONERROR_LABEL = Cleared to show No On error     */
/*      cmnd_optl -  Cleared to default options                 */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a UPDATE command.                      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_UPDATE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */


    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */

    (*tpa_ptr).cmnd_optl  = 0;          /* Clear options mask           */


    /*
     *    Clear the ONERROR flags to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ONERROR Label buffer   */

    return PSL_NORMAL;
}
/*
 * 
 */

/************************************************************************/
/*                                  */
/*    INI_VAL_LIST - Init. routine for value list parsing       */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   OBJ_VALUES = Value list table               */
/*          OBJ_VALUES_PTR = Value list pointer         */
/*    Outputs:  DEF_VALUES_OFFL = Value list offset set         */
/*                                  */
/*    This routine initialises a value list. It retrieves the offset    */
/*    of the next FREE byte in the value list buffer and saves it in    */
/*    the variable DEF_VALUES_OFFL.                 */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_VAL_LIST (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Set the DEF value list offset to be the current offset in the value list
     *    buffer.
     */
    (*tpa_ptr).def_values_offl = (unsigned long) (*tpa_ptr).values_nxtpc -
                                 (unsigned long) (*tpa_ptr).values_begpc;
    /* Calc current offset in buffer */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_VPSTATUS - Initialization routine for VP$STATUS variable  */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                            */
/*    Outputs:  VP$STATUS Global variable added to table.           */
/*                                      */
/*    This routine adds the internal global variable VP$STATUS to the   */
/*    global variables table.                       */
/*                                  */
/*    We set the variable to be a CHARACTER:255 variable as this is all */
/*    that can be set as a logical.                 */
/*                                      */
/*----------------------------------------------------------------------*/

void        INI_VPSTATUS (struct scl_com *com_ptr)

{
    /*
     *    Local data
     */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Now create a new global variable entry for VP$STATUS
     */
    INI_DEF (com_ptr);          /* Clear all work variables */
    (void) strcpy ((*tpa_ptr).def_namec64, "VP$STATUS");
    /* Set variable name = VP$STATUS */
    (*tpa_ptr).def_nlengthl = strlen ((*tpa_ptr).def_namec64);
    /* set length of variable name  */
    (*tpa_ptr).def_addrl = ADDR_GLOBAL; /* Set to be global variable    */
    (*tpa_ptr).def_dtypl = DTYP_CHARACTER;
    /* Set data type = character    */
    (*tpa_ptr).def_typl = DEF_VARIABLE; /* ...type = variable       */
    (*tpa_ptr).def_sizel = 255;     /* Data size = 255 characters   */
    return;
}

/************************************************************************/
/*                                  */
/*    INI_VP_OUTPUT - Init. routine for VP$OUTPUT parsing       */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  DEC_INTEGERL = 512                  */
/*                                  */
/*    This routine initializes the variables used to parse a VP$OUTPUT  */
/*    referenmce.                           */
/*                                      */
/*----------------------------------------------------------------------*/

unsigned long   INI_VP_OUTPUT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    ...all that is required is to set the VP$OUTPUT modifier to its default
     *    size of 512 bytes.
     */
    (*tpa_ptr).dec_integerl = 512;  /* Modifier = 512 bytes     */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_VP_SCREEN - Init. routine for VP$SCREEN parsing       */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  VP$SCREEN operand descriptors initialized       */
/*                                  */
/*    This routine initializes the variables used to parse a VP$SCREEN  */
/*    referenmce.                           */
/*                                      */
/*----------------------------------------------------------------------*/

unsigned long   INI_VP_SCREEN (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Check to see if the script is a screen mode script and report an error
     *    if it is not.
     */
    if (((*tpa_ptr).mode_flagl & MODE_SCREEN) == 0)
    {
        REPORT_ERROR (argblk_ptr, SCL_NOTSCREENMODE);
        /* Report not screen mode   */
    }

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_WAIT - Init. routine for the WAIT commands            */
/*    =========                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      ONTIMEOUT_LABEL = Cleared to show No on Timeout     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*      PIPE_FLAGSL = Cleared to show no pipe flags yet     */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a WAIT command.                                            */
/*                                  */
/*    Operands/variables specific to a an individual WAIT command are   */
/*    initialized in that commands initialization routine       */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_WAIT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Clear the ON TIMEOUT label to show not specified.
     */
    (void) memset ((*tpa_ptr).ontimeout_labelc20 ,'\0'
                   ,sizeof ((*tpa_ptr).ontimeout_labelc20));
    /* Clear ON TIMEOUT Label buffer */

    /*
     *
     *
     *    Initialize the Pipe Flags
     */
    (*tpa_ptr).pipe_flagsl = 0;     /* Clear pipe flags */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    /*
     *
     *
     *    Clear the command options specified mask.
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */
    /*
     *
     *
     *    Initialize the unknown string flag
     */
    (*tpa_ptr).unknown_strl = 0;    /* Clear unknown string flag */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    INI_WAIT_STRING - Init. routine for the WAIT for STRING command   */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  (*tpa_ptr).dec_integerl - Set to 512 (VP$OUTPUT mod)    */
/*      CMND_OPTL - initialise option mask                */
/*    This routine initialises the variablles required during parsing   */
/*    of a WAIT command.                                            */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_WAIT_STRING (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    (*tpa_ptr).dec_integerl = 512;  /* Set modifier size = 512  */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */

    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */


    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    INI_WAIT_TST - Init. routine for the WAIT FOR TEST command    */
/*    ============                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  Default TOF line buffer initialised.            */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a WAIT FOR TEST command. This currently consists only of   */
/*    initialising the default TOF line buffer with the Wtest record's  */
/*    waitfor sub-record header.                                */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long INI_WAIT_TST (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* SCL common data */
    struct scl_tpa  *tpa_ptr;       /* SCL Tparse data */
    char   *lin_ptr;       /* Pointer to line buffer */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;
    lin_ptr = com_ptr->toflinc1024; /* 'tof' line buffer pointer */


    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    INI_WHEN  Init. Routine for the WHEN command          */
/*    ========                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  None                            */
/*                                  */
/*    This routine initialises the data fields used in parsing a WHEN   */
/*    command. It initialises the default 'TOF' line buffer with the    */
/*    "expr " prefix in anticipation of appending the complex       */
/*    conditional expression to this buffer.                    */
/*                                  */
/************************************************************************/
unsigned long INI_WHEN (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* SCL Common data */
    char   *lin_ptr;       /* Pointer to line buffer */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    lin_ptr = com_ptr->toflinc1024; /* 'tof' line buffer pointer*/

    (void) strcpy (lin_ptr, "expr ");

    return PSL_NORMAL;          /* ...and return success */
}



/************************************************************************/
/*                                  */
/*    INI_WRITE - Initialization routine for the WRITE command      */
/*    =========                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a WRITE PIPE command and calls INI_DATAFILE to initialize  */
/*    those needed by a WRITE FILE command.                 */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_WRITE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the Pipe Id and command options
     */
    (*tpa_ptr).open_fididxl = 0;    /* Clear Pipe Id        */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    /*
     *    Clear the ON TIMEOUT label to show not specified.
     */
    (void) memset ((*tpa_ptr).ontimeout_labelc20 ,'\0'
                   ,sizeof ((*tpa_ptr).ontimeout_labelc20));
    /* Clear ON TIMEOUT Label buffer */
    /*
     *    Call INI_DATAFILE and return its status.
     */
    return INI_DATAFILE (argblk_ptr);   /* ...and return success    */
}




/************************************************************************/
/*                                  */
/*    INI_MAP - Initialization routine for the MAP command      */
/*    =======                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a MAP command and calls INI_DATAFILE to initialize those   */
/*    needed by a MAP FILE command.                     */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_MAP     (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the command options
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    /*
     *    Clear the ON ERROR label to show not specified.
     */
    (void) memset ((*tpa_ptr).onerror_labelc16 ,'\0'
                   ,sizeof ((*tpa_ptr).onerror_labelc16));
    /* Clear ON ERROR Label buffer  */

    /*
     *    Clear the ON TIMEOUT label to show not specified.
     */
    (void) memset ((*tpa_ptr).ontimeout_labelc20 ,'\0'
                   ,sizeof ((*tpa_ptr).ontimeout_labelc20));
    /* Clear ON TIMEOUT Label buffer */
    /*
     *    Call INI_DATAFILE and return its status.
     */
    return INI_DATAFILE (argblk_ptr);   /* ...and return success    */
}



/************************************************************************/
/*                                  */
/*    INI_MATCH - Initialization routine for the MATCH command      */
/*    =========                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*      (Parameter field contains initial command options). */
/*    Outputs:  ONERROR_LABEL = Cleared to show No on error     */
/*      CMND_OPT_SPECIFIED = Cleared to show no options yet.    */
/*      CMND_OPTL = Cleared to default options          */
/*                                  */
/*    This routine initialises the variables required during parsing    */
/*    of a MATCH command.                       */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INI_MATCH   (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Initialize the command options
     */
    (*tpa_ptr).cmnd_opt_specifiedl = 0; /* No input command options specified */
    (*tpa_ptr).cmnd_optl = 0;       /* Clear options bitmask.   */

    return PSL_NORMAL;
}

