/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*
 * scl_sav_routines.c:
 * ===================
 *
 * This module contains TPA action routines, and other routines, used to 
 * save data to various data structures during compilation.
 */

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 *  PSL Macros.
 */
#include <psl_macros.h>
#include <vtest_macros.h>
#include <psl_methods.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines */
#include <tofprmdef.h>          /* TOF parameters */
#include <trc_rtndef.h>         /* Tracing Routines */
#include "scl_rtndef.h"         /* SCL Routines */
#include "scl_def_tbl_rtndef.h"     /* SCL definitions routines */
#include "scl_env_tbl_rtndef.h"     /* SCL environment routines */
#include "scl_cod_tbl_rtndef.h"     /* SCL code  routines */
#include "scl_sub_tbl_rtndef.h"     /* SCL subexpression routines */

static unsigned long    GET_SAVOPD_COUNT    (struct scl_tpa *tpa_ptr);

/*
 *  Structure definitions.
 */
#include <shrmsg.h>         /* PSL Messages         */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block    */
#include <scpdef.h>         /* SCP Header record        */
#include "scl_limits.h"         /* SCL limits           */
#include <psl_trc_prmdef.h>     /* Tracing parameters       */
#include <psl_vm_prmdef.h>      /* Linked list Parameters   */
#include <pslerr_prmdef.h>      /* Error routine Parameters     */
#include "scl_tbldef.h"         /* SCL Table Definitions    */
#include <vtest_prmdef.h>       /* VTEST Parameters         */
#include "scl_comdef.h"         /* Common Data          */
#include "scl_tpacom.h"         /* Tparse common Data       */
#include "scl_tlodef.h"         /* Test structures      */

/************************************************************************/
/*                                  */

/*                                  */
/*    ALLOCATE_OPERAND - Allocate space for an operand.         */
/*    ================                              */
/*                                  */
/*    Type: Service routine                     */
/*    Status:   N/A - Errors signalled                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   tpa_ptr - pointer to TPARSE data            */
/*                                  */
/*    Outputs:                              */
/*      oph_pptr - Address of pointer to OPH allocated      */
/*      opd_pptr - Address of pointer to OPD allocated      */
/*                                  */
/*    This routine is called whenever we have parsed an operand to  */
/*    allocate space for the OPH and OPD structures required to define  */
/*    the operands and to save the addresses of these structures in the */
/*    linked list of parsed operands.                   */
/*                                  */
/************************************************************************/

void    ALLOCATE_OPERAND    (struct scl_tpa *tpa_ptr
                             ,struct oph     **oph_pptr
                             ,struct opd     **opd_pptr)

{
    /*
     *    Local data
     */
    unsigned long  iret;            /* Status return variable */

    /*
     *    If the ARRAY_LEVEL variable is non-zero then this means that we are
     *    within an array expression and are therefore NOT parsing the primary
     *    operand for a command.
     *
     *    If the ARRAY_LEVEL is zero then this must be a PRIMARY operand and
     *    therefore we should pick up the position in the LL of parsed operands
     *    of the entry prior to the new operand.
     *
     *    The count is incremented when we parse the '[' array index identifier
     *    and decremented at the end of each array parsed. It is zeroed for safety
     *    in the RESET_COMMAND_STATE routine.
     */
    if ((*tpa_ptr).array_levell == 0)   /* If a PRIMARY operand */
    {
        (*tpa_ptr).primary_operandpc = NULL;
        /* Assume no previous operand   */
        iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                            ,NULL ,NULL ,&(*tpa_ptr).primary_operandpc
                            ,PSL_VM_MODE_LAST);
        /* Get context of last item */
        if  ((!(psl_test_status (iret))) && (iret != PSL_RECNOTFOU))
        {               /* If we have an error      */
            PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"PSL_VM_COPY (in SAVE_VAR)", iret);
        }
    }

    /*
     *    Allocate space for the OPH and save the address of the allocated structure
     *    into the LL of parsed operands signalling any errors encountered.
     */
    *oph_pptr = (struct oph *) CALLOC (sizeof (struct oph) ,1);
    if (*oph_pptr == NULL)
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"MALLOC (allocating space for an OPH)", 0);
    }

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER5))
    {
        PSL_OUTPUT_TRACE ("MALLOC: oph_ptr - '%d/%p' - ALLOCATE_OPERAND\n",
                          sizeof (struct oph),
                          *oph_pptr);
    }
#endif  /* PSL_TRACING */


    iret = PSL_VM_INSERT ((*tpa_ptr).parsed_operand_llps
                          ,(char *) oph_pptr ,NULL, NULL, PSL_VM_POS_END);
    /* ...insert into LL of parsed  */
    if  (!(psl_test_status (iret)))     /* ...report any errors     */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_INSERT (in ALLOCATE_OPERAND)", iret);
    }

    /*
     *    ...and then allocate space for the OPD
     */
    *opd_pptr = (struct opd *) CALLOC (sizeof (struct opd) ,1);
    if (*opd_pptr == NULL)
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"MALLOC (allocating space for an OPD)", iret);
    }

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER5))
    {
        PSL_OUTPUT_TRACE ("MALLOC: opd_ptr - '%d/%p' - ALLOCATE_OPERAND\n",
                          sizeof (struct opd),
                          *opd_pptr);
    }
#endif  /* PSL_TRACING */


    iret = PSL_VM_INSERT ((*tpa_ptr).parsed_operand_llps
                          ,(char *) opd_pptr ,NULL, NULL, PSL_VM_POS_END);
    /* ...insert into LL of parsed  */
    if  (!(psl_test_status (iret)))     /* ...report any errors     */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_INSERT (in ALLOCATE_OPERAND)", iret);
    }

    /*
     *    ...and then simply return
     */
    return;
}

/************************************************************************/
/*                                  */
/*    SAVE_HDR_CNT - Save HTTP Request HEADER value count       */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   tpa_ptr->char_list_cnt                  */
/*    Outputs:  tpa_ptr->http_hdr_cnt = tpa_ptr->char_list_cnt      */
/*                                  */
/*    This routine saves the count of values in the HTTP request    */
/*    HEADER clause that has just been parsed.              */
/*                                  */
/************************************************************************/

unsigned long   SAVE_HDR_CNT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *    Save the value count.
     */
    tpa_ptr->http_hdrval_cntl = tpa_ptr->char_list_val_cntl;

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                  */
/*    SAVE_NULL_OPD - Save a NULL operand into the linked list      */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  Null Operand added to LL of parsed operands     */
/*                                  */
/*    This routine creates a NULL operand and adds it to the linked     */
/*    of parsed operands. This is required in instances when a value    */
/*    has been omitted from a command to act as a placeholder.      */
/*                                  */
/*    For example, in the EXIT command, the exit status variable is     */
/*    optional but we must add a NULL operand to the code so that LGR   */
/*    knows that the status was omitted, otherwise it would simply  */
/*    read the next operand and use that instead.           */
/*                                  */
/************************************************************************/

unsigned long   SAVE_NULL_OPD       (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */
    char     *prmopd_ptr;        /* Primary operand pointer  */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a NULL operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     * 
     *
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list. This operand should not
     *    change the primary operand - which may be changed by ALLOCATE_OPERAND -
     *    to ensure this we save its value before allocating an operand and
     *    restore it afterwards. This is very untidy and the mechanism for handling
     *    'primary' operands could do with re-engineering.
     */
    prmopd_ptr = (*tpa_ptr).primary_operandpc;

    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);
    /* ...allocate a new operand */
    (*tpa_ptr).primary_operandpc = prmopd_ptr;

    (*oph_ptr).addrl = ADDR_NONE;   /* Clear addressing mode */
    (*oph_ptr).dtypl = 0;       /* Clear data type */
    (*oph_ptr).sizel = 0;       /* Clear data size */
    (*oph_ptr).optl = 0;        /* Clear operand options */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    /*
     *    ...and then return. This operand will be encoded later
     */
    return PSL_NORMAL;
}

/************************************************************************/

/*    SAVE_NUMERIC_OPD - Save a Numeric operand into the linked list    */
/*    ================                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*      (*argblk_ptr).paraml = Value for operand        */
/*    Outputs:  Numeric Operand added to LL of parsed operands      */
/*                                  */
/*    This routine creates a numeric operand with a value as passed to  */
/*    the routine and adds it to the linked list of parsed operands.    */
/*                                  */
/************************************************************************/

unsigned long   SAVE_NUMERIC_OPD    (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a Numeric operand of value %ld to the LL\n"
                          ,(*argblk_ptr).paraml);
    }
#endif

    /*
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);
    (*oph_ptr).addrl = ADDR_IMMEDIATE;  /* Immediate addressing mode.   */
    (*oph_ptr).dtypl = DTYP_INTEGER;    /* Integer data type        */
    (*oph_ptr).sizel = sizeof (long);   /* ...which is a long       */
    (*oph_ptr).optl = 0;        /* Clear operand options    */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    (*opd_ptr).uopd.imi.vall = (*argblk_ptr).paraml;
    /* ...of the specified value    */

    /*
     *    ...and then return. This operand will be encoded later
     */
    return PSL_NORMAL;
}
/*
 * 
 */

/************************************************************************/

/*    SAVE_ZERO_OPD - Save a ZERO integer operand into the linked list  */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  Zero integer Operand added to LL of parsed operands */
/*                                  */
/*    This routine creates a zero operand and adds it to the linked     */
/*    of parsed operands.                       */
/*                                  */
/************************************************************************/

unsigned long   SAVE_ZERO_OPD       (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a ZERO operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);
    (*oph_ptr).addrl = ADDR_IMMEDIATE;  /* Immediate addressing mode.   */
    (*oph_ptr).dtypl = DTYP_INTEGER;    /* Integer data type        */
    (*oph_ptr).sizel = sizeof (long);   /* ...which is a long       */
    (*oph_ptr).optl = 0;        /* Clear operand options    */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    (*opd_ptr).uopd.imi.vall = 0;   /* ...of value 0.       */

    /*
     *    ...and then return. This operand will be encoded later
     */
    return PSL_NORMAL;
}
/*
 * 
 */

/************************************************************************/
/*                                  */
/*    SAVE_CHR - Setup for a character variable.                */
/*    ========                                  */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*      CHR_STRING = Character string               */
/*      CHR_STRING_LENGTH = Character string length     */
/*    Outputs:  Immediate character operand created and added to the    */
/*      LL of parsed operands.                          */
/*                                  */
/*    This routine creates an operand descriptor for an immediate   */
/*    character and adds it to the LL o f parsed operands.              */
/*                                  */
/*    The character operand string and its length are held in       */
/*    CHR_STRING and CHR_STRING_LENGTH respectively on entry.       */
/*                                  */
/************************************************************************/

unsigned long   SAVE_CHR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a CHARACTER LITERAL operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     * 
     *
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    /*
     *    Check the length of the character string parsed and report if it
     *    exceeds the maximum permitted length.
     */
    if ((*tpa_ptr).chr_strlenl > MAX_LITERAL_SIZE)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXSTRLEN);
        /* Pass to underline routine    */
    }

    else                /* if value O.k.        */
    {
        (*oph_ptr).addrl = ADDR_IMMEDIATE;
        /* Set operand = Immediate  */
        (*oph_ptr).dtypl = DTYP_CHARACTER;
        /* Set character data type  */
        (*oph_ptr).sizel = (*tpa_ptr).chr_strlenl;
        /* ...and character size    */
        (*oph_ptr).optl = 0;        /* Clear operand options    */
        (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
        /* Set operand id. */

        /* Allocate memory for the immediate character value */
        (*opd_ptr).uopd.imc.val_pc = (char *) MALLOC ((*tpa_ptr).chr_strlenl);
        if ((*opd_ptr).uopd.imc.val_pc == NULL)
        {
            PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"MALLOC (allocating space for immediate character operand)", 0);
        }

        (void) memcpy ((*opd_ptr).uopd.imc.val_pc,
                       (*tpa_ptr).chr_stringc65536,
                       (*tpa_ptr).chr_strlenl);
        /* Copy in character value  */
    }

    return PSL_NORMAL;          /* return success       */
}
/*
 * 
 */

/*======================================================================*/
/*                                  */
/*    SAVE_CHR_CON - Save Character Constant                */
/*    ============                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*          CHR_STRING_LENGTH = Character string length     */
/*    Outputs:  DEF_* set up from information returned          */
/*                                  */
/*   This routine will set the DEF_* values for a character constant    */
/*   variable within SCL.                       */
/*                                  */
/*======================================================================*/

unsigned long   SAVE_CHR_CON (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    if ((*tpa_ptr).chr_strlenl > MAX_LITERAL_SIZE)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXSTRLEN);
        /* Pass to underline routine    */
    }

    else                /* if value O.k.        */
    {
        (*tpa_ptr).def_addrl = ADDR_CONSTANT;
        /* Set operand = Constant   */
        (*tpa_ptr).def_dtypl = DTYP_CONCHARACTER;
        /* Set constant character data type */
        (*tpa_ptr).def_sizel = (*tpa_ptr).chr_strlenl;
        /* ...and character size    */
    }

    return PSL_NORMAL;          /* return success       */
}

/*======================================================================*/

/*    RESTORE_PRIMARY_OPERAND - Restore the Saved primary operand ptr   */
/*    =======================                       */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                              */
/*    Outputs:  Primary operand pointer restored to its saved value */
/*      LL of parsed operands.                          */
/*                                  */
/*    This routine is called to restore the primary operand pointer */
/*    to a previously saved value.                                      */
/*                                  */
/*    Its only use currently is at the end of an EXTRACT clause where   */
/*    the primary operand should be restored to be the EXTRACT operand  */
/*    and not be that of any operands parsed for the EXTRACT function.  */
/*                                  */
/*======================================================================*/

unsigned long   RESTORE_PRIMARY_OPERAND (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nRestoring Saved Primary Operand Pointer\n");
    }
#endif

    /*
     *    Restore the primary operand poinetr to its saved value.
     */
    (*tpa_ptr).primary_operandpc = (*tpa_ptr).save_primary_operandpc;

    return PSL_NORMAL;          /* Return Success Status    */
}
/*
 * 
 */

/************************************************************************/
/*                                  */
/*    SAVE_FILE - Save a FILE Operand                   */
/*    ==========                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  FILE operand created and added to the LL of parsed  */
/*              operands.                               */
/*                                  */
/*    This routine creates an operand descriptor for a FILE reference   */
/*    and adds it to the LL of parsed operands.                     */
/*                                  */
/*    This rtoutine is ONLY called as part of the processing on a   */
/*    parameter list for a CALL SCRIPT when a FILE parameter has been   */
/*    detected.                             */
/*                                  */
/*    If the file is identified as having been passed to this script    */
/*    we must remember that the file index which we add to the operand  */
/*    header must have the base number for shared files added to it */
/*    to identify it as a passed file as opposed to a local file    */
/*                                  */
/************************************************************************/

unsigned long   SAVE_FILE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct fattbl   *fat_ptr;       /* File Access Table pointer    */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a FILE operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     * 
     *
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    (*oph_ptr).addrl = ADDR_FILE;   /* Addressing mode = FILE   */
    (*oph_ptr).dtypl = DTYP_FILE;   /* Data Type = File     */
    (*oph_ptr).optl = 0;        /* Clear operand options  */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    /*
     *    Save the file index into the operand header in place of the data size.
     *    We MUST remember that the file id is incremented by the PASSED_FILE_BASE
     *    if this file has been passed to this script rather than lbeing local to it.
     */
    fat_ptr = (*tpa_ptr).fat_begps + (*tpa_ptr).open_fididxl - 1;
    /* Point to current entry in table */
    if ((*fat_ptr).fat_refflgl == PASSED_FILE)
    {                   /* ...else if a passed file */
        (*oph_ptr).sizel = (*tpa_ptr).open_fididxl + PASSED_FILE_BASE;
        /* ...set index of passed file  */
    }

    else
    {
        (*oph_ptr).sizel = (*tpa_ptr).open_fididxl;
        /* Save index as size       */
    }

    /*
     *  Save File operand data.
     */
    (*opd_ptr).uopd.fat.fattbl_ptrps = fat_ptr;
    /* ...work pointer to FATTBL */
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    SAVE_FNC_EXTRACT - Save EXTRACT Function Operand          */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                              */
/*    Outputs:  Extract function operand created and added to the   */
/*      LL of parsed operands.                          */
/*                                  */
/*    This routine creates an operand descriptor for an EXTRACT         */
/*    function and adds it to the LL of parsed operands.                */
/*                                  */
/*======================================================================*/

unsigned long   SAVE_FNC_EXTRACT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand Header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving an EXTRACT operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    /*
     *    Now set up this operand descriptor.
     */
    (*oph_ptr).addrl = ADDR_FUNCTION;   /* Function addressing mode */
    (*oph_ptr).dtypl = DTYP_CHARACTER;  /* Character data type      */
    (*oph_ptr).optl = 0;        /* Clear operand options  */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    /*
     *  Save Function operand data.
     */
    (*opd_ptr).uopd.fnc.idl = TOF_FNCID_EXTRACT;
    (*opd_ptr).uopd.fnc.prmcntl = 3;    /* EXTRACT has 3 parameters */

    /*
     *    We must be very careful here as the primary operand at this point is the
     *    EXTRACT operand itself. This will however be over-ridden as we parse the
     *    operands for the extract. We need the primary operand on completion of the
     *    EXTRACT clause to point to the EXTRACT operand itself NOT to any of its
     *    operands as clauses (such as CASE_BLIND) relate to the EXTRACT
     *    operand itself and not to any of its operands.
     *
     *    We therefore save the PRIMARY_OPERAND field (which is now set to the
     *    EXTRACT operand) and will restore it at the end of the EXTRACT clause in
     *    the END_EXTRACT routine (held in this module for convenience).
     */
    (*tpa_ptr).save_primary_operandpc = (*tpa_ptr).primary_operandpc;

    return PSL_NORMAL;          /* Return Success Status    */
}
/*
 * 
 */

/*======================================================================*/
/*                                  */
/*    SAVE_FNC_LOCATE - Save LOCATE Function Operand            */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                                  */
/*    Outputs:  LOCATE function operand created and added to the    */
/*      LL of parsed operands.                          */
/*                                  */
/*    This routine creates an operand descriptor for a LOCATE function  */
/*    and adds it to the LL of parsed operands.                     */
/*                                  */
/*======================================================================*/

unsigned long   SAVE_FNC_LOCATE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand Header pointer   */
    struct opd  *opd_ptr;       /* Operand Descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a LOCATE operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /* =
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    /*
     *    Now set up this operand descriptor.
     */
    (*oph_ptr).addrl = ADDR_FUNCTION;   /* Function addressing mode */
    (*oph_ptr).dtypl = DTYP_INTEGER;    /* Integer data type */
    (*oph_ptr).optl = 0;        /* Clear operand options */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    /*
     *  Save Function operand data.
     */
    (*opd_ptr).uopd.fnc.idl = TOF_FNCID_LOCATE;
    (*opd_ptr).uopd.fnc.prmcntl = 2;    /* LOCATE has 2 parameters */

    /*
     *    We must be very careful here as the primary operand at this point is the
     *    LOCATE operand itself. This will however be over-ridden as we parse the
     *    operands for the LOCATE. We need the primary operand on completion of the
     *    LOCATE clause to point to the LOCATE operand itself NOT to any of its
     *    operands as clauses (such as CASE_BLIND) relate to the LOCATE
     *    operand itself and not to any of its operands.
     *
     *    We therefore save the PRIMARY_OPERAND field (which is now set to the
     *    LOCATE operand) and will restore it at the end of the LOCATE clause in
     *    the END_LOCATE routine (held in this module for convenience).
     */
    (*tpa_ptr).save_primary_operandpc = (*tpa_ptr).primary_operandpc;

    return PSL_NORMAL;          /* Return Success Status */
}

/*======================================================================*/
/*                                  */
/*    SAVE_FNC_NEG_EXP_DIST - Save NEG_EXP_DIST Function Operand            */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                                  */
/*    Outputs:  NEG_EXP_DIST function operand created and added to the    */
/*      LL of parsed operands.                          */
/*                                  */
/*    This routine creates an operand descriptor for a NEG_EXP_DIST function  */
/*    and adds it to the LL of parsed operands.                     */
/*                                  */
/*======================================================================*/

unsigned long   SAVE_FNC_NEG_EXP_DIST (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand Header pointer   */
    struct opd  *opd_ptr;       /* Operand Descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a NEG_EXP_DIST operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /* =
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    /*
     *    Now set up this operand descriptor.
     */
    (*oph_ptr).addrl = ADDR_FUNCTION;   /* Function addressing mode */
    (*oph_ptr).dtypl = DTYP_INTEGER;    /* Integer data type */
    (*oph_ptr).optl = 0;        /* Clear operand options */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    /*
     *  Save Function operand data.
     */
    (*opd_ptr).uopd.fnc.idl = TOF_FNCID_NEG_EXP_DIST;
    (*opd_ptr).uopd.fnc.prmcntl = 2;    /* NEG_EXP_DIST has 2 parameters */

    /*
     *    We must be very careful here as the primary operand at this point is the
     *    NEG_EXP_DIST operand itself. This will however be over-ridden as we parse the
     *    operands for the NEG_EXP_DIST. We need the primary operand on completion of
     *    the NEG_EXP_DIST clause to point to the NEG_EXP_DIST operand itself.
     *
     *    We therefore save the PRIMARY_OPERAND field (which is now set to the
     *    NEG_EXP_DIST operand) and will restore it at the end of the LOCATE clause in
     *    the END_NEG_EXP_DIST routine (held in this module for convenience).
     */
    (*tpa_ptr).save_primary_operandpc = (*tpa_ptr).primary_operandpc;

    return PSL_NORMAL;          /* Return Success Status */
}

/************************************************************************/
/*                                  */
/*    SAVE_PIPE - Save a PIPE Operand                   */
/*    ==========                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  PIPE operand created and added to the LL of parsed  */
/*              operands.                               */
/*                                  */
/*    This routine creates an operand descriptor for a PIPE reference   */
/*    and adds it to the LL of parsed operands.                     */
/*                                  */
/*    This routine is ONLY called as part of the processing on a    */
/*    parameter list for a CALL SCRIPT when a PIPE parameter has been   */
/*    detected.                             */
/*                                  */
/*    If the pipe is identified as having been passed to this script    */
/*    we must remember that the pipe index which we add to the operand  */
/*    header must have the base number for shared pipes added to it */
/*    to identify it as a passed pipe as opposed to a local pipe    */
/*                                  */
/************************************************************************/

unsigned long   SAVE_PIPE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct fattbl   *fat_ptr;       /* Pipe Access Table pointer    */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a PIPE operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    (*oph_ptr).addrl = ADDR_PIPE;   /* Addressing mode = PIPE   */
    (*oph_ptr).dtypl = DTYP_PIPE;   /* Data Type = Pipe     */
    (*oph_ptr).optl = 0;        /* Clear operand options */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    /*
     *    Save the pipe index into the operand header in place of the data size.
     *    We MUST remember that the pipe id is incremented by the PASSED_PIPE_BASE
     *    if this pipe has been passed to this script rather than being local to it.
     */
    fat_ptr = (*tpa_ptr).fat_begps + (*tpa_ptr).open_fididxl - 1;
    /* Point to current entry in table */
    if (((*fat_ptr).fat_refflgl & PIPATT_PRM) != 0)
    {                   /* ...else if a passed pipe */
        (*oph_ptr).sizel = (*tpa_ptr).open_fididxl + PASSED_PIPE_BASE;
        /* ...set index of passed pipe  */
    }

    else
    {
        (*oph_ptr).sizel = (*tpa_ptr).open_fididxl;
        /* Save index as size       */
    }
    /*
     *  Save File operand data.
     */
    (*opd_ptr).uopd.fat.fattbl_ptrps = fat_ptr;

    return PSL_NORMAL;          /* ...and return success    */
}


/*======================================================================*/
/*                                  */
/*    SAVE_INT - Save Integer Operand                   */
/*    ========                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                               */
/*    Outputs:  Operand = Immediate integer operand added to LL of  */
/*                parsed operands.              */
/*                                  */
/*    This routine creates an operand descriptor for an immediate   */
/*    integer operand and adds it to the LL of parsed operands.     */
/*                                  */
/*    The integer operand value is held in (*tpa_ptr).int_valuel    */
/*                                  */
/*======================================================================*/

unsigned long   SAVE_INT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving an INTEGER LITERAL operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    /*
     *    Now set up this operand descriptor.
     */
    (*oph_ptr).addrl = ADDR_IMMEDIATE;  /* Immediate addressing mode    */
    (*oph_ptr).dtypl = DTYP_INTEGER;    /* Integer data type        */
    (*oph_ptr).sizel = sizeof (long);   /* Size = (long) bytes      */
    (*opd_ptr).uopd.imi.vall = (*tpa_ptr).int_valuel;
    /* Save integer value       */
    (*oph_ptr).optl = 0;        /* Clear operand options  */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    return PSL_NORMAL;          /* Return Success Status    */
}
/*
 * 
 */

/*======================================================================*/
/*                                  */
/*    SAVE_INT_CON - Save Integer Constant              */
/*    ============                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Operand descriptor address or  */
/*          parameter operand flag                  */
/*          PRM_CNT = Parameter count               */
/*    Outputs:  DEF_* set up from information returned          */
/*                                  */
/*   This routine will set the DEF_* values for an integer constant */
/*   variable within SCL.                       */
/*                                  */
/*======================================================================*/

unsigned long   SAVE_INT_CON (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Set up the DEF_* values then
     */
    (*tpa_ptr).def_addrl = ADDR_CONSTANT;
    /* Constant mode addressing */
    (*tpa_ptr).def_dtypl = DTYP_CONINTEGER;
    /* Constant Integer data type   */
    (*tpa_ptr).def_sizel = sizeof (long);
    /* Length for Integer (long) bytes */

    return PSL_NORMAL;          /* Return Success Status    */
}

/************************************************************************/
/*                                  */
/*    SAVE_INPUT_RATE - Save the parsed input rate.             */
/*    ================                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Flag indicating type of        */
/*          input rate specifier:               */
/*          INP_RATE_DEF                    */
/*          INP_RATE_MAX                    */
/*          INP_RATE_VAL                    */
/*      DEC_INTEGER = Integer part of number            */
/*      DEC_FRACTION = ASCII code of fractional digit       */
/*    Outputs:  Parsed operand added to LL of operands          */
/*          (*tpa_ptr).input_delayl = Literal value of input rate   */
/*                                  */
/*    This routine is called after parsing a NUMERIC input rate or  */
/*    the MAXIMUM or DEFAuLT keywords. Its purpose is to set up OPD1    */
/*    to reflect the input rate specified.              */
/*                                  */
/*    MAXIMUM is flagged by an INTEGER operand of value INP_RATE_MAX    */
/*    DEFAULT is flagged by an INTEGER operand of value INP_RATE_DEF    */
/*    VALUE is flagged as an INTEGER operand of value DEC_INTEGER*10    */
/*  + DEC_FRACTION.                         */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   SAVE_INPUT_RATE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph      *oph_ptr;       /* Operand header pointer   */
    struct opd      *opd_ptr;       /* Operand descriptor pointer   */
    char *char_ptr;      /* Work pointer to char     */
    char fraction [2];       /* Fractional digit     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving an INPUT RATE operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     * 
     *
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    (*oph_ptr).addrl = ADDR_IMMEDIATE;  /* Immediate addressing mode    */
    (*oph_ptr).dtypl = DTYP_INTEGER;    /* Integer data type        */
    (*oph_ptr).sizel = sizeof (long);   /* data size = (long) bytes */
    (*oph_ptr).optl = 0;        /* Clear operand options  */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    /*
     *    Check to see if the input rate is default/maximum/specified
     *    and store the appropriate input rate.
     */
    if ((*argblk_ptr).paraml == INP_RATE_DEF)
        /* If input rate defaulted  */
    {
        (*opd_ptr).uopd.imi.vall = DEF_INP_DELAY;
        /* Value = DEFAULT      */
    }

    else if ((*argblk_ptr).paraml == INP_RATE_MAX)
        /* Input rate = MAXIMUM     */
    {
        /*
         *  On some (SVR4 & OSF) platforms some shells cause us synchronisation
         *  problems at very high speeds. The symtom is that the shell outputs
         *  the entire input line at the same time as the output from the command
         *  and so vp$output contains too much output. This problem does not
         *  occur so long as we use some input timing between characters.
         */
#ifdef MIR_THROTTLE_REQUIRED
        (*opd_ptr).uopd.imi.vall = 500;
#else
        (*opd_ptr).uopd.imi.vall = MIN_INP_DELAY;
        /* Store input rate = MAXIMUM   */
#endif
    }

    else                /* Input rate = Specified   */
    {
        (*opd_ptr).uopd.imi.vall = ((*tpa_ptr).dec_integerl * 10);
        /* Save input rate specified    */
        if ((*tpa_ptr).dec_fractionl != 0)
        {               /* If fractional part specified */
            char_ptr = (char *) &(*tpa_ptr).dec_fractionl;
            /* Get work pointer to fraction */
            fraction [0] = *char_ptr;   /* Copy digit           */
            fraction [1] = '\0';    /* Null-terminate       */
            (void) sscanf (fraction,"%ld",&(*tpa_ptr).dec_fractionl);
            /* Convert as decimal       */
            (*opd_ptr).uopd.imi.vall += (*tpa_ptr).dec_fractionl;
            /* ...and add on fraction   */
        }

        /*
         *    Validate the input rate to ensure that it is within bounds.
         */
        if (((*opd_ptr).uopd.imi.vall > MAX_INPUT_RATE)
                ||  ((*opd_ptr).uopd.imi.vall <= 0))
            /* If specified rate too big    */
            /* ...or too small      */
        {
            REPORT_ERROR (argblk_ptr, VTEST_INPOUTRANGE);
            /* ...then report out of range  */
        }
    }

    /*
     *    Save the input rate value in an integer value as well as this is where we
     *    expect it to be if INPUT RATE is specified in the header as there are no
     *    variables defined and therefore the rate must have been a literal.
     */
    (*tpa_ptr).input_delayl = (*opd_ptr).uopd.imi.vall;
    /* Save value in local data as well */

    return PSL_NORMAL;
}

/*======================================================================*/

/*    SAVE_OPR - Save Operator Operand                  */
/*    ========                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                          */
/*      (*argblk_ptr).paraml = Operator to be saved     */
/*    Outputs:  Operand = Operator operand added to LL of parsed    */
/*                operands.                 */
/*                                  */
/*    This routine creates an operand descriptor for an operator    */
/*    operand and adds it to the LL of parsed operands.         */
/*                                  */
/*======================================================================*/
unsigned long SAVE_OPR (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving an OPERATOR operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Call a support routine to do the work.
     */
    ADD_OPR (com_ptr, (*argblk_ptr).paraml);

    return PSL_NORMAL;          /* Return Success Status */
}


/************************************************************************/
/*                                  */
/*    SAVE_OPTS - Save Options mask into the PRIMARY operand        */
/*    =========                                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Operand                */
/*          OPT_OPTL = Current options bitmask          */
/*    Outputs:  Operand descriptor set up with options mask     */
/*          OPT_OPTL = Cleared                  */
/*          OPT_OPT_SPECIFIED = Cleared to show no options yet. */
/*                                  */
/*    This routine is called when we have to move variable options  */
/*    from the options specified bitmask into an operand descriptor     */
/*    when the options are operand rather than command specific, for    */
/*    instance when parsing the SIMPLE_COND sub-expression the {NO}     */
/*    TRAILING SPACES clause is specific to each operand not to the     */
/*    command.                              */
/*                                  */
/*    We always set the options bitmask for the PRIMARY operand that    */
/*    is the base variable specified in the command rather than a   */
/*    variable specified as an index to it.             */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   SAVE_OPTS (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    unsigned long    iret;          /* Status return variable   */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   oph     *oph_ptr;      /* Operand header pointer   */
    char    *llb_ctx;      /* Linked list context      */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     * 
     *
     *    OPTIONS in the scope of this routine are VARIABLE options. These
     *    will always be set in the PRIMARY operand for this command.
     *    We will therefore pickup the operand header for the PRIMARY operand
     *    located by the saved posituion held in the TPA_COM structure which is the
     *    LL position of the preceding operand (if any).
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                        ,(char *) &oph_ptr
                        ,(*tpa_ptr).primary_operandpc
                        ,&llb_ctx ,PSL_VM_MODE_NEXT);
    if  (!(psl_test_status (iret)))     /* ...report any errors     */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY oph from list (in SAVE_OPTS)", iret);
    }

    /*
     * 
     *
     *    If the TRAILING SPACES option bit is set then set it in the operand
     *    descriptor.
     */
    if (((*tpa_ptr).opt_optl & VOPT_TRAILING) != 0)
    {                   /* If trailing spaces set   */
        (*oph_ptr).optl |= VOPT_TRAILING;
        /* ...set trailing spaces in OPD */
    }

    /*
     *    If the ONE TRAILING SPACE option bit is set then set it in the operand
     *    descriptor.
     */
    if (((*tpa_ptr).opt_optl & VOPT_ONE_TRAIL) != 0)
    {                   /* If trailing spaces set   */
        (*oph_ptr).optl |= VOPT_ONE_TRAIL;
        /* ...set trailing spaces in OPD */
    }

    /*
     *    If the CASE BLIND option bit is set then set it in the operand
     *    descriptor.
     */
    if (((*tpa_ptr).opt_optl & VOPT_CASEBLIND) != 0)
    {                   /* If case blind set        */
        (*oph_ptr).optl |= VOPT_CASEBLIND;
        /* ...set trailing spaces in OPD */
    }

    /*
     *    
     *    If the Atomic input option bit is set then set it in the operand
     *    descriptor.
     */
    if (((*tpa_ptr).opt_optl & VOPT_ATOMIC_INPUT) != 0)
    {                   /* If Atomic input set      */
        (*oph_ptr).optl |= VOPT_ATOMIC_INPUT;
    }

    /*
     * 
     */

    /*
     *    ...and copy the OPH back into the LL.
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                        ,(char *) &oph_ptr ,llb_ctx ,NULL ,PSL_VM_MODE_INTO);
    if  (!(psl_test_status (iret)))     /* ...report any errors     */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY oph back into list (in SAVE_OPTS)", iret);
    }

    (*tpa_ptr).opt_optl = 0;        /* Clear options bit mask.  */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    SAVE_SEED - Save the SEED operand.                */
/*    =========                                 */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   INT_VALUE = Integer value               */
/*    Outputs:  DEF_SEEDL set up with specified seed value      */
/*                                  */
/*    This routine save the specified seed value into the variable  */
/*    definition work variables.                    */
/*                                  */
/************************************************************************/

unsigned long   SAVE_SEED (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Save the parsed integer value as the seed
     */
    (*tpa_ptr).def_seedl = (*tpa_ptr).int_valuel;
    return PSL_NORMAL;          /* Return success       */
}

/************************************************************************/

/*    SAVE_SQL_DATABASE - Save the SQL Database type            */
/*    =================                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   TPARSE argumenr block, type in parameter        */
/*    Outputs:  sql_database_typel set up eith th supplied value    */
/*                                  */
/*    This routine save the specified SQL database type into the work   */
/*    structure                             */
/*                                  */
/************************************************************************/

unsigned long   SAVE_SQL_DATABASE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Save the parsed integer value as the seed
     */
    (*tpa_ptr).sql_database_typel = (*argblk_ptr).paraml;
    return PSL_NORMAL;          /* Return success       */
}
/*
 * 
 */

/************************************************************************/

/*    SAVE_SUBROUTINE - Save subroutine name                */
/*    ===============                           */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  (*tpa_ptr).subroutine_namec20 - Subroutine name.    */
/*                                  */
/*    This routine is called from a PSLTPA_DATANAME transition to save  */
/*    the current token as a subroutine name.               */
/*                                  */
/************************************************************************/
unsigned long SAVE_SUBROUTINE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Saving a Subroutine Name\n");
    }
#endif

    /*
     *    Save the subroutine name into global data for later reference.
     */
    (void) memcpy ((*tpa_ptr).subroutine_namec20,
                   (*argblk_ptr).token_ptrpc,
                   (*argblk_ptr).tokencntl);
    (*tpa_ptr).subroutine_namec20 [(*argblk_ptr).tokencntl] = '\0';

    return PSL_NORMAL;          /* ...and return success */
}


/************************************************************************/

/*    SAVE_SUBTST - Save sub-test name                  */
/*    ===========                           */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  (*tpa_ptr).subtst_namec20 - Sub-test name.      */
/*                                  */
/*    This routine is called from a PSLTPA_DATANAME transition to save  */
/*    the current token as a sub-test name.             */
/*                                  */
/************************************************************************/
unsigned long SAVE_SUBTST (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct   scl_com *com_ptr;      /* SCL common data */
    struct   scl_tpa *tpa_ptr;      /* SCL TPARSE data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Saving a Sub-test Name\n");
    }
#endif

    /*
     *  Save the sub-test name into global data for later reference.
     */
    (void) memcpy ((*tpa_ptr).subtst_namec20,
                   (*argblk_ptr).token_ptrpc,
                   (*argblk_ptr).tokencntl);
    (*tpa_ptr).subtst_namec20 [(*argblk_ptr).tokencntl] = '\0';

    return PSL_NORMAL;          /* ...and return success */
}


/************************************************************************/

/*    SAVE_SYMBOL_NAME - Save a symbol name (definition)        */
/*    ================                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr                      */
/*    Outputs:  Name of symbol saved in def_namec64 and def_nlengthl    */
/*                                  */
/*    This routine is called after parsing the name of a symbol being   */
/*    declared (timer/file/pipe/variable). Its purpose is to save the   */
/*    name of the symbol into the 'def_' fields.            */
/*                                  */
/************************************************************************/

unsigned long   SAVE_SYMBOL_NAME    (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Save the parsed symbol name into the 'def_' fields.
     */
    (*tpa_ptr).def_nlengthl = (*argblk_ptr).tokencntl;
    /* Save length of symbol    */
    (void) memcpy ((*tpa_ptr).def_namec64
                   ,(*argblk_ptr).token_ptrpc
                   ,(*argblk_ptr).tokencntl);
    /* ...and save the symbol value */
    return PSL_NORMAL;          /* Return success       */
}

/************************************************************************/

/*    SAVE_TEST - Save Test Operand                 */
/*    =========                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   (*tpa_ptr).tst_ptrps = Address of test entry in the */
/*          Test Symbol Table.              */
/*    Outputs:  Operand for test added to operands list.        */
/*                                  */
/*    This routine adds a TEST operand descriptor to the operands   */
/*    linked list for the specified test and sets the test reference    */
/*    flag.                             */
/*                                  */
/************************************************************************/
unsigned long SAVE_TEST (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */
    struct tsttbl   *tst_ptr;       /* Work test pointer */
    struct oph      *oph_ptr;       /* Work operand header */
    struct opd      *opd_ptr;       /* Work operand descriptor */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */
    tst_ptr = (*tpa_ptr).tst_ptrps; /* Get pointer to current test */

    /*
     *    If we have a current test, then add an operand descriptor for it and set the
     *    test reference flag.
     */
    if (tst_ptr != NULL)
    {
        /*
         *    Output tracing information if required.
         */
#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER1))
        {
            PSL_OUTPUT_TRACE ("\nSaving a TEST operand to the LL\n");
            PSL_OUTPUT_TRACE ("Operand Count = %ld\n",GET_SAVOPD_COUNT (tpa_ptr));
        }
#endif

        /*
         *    Allocate operand header and operand descriptor records for the TEST.
         */
        ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

        /*
         *    Set up the TEST operand header (OPH) and operand descriptor (OPD).
         */
        (*oph_ptr).addrl = ADDR_TST;    /* Show this is a test OPD */
        (*oph_ptr).optl = 0;        /* Clear operand options */
        (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
        /* Set operand id. */
        (*opd_ptr).uopd.tst.tsttbl_ptrps = tst_ptr;
        /* ...and work pointer to TSTTBL */
    }

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SAVE_TEST_THREAD - Save Test Thread Operand           */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   (*tpa_ptr).tstthr_ptrps = Address of test thread entry  */
/*          Test in the Test Thread Symbol Table.       */
/*    Outputs:  Operand for test thread added to operands list.     */
/*                                  */
/*    This routine adds a TEST THREAD operand descriptor to the     */
/*    operands linked list for the specified test thread and sets   */
/*    the test thread reference flag.                   */
/*                                  */
/************************************************************************/
unsigned long SAVE_TEST_THREAD (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */
    struct tstthrtbl
                *tstthr_ptr;        /* Work test thread pointer */
    struct oph      *oph_ptr;       /* Work operand header */
    struct opd      *opd_ptr;       /* Work operand descriptor */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Common data pointer */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */
    tstthr_ptr = (*tpa_ptr).tstthr_ptrps;
    /* Get pointer to current test thread */
    /*
     *    If we have a current test thread, then add an operand descriptor for it and
     *    set the test thread reference flag.
     */
    if (tstthr_ptr != NULL)
{
        /*
         *    Output tracing information if required.
         */
#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER1))
        {
            PSL_OUTPUT_TRACE ("\nSaving a TEST THREAD operand to the LL\n");
            PSL_OUTPUT_TRACE ("Operand Count = %ld\n",GET_SAVOPD_COUNT (tpa_ptr));
        }
#endif

        /*
         *    Allocate operand header and descriptor records for the TEST THREAD.
         */
        ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

        /*
         *    Set up the TEST THREAD operand header (OPH) and operand descriptor (OPD).
         */
        (*oph_ptr).addrl = ADDR_TSTTHR; /* Show this is a test thread OPD */
        (*oph_ptr).optl = 0;        /* Clear operand options */
        (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
        /* Set operand id. */
        (*opd_ptr).uopd.tstthr.tstthrtbl_ptrps = tstthr_ptr;
        /* ...and work pointer to TSTTHRTBL */
    }

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/
/*                                  */
/*    SAVE_TIMER - Save Timer Operand                   */
/*    ==========                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Operand descriptor address or  */
/*          parameter operand flag                  */
/*          PRM_CNT = Parameter count               */
/*          TIM_PTR = Address of timer entry in the Timers Symbol   */
/*            Table                     */
/*    Outputs:  Operand = Timer operand descriptor set up at        */
/*                    address indicated by PSLTPA$L_PARAM(AP)   */
/*                                  */
/*    This routine creates an operand descriptor for a timer operand    */
/*    at the address indicated by PSLTPA$L_PARAM(AP). The address of    */
/*    timer's entry in the Timers Symbol Table is held in TIM_PTR on    */
/*    entry.                                */
/*                                  */
/************************************************************************/
unsigned long   SAVE_TIMER (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct timtbl   *tim_ptr;       /* Work timer pointer       */
    struct oph      *oph_ptr;       /* Work operand header      */
    struct opd      *opd_ptr;       /* Work operand descriptor  */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    tim_ptr = (*tpa_ptr).tim_ptrps; /* Get pointer to current timer */
    /* definition, exiting if we do */
    if (tim_ptr != NULL)        /* do not have one.     */
    {
        /*
         *    Output tracing information if required.
         */
#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER1))
        {
            PSL_OUTPUT_TRACE ("\nSaving a TIMER operand to the LL\n");
            PSL_OUTPUT_TRACE ("Operand Count = %ld\n",GET_SAVOPD_COUNT (tpa_ptr));
        }
#endif

        /*
         * 
         *
         *    Allocate sufficient space to hold an operand header record and save its
         *    address into the parsed operands linked list.
         */
        ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

        /*
         *    Now set up the timer operand descriptor (OPD) and Operand header (OPH)
         */
        (*oph_ptr).addrl = ADDR_TIMER;  /* Show this is a timer OPD */
        (*oph_ptr).optl = 0;        /* Clear operand options  */
        (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
        /* Set operand id. */
        (*opd_ptr).uopd.tim.timtbl_ptrps = tim_ptr;
        /* ...and work pointer to TIMTBL */
    }

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/

/*    SAVE_VARIANT - Save a variant name                */
/*    =============                                 */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                              */
/*    Outputs:  (*tpa_com).variant_namec20 set up with name     */
/*                                  */
/*    This routine saves the dataname entered as the variant value  */
/*    into local data null terminating it as we go.         */
/*                                  */
/*    NOTE: We pick up the variant name from the source line retaining  */
/*      the case as specified by the user.              */
/*                                  */
/************************************************************************/

unsigned long   SAVE_VARIANT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    unsigned long offset;           /* Offset in line.      */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Pick up the offset of the current token as we will actually pick up the
     *    variant name from the source line as input rather than from the parsable
     *    line so that the case of the variant is retained.
     */
    offset = (long) ((*argblk_ptr).token_ptrpc - (*tpa_ptr).parsable_linec136);
    /* Current offset in line   */
    (void) memcpy ((*tpa_ptr).variant_namec20
                   ,&(*tpa_ptr).source_linec136 [offset]
                   ,(*argblk_ptr).tokencntl);
    /* Copy in variant name     */
    (*tpa_ptr).variant_namec20 [(*argblk_ptr).tokencntl] = '\0';
    /* ...and null terminate it.    */
    return PSL_NORMAL;
}

/************************************************************************/

/*    SAVE_VAR - Save a PRIMARY variable operand.           */
/*    ========                                  */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   VAR_PTR = Address of variable's entry in the        */
/*                Variables Symbol Table            */
/*    Outputs:  Operand = Variable operand descriptor set up        */
/*                and added to LL of parsed operands.           */
/*                                  */
/*    This routine creates an operand descriptor for a variable     */
/*    operand and saves the OPD and OPH address in the linked list of   */
/*    of operands.                          */
/*                                  */
/*    In this instance, the operand being saved MUST be a PRIMARY   */
/*    operand. It must not be an array index.               */
/*                                  */
/*    This routine will save the LL position of the primarey operand    */
/*    so that it can subsequently be found for example when we need */
/*    to add an option such as trailking spaces to the options bitmask. */
/*                                  */
/************************************************************************/

unsigned long   SAVE_VAR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct symtbl   *var_ptr;       /* Pointer to symbol table entry */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    var_ptr = (*tpa_ptr).var_ptrps; /* Variable in symbol table */

    /*
     *    ...then add the operand to the LL of parsed operands.
     */
    if (var_ptr != NULL)        /* If variable defined      */
    {
        CREATE_VAR_OPD (tpa_ptr, var_ptr);
        /* Create variable operand desc */
    }

    return PSL_NORMAL;          /* Return success       */
}
/*
 * 
 */

/************************************************************************/
/*                                  */
/*    SAVE_CONSTANT - Save a CONSTANT operand.              */
/*    =============                                 */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   VAR_PTR = Address of CONSTANTS entry in the     */
/*                Variables Symbol Table            */
/*    Outputs:  Immediate Integer or Immediate Character operand added  */
/*          to the LL of parsed operands as required.               */
/*      int_value - Set up if an integer constant       */
/*      chr_strlenl/chr_stringc65536 set up if character    */
/*          constant.                       */
/*                                  */
/*    This routine adds an operand descriptor for an Immediate Integer  */
/*    or character to the LL of parsed operands after parsing a         */
/*    constant.                             */
/*                                  */
/*    We may need the integer and character fields setting up as well   */
/*    so we will do this here also.                 */
/*                                  */
/************************************************************************/

unsigned long   SAVE_CONSTANT    (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct symtbl   *var_ptr;       /* Pointer to symbol table entry */
    struct oph  *oph_ptr;       /* Pointer to OPH       */
    struct opd  *opd_ptr;       /* Pointer to OPD       */
    char *const_ptr;     /* Pointer to constant value    */
    long *int_ptr;       /* Temporary pointer to long    */
    void *void_ptr;      /* WORK POINTER TO VOID     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    var_ptr = (*tpa_ptr).var_ptrps; /* Variable in symbol table */

    if (var_ptr != NULL)        /* If variable defined      */
    {

        /*
         *    Output tracing information if required.
         */
#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER1))
        {
            PSL_OUTPUT_TRACE ("\nSaving a CONSTANT operand to the LL\n");
            PSL_OUTPUT_TRACE ("Operand Count = %ld\n",GET_SAVOPD_COUNT (tpa_ptr));
        }
#endif

        /*
         *    Allocate sufficient space to hold an operand header record and save its
         *    address into the parsed operands linked list.
         */
        ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

        /*
         *    ...and set this up to be either an immediate character or integer value.
         */
        const_ptr = (*tpa_ptr).constant_begpc + (*var_ptr).var_offl;
        /* Get the offset of the data   */
        /* ...in the constants data area*/
        (*oph_ptr).addrl = ADDR_IMMEDIATE;
        /* Set operand = Immediate  */
        (*oph_ptr).sizel = (*var_ptr).var_sizel;
        /* ...and variable size     */
        (*oph_ptr).optl = 0;        /* Clear operand options */
        (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
        /* Set operand id. */

        /*
         *    ...and the data type specific fields.
         */
        if ((*var_ptr).var_typel == DTYP_CONCHARACTER)
        {               /* If character constant    */
            (*oph_ptr).dtypl = DTYP_CHARACTER;
            /* Set character data type  */

            /* Allocate memory for the immediate character value */
            (*opd_ptr).uopd.imc.val_pc = (char *) MALLOC ((*oph_ptr).sizel);
            if ((*opd_ptr).uopd.imc.val_pc == NULL)
            {
                PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                             ,"MALLOC (allocating space for character constant operand)", 0);
            }

            (void) memcpy ((*opd_ptr).uopd.imc.val_pc
                           ,const_ptr ,(*oph_ptr).sizel);
            /* Copy in character value  */
            (void) memcpy ((*tpa_ptr).chr_stringc65536
                           ,const_ptr ,(*oph_ptr).sizel);
            /* ...and into work variables   */
            (*tpa_ptr).chr_strlenl = (*oph_ptr).sizel;
            /* ...and set up size as well   */
        }
        else
        {
            (*oph_ptr).dtypl = DTYP_INTEGER;
            /* Set INTEGER data type    */
            void_ptr = (void *) const_ptr;
            /* Convert to void * first ...  */
            int_ptr = (long *) void_ptr;/* ...then to pointer to long   */
            (*opd_ptr).uopd.imi.vall = *int_ptr;
            /* Get LONG value from table    */
            (*tpa_ptr).int_valuel = *int_ptr;
            /* ...save in work as well  */
            (*argblk_ptr).numberl = *int_ptr;
            /* SAVE IN ARG BLOCK ALSO */
        }
    }

    return PSL_NORMAL;          /* Return success       */
}


/************************************************************************/
/*                                  */
/*    SAVE_TESTCASE - Store that TEST-CASE is active            */
/*    =============                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = SAVE_TESTCASE (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL                      */
/*    Pointers: Unchanged                       */
/*    Outputs:  tpa_ptr->testcase_flagsl                */
/*                                  */
/*    This routine sets the test-case is Active flag            */
/*                                  */
/************************************************************************/

unsigned long   SAVE_TESTCASE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Setting TEST-CASE active\n");
    }
#endif

    /*
     *  Now encode the START TEST-CASE opcode
     */
    ENC_INTEGER (tpa_ptr, OPC_START_TC);

    /*
     *    Set the test-case active flag.
     */
    (*tpa_ptr).testcase_flagsl |= TCAS_ACTIVE_TC;
    return PSL_NORMAL;
}


/************************************************************************/

/*    SAVE_THISTST - Save test file test data               */
/*    ============                          */
/*                                  */
/*    Type: Action routine.                     */
/*    Status:   PSL_NORMAL                      */
/*    Pointers: Unchanged.                      */
/*    Inputs:   argblk_ptr - Tparse argument block.         */
/*      tpa_ptr - Common tparse data.               */
/*    Outputs:  tpa_ptr - Common tparse data reset for next definition  */
/*      Thistst - 'This test' data set.             */
/*                                  */
/*    This routine sets the data for this test file using the data  */
/*    held in the definition 'work' variables.              */
/*                                  */
/************************************************************************/
unsigned long SAVE_THISTST (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */
    struct tsttbl   *tst_ptr;       /* Pointer to work test data */
    struct tsttbl   *thistst_ptr;       /* Pointer to this test data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    thistst_ptr = (*tpa_ptr).tst_thisps;

    /*
     *    If not a test type script, all done.
     */
    if (!((*com_ptr).type_testn)) return PSL_NORMAL;

    /*
     *    Copy the TEST data into the test data structure for this test file.
     */
    *thistst_ptr = *tst_ptr;        /* Copy data from work record */

    /*
     *    Show that we have finished with the structures in the work structure.
     */
    (*tst_ptr).tst_prm_ptrps = NULL;
    (*tst_ptr).tst_tlh_ptrps = NULL;

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    SAVE_VPO - Save VP$OUTPUT Operand                 */
/*    ========                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  Operand = VP$OUTPUT operand descriptor set up       */
/*                and added to LL of parsed operands.       */
/*                                  */
/*    This routine creates an operand descriptor for a VP$OUTPUT    */
/*    operand and adds it to the LL of parsed operands.             */
/*                                  */
/*    On entry (*tpa_com).dec_integerl is the VP$OUTPUT modifier value. */
/*                                  */
/************************************************************************/

unsigned long   SAVE_VPO (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a VP$OUTPUT operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     * 
     *
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    (*oph_ptr).addrl = ADDR_VP_OUTPUT;  /* Set addressing mode in OPD   */

    (*oph_ptr).dtypl = DTYP_CHARACTER;  /* Set Data type in OPD     */
    (*oph_ptr).sizel = 0;       /* Size is defined later    */
    (*oph_ptr).optl = 0;        /* Clear operand options */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    /*
     *    Set the trailing spaces bit as trailing spaces are always significant
     *    for VP$OUTPUT operands.
     */
    (*oph_ptr).optl |= VOPT_TRAILING;   /* Set trailing spaces significant */

    /*
     *    ...and return success
     */
    return PSL_NORMAL;          /* ...and return success    */
}
/*
 * 
 */

/************************************************************************/
/*                                  */
/*    SAVE_VPS - Save VP$SCREEN Operand                 */
/*    ========                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  VP$SCREEN operand descriptor set up and added to the LL */
/*          of parsed operands.                     */
/*                                  */
/*    This routine creates an operand descriptor for a VP$SCREEN    */
/*    operand and adds it to the LL of parsed operands          */
/*                                  */
/*    This routine is alos responsible for ensuring that all operands   */
/*    for this command have a datatype of integer literal if not    */
/*    already specified.                        */
/*                                  */
/************************************************************************/

unsigned long   SAVE_VPS (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   oph     *oph_ptr;      /* Operand header pointer   */
    struct   opd     *opd_ptr;      /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a VP$SCREEN operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     * 
     *
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    (*oph_ptr).addrl = ADDR_VP_SCREEN;  /* Set addressing mode in OPD   */

    (*oph_ptr).dtypl = DTYP_VP_SCREEN;  /* Set Data type in OPD     */

    (*oph_ptr).sizel = 0;       /* Size is defined later    */
    (*oph_ptr).optl = 0;        /* Clear operand options */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    /*
     *    Set the trailing spaces bit as trailing spaces are always significant
     *    for VP$SCREEN operands.
     */
    (*oph_ptr).optl |= VOPT_TRAILING;   /* Set trailing spaces significant */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    SAVE_VP_COLUMN - Save VP$COLUMN Operand               */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  Immediate integer operand descriptor set up and added   */
/*          to the LL of parsed operands.               */
/*                                  */
/*    This routine is responsible for creating an operand for a     */
/*    VP$COLUMN reference and adding the operand to the LL of parsed    */
/*    operands.                             */
/*                                  */
/************************************************************************/

unsigned long   SAVE_VP_COLUMN (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a VP$COLUMN operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     * 
     *
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    (*oph_ptr).addrl = ADDR_VP_COLUMN;  /* Addressing mode = VP$COLUMN  */
    (*oph_ptr).dtypl = DTYP_INTEGER;    /* Data Type = Integer      */
    (*oph_ptr).optl = 0;        /* Clear operand options  */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    return PSL_NORMAL;          /* ...and return success    */
}
/*
 * 
 */

/************************************************************************/

/*    SAVE_VP_HTTP_CODE - Save _HTTP_CODE Operand           */
/*    =================                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                              */
/*    Outputs:  _HTTP_CODE operand descriptor set up and added      */
/*          to the LL of parsed operands.               */
/*                                  */
/*    This routine creates an operand for an _HTTP_CODE reference and   */
/*    adds it to the LL of parsed operands.             */
/*                                  */
/************************************************************************/
unsigned long SAVE_HTTP_CODE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */
    struct oph  *oph_ptr;       /* Operand Header pointer */
    struct opd  *opd_ptr;       /* Operand descriptor pointer */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving an _HTTP_CODE operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    (*oph_ptr).addrl = ADDR_HTTP_CODE;    /* Addressing mode = _HTTP_CODE */
    (*oph_ptr).dtypl = DTYP_INTEGER;        /* Data Type = Integer */
    (*oph_ptr).optl = 0;            /* Clear operand options */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    return PSL_NORMAL;              /* ...and return success */
}


/************************************************************************/

/*    SAVE_VP_HTTP_STATUS - Save _HTTP_STATUS Operand           */
/*    ===================                       */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                              */
/*    Outputs:  HTTP_STATUS operand descriptor set up and added     */
/*          to the LL of parsed operands.               */
/*                                  */
/*    This routine creates an operand for an _HTTP_STATUS reference and */
/*    adds it to the LL of parsed operands.             */
/*                                  */
/************************************************************************/
unsigned long SAVE_HTTP_STATUS (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */
    struct oph  *oph_ptr;       /* Operand Header pointer */
    struct opd  *opd_ptr;       /* Operand descriptor pointer */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving an _HTTP_STATUS operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    (*oph_ptr).addrl = ADDR_HTTP_STATUS;    /* Addressing mode = HTTP_STATUS */
    (*oph_ptr).dtypl = DTYP_INTEGER;        /* Data Type = Integer */
    (*oph_ptr).optl = 0;            /* Clear operand options */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    return PSL_NORMAL;              /* ...and return success */
}


/************************************************************************/
/*                                  */
/*    SAVE_VP_LINE - Save VP$LINE Operand               */
/*    ============                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  Immediate integer operand descriptor set up and added   */
/*          to the LL of parsed operands.               */
/*                                  */
/*    This routine is responsible for creating an operand for a     */
/*    VP$LINE reference and adding the operand to the LL of parsed  */
/*    operands.                             */
/*                                  */
/************************************************************************/

unsigned long   SAVE_VP_LINE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a VP$LINE operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     * 
     *
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    (*oph_ptr).addrl = ADDR_VP_LINE;    /* Addressing mode = VP$LINE    */
    (*oph_ptr).dtypl = DTYP_INTEGER;    /* Data Type = Integer      */
    (*oph_ptr).optl = 0;        /* Clear operand options  */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */

    return PSL_NORMAL;          /* ...and return success    */
}
/*
 * 
 */

/************************************************************************/
/*                                  */
/*    SAVE_VP_RESPONSE - Save VP$RESPONSE Operand           */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                              */
/*    Outputs:  Immediate integer operand descriptor set up and added   */
/*          to the LL of parsed operands.               */
/*                                  */
/*    This routine is responsible for creating an operand for a     */
/*    VP$RESPONSE reference and adding the operand to the LL of parsed  */
/*    operands.                             */
/*                                  */
/************************************************************************/

unsigned long   SAVE_VP_RESPONSE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand Header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a VP$SRESPONSE operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     * 
     *
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    (*oph_ptr).addrl = ADDR_VP_RESPONSE;/* Addressing mode = VP$RESPONSE */
    (*oph_ptr).dtypl = DTYP_INTEGER;    /* Data Type = Integer      */
    (*oph_ptr).optl = 0;        /* Clear operand options  */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    return PSL_NORMAL;          /* ...and return success    */
}
/*
 * 
 */

/************************************************************************/
/*                                  */
/*    SAVE_VP_STATUS - Save VP$STATUS Operand               */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                              */
/*    Outputs:  Immediate integer operand descriptor set up and added   */
/*          to the LL of parsed operands.               */
/*                                  */
/*    This routine is responsible for creating an operand for a     */
/*    VP$STATUS reference and adding the operand to the LL of parsed    */
/*    operands.                             */
/*                                  */
/************************************************************************/

unsigned long   SAVE_VP_STATUS (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    All this routine does is return the address of the symbol table in
     *    VAR_PTR as VP$STATUS is the first variable to be defined.
     */
    (*tpa_ptr).var_ptrps = (*tpa_ptr).variable_begps;
    /* Set variable = first in table */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    SAVE_VP_TESTRESULTS - Save VP$TEST_RESULTS Operand        */
/*    ===================                       */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  Immediate integer operand descriptor set up and added   */
/*          to the LL of parsed operands.               */
/*                                  */
/*    This routine is responsible for creating an operand for a     */
/*    VP$TEST_RESULTS reference and adding the operand to the LL of     */
/*    parsed operands.                          */
/*                                  */
/************************************************************************/

unsigned long   SAVE_VP_TESTRESULTS (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a VP$TEST_RESULTS operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    (*oph_ptr).addrl = ADDR_VP_TEST_RESULTS;
    /* Addressing mode = VP$TEST_RESULTS */
    (*oph_ptr).dtypl = DTYP_CHARACTER;  /* Data Type = Character    */
    (*oph_ptr).optl = 0;        /* Clear operand options    */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/

/*    SAVE_SWITCH - Save SWITCH command                 */
/*    ===========                           */
/*                                  */
/*    Calling Format:                           */
/*  iret = SAVE_SWITCH (argblk_ptr)                 */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*      SW_DEPTHL - Index of next FREE switch stack entry   */
/*    Outputs:  SWSTACK - Entry on SWITCH stack initialized         */
/*                                  */
/*    This routine saves the switch variable data type (used to ensure  */
/*    valid case values are used.                                       */
/*                                      */
/*    It also saves a pointer to the current operand LList - this is    */
/*    referenced later when encoding switch operands in the case        */
/*    statement.                            */
/*                                      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   SAVE_SWITCH (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com    *com_ptr;       /* Common Data          */
    struct   scl_tpa    *tpa_ptr;       /* TPARSE Data          */
    struct   swstack    *sw_ptr;        /* Pointer to SWITCH stack entry*/
    struct   symtbl     *symtbl_ptr;    /* Pointer to symbol table  */
    unsigned long       iret;       /* status return variable   */


    struct   oph        *oph_ptr;       /* Pointer to primary operand   */
    char    *llb_ctx;      /* Linked list context      */


    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    sw_ptr = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl - 1;
    /* Point to current entry   */
    symtbl_ptr = (*tpa_ptr).var_ptrps;


    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a SWITCH Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     * 
     *
     *  Save data type of switch variable from symbol table.
     *    To do this we must recover the primary operand which we have just parsed
     *    and determine its data type as this is the variable in which we are most
     *    interested
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                        ,(char *) &oph_ptr
                        ,(*tpa_ptr).primary_operandpc
                        ,&llb_ctx ,PSL_VM_MODE_SPEC);
    if  (!(psl_test_status (iret)))     /* ...report any errors     */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY (in SAVE_SWITCH)", iret);
    }

    (*sw_ptr).sw_dtypl  = (*oph_ptr).dtypl;


#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE("\nSAVE_SWITCH*Switch datatype = %ld \n"
                         ,(*sw_ptr).sw_dtypl);
    }
#endif


    /*
     * So not to lose the SWITCH operand initialise another LL for further
     * compilation and save original list on the SWITCH stack fo later use.
     */
    (*sw_ptr).sw_operand_llps = (*tpa_ptr).parsed_operand_llps;
    /* Saving old LL pointer */
    iret = PSL_VM_INITIALIZE (sizeof (long) ,&(*tpa_ptr).parsed_operand_llps);
    if   (!(psl_test_status (iret)))
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_INITIALISE (in SAVE_SWITCH)", iret);
    }
end:
    ;

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    SAVE_WITH_CNT - Save HTTP Request WITH value count        */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   tpa_ptr->char_list_val_cntl             */
/*    Outputs:  tpa_ptr->http_withval_cntl set                  */
/*                                  */
/*    This routine saves the count of values in the HTTP request    */
/*    WITH clause that has just been parsed.                */
/*                                  */
/************************************************************************/
unsigned long SAVE_WITH_CNT (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *    Save the value count.
     */
    tpa_ptr->http_withval_cntl = tpa_ptr->char_list_val_cntl;

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                  */
/*    SAVE_WOUT_CNT - Save HTTP Request WITHOUT value count     */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   tpa_ptr->char_list_val_cntl             */
/*    Outputs:  tpa_ptr->http_wout_cntl set                 */
/*                                  */
/*    This routine saves the count of values in the HTTP request    */
/*    WITHOUT clause that has just been parsed.             */
/*                                  */
/************************************************************************/
unsigned long SAVE_WOUT_CNT (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *    Save the value count.
     */
    tpa_ptr->http_woutval_cntl = tpa_ptr->char_list_val_cntl;

    return PSL_NORMAL;          /* Return success */
}

/*======================================================================*/
/*                                  */

/*                                  */
/*    CREATE_VAR_OPD - Create Variable Operand              */
/*    ==============                            */
/*                                  */
/*    Type: Service routine                     */
/*    Status:   Not applicable                      */
/*    Pointers: None                            */
/*    Inputs:   *var_ptr - Pointer to variable entry in Variables   */
/*                 Symbol Table                 */
/*    Outputs:  Variable symbol table entry set up.         */
/*                                  */
/*    This routine creates an operand descriptor from an entry in the   */
/*    Variables Symbol Table. The routine is supplied with pointers */
/*    to variable symbol table entry.                   */
/*                                  */
/*======================================================================*/

void        CREATE_VAR_OPD  (struct   scl_tpa *tpa_ptr,
                             struct   symtbl  *var_ptr)

{
    /*
     *    Local data
     */
    struct oph *oph_ptr;            /* Pointer to OPH       */
    struct opd *opd_ptr;            /* Pointer to OPD       */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nSaving a VARIABLE operand to the LL\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_SAVOPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    /*
     *    ...and fill in the fields in the OPH and OPD
     */
    (*oph_ptr).addrl = (*var_ptr).var_addrl;
    /* Save addressing mode in OPH  */
    (*oph_ptr).dtypl = (*var_ptr).var_typel;
    /* Save Data type in OPH    */
    (*oph_ptr).sizel = (*var_ptr).var_sizel;
    /* Save variable size in OPH    */
    (*oph_ptr).optl = (*var_ptr).var_optl;
    /* Save variable options in OPH */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    /*
     *    If this variable is an array reference then we should flag the fact in
     *    the operand header.
     */
    if ((*var_ptr).var_array_dimcntl != 0)
    {                   /* If variable is an array  */
        (*oph_ptr).optl |= VOPT_ARRAY; /* Flag as an array in the OPH   */
    }

    /*
     *    ...and set up the rest of the OPD for the variable.
     */
    (*opd_ptr).uopd.var.offl = (*var_ptr).var_offl;
    /* Save variable offset in OPD  */
    (*opd_ptr).uopd.var.dimcntl = (*var_ptr).var_array_dimcntl;
    /* Save array dimensions count  */
    (*opd_ptr).uopd.var.symtbl_ptrps = var_ptr;
    /* ...and work pointer to SYMTBL */
    /*
     * 
     * Set operand descriptor for variable depending if it is a record
     * or non-record variable.
     * Record variables cannot have value lists or contain arrays - they
     * only require an index value. This index will be used later for accessing
     * the format definitions array.
     */

    if (((*oph_ptr).dtypl == DTYP_CHARACTER)
            ||((*oph_ptr).dtypl == DTYP_INTEGER))
        /* If character var (not record) */
    {
        (*opd_ptr).uopd.var.uvar.nrec.values_offl = (*var_ptr).var_values_offl;
        /* Save value list offset   */
        (*opd_ptr).uopd.var.uvar.nrec.values_endl = (*var_ptr).var_values_endl;
        /* ...and offset of byte beyond */
        (*opd_ptr).uopd.var.uvar.nrec.values_cntl = (*var_ptr).var_values_cntl;
        /* Add values count         */
        (*opd_ptr).uopd.var.uvar.nrec.low_rangel = (*var_ptr).var_low_rangel;
        /* Save low range value     */
        (*opd_ptr).uopd.var.uvar.nrec.high_rangel = (*var_ptr).var_high_rangel;
        /* Save high range value    */
    }
    if ((*oph_ptr).dtypl == DTYP_RECORD)  /* If character  record */

    {
        (*opd_ptr).uopd.var.uvar.rec.indexl = (*var_ptr).var_indexl;
        /* Save record index        */
    }
    /*
     * 
     */

    return;             /* ...and return        */
}
/*
 * 
 */

/*======================================================================*/

/*    GET_SAVOPD_COUNT - Recover and return count of operands In LL */
/*    ================                          */
/*                                  */
/*    Calling Format:                           */
/*  count = GET_SAVOPD_COUNT (tpa_ptr)              */
/*                                  */
/*    Where:                                */
/*  count = returned UL count of operands in the linked list    */
/*  tpa_ptr = pointer to the Tparse data                */
/*                                  */
/*    This routine scans through the linked list of parsed operands */
/*    and returns a count of those found.               */
/*                                  */
/*----------------------------------------------------------------------*/
static unsigned long GET_SAVOPD_COUNT (struct scl_tpa *tpa_ptr)
{
    /*
     *    Local data
     */
    struct   oph      *oph_ptr;     /* Pointer to OPH       */
    unsigned long     count = 0;        /* Count of operands found  */
    unsigned long     iret;         /* Status return        */
    char     *llb_pos;     /* Linked list position     */
    char     *llb_ctx;     /* Linked list context      */

    llb_pos = NULL;         /* Position at start of list    */

    for ( ; ; count++)
    {
        iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                            ,(char *) &oph_ptr ,llb_pos ,&llb_ctx
                            ,PSL_VM_MODE_NEXT);
        if (!(psl_test_status (iret))) return count;
        llb_pos = llb_ctx;      /* Update the context       */

        /*
         *    There should always be an OPD following an OPH so pick it up to
         *    complete the operand.
         */
        iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                            ,(char *) &oph_ptr ,llb_pos ,&llb_ctx
                            ,PSL_VM_MODE_NEXT);
        if (!(psl_test_status (iret))) return count;
        llb_pos = llb_ctx;      /* Update the context       */
    }
}
/*
 * 
 */

/*======================================================================*/

/*    SAVE_TARGET - Save TARGET environment             */
/*    ===========                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*                                  */
/*   This routine will set the TARGET environment parsed.       */
/*                                  */
/*======================================================================*/

unsigned long   SAVE_TARGET (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    (*tpa_ptr).sql_database_typel = (*argblk_ptr).paraml;

    return PSL_NORMAL;          /* return success       */
}
