/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/* scl_subs.c:
 * ===========
 *
 * This module contains all the platform independent routines used by scl that
 * do not belong in any other SCL module.
 */

#include <psl_macros.h>
#include <vtest_macros.h>

/*
 *  Standard ANSI headers.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>


/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>
#include <trc_rtndef.h>         /* Tracing routines */
#include "scl_rtndef.h"         /* SCL routines */

static void SAVE_SCRIPT_NAME (struct scl_com *com_ptr);

/*
 *  Structure definitions and parameters.
 */
#include <shrmsg.h>         /* PSL messages */
#include "scl_cmd.h"            /* Command Line Data Structure  */
#include "scl_comdef.h"         /* Global Data structure */
#include "scl_tbldef.h"         /* SCL table definitions */
#include <psldiag_prmdef.h>     /* Diagnostics paramters */
#include <pslspn_prmdef.h>      /* PSL_SPAWN symbols */
#include <pslfil_prmdef.h>      /* PSL_FILE symbols */
#include <psl_trc_prmdef.h>     /* PSL tracing parameters */
#include <vtest_prmdef.h>       /* VTEST parameters */

#define BLOCK_SIZE 512          /* Bytes per block */

/******************************************************************************
 * @func    This is a generic routine for deallocating all the memory allocated
 *      to the supplied operand header and operand data structures. This
 *      currently includes the memory for the structures themselves and the
 *      memory for immediate character and character constant operands.
 *
 * @parm    Pointer to operand header.
 * @parm    Pointer to operand data.
 */
void FREE_OPH_OPD (struct oph *oph_ptr,
                   struct opd *opd_ptr)
{

    /*
     *   If the operand is an immediate character operand or a character constant
     *   operand free up the memory allocated for the data value.
     */
    if ((oph_ptr->addrl == ADDR_IMMEDIATE)
            &&  ((oph_ptr->dtypl == DTYP_CHARACTER)
                 ||   (oph_ptr->dtypl == DTYP_CONCHARACTER)))
    {
        FREE (opd_ptr->uopd.imc.val_pc);
        opd_ptr->uopd.imc.val_pc = NULL;
    }

    /* Free the memory allocated for the operand header and operand data structures. */
    FREE (oph_ptr);
    FREE (opd_ptr);

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER5))
    {
        PSL_OUTPUT_TRACE ("FREE: oph ('%p') and opd ('%p')\n", oph_ptr, opd_ptr);
    }
#endif  /* PSL_TRACING */
    return;
}

/************************************************************************/
/*                                  */
/*  SAVE_SCRIPT_NAME - Save the script name             */
/*  ================                        */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     SAVE_SCRIPT_NAME (com_ptr);                  */
/*                                  */
/*  Where:                              */
/*                                  */
/*     com_ptr  = Pointer to scl common data            */
/*                                  */
/*  The script name is taken from the name of the script file. This */
/*  routine extracts the script name from the expanded script   */
/*  source file name, by stripping off the path, extension and  */
/*  and stores it in the scl common data structure.         */
/*                                  */
/************************************************************************/
void SAVE_SCRIPT_NAME (struct scl_com *com_ptr)
{
    /*
     *  Local data
     */
    char     *nam_ptr;     /* Temporary pointer to filename */
    char     *namend_ptr;      /* Temporary pointer to filename end */
    unsigned long     namlen;       /* File name length */

    /*
     *  Extract the filename/script name from the script source file name.
     *  The following filename formats are assumed:
     *
     *  WIN32: ...\scriptname{.extension}
     *     scriptname is the string between the last '\' and the next '.'.
     *  UNIX: .../scriptname{.extension}
     *     scriptname is the string between the last '/' and the next '.'.
     */
#ifdef WIN32
    nam_ptr = strrchr ((*com_ptr).scl_namec256, '\\');
    if (nam_ptr != NULL)
    {
        nam_ptr++;
        namend_ptr = strchr (nam_ptr, '.');
        if (namend_ptr != NULL)
        {
            namlen = namend_ptr - nam_ptr;
            (void) memcpy ((*com_ptr).scrnam_c20, nam_ptr, namlen);
            (*com_ptr).scrnam_c20 [namlen] = '\0';
        }
    }
#elif UNIX
nam_ptr = strrchr ((*com_ptr).scl_namec256, '/');
    {
        nam_ptr++;
        namend_ptr = strchr (nam_ptr, '.');
        if (namend_ptr != NULL)
        {
            namlen = namend_ptr - nam_ptr;
            (void) memcpy ((*com_ptr).scrnam_c20, nam_ptr, namlen);
            (*com_ptr).scrnam_c20 [namlen] = '\0';
        }
    }
#else
#  error Platform not supported at line __LINE__ (Cannot obtain script name)
#endif

    return;
}

/************************************************************************/
/*                                  */
/*  SCL_OPEN_SOURCE - Open script source file.          */
/*  ================                        */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = SCL_OPEN_SOURCE (scl_com_ptr, file_idx);      */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret     = unsigned long routine status          */
/*          = PSL_NORMAL - if success           */
/*          = PSL_OPENIN - if unable to open script source  */
/*     scl_com_ptr  = Pointer to the SCL Common Data Record     */
/*     file_idx = Unsigned long Source file index       */
/*                                  */
/*  This routine is called to open the script source file.      */
/*                                  */
/************************************************************************/
unsigned long SCL_OPEN_SOURCE (struct scl_com *com_ptr,
                               unsigned long file_idx)
{
    /*
     *  Local data
     */
    char     *filnam_ptr;       /* Ptr to filename */

    /*
    *  Set up the file specification for the input source file.
    *  There is NO default nor related file specification string (NULLs),
    *  and no flags are set.
    */
    filnam_ptr = (*com_ptr).scl_namec256;   /* Identify file for "open" */
    PSL_FILE_PARSE (filnam_ptr, NULL, NULL, NULL,
                    filnam_ptr, sizeof ((*com_ptr).scl_namec256)
                    ,FILPRS_INPUT);
    /* Get expanded filespec */
    /*
     *  Open the input source file thus identified.
     */
    (*com_ptr).files.ptrpf14 [file_idx] = fopen (filnam_ptr, "r");
    /* Open source file for "read"*/

    if ((*com_ptr).files.ptrpf14 [file_idx] == NULL)
    {                       /* File not opened? */
        PSL_LOG_CIOERR ((*com_ptr).diag_ptrps,
                        PSL_OPENIN,
                        (*com_ptr).scl_namec256);
        /* Setup "open" diagnostic */
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        /* Report "open" error */
        return PSL_OPENIN;
    }

    /*
     *  Now get and save the script name.
     */
    SAVE_SCRIPT_NAME (com_ptr);

    /*
     *  File opened O.k.
     */
    return PSL_NORMAL;
}

/*************************************************************************/
/*                                                                      */
/*  SCL_OPEN_INCLUDE - Open INCLUDE-d script source file                */
/*  =================                                                   */
/*                                                                      */
/*  Calling Convention:                                                 */
/*                                                                      */
/*     iret = SCL_OPEN_INCLUDE (scl_com_ptr, incnam_ptr, namlen,        */
/*                   file_idx)                                          */
/*                                                                      */
/*  Where:                                                              */
/*                                                                      */
/*     iret     = unsigned long routine status                          */
/*          = PSL_NORMAL - if success                                   */
/*          = PSL_OPENIN - if unable to open INCLUDE script             */
/*     scl_com_ptr  = Pointer to the SCL Common Data Record             */
/*     incnam_ptr   = Pointer to include filename string                */
/*     namlen   = Unsigned long length of above                         */
/*     file_idx = Unsigned long Include file index                      */
/*                                                                      */
/*  This routine is called to open the 'include' file specified         */
/*  in the script.                                                      */
/*                                                                      */
/************************************************************************/

unsigned long  SCL_OPEN_INCLUDE (struct scl_com *scl_com_ptr,
                                 char *incnam_ptr,
                                 unsigned long name_length,
                                 unsigned long file_idx)
{
    /*
     *  Local data
     */
    char    *expscrnam_ptr;         /* Ptr to Expanded script name*/
    char    script_name [256];      /* Script name (copy)   */
    char    exp_script_name [256];  /* Expanded Script name */
    char    *exp_file_ptr;          /* Pointer common expand script*/
struct scl_cmd
            *scl_cmd_ptr;            /* Pointer to command line structure */

    /*
     *  Copy pointers for ease of reference.
     */
    scl_cmd_ptr = (*scl_com_ptr).cmd_ptrps;

    /*
     *  Set up the file specifications for the include source file.
     *  The default file is "SYS$DISK:[].SCL" (or UNIX equivalent), there is NO
     *  related filespec (NULL), and no flags.
     */
    (void) memcpy (script_name, incnam_ptr, name_length);
    /* Copy Include filename */
    script_name [name_length] = '\0';       /* and null-terminate   */

    expscrnam_ptr = exp_script_name;        /* Setup pointer    */

    PSL_FILE_PARSE (script_name, ".scl", NULL, NULL,
                    expscrnam_ptr, 255
                    ,FILPRS_INPUT);     /* Get expanded filespec */

     /* Open the include source file thus identified for 'read' access. */
    (*scl_com_ptr).files.ptrpf14 [file_idx] = fopen (expscrnam_ptr, "r");

    /* If 'open' failed, see if the include file exists in the alternate include */
    /* directory specified on the command line using --include (-I) - if any */
    if ((*scl_com_ptr).files.ptrpf14 [file_idx] == NULL)    /* If File not opened */
    {
        if ((*scl_cmd_ptr).includen)        /* If alternative include path specified */
        {
            PSL_FILE_PARSE ((*scl_com_ptr).include_valuec256, NULL,
                    script_name, NULL,
                    expscrnam_ptr, 255,
                    FILPRS_INPUT);  /* Get expanded filespec */
            /* Open the include source file thus identified for 'read' access. */
            (*scl_com_ptr).files.ptrpf14 [file_idx] = fopen (expscrnam_ptr, "r");
        }

        if ((*scl_com_ptr).files.ptrpf14 [file_idx] == NULL)
        {                                   /* If file not opened - report error */
            PSL_LOG_CIOERR ((*scl_com_ptr).diag_ptrps,
                        PSL_OPENIN,
                        script_name);     /* Setup "open" diagnostic */
            PSL_OUTPUT_DIAG ((*scl_com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
            return PSL_OPENIN;              /* Report "open" error */
        }
    }

    /*
     *  Save expanded include file name for use in debugging
     */

    exp_file_ptr = (*scl_com_ptr).exp_src_filc256;

    (void) strcpy (exp_file_ptr,expscrnam_ptr);

    /*
     *  File opened O.k.
     */
    return PSL_NORMAL;
}

/************************************************************************/
/*                                                                      */
/*  SCL_CLOSE_SOURCES - Close an SCL Source file.                       */
/*  ==================                                                  */
/*                                                                      */
/*  Calling Convention:                                                 */
/*                                                                      */
/*     SCL_CLOSE_SOURCE (scl_com_ptr, file_idx);                        */
/*                                                                      */
/*  Where:                                                              */
/*                                                                      */
/*     scl_com_ptr  = Pointer to the SCL Common Data Record             */
/*     file_idx = Unsigned long index of Source File to close           */
/*                                                                      */
/*  This routine is called to close a SOURCE file.                      */
/*                                                                      */
/************************************************************************/

void SCL_CLOSE_SOURCE (struct scl_com *scl_com_ptr,
                       unsigned long file_idx)

{
    /*
     *  Local data.
     */
    long       close_sts;           /* Status from fclose */

    /*
     *  Close the Source file.
     */
    close_sts = fclose ((*scl_com_ptr).files.ptrpf14 [file_idx]);
    /* Close the Source file */
    if (close_sts == EOF)           /* Unable to close file? */
    {
        PSL_LOG_CIOERR ((*scl_com_ptr).diag_ptrps,
                        PSL_CLOSEIN,
                        (*scl_com_ptr).scl_namec256);
        /* Setup error diagnostic */
        PSL_OUTPUT_DIAG ((*scl_com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
    }

    return;
}

/************************************************************************/
/*                                  */
/*  READ_NEXT_LINE - Read next line from Source File        */
/*  ===============                         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  READ_NEXT_LINE (scl_com_ptr, line_size, source_line,        */
/*          parsable_line, error_line, eof_flag, file_idx); */
/*                                  */
/*    Where:                                */
/*                                  */
/*  scl_com_ptr   = Pointer to Common Data Structure        */
/*  line_size     = Pointer to unsigned long line size      */
/*  source_line   = Pointer to the source line be read      */
/*  parsable_line = Pointer to parsed line, minus comments, U/case  */
/*  error_line    = Pointer to error line               */
/*  eof_flag      = Pointer to unsigned long end-of-file flag   */
/*  file_idx       = unsigned long source file identifier       */
/*                                  */
/*  This routine reads the next line from the source file       */
/*  identified, and outputs that line to the Listing file, if   */
/*  required.                           */
/*                                  */
/*  The routine also initializes the Error underline to the same    */
/*  value as the original source text line, to ease the process of  */
/*  underlining errors.                     */
/*                                  */
/*  The routine also converts the input text line to Uppercase, */
/*  and strip off Comments, leaving only the parsable text.     */
/*                                  */
/************************************************************************/

void        READ_NEXT_LINE (struct scl_com *scl_com_ptr,
                            unsigned long *line_size,
                            char *source_line,
                            char *parsable_line,
                            char *error_line,
                            unsigned long *eof_flag,
                            unsigned long file_idx)

{
    /*
     *  Local Data.
     */
    char   *read_ptr;          /* Returned pointer from read */
    unsigned long   isub;               /* Loop control */
    Boolean   lin2big = PFALSE;       /* Line too big flag */

    /*
     *  If we have already found the End of File on a previous parsing
     *  operation, return straight away. This will enable this routine to be
     *  called at lots of places in the TPARSE state table, and will always
     *  produce the same result.
     */
    if (*eof_flag != 0) goto end;

    /*
     *  Read the next record sequentially from the file.
     */
next_line:
    clearerr ((*scl_com_ptr).files.ptrpf14 [file_idx]);
    /* Clear EOF and error bits */
    read_ptr = fgets (source_line, 134, (*scl_com_ptr).files.ptrpf14 [file_idx]);

    /*
     * Save end of line offset. Start of line offset is then
     * worked out from the end offset - line size.
     */
    (*scl_com_ptr).src_line_offsetl =
        ftell((*scl_com_ptr).files.ptrpf14 [file_idx]);
    /* Save source line offset */
    if (read_ptr == NULL)           /* A problem? */
    {
        if ((feof ((*scl_com_ptr).files.ptrpf14 [file_idx])) == 0)
            /* Some error other than EOF? */
        {
            PSL_LOG_CIOERR ((*scl_com_ptr).diag_ptrps,
                            PSL_READERR,
                            (*scl_com_ptr).scl_namec256);
            /* Setup error diagnostic */
            PSL_OUTPUT_DIAG ((*scl_com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
            /* Output diagnostic */
            goto abort_end;         /* report 'abort' and exit */
        }
        else                        /* Must be End-Of-File */
        {
            *line_size = 0;         /* Return line length of ZERO */
            *eof_flag = 1;          /* Return End-Of-File status */
            goto end;               /* Exit */
        }
    }

    /*
     *  We have read up to the number of characters specified in the fgets call
     *  above. This INCLUDES any trailing 'newline' character AND the terminating
     *  NULL character. If the last character is a 'newline' character the
     *  line is OK and we strip off the 'newline', adjust the line size
     *  accordingly and continue. If the last character is NOT a 'newline', the
     *  source line is too long, so we discard the rest of the line and read
     *  more lines until we find a line that is NOT too long.
     */
    *line_size = strlen (source_line);      /* Set length of record read */

    if ((source_line[(*line_size) - 1]) == '\n')
    {
        *line_size -= 1;            /* Remove trailing newline */
        source_line[(*line_size)] = '\0';   /* ... */

        if (lin2big)
        {
            lin2big = PFALSE;
            goto next_line;
        }
    }
    else if ((*line_size) == 133)       /* Else if line too big */
    {
        if (lin2big) goto next_line;

        lin2big = PTRUE;

        PSL_PUTDIAG ((*scl_com_ptr).diag_ptrps, 2,
                     PSL_READERR, (*scl_com_ptr).scl_namec256,
                     PSL_BADRECSIZ);
        /* Setup error diagnostic */
        PSL_OUTPUT_DIAG ((*scl_com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        /* Output diagnostic */
        goto next_line;             /* and read next 'line' */
    }

    (*scl_com_ptr).srclinl10 [(*scl_com_ptr).include_depthl] +=  1;
    /* Count this line */
    if (*line_size <= 0) goto next_line;    /* Loop for another if empty */

    /*
     *    Check the line to see if there are any printing characters on it as 
     *    if there are not we will throw away the line and get the next one.
     */
    for (isub = 0; isub < *line_size; isub++)
    {
        if (isgraph (source_line [isub])) goto line_not_empty;
    }

    goto next_line;

    /*
     *    LINE_NOT_EMPTY
     *    ==============
     *
     *  Create a copy of the source line with all lowercase characters
     *  converted to uppercase.
     */
line_not_empty:
    ;

    for (isub = 0; isub <= *line_size; isub++)
    {                       /* For each char in the line */
        if (isalpha(source_line [isub]))    /* If it is alphabetic ... */
        {
            parsable_line [isub] = toupper (source_line [isub]);
        }                   /* copy it (in uppercase) */
        else                    /* anything else ... */
        {
            parsable_line [isub] = source_line [isub];
        }                   /* gets copied as is */

        /*
         *  Set up Error Line as a mask of the parsable line with Tabs in the
         *  same place.
         */
        if (parsable_line [isub] == '\t')   /* If it was a TAB character */
        {
            error_line [isub] = '\t';       /* mask errror line similarly */
        }
        else                    /* anything else ... */
        {
            error_line [isub] = ' ';        /* is masked with a SPACE */
        }

    }

    /*
     * Have to check for rubbish at the start of the line (from the
     * end of the previous line.)
     * Line size + 1 (ADD 1 as newline character has been removed from the 
     * source line string size).
     */
    (*scl_com_ptr).src_line_offsetl -= (*line_size + 1);

    /*
     *  All done - exit here.
     */
end:
    return;

    /*
     *  We have a severe error from which we cannot safely continue.
     */
abort_end:
    PSL_PUTDIAG ((*scl_com_ptr).diag_ptrps, 1, PSL_ENDABORT,
                 (*scl_com_ptr).scl_namec256);
    /* Setup error diagnostic */
    PSL_OUTPUT_DIAG ((*scl_com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
    /* Output diagnostic */

    PSL_EXIT();                 /* Can do nought else - abort!*/
}
