/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/**
 * Define the Methods for the AtvfKey class
 * Keys don't begin with whitespace so any given will be ignored.
 **/

// Local includes
#include "AtvfKey.h"

/**
 * Constructor
 **/
AtvfKey::AtvfKey(const string& token)
{
    set(token);
}

/**
 * Copy constructor
 **/
AtvfKey::AtvfKey(const AtvfKey& rhs)
{
    for(unsigned int i = 0; i < NUSED; i++)
    {
        _key[i] = rhs._key[i];
    }
}

/**
 * Accessor to make sure we can do less than comparisons
 **/
bool AtvfKey::operator<(const AtvfKey& rhs) const
{
    // return true if last char tested was less than rhs
    return (compare(rhs) < 0) ? true : false;
}

/**
 * Accessor to allow us to test equivelence of keys
 **/
bool AtvfKey::operator==(const AtvfKey& rhs) const
{
    return (compare(rhs) == 0) ? true : false;
}

/**
 * Accessor to compare this key with another
 **/
int AtvfKey::compare(const AtvfKey& cmpKey) const
{
    int diff = 0;
    // test characters, if they match continue to next
    for(unsigned int i = 0;
            (diff == 0) && (i < NUSED); // stop at end sequence
            i++)
    {
        diff = _key[i] - cmpKey._key[i];
    }
    return diff;
}

/**
 * accessor to compare this key with a string
 **/
int AtvfKey::compare(const string& cmpStr) const
{
    return compare(AtvfKey(cmpStr));
}

/**
 * Mutator for changing the value of the key
 * Initial whitespace in a token is ignored. 
 **/
AtvfKey& AtvfKey::set(const string &token)
{
    for(unsigned int tokidx = 0, keyidx = 0; keyidx < NUSED; tokidx++)
    {
        if(token.length() < tokidx)
        {
            // null fill if string not long enuf
            _key[keyidx] = '\0';
            keyidx++;
        }
        else
        {
            char curToken;
            if(tokidx >= token.size())
                curToken = '\0';
            else
                curToken = token[tokidx];

            if(!(keyidx == 0 // empty key
                    && (curToken == ' ' || curToken == '\t')))
            {
                // If char is uppercase then lowercase, only works when ASCII
                if((curToken < 91) && (curToken > 64))
                {
                    _key[keyidx] = curToken + 32;
                }
                else
                {
                    _key[keyidx] = curToken;
                }
                keyidx++;
            }
        }
    }

    return *this;
}

/**
 * Access the key as a string
 **/
string AtvfKey::toString() const
{
    return string(_key, NUSED);
}

/**
 * Provide method of outputting an AtvfKey
 **/
ostream& operator<<(ostream& os, const AtvfKey& k)
{
    os << k.toString();
    return os;
}
