/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
#include <stdio.h>
#include "CECSpawnObject.h"
#include "Factory/ParamList_i.h"
#include <ctime.hpp>
CECSpawnObject::CECSpawnObject  (const CECNodeDefinition& Node ,const string& product
                                 ,TARGETIMAGE imageid ,TARGETID targetid
                                 ,const string& params
                                 ,const string& ClassName
                                 ,const string& ExpectedName
                                 ,CORBAServer* CS)
        :m_NodeDefinition(Node)
        ,m_Product(product)
        ,m_Exe(imageid)
        ,m_Target(targetid)
        ,m_Params(params)
        ,m_ExpectedName(ExpectedName)
        ,m_ClassName(ClassName)
        ,m_Object(CORBA::Object::_nil())
{
    if(CS!=NULL)
    {
        m_pCORBAServer=new CORBAServer(*CS);    
    }
    else
    {
        m_pCORBAServer=new CORBAServer((int)NULL,NULL);
    }
}

CECSpawnObject::~CECSpawnObject()
{
    delete m_pCORBAServer;
}

CORBA::Object_ptr CECSpawnObject::RetrieveObject()
{
    bool debug=FindDebugProcess();
    if(!debug)//if no debug process found
    {
        AttachToFactory();
        CreateSpawnProcess();
        AttachToObject();
    }
    return CORBA::Object::_duplicate(m_Object);
}

bool CECSpawnObject::FindDebugProcess()
{
    char * exe = new char[32];
    Tgt_GetTargetExecutable(m_Target,m_Exe,exe,32);
    string l_ExpectedName="DEBUG_"+m_NodeDefinition.GetNodeIPAddrAsString()+"_"+exe;    
    delete[] exe;
    try
    {
        CORBA::Object_ptr obj;
        obj = m_pCORBAServer->FindObjectWithName(m_ClassName.c_str(),l_ExpectedName.c_str());
        if( CORBA::is_nil(obj) )
        {
            return false;
        }
        m_Object=obj;
        m_pCORBAServer->UnNameObject(m_ClassName.c_str(),l_ExpectedName.c_str());   
    }
    catch(...)
    {
        return false;
    }
    return true;
}
void CECSpawnObject::AttachToFactory()
{
    const char* where="CECSpawnObject::AttachToFactory";
    try
    {
        CORBA::Object_ptr obj;
        // Get pointer to main factory object
        obj = m_pCORBAServer->FindObjectWithName( CyrFactory::CLASS_NAME, m_NodeDefinition.GetNodeIPAddrAsString().c_str());   
        m_Factory = CyrFactory::_narrow(obj);   
    }
    catch(CORBA::Exception&e)
    {
        const char* err=CorbaExceptionString(e);
        throw CyrException(err,where,-1);
    }
}
void CECSpawnObject::CreateSpawnProcess()
{
    const char* where="CECSpawnObject::CreateSpawnProcess";

    //We have a connection to the factory so create the new object
    SpawnProcess_var _SpawnProcess;
    int _PID = 0;
    CORBA::Object_ptr obj;

    // Create a new SpawnProcess object
    // Create a new parameter list
    FactoryParamList* params;
    try
    {
        params = new FactoryParamList( SpawnProcess::CLASS_NAME, m_ExpectedName.c_str(), 10 );
        m_pCORBAServer->ActivateObject( params );
        obj = m_Factory->New (params->_this());
        if( CORBA::is_nil(obj) )
        {
            throw CyrException("New Spawned process failed",where,-1);
        }
        _SpawnProcess = SpawnProcess::_narrow(obj); 
        }
    catch(CORBA::Exception&e)
    {
        const char* err=CorbaExceptionString(e);
        throw CyrException(err,where,-1);
    }
    params->_remove_ref();  
    try
    {
        //    ...and create the spawned process
        _PID = _SpawnProcess->StartProductImage((CORBA::Long)m_Target
                                                ,(CORBA::Long)m_Exe
                                                ,m_Product.c_str()
                                                ,m_Params.c_str());
        char* c_pid=new char[10];
        sprintf(c_pid,"_%8.8lX",_PID);
        m_ExpectedName+=c_pid;
        delete c_pid;
    }
    catch(const CyrSpawnException &e)
    {
        throw CyrException((const char*)e.what,(const char*)e.where,e.status);
    }
    catch(CORBA::Exception&e)
    {
        const char* err=CorbaExceptionString(e);
        throw CyrException(err,where,-1);
    }
    catch(...)
    {
        throw CyrException("Unhandled exception",where,-1);
    }
}
void CECSpawnObject::AttachToObject()
{
    const char* where="CECSpawnObject::AttachToObject";
    bool    Connected = false;
    int     RetryAttempts = 0;
    CORBA::Object_var obj;
    while ((!Connected) && (RetryAttempts < 60))
    {
        CORBA::String_var classname=m_ClassName.c_str();
        CORBA::String_var exname=m_ExpectedName.c_str();
        try
        {
            CORBA::Object_ptr obj;
            obj = m_pCORBAServer->FindObjectWithName(m_ClassName.c_str(),m_ExpectedName.c_str());   

            if( CORBA::is_nil(obj) )
            {
                string err="Unable to find " + m_ClassName + " running registered as " + m_ExpectedName + " on Node " + m_NodeDefinition.GetNodeIPAddrAsString();
                throw CyrException(err.c_str(),where,-1);
            }
            m_Object=obj;
            Connected = true;
        }
        catch (CyrCORBAException)
        {
            CyrSleep(1000);
            RetryAttempts++;
        }
    }
    if(RetryAttempts>=60)throw CyrException(string("Unable to connect to " + m_ExpectedName).c_str() ,where,-1);
}
