/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     TestExecuterImpl
// @module  TestExecuterImpl |
//
///////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
#   pragma warning(disable:4786)
#   pragma warning(disable:4503)
#endif

#ifdef TH_POOL_VER
#include <Winsock2.h>
#endif
#include <stdio.h>
#ifdef TH_POOL_VER
#include <ctime>
#else
#include <ctime.hpp>
#include <cassert>
#endif

#include "TECORBARepository.h"
#include "TestExecuter_i.h"
#include "TPLog.h"

#ifdef TH_POOL_VER
#   include "AsyncVirtualUserController.h"
extern CAsyncVirtualUserController* _threadPoolController;
#else
#   include "TEthreadcontrol.h"
#endif

extern "C"
{
#   include <psl_macros.h>
#   include <psl_rtndef.h>
#   include <target_rtndef.h>
}

#ifdef TH_POOL_VER
#   include "TScriptRule.h"
#   include "mutex.hpp"    // Provides CTSemaphore.
#endif

CTSemaphore blocking_semaphore(0);
extern string exedir;
CORBAServer*    pCServer=NULL;


/////////////////////////////////////////////////////////////////////////////
// TestExecuterImpl  Constructor
///////////////////////////////////////////////////////////////////////////////
// @mfunc Constructor
#ifndef TH_POOL_VER
TestExecuterImpl::TestExecuterImpl(CORBAServer* pserver,string &regname,const string& node,int pid,const string& exename,TScriptRule* rule,int max_threads)
#else
TestExecuterImpl::TestExecuterImpl(CORBAServer* pserver,string &regname,const string& node,int pid,const string& exename)
#endif
        :
        m_pServer(pserver)
        ,m_RegisterName(regname)
        ,m_nodename(node)
        ,m_pid(pid)
        ,m_exename(exename)
#ifndef CYRANO_TP_STANDALONE
        ,m_ParentTestLevel(TestLevel::_nil())
#endif
        ,m_closing_down(false)
{
    pCServer=pserver;
#ifndef CYRANO_TP_STANDALONE
    m_pServer->ActivateObject(this);
    m_pServer->NameObject(this->_this(),TestExecuter::CLASS_NAME,m_RegisterName.c_str());
#endif
#ifndef TH_POOL_VER
    m_threadcontrol = new TEthreadcontrol(rule,max_threads);
#endif

    // Global CORBA Respository class
    Repository = NULL;//Make sure is null if not used
}

/////////////////////////////////////////////////////////////////////////////
// TestExecuterImpl  Destructor
///////////////////////////////////////////////////////////////////////////////
// @mfunc Destructor

TestExecuterImpl::~TestExecuterImpl	()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc CloseDown , clear out member data in case Destructor never gets called in abnormal closedown

void TestExecuterImpl::CloseDown()
{
    //first close down threads, these may still be writing to repository
    if(logr) logr->AddTrace(SHUTDOWN,DEBUGLEVEL,"Close down all threads");

#ifndef TH_POOL_VER
    delete m_threadcontrol;//close down the threading capabilities
#endif

    if(logr) logr->AddTrace(SHUTDOWN,DEBUGLEVEL,"Start clear down of Repository");
    if (Repository)
        delete Repository;
    // delete the CORBAinterface pointer
    // dispose of the m_TestExecuterStatus Obj

#ifndef CYRANO_TP_STANDALONE
    //Finished shutting doewn , inform TestLevel that we are complete
    //logr->AddTrace(SHUTDOWN,DEBUGLEVEL,"Inform TestManager of completion");
    if(!CORBA::is_nil(m_ParentTestLevel))
    {
        m_ParentTestLevel->TestExecuterFinished();
    }
    try
    {
        m_pServer->UnNameObject(TestExecuter::CLASS_NAME,m_RegisterName.c_str());
    }
    catch(...)
    {
        //if we can't de-register probably means either everything
        //has gone anyway or we are already de-registered
    }

#endif
}

CORBA::Long  TestExecuterImpl::Initialise(const char*  node,const char*  user, const char*  password, const char* TestRunString)
{
    m_TestRunString=TestRunString;//store for use later
    //Create New repository
    try
    {
        Repository = new TECORBARepository(m_pServer,node,user,password,m_nodename.c_str(),m_pid,m_exename);
        assert(Repository);
        blocking_semaphore.unlock();//release semaphore blocking main thread
    }

    catch(CyrRepositoryException&e)
    {
        Stop();
        throw CyrProcessException(e.what().c_str(),e.where().c_str(),e.status());
    }
    catch(exception&e)
    {
        Stop();
        throw CyrProcessException(e.what(),"TestExecuterImpl::Initialise",-1);
    }
    catch(...)
    {
        Stop();
        throw CyrProcessException("UNKNOWN","TestExecuterImpl::Initialise",-1);
    }
    return 0;
}



/////////////////////////////////////////////////////////////////////////////
// ExecuteThread
///////////////////////////////////////////////////////////////////////////////
// @mfunc
// @param
// @param
// @param
// @rdesc long
long		TestExecuterImpl::ExecuteThread(const char*  ProjectName
                                      ,const char*  ScriptName
#ifndef CYRANO_TP_STANDALONE
                                      ,TestLevel_ptr & Parent
                                      ,const ThreadDef & thrdef
#endif
                                     )
{
#ifndef CYRANO_TP_STANDALONE
    if(m_closing_down)//if this is closing down then should not allow any more
        throw CyrClosingDownException();
    //if first call store test level pointer
    if(CORBA::is_nil(m_ParentTestLevel))
    {
        if(!CORBA::is_nil(Parent))
            m_ParentTestLevel=TestLevel::_duplicate(Parent);
    }
    m_ThreadDef=thrdef;//doesn't matter that this is continually overwritten as they're all the same
#endif

    // msg to identify the start of the execute thread routine.
    if(logr) logr->AddTrace(STARTUP,ERRORLEVEL, (string("Execute Thread, ProjectName: ") +ProjectName +" ScriptName: " +ScriptName).c_str());

    static bool have_test_run=false;
    if(!have_test_run)
    {
        if(logr) logr->AddTrace(STARTUP,DEBUGLEVEL,"Connect to TestRun");
        try
        {
            long status=Repository->Connect(ProjectName, ScriptName,m_TestRunString);
            have_test_run=true;
            if (status == -1)
            {
                // Get Test has failed, set error message and throw exception
                CORBA::String_var where="TestExecuterImpl::ExecuteThread";
                CORBA::String_var what= "GetTest Failed";
#ifndef CYRANO_TP_STANDALONE
                throw CyrProcessException(what,where,-1);
#else
                throw CyrProcessException(what.c_str(),where.c_str(),-1);
#endif
            }
        }
        catch(CyrRepositoryException&e)
        {
            CORBA::String_var what=e.what().c_str();
            CORBA::String_var where=e.where().c_str();
#ifndef CYRANO_TP_STANDALONE
            throw CyrProcessException(what,where,e.status());
#else
            throw CyrProcessException(what.c_str(),where.c_str(),e.status());
#endif
        }
        catch(exception&e)
        {
            CORBA::String_var what=e.what();
            CORBA::String_var where="TestExecuterImpl::ExecuteThread";
#ifndef CYRANO_TP_STANDALONE
            throw CyrProcessException(what,where,-1);
#else
            throw CyrProcessException(what.c_str(),where.c_str(),-1);
#endif
        }
        catch(...)
        {
            CORBA::String_var what="UNKNOWN";
            CORBA::String_var where="TestExecuterImpl::ExecuteThread";
#ifndef CYRANO_TP_STANDALONE
            throw CyrProcessException(what,where,-1);
#else
            throw CyrProcessException(what.c_str(),where.c_str(),-1);
#endif
        }
    }

#ifndef TH_POOL_VER    // In the case of the Thread pool, this will be done by CreateTask.

    // Access the repository and retrieve the TOF test description file
    // by copying it to local disk space. The full path / location of
    // the TOF file is returned.
    string FullScriptName;
    string tmpstring=exedir + FN_PATHSEP + ScriptName + ".tof";

    bool is_file_there=m_threadcontrol->doesTaskExist(tmpstring);
    if(!is_file_there)
    {
        string objnam = string(ScriptName) + ".tof";
        if(logr) logr->AddTrace(DBGSTD,DEBUGLEVEL,(string("Get the File: ") + objnam).c_str());
        try
        {
            FullScriptName = Repository->GetTofFile(objnam);
        }
        catch(...)
        {
            CORBA::String_var what=(string("Unable to retrieve File: ")+ScriptName).c_str();
            CORBA::String_var where="TestExecuterImpl::ExecuteThread";
#   ifndef CYRANO_TP_STANDALONE
            throw CyrProcessException(what,where,-1);
#   else
            throw CyrProcessException(what.c_str(),where.c_str(),-1);
#   endif
        }
    }
    else
    {
        FullScriptName = tmpstring;
    }

    if(logr) logr->AddTrace(DBGSTD,DEBUGLEVEL,string("Using File: "+FullScriptName).c_str());

    if (FullScriptName.size() == 0)
    {
        // check in case nothing is returned.
        CORBA::String_var where="TestExecuterImpl::ExecuteThread";
        CORBA::String_var what= "GetTofFile Failed";
#   ifndef CYRANO_TP_STANDALONE
        throw CyrProcessException(what,where,-1);
#   else
        throw CyrProcessException(what.c_str(),where.c_str(),-1);
#   endif
    }

#endif  // ndef TH_POOL_VER

    // Attach Thread, calls the AttachThread in mdbext , after passing through
    // the threadcontrol class (tc)
    int status=0;
    int threads_attached=0;
    try
    {
        if(logr) logr->AddTrace(DBGSTD,DEBUGLEVEL,"Attach New Thread");

#ifndef TH_POOL_VER

        status = m_threadcontrol->AttachThread(FullScriptName.c_str()
#   ifndef CYRANO_TP_STANDALONE
                                               ,Parent,thrdef
#   endif
                                              );

#else
        status = _threadPoolController->StartNewUsersGroup(1, ScriptName
#   ifndef CYRANO_TP_STANDALONE
                                                             , Parent, thrdef
#   endif
                                                          );
#endif  // ndef TH_POOL_VER

        if (status == -1)
        {
            if(logr) logr->AddTrace(STARTUP,ERRORLEVEL,"Execute Thread Completed FAILED");
        }
        else
        {
            if(logr) logr->AddTrace(STARTUP,ERRORLEVEL,"Returning from ExecuteThread");
        }
    }
#ifndef CYRANO_TP_STANDALONE
    //If cannot add new thread throw exception showning how many are left to do
    catch( const CyrNoFreeThreadException& )
    {
        throw CyrNoFreeThreadException();
    }
#endif
#ifndef TH_POOL_VER
    catch(mdbIOException &e)
    {
        if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, (e.what() + e.ioname()).c_str());
        throw;
    }
#endif
    catch(...)
    {
        throw CyrProcessException("Cannot attach new thread","TestExecuterImpl::ExecuteThread",-1);
    }
    return status;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Stop this Test Executer - tidily.
//
void TestExecuterImpl::Stop()
{
    //deregister self, this stops other people trying to get hold of object
    //and also should get around messing shutdown procedures
#ifndef CYRANO_TP_STANDALONE
    if(!CORBA::is_nil(m_ParentTestLevel))
    {
        m_ParentTestLevel->TestExecuterFinished();
    }
    m_pServer->UnNameObject(TestExecuter::CLASS_NAME,m_RegisterName.c_str());
#endif
    if(logr) logr->AddTrace(SHUTDOWN,ERRORLEVEL,"Stop invoked for Test Executer" );

#ifndef TH_POOL_VER
    //release sempahore blocking main thread
    blocking_semaphore.unlock();
    //release sempahore blocking main thread again in case
    //this is very early in execution (semaphore is acquired maximum of twice)
    blocking_semaphore.unlock();
#else
    // Stop the thread pool.
    _threadPoolController->CancelAllUsers();
    //    _threadPoolController->Stop();
#endif
    return;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Kill this Test Executer.
//
//          Note that no return is made from this routine - an explit exit is
//          executed instead.
//
//          The Stop() method should be used in order to perform an orderly
//          shutdown. This method should be used for emergency shutdowns only.
//
void TestExecuterImpl::Kill()
{
    if (logr) logr->AddTrace(SHUTDOWN, ERRORLEVEL, "Kill invoked for Test Executer");
    if (logr) logr->AddTrace(SHUTDOWN, ERRORLEVEL, "Kill for Test Executer complete - about to exit process");
    exit(0);
}

#ifndef CYRANO_TP_STANDALONE
///////////////////////////////////////////////////////////////////////////////
// @mfunc Get the node that is currently being run on
// @rdesc Node String
char*  TestExecuterImpl::GetNode()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::string_dup(m_nodename.c_str());
}

char*  TestExecuterImpl::GetNodeName()
{
    if(m_closing_down)throw CyrClosingDownException();
    CECInetInfo inet;
    return CORBA::string_dup(inet.GetHostName().c_str());
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc Get Name that binary is registered with nameservice as
// @rdesc Registered Name
char*  TestExecuterImpl::Name()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::string_dup(m_RegisterName.c_str());
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc Get pipe separated list of all unique identifiers for Threads in this Test executer
// @rdesc List
char*  TestExecuterImpl::GetThreadIdents()
{
    if(m_closing_down)
        throw CyrClosingDownException();

    string ilist("");

#ifndef TH_POOL_VER
    Cmdbthread* tp;

    for (int i=0; i<m_threadcontrol->get_nactive(); i++)
    {
        // Get pointer to the thread in this iteration
        tp= m_threadcontrol->get_threadp(i);
        // If that pointer is not NULL add Ident to list
        // before continuing
        if (tp != NULL)
        {
            ilist+=tp->GetThreadID()+"|";
        }

    }
    ilist+="|";
#else
    _threadPoolController->GetThreadIdents(ilist);
#endif

    return CORBA::string_dup(ilist.c_str());
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc Get How many threads are running (optional scriptname)
// @parm  name of script to query
// @rdesc no of threads running script
CORBA::ULong  TestExecuterImpl::GetNoThreads(char* & ScriptName)
{
    if(m_closing_down)throw CyrClosingDownException();

    string sname = ScriptName;
    if(sname.empty())
    {
        return GetNoActiveThreads();
    }
    else
    {
#ifndef TH_POOL_VER
        CORBA::ULong no_threads=0;

        string cur_name("");
        Cmdbthread* tp;
        for (int i=0; i<m_threadcontrol->get_nactive(); i++)
        {
            // Get pointer to the thread in this iteration
            tp= m_threadcontrol->get_threadp(i);
            // If that pointer is not NULL add Ident to list
            // before continuing
            if (tp != NULL)
            {
                cur_name = tp->ScriptName();
                if(cur_name==sname)no_threads++;
            }
        }
        return no_threads;
#else
        return (CORBA::ULong) _threadPoolController->GetMatchingThreadsCount(sname);
#endif

    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc return Total number of threads running
// @rdesc no of threads running
CORBA::ULong  TestExecuterImpl::GetNoActiveThreads()
{
    if(m_closing_down)throw CyrClosingDownException();

#ifndef TH_POOL_VER
    return m_threadcontrol->ThreadMap.size();
#else
    return (CORBA::ULong) _threadPoolController->GetUsersCount();
#endif
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc return thread status based on unique ID
// @parm  unique thread identifier
// @rdesc status of Thread (ACTIVE,SUSPENDED)
ThreadStatus TestExecuterImpl::GetStatus(char* & ident)
{
    if(m_closing_down)throw CyrClosingDownException();

#ifndef TH_POOL_VER
    return (*(m_threadcontrol->ThreadMap.find(ident))).second.Status;
#else
if(_threadPoolController->GetVuStatus(ident) == EXEC_OVER)
    return SUSPENDED;
    else
        return ACTIVE;
#endif
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc Get Target for this executer
// @rdesc Target
char*  TestExecuterImpl::GetTarget()
{
    if(m_closing_down)throw CyrClosingDownException();
    string Target;
    char* c_target=new char[32];
    int	stat=Tgt_GetTarget((TARGETID)m_ThreadDef.Target,c_target,32,PTRUE);
    Target=c_target;
    delete c_target;
    return CORBA::string_dup(Target.c_str());
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc Get TestName for this executer
// @rdesc Test Name
char*  TestExecuterImpl::TestName()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::string_dup(m_ParentTestLevel->TestName());
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc suspend thread status based on unique ID
// @parm  unique thread identifier
// @rdesc true =  success
CORBA::Boolean  TestExecuterImpl::Suspend(char* & ident)
{
    return false;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc Activate thread status based on unique ID
// @parm  unique thread identifier
// @rdesc true =  success
CORBA::Boolean  TestExecuterImpl::Activate(char* & ident)
{
    return false;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc Detatch thread status based on unique ID
// @parm  unique thread identifier
// @rdesc true =  success
CORBA::Boolean  TestExecuterImpl::Detatch(char* & ident)
{
    return false;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc return thread current scriptname based on unique ID
// @parm  unique thread identifier
// @rdesc name of script
char*  TestExecuterImpl::ScriptName(char* & ident)
{
#ifndef TH_POOL_VER
    if(m_closing_down)throw CyrClosingDownException();
    return (*(m_threadcontrol->ThreadMap.find(ident))).second.ScriptName;
#else
    return (char*)_threadPoolController->GetVuScriptName(ident).c_str();
#endif
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc return thread start time based on unique ID
// @parm  unique thread identifier
// @rdesc Time
CORBA::ULong  TestExecuterImpl::StartTime(char* & ident)
{
    if(m_closing_down)throw CyrClosingDownException();
#ifndef TH_POOL_VER
    return (*(m_threadcontrol->ThreadMap.find(ident))).second.StartTime;
#else
return _threadPoolController->GetVuStartTime(ident);
#endif
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc return thread note text based on unique ID
// @parm  unique thread identifier
// @rdesc text
char*  TestExecuterImpl::GetNoteText(char* & ident)
{
#ifndef TH_POOL_VER
    if(m_closing_down)throw CyrClosingDownException();
    return (*(m_threadcontrol->ThreadMap.find(ident))).second.NoteText;
#else
//TODO
    assert(0);
    return NULL;
#endif
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc return number of testcases passed /failed
// @parm  passed/FAILED
// @rdesc number with status
CORBA::Long  TestExecuterImpl::GetTestCaseResults(TestCaseStatus  Status)
{
    if(m_closing_down)
        throw CyrClosingDownException();

#ifndef TH_POOL_VER
    Cmdbthread* tp;
    TEThread* tetp;
    CORBA::Long no_results=0;
    for (int i=0; i<m_threadcontrol->get_nactive(); i++)
    {
        // Get pointer to the thread in this iteration
        tp= m_threadcontrol->get_threadp(i);
        // If that pointer is not NULL add no of test cases with status
        if (tp != NULL)
        {
            tetp=static_cast<TEThread*>(tp);
            if(tetp)no_results+=tetp->GetTestCaseResults(Status);
        }
    }
    return no_results;
#else

return _threadPoolController->GetTestCaseResults(Status);

#endif
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc Retunr Information of ALL threads currently running
// @parm  ThreadInfoList to be populated
// @rdesc number to be found in populated list
CORBA::ULong  TestExecuterImpl::GetThreadInformation(ThreadInfoList & TIList)
{
    if(m_closing_down)
        throw CyrClosingDownException();

#ifndef TH_POOL_VER
    TIList.length(m_threadcontrol->ThreadMap.size());
    ThreadMap_t::iterator i;
    int j;
    for(i=m_threadcontrol->ThreadMap.begin(),j=0;i!=m_threadcontrol->ThreadMap.end();i++,j++)
    {
        TIList[j]=(*i).second;
    }
    return TIList.length();
#else
// Create the data structure.
    return _threadPoolController->GetThreadInformation(TIList);
#endif
}
#endif //#ifndef CYRANO_TP_STANDALONE

bool TestExecuterImpl::ReportOnThreads()
{
#ifndef TH_POOL_VER
    return m_threadcontrol->ReportOnThreads();
#else
    //TODO
    return TRUE;
#endif
}
