/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/


///////////////////////////////////////////////////////////////////////////////
// @doc     TestManagerImpl
// @module  TestManagerImpl |
//
///////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
#include "TPLog.h"
#include "TestManager_i.h"
#include "TMCORBARepository.h"
#include "TMthreadcontrol.h"
#include <strstream>
CORBAServer*    pCServer=NULL;
static bool CanExecuteThread=true;
#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
CTSemaphore blocking_semaphore(0);
/////////////////////////////////////////////////////////////////////////////
// TestManagerImpl  Constructor
///////////////////////////////////////////////////////////////////////////////
// @mfunc Constructor
TestManagerImpl::TestManagerImpl(CORBAServer* pserver,string &regname,string node,int pid,string exename,TScriptRule* rule)
        :m_corba(pserver),
        m_RegisterName(regname),
        m_nodename(node),
        m_pid(pid),
        m_exename(exename),
        m_TMVarFactory(NULL),
        m_closing_down(false)
{
    CTTime tm;
    tm.now();
    m_StartTime=tm.getSeconds();

    pCServer=pserver;
    m_corba->ActivateObject(this);
    m_corba->NameObject(this->_this(),TestManager::CLASS_NAME,m_RegisterName.c_str());
    m_threadcontrol=new TMthreadcontrol(rule);

    // Global CORBA Respository class
    Repository = NULL;//Make sure is null if not used
}
/////////////////////////////////////////////////////////////////////////////
// TestManagerImpl  Destructor
///////////////////////////////////////////////////////////////////////////////
// @mfunc Destructor

TestManagerImpl::~TestManagerImpl()
{
}
///////////////////////////////////////////////////////////////////////////////
//@mfunc clear up member data

void TestManagerImpl::CloseDown()
{
    delete m_threadcontrol;//close down the threading capabilities
    logr->AddTrace(SHUTDOWN,DEBUGLEVEL,"Dispose of Global Variable Factory");
    if(m_TMVarFactory)
        m_TMVarFactory->_remove_ref();      
    logr->AddTrace(SHUTDOWN,DEBUGLEVEL,"Delete of Repository Information");
    if (Repository)
        delete Repository;
    //Deregister TestManager in an attempt to clean up, no more logging unless there is an error
    try
    {
        m_corba->UnNameObject(TestManager::CLASS_NAME,m_RegisterName.c_str());  
    }
    catch(...)
    {
        //if we can't de-register probably means either everything
        //has gone anyway or we are already de-registered
    }

}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Set up information for logging onto repository
// @parm    nodename
// @parm    username
// @parm    password
// @-parm    TestRun string of parent TestRuns(if threre is one)
// @rdesc   0 for success
CORBA::Long  TestManagerImpl::Initialise(const char*  node, const char*  user, const char*  password, const char*  TestRunString)
{
    m_User=user;
    m_TestRunString=TestRunString;//store for use later
    //Create New repository
    try
    {
        logr->AddTrace(STARTUP,DEBUGLEVEL,string("Get Node Information for "+string(node)).c_str());
        Repository = new TMCORBARepository(m_corba,node,user,password,m_nodename,m_pid,m_exename);
        assert(Repository);
        logr->AddTrace(STARTUP,DEBUGLEVEL,"Repository Initiated");
    }
    catch(CyrRepositoryException&e)
    {
        strstream msg;
        msg<<e.what()<<","<<e.where()<<","<<e.status()<<ends;
        logr->AddTrace(STARTUP,ERRORLEVEL,msg.str());
        msg.freeze(false);
        Stop();//not much point in going any further
        throw CyrProcessException(e.what().c_str(),e.where().c_str(),e.status());
    }
    catch(exception&e)
    {
        logr->AddTrace(STARTUP,ERRORLEVEL,e.what());
        Stop();
        throw CyrProcessException(e.what(),"TestManagerImpl::Initialise",-1);
    }
    catch(...)
    {
        logr->AddTrace(STARTUP,ERRORLEVEL,"UNKOWN Exception in TestManagerImpl::Initialise, shutdown");
        Stop();
        throw CyrProcessException("UNKNOWN","TestManagerImpl::Initialise",-1);
    }
    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Call Stop() on all child Test Managers and TestLevels in order to
//          close them down gracefully.
//
//          This method should be used in order to perform an orderly shutdown.
//          The Kill() method should be used for emergency shutdowns only.
//
void TestManagerImpl::Stop()
{
    // Deregister this Test Manager object. This prevents other processes from
    // accessing this object and also should get around messing shutdown procedures
    m_corba->UnNameObject(TestManager::CLASS_NAME, m_RegisterName.c_str());
    logr->AddTrace(SHUTDOWN,ERRORLEVEL,"Stop invoked for Parent Test Manager");

    // Call Stop() on all child Test Managers.
    TestManagerList_t::iterator i;
    for(i=m_TestManagerList.begin();i!=m_TestManagerList.end();i++)
    {
        logr->AddTrace(SHUTDOWN,ERRORLEVEL,"Stop invoked for Child Test Manager");
        (*i)->Stop();
    }

    // Call Stop() on each TestLevel - this should cause Stop() to be called on
    // all child Test Executers.
    if(!(m_TestLevel->_non_existent()))
    {
        for(int i = 0; m_threadcontrol->get_threadp(i) != NULL; ++i)
        {
            TMThread* pTM = (TMThread*)m_threadcontrol->get_threadp(i);
            TestLevelImpl* pTLI = pTM->GetTestLevelImpl();
            if(pTLI)
                pTLI->Stop();
        }
    }

    // Release sempahore blocking main thread.
    logr->AddTrace(SHUTDOWN,ERRORLEVEL,"Release main Test Manager thread");
    blocking_semaphore.unlock();
    blocking_semaphore.unlock();

    return;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Call Kill() on all child Test Managers and TestLevels in order to
//          close them down in an emergency.
//
//          Note that no return is made from this routine - an explit exit is
//          executed instead.
//
//          The Stop() method should be used in order to perform an orderly
//          shutdown. This method should be used for emergency shutdowns only.
//
void TestManagerImpl::Kill()
{
    logr->AddTrace(SHUTDOWN, ERRORLEVEL, "Kill invoked for Parent Test Manager");

    // Deregister this Test Manager object.
    try
    {
        m_corba->UnNameObject(TestManager::CLASS_NAME, m_RegisterName.c_str());
    }
    catch (...)
    {
        // Do nothing.
    }

    // Call Kill() on all child Test Managers.
    TestManagerList_t::iterator i;
    for(i=m_TestManagerList.begin();i!=m_TestManagerList.end();i++)
    {
        logr->AddTrace(SHUTDOWN, ERRORLEVEL, "Kill invoked for Child Test Manager");
        (*i)->Kill();
    }

    // Call Kill() on each TestLevel - should cause Kill() to be called on all child Test Executers.
    if(!(m_TestLevel->_non_existent()))
    {
        for(int i = 0; m_threadcontrol->get_threadp(i) != NULL; ++i)
        {
            TMThread* pTM = (TMThread*)m_threadcontrol->get_threadp(i);
            TestLevelImpl* pTLI = pTM->GetTestLevelImpl();
            if(pTLI)
                pTLI->Kill();
        }
    }

    logr->AddTrace(SHUTDOWN, ERRORLEVEL, "Kill for Test Manager complete - about to exit process");
    exit(0);
}

char*  TestManagerImpl::GetNode()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::string_dup(m_nodename.c_str());
}
char*  TestManagerImpl::GetNodeName()
{
    if(m_closing_down)throw CyrClosingDownException();
    CECInetInfo inet;
    return CORBA::string_dup(inet.GetHostName().c_str());
}
/////////////////////////////////////////////////////////////////////////////
// ExecuteThread
///////////////////////////////////////////////////////////////////////////////
// @mfunc
// @param
// @param
// @param
// @rdesc long
CORBA::Long  TestManagerImpl::ExecuteThread(const char*  ProjectName, const char*  TestName, TestLevel_ptr & Parent)
{
    if(!CanExecuteThread)
        return 0;//if this has been called then a test manager should not allow
    //this to be called again
    // msg to identify the start of the execute thread routine.
    string msg=string("Execute Thread, ProjectName: ") +ProjectName +" TestName: " +TestName ;
    logr->AddTrace(STARTUP,ERRORLEVEL,msg.c_str());

    try
    {
        string strProjectName=ProjectName;
        string strTestName=TestName;
        long status=Repository->Connect(strProjectName, strTestName,m_TestRunString);
        if (status == -1)
        {
            // Get Test has failed, set error message and throw exception
            CORBA::String_var where="TestManagerImpl::ExecuteThread";
            CORBA::String_var what= "GetTest Failed";
            Stop();
            throw CyrProcessException(what,where,-1);
        }
        m_TMVarFactory=new TMVarFactoryImpl(m_corba,Repository->GetTestRunString());
    }
    catch(CyrRepositoryException&e)
    {
        CORBA::String_var what=e.what().c_str();
        CORBA::String_var where=e.where().c_str();
        Stop();
        throw CyrProcessException(what,where,e.status());
    }
    catch(exception&e)
    {
        CORBA::String_var what=e.what();
        CORBA::String_var where="TestManagerImpl::ExecuteThread";
        Stop();
        throw CyrProcessException(what,where,-1);
    }
    catch(...)
    {
        CORBA::String_var what="UNKNOWN";
        CORBA::String_var where="TestManagerImpl::ExecuteThread";
        Stop();
        throw CyrProcessException(what,where,-1);
    }


    // Access the repository and retrieve the TOF test description file
    // by copying it to local disk space. The full path / location of
    // the TOF file is returned.
    string FullTestName;
    try
    {
        FullTestName = Repository->GetTofFile(string(TestName) + ".tso");
    }
    catch(...)
    {
        CORBA::String_var what=(string("Unable to retrieve File: ")+TestName).c_str();
        CORBA::String_var where="TestManagerImpl::ExecuteThread";
        Stop();
        throw CyrProcessException(what,where,-1);
    }

    if (FullTestName.size() == 0)
    {
        // check in case nothing is returned.
        CORBA::String_var where="TestManagerImpl::ExecuteThread";
        CORBA::String_var what= "GetTestFile Failed";
        Stop();
        throw CyrProcessException(what,where,-1);
    }


    // Attach Thread, calls the AttachThread in mdbext , after passing through
    // the threadcontrol class (tc)
    int status=0;
    try
    {
        status = m_threadcontrol->AttachThread(FullTestName.c_str(),Parent,this->_this());

        if (status == -1)
        {
            logr->AddTrace(STARTUP,ERRORLEVEL,"Execute Thread Completed FAILED");
            Stop();//not much point in going any further
        }

        else
        {
            logr->AddTrace(STARTUP,ERRORLEVEL,"Returning from ExecuteThread");
        }
    }
    catch(exception&e)
    {
        CORBA::String_var what=e.what();
        CORBA::String_var where="TestManagerImpl::ExecuteThread,Attach new Thread";
        Stop();
        throw CyrProcessException(what,where,-1);
    }
    catch(...)
    {
        CORBA::String_var what="UNKNOWN";
        CORBA::String_var where="TestManagerImpl::ExecuteThread,Attach new Thread";
        Stop();
        throw CyrProcessException(what,where,-1);
    }
    CanExecuteThread=false;
    blocking_semaphore.unlock();//release semaphore blocking main thread
    return status;
}
void TestManagerImpl::AddSubTestManager(TestManager_ptr & subtest)
{
    m_TestManagerList.push_back(subtest);
}
void TestManagerImpl::RemoveSubTestManager(TestManager_ptr & subtest)
{
    TestManager_var TM_var=subtest;
    TestManagerList_t::iterator i;
    for(i=m_TestManagerList.begin();i!=m_TestManagerList.end();i++)
    {
        if( *i == TM_var )
        {
            m_TestManagerList.erase(i);
            break;
        }
    }
}
CORBA::ULong  TestManagerImpl::GetTestManagerList(TestManagerList & TMlist)
{
    if(m_closing_down)throw CyrClosingDownException();
    TMlist.length(m_TestManagerList.size());
    TestManagerList_t::iterator i;
    int j;
    for(i=m_TestManagerList.begin(),j=0;i!=m_TestManagerList.end();i++,j++)
    {
        TMlist[j]=TestManager::_duplicate(*i);
    }
    return TMlist.length();
}
CORBA::Object_ptr  TestManagerImpl::GetTestRunAsObject()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::Object::_duplicate(Repository->GetTestRun()->GetServant());
}
char*  TestManagerImpl::Name()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::string_dup(m_RegisterName.c_str());
}
CORBA::Long  TestManagerImpl::GetPID()
{
    if(m_closing_down)throw CyrClosingDownException();
    return m_pid;
}
CORBA::Long  TestManagerImpl::GetTotalTestCasePasses()
{
    static CORBA::Long last_results;
    if(m_closing_down)throw CyrClosingDownException();
    try
    {
        if(!(m_TestLevel->_non_existent()))
        {
            if(!CORBA::is_nil(m_TestLevel))
            {
                last_results=m_TestLevel->GetTestCaseResults(PASSED,true);
            }
        }
    }
    catch(const CyrClosingDownException&)//if test level closing down use last known result
    {}

    return last_results;
}
CORBA::Long  TestManagerImpl::GetTotalTestCaseFails()
{
    static CORBA::Long last_results;
    if(m_closing_down)throw CyrClosingDownException();
    try
    {
        if(!(m_TestLevel->_non_existent()))
        {
            if(!CORBA::is_nil(m_TestLevel))
            {
                return m_TestLevel->GetTestCaseResults(FAILED,true);
            }
        }
    }
    catch(const CyrClosingDownException&)//if test level closing down use last known result
    {}

    return last_results;
}
CORBA::Long  TestManagerImpl::GetNoThreads()
{
    if(m_closing_down)throw CyrClosingDownException();
    CORBA::Long  no_threads=0;
    try
    {
        if(!(m_TestLevel->_non_existent()))
        {
            if(!CORBA::is_nil(m_TestLevel))
            {
                no_threads=m_TestLevel->NoChildThreads(true);
            }
        }
    }
    catch(const CyrClosingDownException&)//if closing down can have no threads
    {
        no_threads=0;
    }
    return no_threads;
}
void TestManagerImpl::SetTestLevel(TestLevel_ptr & TL)
{
    m_TestLevel=TestLevel::_duplicate(TL);
}
TestLevel_ptr  TestManagerImpl::GetTestLevel()
{
    if(m_closing_down)throw CyrClosingDownException();
    return TestLevel::_duplicate(m_TestLevel);
}
char*  TestManagerImpl::GetUser()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::string_dup(m_User.c_str());
}
CORBA::ULong  TestManagerImpl::StartTime()
{
    if(m_closing_down)throw CyrClosingDownException();
    return m_StartTime;
}
//wait for completion of threads
bool TestManagerImpl::WaitForCompletion()
{
    return m_threadcontrol->WaitForCompletion();
}
void TestManagerImpl::SetToCloseDown()//Set flag to say closing down is imminent
{
    m_closing_down=true;
    Repository->GetTestRun()->StopTestRun();
    CloseDown();
}
