/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//    CECCommandProcess.cpp                                                 //
//    =====================                                                 //
//                                                                          //
//    This is the implementation file for the CECCommandProcess Class.      //
//                                                                          //
//                                                                          //
//    This class supports the execution of an image or DOSCommand with the  //
//    option of sending the data from the command to a window               //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

// PSDiagnostics.cpp : implementation file
//
#include "stdafx.h"
#include "CyrTstShrRes.h"
#include "CECTrace.h"

#include <filesys.h>
#include <winsock.h>
#include <io.h>

//
//    Class definition files.
//
#include "CECCommandProcess.h"
#include "UserMessages.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define BUFSIZE 4096

/////////////////////////////////////////////////////////////////////////////
// CECCommandProcess

//
//    Default constructor.
//    @mfunc Constructor
//
CECCommandProcess::CECCommandProcess    ()
{
    m_bClearOutputWindow = FALSE    ;
    m_bOutToAWindow = FALSE;
    m_pOutputWindow = NULL;
    m_RemoteChannel = 0;
    m_hChildStdoutWr = NULL;
    m_hChildStdoutRd = NULL;

    //  Initialize for Windows sockets - note that WSAStartup may be called
    //  more than once
    //
    WORD version_requested = MAKEWORD (1, 1);
    // Windows sockets version
    WSADATA wsadata;            // Returned data structure
    WSAStartup (version_requested, &wsadata);
}

//
// @mfunc Destructor
//
//    Destructor
//
CECCommandProcess::~CECCommandProcess()
{

    //    Tidy up our use of Windows Sockets to allow other DLL's to use the WS32 Dll
    //
    WSACleanup ();
}

//
// @mfunc RunImage
//        ========
//
//    This routine allows an image to be executed given the specified command line.
//
//    Calling Format:
//      status = RumImage (ImagePath);
//
//  @parm char* | ImagePath | Pointer to fully qualified specification of image to be executed.
//  @rdesc 0 if success, non-zero otherwise
//
int CECCommandProcess::RunImage (char *pImagePath)
{
    Trace ("CECCommandProcess::RunImage -> Executing Image " ,pImagePath);
    return GoMultiThread (pImagePath ,FALSE ,NULL);
}


//
// @mfunc ExecuteCommand
//        ==============
//
//    This routine allows a DOS Command to be executed, optionally sending the output from the
//    command to a CWnd derived object.
//
//    The parameter bAppendOutput is used to indicate whether the output window will be
//    appended (TRUE) or its previous contents erased (FALSE).
//
//    If a remote node name is specified then the command is executed on the specified remote node.
//    Note however that in this case we shall also have to pass across to this routine a
//    username and password which will be used to log onto the remote node.
//
//    Calling Format:
//      status = ExecuteCommand (Command
//                              ,OutputWindow
//                              ,ClearFirst
//                              ,RemoteNode
//                              ,RemoteUser
//                              ,RemotePassword
//
// @parm    char* | pCommandLine | Pointer to buffer for the command to execute
// @parm    CEdit* | pEditBox | Pointer to CEdit to which output from the command is sent
// @parm    BOOL | bClearOutput | Clear the CEdit buffer first or not
// @parm    CString* | pRemoteNodeName | Name of node to execute the command on. NULL is execute locally
// @parm    CString* | pRemoteUsername | Name of the user to log into the remote node as. NULL if local
// @parm    CString* | pRemotePassword | Password for the above user. NULL if local
// @parm    CString* | pRemoteDirectory | Directory to execute the command in
// @rdesc   0 if success, non-zero otherwise
//
int CECCommandProcess::ExecuteCommand   (char *pCommandLine
        ,CEdit* pEditBox
        ,BOOL bClearOutput
        ,CString* pRemoteNodeName
        ,CString* pRemoteUsername
        ,CString* pRemotePassword
        ,CString* pRemoteDirectory)
{
    Trace ("CECCommandProcess::ExecuteCommand  " ,pCommandLine);

    m_bClearOutputWindow = bClearOutput;
    m_pOutputWindow = pEditBox;
    m_bOutToAWindow = TRUE;
    return GoMultiThread (pCommandLine
                          ,TRUE
                          ,pRemoteNodeName
                          ,pRemoteUsername
                          ,pRemotePassword
                          ,pRemoteDirectory);
}


//
// @mfunc GoMultiThread
//        =============
//
//    This routine is called as a new thread to execute the specified DOSCommand and route its
//    output to the output window.
//
// @parm char* | pCommandLine | Pointer to command line to execute
// @parm BOOL  | DosCommand | TRUE if this command is to be executed as a DOS Command, FALSE otherwise
// @parm    CString* | pRemoteNodeName | Name of node to execute the command on. NULL is execute locally
// @parm    CString* | pRemoteUsername | Name of the user to log into the remote node as. NULL if local
// @parm    CString* | pRemotePassword | Password for the above user. NULL if local
// @parm    CString* | pRemoteDirectory | Directory to execute the command in
// @rdesc   0 if success, non-zero otherwise
//
int     CECCommandProcess::GoMultiThread    (char*    pCommandLine
        ,BOOL     DosCommand
        ,CString* pRemoteNodeName
        ,CString* pRemoteUsername
        ,CString* pPassword
        ,CString* pRemoteDirectory)

{

    //
    //    If we have specified a remote node then we may as well branch off and do our own thing
    //    as the rest of this code is specific to local command execution.
    //
    if (pRemoteNodeName != NULL)
        return ExecuteRemoteCommand (pCommandLine
                                     ,pRemoteNodeName
                                     ,pRemoteUsername
                                     ,pPassword
                                     ,pRemoteDirectory);

    //
    //    The command is being executed locally.
    //
    SECURITY_ATTRIBUTES saAttr;

    //
    // Set the bInheritHandle flag so pipe handles are inherited.
    //
    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = NULL;

    //
    // The steps for redirecting child's STDOUT:
    //     2.  Create anonymous pipe to be STDOUT for child.
    //     3.  Set STDOUT of parent to be write handle of pipe, so
    //         it is inherited by child.
    //

    //
    // Create a pipe for the child's STDOUT.
    //
    if (!CreatePipe (&m_hChildStdoutRd, &m_hChildStdoutWr, &saAttr, 0))
        ErrorExit("Stdout pipe creation failed\n");

    //
    // Now create the child process to execute the command
    //
    InvokeImage (pCommandLine ,DosCommand);

    //
    //  Read from pipe that is the standard output for child process.
    //
    ReadFromChild();

    return 0;
}

//
//    ReadFromChild
//    =============
//
//    This routine is called from a multi-threaded operation to read from the pipe
//    created to a child process which is executing a DOS command for us. We send
//    the output recovered to the output window
//
void    CECCommandProcess::ReadFromChild    ()
{
    CString TextBuffer;
    CString s;

    DWORD dwRead;
    CHAR chBuf[BUFSIZE];

    Trace ("CECCommandProcess::ReadFromChild -> Recovering output from the last command","");

    //
    // Close the write end of the pipe before reading from the
    // read end of the pipe.
    //
    if (!CloseHandle(m_hChildStdoutWr))
        ErrorExit("Closing handle failed");

    //
    // Read output from child, and write it to the output window edit box.
    //
    TextBuffer.Empty();
    for (;;)
    {
        if ((!ReadFile(m_hChildStdoutRd, chBuf, BUFSIZE, &dwRead, NULL)) || (dwRead == 0)) break;
        chBuf[dwRead] = '\0';
        TextBuffer += chBuf;
    }

    m_strOutput = TextBuffer;

    if (m_bOutToAWindow && (m_pOutputWindow != NULL) && (m_pOutputWindow->m_hWnd != NULL))
    {

        //
        // Output is being collated in a window
        //
        ::SendMessage (m_pOutputWindow->m_hWnd ,WMU_SETOUTPUTTEXT
                       ,(WPARAM) &TextBuffer
                       ,(LPARAM) m_bClearOutputWindow);
        Trace ("...Output Recovered > " ,TextBuffer.GetBuffer(0));
    }
}

void CECCommandProcess::ErrorExit (LPTSTR lpszMessage)
{
    fprintf(stderr, "%s\n", lpszMessage);
    ExitProcess(0);
}

//
//    CECCommandProcess::InvokeImage
//    ==============================
//
//    This routine is used to invoke the specified image. or execute the specified DOS Command
//
int CECCommandProcess::InvokeImage  (char*  pImageName ,BOOL DOSCommand)
{
    PROCESS_INFORMATION ProcInfo;
    CString     ErrorBuffer;
    STARTUPINFO StartInfo = {sizeof(StartInfo)};
    BOOL        InheritHandles = FALSE;
    int         CreateFlags = 0;

    Trace ("Invoking Image ", pImageName);

    //
    //    If we have requested that the window should be hidden then set up the STARTUPINFO
    //
    if (DOSCommand)
    {
        StartInfo.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
        StartInfo.hStdOutput = m_hChildStdoutWr;
        StartInfo.hStdError = m_hChildStdoutWr;
        StartInfo.wShowWindow = SW_HIDE;
        InheritHandles = TRUE;
        CreateFlags = CREATE_NEW_CONSOLE;
    }

    //
    //    Create a process to execute the specified image.
    //
    if (!CreateProcess (NULL
                        ,pImageName                      // Command Line
                        ,NULL                            // Process Security attributes
                        ,NULL                            // Primary thread security attributes
                        ,InheritHandles                  // Handles are {NOT} inherited
                        ,CreateFlags                     // Creation flags
                        ,NULL                            // Use Parents environment
                        ,NULL                            // Use Parents current directory
                        ,&StartInfo                      // STARTUPINFO pointer
                        ,&ProcInfo))                     // Receives Process Information
    {
        ErrorBuffer.Format ("Error invoking the image %s" ,pImageName);
        LPTSTR lpMsgBuf;

        FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                       NULL,
                       GetLastError(),
                       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                       (LPTSTR) &lpMsgBuf,
                       0,
                       NULL);

        //
        //    Output the error message via a message box if we have requested the child process output
        //    be directed to a window, If we have requested that child process output be sent to a string
        //    then the error should be logged into that string
        //
        if ( m_bOutToAWindow == TRUE )
        {
            AfxMessageBox (lpMsgBuf);
        }
        else if ( m_bOutToAWindow == FALSE)
        {
            m_strOutput = lpMsgBuf;
        }


        //
        //    Free the buffer.
        //
        LocalFree( lpMsgBuf );
        return -1;

    }
    return 0;
}

//
//    ExecuteRemoteCommand
//    ====================
//
//    This routine is called to execute a command on a remote node via REXEC.
//
int CECCommandProcess::ExecuteRemoteCommand (char*    pCommandLine
        ,CString* pRemoteHostName
        ,CString* pRemoteUsername
        ,CString* pRemotePassword
        ,CString* pRemoteDirectory)

{
    Trace ("CECCommandProcess::ExecuteRemoteCommand Executing the remote command " ,pCommandLine);
    Trace ("...on host " ,pRemoteHostName);
    Trace ("...as user " ,pRemoteUsername);
    Trace ("...password " ,pRemotePassword);
    Trace ("...Initial directory ",pRemoteDirectory);

    //
    //    Open the connection to the remote host first.
    //
    m_RemoteChannel = -1;
    int status = OpenRemoteConnection (pRemoteHostName
                                       ,pRemoteUsername
                                       ,pRemotePassword
                                       ,pRemoteDirectory
                                       ,m_pOutputWindow
                                       ,FALSE);
    if (status == -1) goto end;

    //
    //    We then need to send the formatted command to the remote server to execute and read the output
    //    from it.
    //
    status = SendRemoteCommand (pCommandLine);
    if (status == -1) goto cerror;

    //
    //    ...and close the link
    //
    CloseRemoteConnection (FALSE);

    //
    //    The command has been issued and all output recovered.
    //    Ensure that we close the socket, assuming that we opened it.
    //
end:
    ;
    return status;

    //
    //    report a generic error
    //
cerror:
    ;
    CString  Error;
    LPTSTR lpMsgBuf;
    Error.LoadString(IDS_ERR_REMOTECOMMANDFAIL);
    FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                   NULL,
                   WSAGetLastError(),
                   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                   (LPTSTR) &lpMsgBuf,
                   0,
                   NULL);
    Error += lpMsgBuf;
    AfxMessageBox (Error);
    LocalFree( lpMsgBuf );
    status = -1;
    CloseRemoteConnection (FALSE);
    goto end;
}

//
// CheckForCYRRSH
// ==============
//
// This function checks whether the CYRANO RSH port has been setup.
//
// Returns TRUE if it is, FALSE otherwise.
//
BOOL CECCommandProcess::CheckForCYRRSH(CString* /*pRemoteHostName*/, CString& strErrorMessage)
{
    strErrorMessage = "Not Checked";
    return TRUE;
} // CECCommandProcess::CheckForCYRRSH


//
//    OpenRemoteConnection
//    ====================
//
//    This routine initiates a connection to the CYRRSH daemon running on the specified host
//    in preparation for the remote execution of commands.
//
//    Returns:
//      channel number
//      -1 if an error
//
int  CECCommandProcess::OpenRemoteConnection (CString* pRemoteHostName
        ,CString* pRemoteUsername
        ,CString* pRemotePassword
        ,CString* pRemoteDirectory
        ,CEdit*   pEditBox
        ,BOOL     ClearOutput)
{
#define CYRANORSHPORT "cyrrsh"
#define MAX_CON_ATTEMPTS 10

    unsigned long         ip_addr;                      // IP Address
    struct   sockaddr_in  remote_portaddr =
        {
            0
        }
        ;        // Remote port address
    struct   sockaddr_in  local_portaddr =
        {
            0
        }
        ;         // Local port address
    struct   hostent      *hostent_ptr;                 // Pointer to host structure
    struct   servent      *servent_ptr;                 // Pointer to service structure
    struct   timeval      select_timeout=
        {
            10,0
        }
        ;        // 10 second select timeout
    int          con_attempts = 0;             // Connection attempts made
    long         status;                       // Routine status
    fd_set       write_fds = {0};              // Write channel set
    CString  Error;

    //
    //    Before we proceed make sure that we have set the ClearOutputWindow flag correctly as we do not
    //    want its state to be undefined and save any pointer to the output window for later use.
    //
    m_bClearOutputWindow = ClearOutput;
    m_pOutputWindow = pEditBox;
    m_bOutToAWindow = TRUE;

    //
    //    Construct the message whicg we shall send to the remote cyrano service to execute the command.
    //    This should have the form :-
    //
    //      username\npassword\nDIR:RemoteDirectory\nCMD:commandtoexecute\n
    //
    CString RshCommand;
    ASSERT (pRemoteUsername != NULL);
    ASSERT (pRemotePassword != NULL);

    //
    //    Create a socket, for stream communication using Internet addresses
    //
recreate:
    WSASetLastError (0);
    m_RemoteChannel = socket (AF_INET, SOCK_STREAM, 0);
    if (m_RemoteChannel == -1) goto cerror;

    //
    //    Bind socket to local address
    //
    local_portaddr.sin_family = AF_INET;
    local_portaddr.sin_addr.s_addr= htonl (INADDR_ANY);
    local_portaddr.sin_port = 0;

    status = bind (m_RemoteChannel ,(struct sockaddr *) &local_portaddr, sizeof (local_portaddr));
    // Bind socket to address
    if (status == -1) goto cerror;          // ...branch to report errors

    //
    //    Set up remote address of the host we are connecting to
    //
    remote_portaddr.sin_family = AF_INET;   // Internet address
    ip_addr = inet_addr (pRemoteHostName->GetBuffer(0));
    // Try dotted-decimal format
    if (ip_addr != ((unsigned long) -1))    // If it succeeded...
    {
        (void) memcpy (&remote_portaddr.sin_addr, &ip_addr, sizeof (ip_addr));
        // Copy to address structure
    }
    else
    {
        hostent_ptr = gethostbyname (pRemoteHostName->GetBuffer(0));
        // Try by name
        if (hostent_ptr ==NULL)             // If name not found
        {
            AfxMessageBox ("Remote Host Not Known");
            status = -1;
            goto end;
        }

        (void) memcpy (&remote_portaddr.sin_addr, (*hostent_ptr).h_addr, (*hostent_ptr).h_length);
        // Copy to address structure
    }

    //
    //    Identify the port to which we are to connect knowing that it will be the SYRANO RSH service port
    //

    servent_ptr = getservbyname (CYRANORSHPORT, "tcp");
    // Get service entry
    if (servent_ptr == NULL)                // If not found
    {
        AfxMessageBox ("Failed to locate the CYRANO RSH Service port, Check the services file");
        status = -1;
        goto end;
    }

    remote_portaddr.sin_port = (*servent_ptr).s_port;
    // Store port number

    //
    //  Connect the socket to the remote port
    //
    status = connect (m_RemoteChannel ,(struct sockaddr *) &remote_portaddr ,sizeof (remote_portaddr));

    //
    //    If we failed to connect then we should check the error to see if it is one from which we can
    //    continue and try again for the connection.
    //
    if ((status == -1) && (WSAGetLastError() != WSAEISCONN))
    {

        //
        //    If the connection was refused try again until we have attempted the maiximum number of times,
        //
        if((WSAGetLastError() == WSAECONNREFUSED) && (con_attempts < MAX_CON_ATTEMPTS))
        {
            con_attempts += 1;
            (void)close(m_RemoteChannel);
            goto recreate;
        }

        //
        //    If the connection was blocked then we shall sit on a select waiting for the connection to complete.
        //    We specify a 10 second timeout for the select after which time we give up.
        //
        else if (WSAGetLastError() == WSAEWOULDBLOCK)
            // If connect not completed yet
        {
            FD_SET (m_RemoteChannel, &write_fds);       // Include connecting socket
            select_timeout.tv_sec = 10;         // Set 10 sec timeout
            select_timeout.tv_usec = 0;

            status = select (m_RemoteChannel + 1, NULL, &write_fds, NULL, &select_timeout);
            // Wait for connection

            //
            //    If the connection request timed out then report the error and exit.
            //
            if (FD_ISSET (m_RemoteChannel ,&write_fds) == 0)
            {
                AfxMessageBox ("Failed to connect to the CYRANO RSH Service port");
                status = -1;
                goto end;
            }
        }

        else
        {
            AfxMessageBox ("Failed to connect to the CYRANO RSH Service port");
            status = -1;
            goto end;
        }
    }

    //
    //    And send the initial logon data.
    //
    RshCommand = *pRemoteUsername + '\n' + *pRemotePassword + '\n';
    if (pRemoteDirectory != NULL) RshCommand += "DIR:" + *pRemoteDirectory + '\n';
    status = send (m_RemoteChannel ,RshCommand.GetBuffer(0) ,RshCommand.GetLength() ,0);
    if (status == -1) goto cerror;

    //
    //    Recover the output from this (logon) process and check to ensure that we receive NO output
    //    Any output from this stage indicates an error that will need reporting.
    //
    RecoverRemoteOutput(FALSE);
    if (!m_strOutput.IsEmpty())
    {
        CString ErrorText;
        AfxFormatString1(ErrorText ,IDS_ERR_CONNECT_CYRRSH ,m_strOutput);
        AfxMessageBox (ErrorText);
        status = -1;
    }
    else
    {
        status = 0;
    }

    //
    //    The command has been issued and all output recovered.
    //    Ensure that we close the socket, assuming that we opened it.
    //
end:
    ;
    if (status != 0)
    {
        return -1;
    }
    return m_RemoteChannel;

    //
    //    report a generic error
    //
cerror:
    ;
    LPTSTR lpMsgBuf;
    Error.LoadString(IDS_ERR_REMOTECOMMANDFAIL);
    FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                   NULL,
                   WSAGetLastError(),
                   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                   (LPTSTR) &lpMsgBuf,
                   0,
                   NULL);
    Error += lpMsgBuf;
    AfxMessageBox (Error);
    LocalFree( lpMsgBuf );
    return -1;
}


//
//    SendRemoteCommand
//    =================
//
//     Send a command to be executed by the CYRRSH daemon on the host to which we should already
//     have an open channel.
//
int CECCommandProcess::SendRemoteCommand (char* pCommandLine ,BOOL Flush)
{
    CString RshCommand;
    CString CommandLine = pCommandLine;

    //
    //    Construct the command to send and send it.
    //
    RshCommand += "CMD:" + CommandLine + "\n";
    int status = send (m_RemoteChannel ,RshCommand.GetBuffer(0) ,RshCommand.GetLength() ,0);

    //
    //    If we have an error back from the send then we may have some output which will explain why we had
    //    the error so try and output it.
    //
    if (status == -1)
    {
        RecoverRemoteOutput (m_bOutToAWindow);
        CloseRemoteConnection(FALSE);
    }
    else
    {

        //
        //    If required, flush the command immediately.
        //
        if (Flush) FlushRemoteConnection ();
    }

    //
    //    ...and return the status.
    //
    return status;
}


//
//    CloseRemoteConnection
//    =====================
//
//    This routine initiates a connection to the CYRRSH daemon running on the specified host
//    in preparation for the remote execution of commands.
//
//    Returns:
//      channel number
//      -1 if an error
//
int  CECCommandProcess::CloseRemoteConnection (BOOL FlushFirst)
{

    //
    //    Ensure that the connection is open
    //
    ASSERT (m_RemoteChannel != -1);

    //
    //    If requested, flush the remote commands to the daemon and recover any output from the command,
    //    sending it to the edit box specified if any.
    //
    if (FlushFirst)
    {
        FlushRemoteConnection ();
    }

    //
    //    ...and close the connection to the remote CYRRSH daemon process.
    //
    (void) closesocket (m_RemoteChannel);
    return 0;
}

//
//    FlushRemoteConnection
//    =====================
//
//    Flush any commands already sent to the CYRRSH daemon process, sending the output to the
//    specified edit window if required.
//
void    CECCommandProcess::FlushRemoteConnection    ()
{

    //
    //    Ensure that the connection is open
    //
    ASSERT (m_RemoteChannel != -1);

    //
    //    Send a terminating newline to cause the connection to be flushed.
    //
    CString RshCommand;
    RshCommand = "\n";
    (void) send (m_RemoteChannel ,RshCommand.GetBuffer(0) ,RshCommand.GetLength() ,0);

    //
    //    Recover any output from CYRRSH.
    //
    RecoverRemoteOutput (m_bOutToAWindow);
}


//
//    RecoverRemoteOutput
//    ===================
//
//    Recover output from a previously opened remote connection to the CYRRSH daemon.
//
void    CECCommandProcess::RecoverRemoteOutput (BOOL bOutputNow)
{
    int      LengthRead;
    char     OutputBuffer[512];
    CString  OldOutput;
    char*    pDelimiter;

#define DELIMITERSIZE 4
#define DELIMITER "\xff\xff\xff\xff"

    //
    //    Ensure that the connection is open
    //
    ASSERT (m_RemoteChannel != -1);
    Trace ("CECCommandProcess::RecoverRemoteOutput -> recovering remote commmand output","");

    //
    //    ...and then read the data coming back from the socket.
    //
    m_strOutput.Empty();

    BOOL DelimiterFound = FALSE;
    do
    {
        LengthRead = recv (m_RemoteChannel, OutputBuffer ,sizeof(OutputBuffer) ,0);
        if (LengthRead > 0)
        {

            //
            //    Seeing as how the socket is not going to be closed we must check for the delimiter which
            //    is hardwired to 4* FF characters.
            //
            pDelimiter = OutputBuffer + LengthRead - DELIMITERSIZE;
            if (memcmp(pDelimiter ,DELIMITER ,DELIMITERSIZE) == 0)
            {
                LengthRead -= DELIMITERSIZE;
                DelimiterFound = TRUE;
            }
            OutputBuffer[LengthRead] = '\0';
            m_strOutput += OutputBuffer;
        }
    }
    while ((LengthRead > 0) && (!DelimiterFound));

    //
    //    All output recovered, close the remote link and then send the output to the window
    //    specified (if any)
    //
    if (bOutputNow && (m_pOutputWindow != NULL) && (m_pOutputWindow->m_hWnd != NULL))
    {
        //
        // Output is being collated in a window
        ::SendMessage (m_pOutputWindow->m_hWnd ,WMU_SETOUTPUTTEXT
                       ,(WPARAM) &m_strOutput
                       ,(LPARAM) m_bClearOutputWindow);
        Trace ("...Output Recovered > " ,m_strOutput.GetBuffer(0));
    }
}

//
//    DidCommandSucceed
//    =================
//
//    This is an over-ridable routine which can be called to determine if the output from the
//    last command executed indicated that the command was successful. The success criteria string
//    is passed through to us and is a standard text string, delimited by newline characters of words
//    or phrases which, if they appear in the last output, indicate that the command was sucessful.
//
//    Return Status:
//      TRUE if any of the keywords or phrases appear in the last output string recovered
//      FALSE otherwise
//
//    Parameters:
//      SuccessText - CString passsed by reference. This should be a list of the
//                  phrases or strings to search for separated by newline ("\n") characters.
//
BOOL CECCommandProcess::DidCommandSucceed(CString & SuccessText) const
{
    int SubStrIdx = SuccessText.Find('\n');

    while (SubStrIdx != -1)
    {
        CString SubString;

        //
        //    We have detected a delimiter of a substring so extract the substring itself.
        //
        SubString = SuccessText.Left (SubStrIdx);

        //
        //    Scan the output buffer to determine if this sub-string is present in it.
        //
        if (m_strOutput.Find (SubString) != -1)
        {
            return TRUE;
        }

        //
        //    Substring not found so move on to the next chunk.
        //
        SuccessText = SuccessText.Mid (SubStrIdx + 1);
        SubStrIdx = SuccessText.Find('\n');
    }

    //
    //    At this point we are left with (hopefully) the final chunk of the string to check.
    //
    if ((!SuccessText.IsEmpty()) && (m_strOutput.Find (SuccessText) != -1))
    {
        return TRUE;
    }

    return FALSE;
}


//
//    DidCommandSucceed
//    =================
//
//    This is an over-ridable routine which can be called to determine if the output from the
//    last command executed indicated that the command was successful. The success criteria string
//    is passed through to us and is a standard text string, delimited by newline characters of words
//    or phrases which, if they appear in the last output, indicate that the command was sucessful.
//
//    This is a shell around the previous routine as we simply translate the string ID into a string
//    and pass it back up.
//
//    Return Status:
//      TRUE if any of the keywords or phrases appear in the last output string recovered
//      FALSE otherwise
//
//    Parameters:
//      SuccessID - integer ID of a string in the string table. This should be a list of the
//                  phrases or strings to search for separated by newline ("\n") characters.
//
BOOL CECCommandProcess::DidCommandSucceed(int SuccessId) const
{
    CString SuccessText;

    SuccessText.LoadString (SuccessId);
    return DidCommandSucceed (SuccessText);
}


//
//    Trace
//    =====
//
//    Trace to both a file and the output window (if specified).
//
void CECCommandProcess::Trace(CString& TraceMsg)
{
    CECTrace TraceWindow (m_pOutputWindow);
    CECTrace TraceFile;

    TraceWindow.Trace (TraceMsg);
    TraceFile.Trace (TraceMsg);
}

void CECCommandProcess::Trace(char* pMsg1, CString* pMsg2)
{
    CString Msg = pMsg1;
    if (pMsg2 != NULL) Msg += *pMsg2;
    Trace (Msg);
}

void CECCommandProcess::Trace(char* pMsg1, char* pMsg2)
{
    CString Msg = pMsg1;
    if (pMsg2 != NULL) Msg += pMsg2;
    Trace (Msg);
}
