/* Copyright (C) 2000 Cyrano SA

   This file is part of OpenSTA

   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.

   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html

   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//========================================================================
//
// TEHttpRecords.cpp
// -----------------
//=========================================================================

#ifdef WIN32
#   pragma warning(disable:4786)
#   pragma warning(disable:4503)
#endif

#include  "RegExp.h"

#ifdef TH_POOL_VER
#   include "TRecMake.h"
#   include "TContext.h"
#   include "tools.hpp"
#   include "AsyncVirtualUser.h"
#   include "Address.hpp"
#   include "PString.hpp"
#   include "PStringSpecialised.inl"
#endif

#include "TEHttpRecords.h"
#include "TEHttpRecords.inl"

static const char HTTP_NONE       [] = "None";
static const char HTTP_CONNECT    [] = "Connect";
static const char HTTP_REQUEST    [] = "Request";
static const char HTTP_DISCONNECT [] = "Disconnect";
static const char HTTP_RESPONSE	  [] = "Response";
static const char HTTP_SYNCHRONIZE[] = "Synchronize";
static const char HTTP_WAIT       [] = "Wait";


//
// TRcrdConnect::prepare
// =====================
//
// perform any preparation before conecting the threads
int TRcrdConnect::prepare(TScript* script, int objIndex, bool verbose) 
{
	const TParam* prm = param(TParam::stServer);
    if(prm == NULL) 
		return -1;
    string l_data = prm->strVal();
    _exprP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_exprP->Parse(l_data, 0)))
    {
		MSGERR("Connect : " + l_data +
			   " Failed during 'server' PREPARE with: " + _exprP->Reason());
		return -1;
    }
    return 0;
}

//
// TRcrdConnect::execute
// =====================
//
// execute the threads in turn
#ifndef TH_POOL_VER

int TRcrdConnect::execute(TContext& ctxt)
{
	ctxt.set_phase_text(PHASE_STRT_CONNECT);

	// Get the variables from the tof
	int			nConid		=   intPrmVal(TParam::stConnid, ctxt);
	string		szServer	=   "";
	{
		CTLockGuard<CTMutex> guard(m_mutex);
		szServer=_exprP->Evaluate(ctxt);
	}

    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("CONNECT(" + int_to_string(nConid) + "): Server: " + szServer, TRACEMSG);


	// ------------------------------
    // Get connection object, create new if necessary.
	// If it already exists, receive all remaining datas, disconnect and reconnect
    CHTTPconn* pConnection = static_cast<CHTTPconn*>(ctxt.connection(nConid));
    if(pConnection == NULL)
	{
		string statLevel;
		try
		{
			statLevel = ctxt.get_variable(STAT_LEVEL);
		}
		catch(mdbParseException &)
		{
			statLevel = "L";
		}

        pConnection = new CHTTPconn(nConid, ctxt.GetThreadID(), statLevel);
        if(pConnection == NULL)
        {
			// print statistics before exiting
			CHTTPconn::printAllStatistics(ctxt);

            ctxt.log("Connection alloc failed(" + int_to_string(nConid) + ") for this Connect object", ERRMSG);
			throw mdbParseException("Connection alloc failed(" + int_to_string(nConid) + ") for this Connect object");
        }
        // Add new connection to list
        ctxt.addCon(nConid, pConnection);
    }
	else
	{
		// Get all remaining data on this connection
		if ( !(pConnection->isReceiveFinished()) )
		{
			if (CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
			{
				return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
			}
		}

		// Disconnect this connection object
		if( pConnection->disconnect() != 0 )
		{
			// If error
            if(ctxt.GetVerbose(ERRMSG))
   		    	ctxt.log("Disconnect failed:" + pConnection->getLastError(), ERRMSG);
		}
	}
	// ------------------------------


	// ------------------------------
    // Perform connection
	string szHost;
	int	nPort = 80;
	int endServer = szServer.find(":");
	if (endServer != string::npos)
	{
		szHost = szServer.substr(0, endServer).c_str();
		nPort = atoi(szServer.substr(endServer + 1, szServer.size() - endServer - 1).c_str());
	}

	if((pConnection->connect(szHost, nPort)) != 0)
	{
        string lastError = pConnection->getLastError();
        // Log client error
        ctxt.log("Connection Failed:" + lastError, ERRMSG);
        // print statistics before exiting
        CHTTPconn::printAllStatistics(ctxt);
        // Remove connection again
        ctxt.remCon(nConid);

		throw mdbParseException("Connection Failed:" + lastError);
	}
	// ------------------------------

	// --- A Connect object implies that connection was made through a proxy
	pConnection->setProxy(true);

	ctxt.set_phase_text(HTTP_NONE);

	return 0;
}

#else   // TH_POOL_VER

// Return ETofResult::TOF_RES_ASYNC_STARTED if must be called again to complete execution.
int TRcrdConnect::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);
    int nConid;
    CAsyncHTTPRequest* pConnection = NULL;

	ctxt.set_phase_text(PHASE_STRT_CONNECT);

    if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
    {
        GetTraceStream(VU.GetTrace()) << "TRcrdConnect: entering step " << VU.GetTRecordCallStep() << ".\r\n";
        ReleaseTraceStream();
    }

    switch(VU.GetTRecordCallStep())
    {
    case 0:
        {
            // Synchronise on nConid in case it would still be in use (TOF anomaly).

        	// Get the variables from the tof
            nConid = intPrmVal(TParam::stConnid, ctxt);
            VU.SetCurrentConid(nConid);

            VU.SetTRecordCallStep(1);   // Set before the test to avoid run-conditions.
            if(VU.SynchroniseHTTPClient(nConid) == FALSE)
            {
                if(ctxt.GetVerbose(TRACEMSG))
                    ctxt.log("WARNING: CONNECTing with a connection ID that is still in use: " + int_to_string(nConid) + ". Synchronising.", TRACEMSG);

                return TOF_RES_ASYNC_STARTED;
            }
        }
        // No break since should continue if no need to synchronize.

    case 1:
        {
			// Eventually break the previous connection (if there's already a connection with that ID that 
            // received a Connection:close or if that connection was established to another server/port).

            nConid = VU.GetCurrentConid();

            string szServer;
            {
#ifndef DONTLOCKEVAL
                CLock guard (m_pFastLock);
#endif
                szServer = _exprP->Evaluate(ctxt);
            }

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("connection"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdConnect: Connecting " << nConid << " to server " << szServer << ".\r\n";
                ReleaseTraceStream();
            }
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("CONNECT(" + int_to_string(nConid) + ") to proxy server " + szServer, TRACEMSG);

            // Get connection object, create new if necessary.
            pConnection = VU.ObtainRequestStruct(nConid);
            if(pConnection == NULL)
            {
                // All the request slots are already used, and was not able to allocate a new one. Fatal error.
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors connection"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdConnect: Could not allocate a new connection.\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("Connection failed.", ERRMSG);

                assert(0);
                return TOF_RES_FAILED;
            }

	        string szHost;
	        int	nPort = 80;
	        unsigned int endServer = szServer.find(':');
	        if(endServer != string::npos)
	        {
		        szHost = szServer.substr(0, endServer).c_str();
		        nPort = atoi(szServer.substr(endServer + 1, szServer.size() - endServer - 1).c_str());
	        }

            // Store connection parameters into the VU for future use.
            VU.SetProxyServerName(szHost);
            VU.SetProxyServerPort(nPort);

            if(pConnection->IsTCPConnected())
            {
                // Disconnect if necessary.
                BOOL bMustDisconnect = FALSE;

                if(pConnection->GetConnectionType() == CAsyncHTTPRequest::CNX_DISCONNECTS)
                    bMustDisconnect = TRUE;
                else
                {
                    string  zCurrentServer;
                    DWORD   dwCurrentPort;

                    pConnection->GetConnParams(zCurrentServer, dwCurrentPort);

                    if(zCurrentServer != szHost || dwCurrentPort != (DWORD)nPort)
                    {
                        // Must connect to another server/port.
                        if(ctxt.GetVerbose(TRACEMSG))
                            ctxt.log("WARNING: CONNECTing with a connection ID that is still connected to another server: " + int_to_string(nConid) + ". Reconnecting.", TRACEMSG);

                        bMustDisconnect = TRUE;
                    }
                }

                if(bMustDisconnect)
                {
                    VU.SetTRecordCallStep(2);   // Will enter the "case 2" section next time execute is called.
                    if(VU.DisconnectClient(nConid) == TOF_RES_ASYNC_STARTED)    // Can't use CAsyncHTTPRequest::Disconnect().
                    {
                        // That case occurs when using SSL protocol.
                        return TOF_RES_ASYNC_STARTED;
                    }
                    else
                    {
                        pConnection->SetConnectionId(nConid);
                    }
                }
            }
        }
        // No break since should continue if no async. op.

    case 2:
        {
            // Perform the connection.

            nConid = VU.GetCurrentConid();

            if(pConnection == NULL)
                pConnection = VU.ObtainRequestStruct(nConid);
            assert(pConnection != NULL);

            // Get the values computed in case 1.
            string szHost = VU.GetProxyServerName();
	        int	nPort = VU.GetProxyServerPort();

            // Read the statistics verbose level flag. Tell the request structures what they're gonna have to do of the stats.
            pConnection->SetWatchStatsTo(VU.MustDumpStats());

            // Perform connection

	        // Note that a CONNECT object should not be asynchronous...
            DWORD dwRes = pConnection->Connect(szHost.c_str(), nPort, (IocpEvent*) pConnection);
            switch(dwRes)
	        {
	        case TOF_RES_OK:
		        {
			        // --- A Connect object implies that connection was made through a proxy
			        pConnection->SetUseProxyTo(TRUE);
			        ctxt.set_phase_text(HTTP_NONE);

			        return dwRes;
		        }

	        case TOF_RES_ASYNC_STARTED:
	        default:
                // Log client error
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors connection"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdConnect: Connection Failed for " << nConid << ". Error is " << dwRes << ".\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("Connection ERROR for " + int_to_string(nConid) + ". Error is " + int_to_string((unsigned int)dwRes), ERRMSG);

                // Release the connection.
                VU.ReleaseRequest(pConnection);

		        return TOF_RES_FAILED;
	        };
        }   // Case 2.

    default:
        {
            assert(0);
            return TOF_RES_FAILED;
        }
    };

	ctxt.set_phase_text(HTTP_NONE);
    return TOF_RES_OK;
}

#endif

//
// TRcrdRequest::prepare
// =====================
//
// perform any preparation before sending the request
int TRcrdRequest::prepare(TScript* script, int objIndex, bool verbose) 
{
	const TParam* prm = param("uri");
    if(prm == NULL)
    {
		MSGERR("Request:PREPARE: could not find the \'uri\'");
		return -1;
    }
    string l_uri = prm->strVal();
    _uriP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_uriP->Parse(l_uri, 0)))
    {
		MSGERR("Request : " + l_uri +
			   " Failed during 'uri' PREPARE with: " + _uriP->Reason());
		return -1;
    }


	prm = param("header");
    if(prm == NULL) 
		return -1;
    string l_header = prm->strVal();
    _headerP = new TtValList(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_headerP->Parse(l_header, 0)))
    {
		MSGERR("Request : " + l_header +
			   " Failed during 'header' PREPARE with: " + _headerP->Reason());
		return -1;
    }


	prm = param("with");
    if(prm == NULL) 
		return -1;
    string l_with = prm->strVal();
    _withP = new TtValList(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_withP->Parse(l_with, 0)))
    {
		MSGERR("Request : " + l_with +
			   " Failed during 'with' PREPARE with: " + _withP->Reason());
		return -1;
    }


	prm = param("wout");
    if(prm == NULL) 
		return -1;
    string l_wout = prm->strVal();
    _woutP = new TtValList(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_woutP->Parse(l_wout, 0)))
    {
		MSGERR("Request : " + l_wout +
			   " Failed during 'wout' PREPARE with: " + _woutP->Reason());
		return -1;
    }


	prm = param("body");
    if(prm == NULL) 
		return -1;
    string l_body = prm->strVal();
    _bodyP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_bodyP->Parse(l_body, 0)))
    {
		MSGERR("Request : " + l_body +
			   " Failed during 'body' PREPARE with: " + _bodyP->Reason());
		return -1;
    }

    prm = param(TParam::stConnid);
    string l_conid = prm->strVal();
    _iConid = atoi(l_conid.c_str());


	prm = param("certificate");
    if(prm == NULL) 
    {
		MSGERR("Request : " + l_uri +
			   " Failed during 'certificate' PREPARE with: Parameter missing from ODL");
		return -1;
    }
	string l_cert = prm->strVal();
    _certName = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_certName->Parse(l_cert, 0)))
    {
		MSGERR("Request : " + l_cert +
			   " Failed during 'uri' PREPARE with: " + _certName->Reason());
		return -1;
    }

	prm = param(TParam::stRetCod);
    if(prm == NULL) 
		return -1;
    string szCode = prm->strVal();
    if(!szCode.empty())
    {
        _retCode = new TtVarRef(script);
        if (!(_retCode->Parse(szCode, 0)))
        {
		    MSGERR("Request : " + szCode +
			       " Failed during 'retCode' PREPARE with: " + _retCode->Reason());
		    return -1;
        }
    }

    return 0;
}

#ifndef TH_POOL_VER

//
// TRcrdRequest::getRequestType
// ============================
//
// translate the type into the actual Command string
string TRcrdRequest::getRequestString(int eType) const
{
	string szCommand;

	switch (eType)
	{
		case HTTP_REQUEST_GET:
		{
			szCommand = "GET";
			break;
		}
		case HTTP_REQUEST_HEAD:
		{
			szCommand = "HEAD";
			break;
		}
		case HTTP_REQUEST_POST:
		{
			szCommand = "POST";
			break;
		}
		default:
		{
			szCommand = "GET";
		}
	}
 
	return szCommand;
}

//
// TRcrdRequest::execute
// =====================
//
// execute the request
int TRcrdRequest::execute(TContext& ctxt)
{
    ctxt.set_phase_text(HTTP_REQUEST);

    // --- Get the variables from the tof
    int		nType		=   intPrmVal(TParam::stType, ctxt);
    string	szUri		=   "";
    int		nConid		=   intPrmVal(TParam::stConnid, ctxt);
    string	szBody		=   "";
    string	szTimer		=   strPrmVal(TParam::stTimer, ctxt);
    string	szStatus	=	strPrmVal(TParam::stRetSts, ctxt);
    string	szCode		=	strPrmVal(TParam::stRetCod, ctxt);
	string	szCertName	=	"";
    int     nOptions	=   intPrmVal("options", ctxt);

    list<string> listHeader;
    list<string> listWithHeader;
    list<string> listWoutHeader;
    {
		CTLockGuard<CTMutex> guard(evaluateMtx);
		szUri		=   _uriP->Evaluate(ctxt);
		szCertName	=   _certName->Evaluate(ctxt);
		szBody		=   _bodyP->Evaluate(ctxt);
		_headerP->EvaluateValues(ctxt, listHeader);
		_withP->EvaluateValues(ctxt, listWithHeader);
		_woutP->EvaluateValues(ctxt, listWoutHeader);
    }

    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("REQUEST(" + int_to_string(nConid) + "): "	
	    				+ getRequestString(nType) + " " + szUri, TRACEMSG);
	
	// --- Is it a http or a https request ?
	bool secureRequest = (begins_with(szUri, "HTTPS://"))
		? true 
		: false;

	// ------------------------------
    // Get connection object, create new if necessary (if there is no proxy)
    CHTTPconn* pConnection = static_cast<CHTTPconn*>(ctxt.connection(nConid));

	if(pConnection == NULL)
	{
		string statLevel;
		try
		{
			statLevel = ctxt.get_variable(STAT_LEVEL);
		}
		catch(mdbParseException &)
		{
			statLevel = "L";
		}

        pConnection = new CHTTPconn(nConid, ctxt.GetThreadID(), statLevel);
        if(pConnection == NULL) 
        {
			// print statistics before exiting
			CHTTPconn::printAllStatistics(ctxt);

            ctxt.log("Connection alloc failed(" + int_to_string(nConid) + ") for this Connect object", ERRMSG);
			throw mdbParseException("Connection alloc failed(" + int_to_string(nConid) + ") for this Connect object");
		}		
        // --- Add new connection to list
        ctxt.addCon(nConid, pConnection);


		// --- Is it a http or a https request ?
		int nPort;
		string RequestProtocol;
		if (secureRequest)
		{
			RequestProtocol = "https://";
			nPort = 443; // Default port for SSL
		}
		else
		{
			RequestProtocol = "http://";
			nPort = 80;
		}

		// --- Extract server name from request line
		int server_begin	= find_in_string(szUri, RequestProtocol, false, 0) + RequestProtocol.size();
		int server_end		= szUri.find_first_of(":/ ", server_begin);
		string server_name	= szUri.substr(server_begin, server_end - server_begin);

		// --- If port is specified
		if (szUri[server_end] == ':')
		{
			int port_begin	= server_end + 1;
			int port_end	= szUri.find_first_of("/ ", port_begin);
			string szPort	= szUri.substr(port_begin, port_end - port_begin);
			nPort			= atoi(szPort.c_str());
		}

		// --- Perform connection
		if((pConnection->connect(server_name, nPort)) != 0)
		{
            string lastError = pConnection->getLastError();
			// Log client error
			ctxt.log("Connection Failed:" + lastError, ERRMSG);      

			// print statistics before exiting
			CHTTPconn::printAllStatistics(ctxt);
			
			// Remove connection again
			ctxt.remCon(nConid);

            throw mdbParseException("Connection Failed:" + lastError);
		}
	}
	// ------------------------------

	// ------------------------------
    // Get all remaining data on this connection
	if ( !(pConnection->isReceiveFinished()) )
	{
		if (CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
		{
			return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
		}
	}
	// ------------------------------

	// If some variables names are here
	pConnection->setTimerName(szTimer);
	pConnection->setStatusName(szStatus);
	pConnection->setCodeName(szCode);

	// ------------------------------
    // Reconnect if necessary
    if (!pConnection->isConnected())
    {
        if(ctxt.GetVerbose(TRACEMSG))
	    	ctxt.log("Have to reconnect", TRACEMSG);

		if (pConnection->connect(pConnection->getHost(),pConnection->getPort()) != 0)
		{
            string lastError = pConnection->getLastError();
			// print statistics before exiting
			CHTTPconn::printAllStatistics(ctxt);

			ctxt.log("Connection Failed:" + lastError, ERRMSG);      
			throw mdbParseException("Connection Failed:" + lastError);
		}
    }
	// ------------------------------

	
	if (secureRequest)
		pConnection->setProtocol(HTTP_SECURE);


	// ------------------------------
	// Convert body to binary if necessary
	string newBody;
	if (szBody.length())
	{
		Ascii2Binary(szBody, newBody);
	}
	// ------------------------------


	// ------------------------------
    // Format headers
	string szHeader;
	string szWithHeader;
	string szWoutHeader;
	szHeader.reserve(DEFAULT_HEADER_SIZE);

	list<string>::iterator j;
	for(j = listHeader.begin(); j != listHeader.end(); j++) 
	{
		szHeader.append(*j);
		remove_trailing(szHeader, "\n");
		szHeader.append("\r\n");
	}

	for(j = listWithHeader.begin(); j != listWithHeader.end(); j++) 
	{
		szWithHeader.append(*j);
		remove_trailing(szWithHeader, "\n");
		szWithHeader.append("\r\n");
	}

	for(j = listWoutHeader.begin(); j != listWoutHeader.end(); j++) 
	{
		szWoutHeader.append(*j);
		remove_trailing(szWoutHeader, "\n");
		szWoutHeader.append("\r\n");
	}

	string_lf2crlf(szHeader);
	string_cr2crlf(szHeader);
	string_lf2crlf(szWithHeader);
	string_cr2crlf(szWithHeader);
	string_lf2crlf(szWoutHeader);
	string_cr2crlf(szWoutHeader);

    // Remove the "with" header fields
	int offset = 0;
	int endLine = 0;
	int beginField = 0;
	int endField = 0;
	while ( (offset != string::npos) && ((endField = szWithHeader.find_first_of(" :", offset)) != string::npos) )
	{
		beginField = find_in_string(szHeader, szWithHeader.substr(offset, endField - offset), false, 0);
		endLine = szHeader.find("\r\n", beginField);
		if ( (beginField != string::npos) && (endLine != string::npos) )
			szHeader.erase(beginField, endLine - beginField + 2);
		offset = szWithHeader.find("\r\n", offset);
		offset += 2;
	}
    
	// Add the "with" header fields
    szHeader.append(szWithHeader);
    

	// If it is a POST request, calculate the Content-Length
	if (nType == 2) 
	{
		beginField = find_in_string(szHeader, "Content-Length", false, 0);
		endLine = szHeader.find("\r\n", beginField);
		if ( (beginField != string::npos) && (endLine != string::npos) )
			szHeader.erase(beginField, endLine - beginField + 2);
		// Add the new Content-Length
		char buf[30];
		sprintf(buf, "Content-Length: %d\r\n", newBody.size());
		szHeader.append(buf);
	}

	
	// And remove the "wout" ones
	offset = 0;
	endLine = 0;
	beginField = 0;
	endField = 0;
	while ( (offset != string::npos) && ((endField = szWoutHeader.find_first_of(" \r\n", offset)) != string::npos) )
	{
		beginField = find_in_string(szHeader, szWoutHeader.substr(offset, endField - offset), false, 0);
		endLine = szHeader.find("\r\n", beginField);
		if ( (beginField != string::npos) && (endLine != string::npos) )
			szHeader.erase(beginField, endLine - beginField + 2);
		offset = szWoutHeader.find("\r\n", offset);
		offset += 2;
	}

	// Add an empty line
    szHeader.append("\r\n");
	// ------------------------------


	// ------------------------------
	// Remove server name if there is no proxy
	pConnection->setInfo(getRequestString(nType) + " " + szUri);

	if ( !pConnection->isProxy() )
	{
		int endServerName = szUri.find_first_of("/", strlen("http://") + 2) ;
		if (endServerName != string::npos)
			szUri.erase(0, endServerName);
		else
			szUri = "/";
	}
	// ------------------------------


    // ------------------------------
    // Is there a certificate to send to the server ?
	if ( nOptions & TOF_REQOPT_CERTFIL )
		pConnection->setCertificate(szCertName);
	else
		pConnection->setCertificate("");
    // ------------------------------


	// ------------------------------
    // Do the request
	if (!( pConnection->getTimerName().empty()) )
		ctxt.StartTimer(pConnection->getTimerName());

    if ((pConnection->request(nType, pConnection->getInfo() + "\r\n" + szHeader, newBody, _bSaveDocument) == -1))
    {
		// print statistics before exiting
		CHTTPconn::printAllStatistics(ctxt);

        // Log client error
        ctxt.log("Request Failed:" + pConnection->getLastError(), ERRMSG);
		throw mdbParseException("Request Failed:" + pConnection->getLastError());
		// return -1;
    }
	// ------------------------------
	

	// ------------------------------
	if ( (pConnection->getCodeName() != "") || (pConnection->getStatusName() != "") )
	{
		if (CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection, 0) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
		{
			return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
		}
		// CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection, 0);
	}
	// ------------------------------


    ctxt.set_phase_text(HTTP_NONE);

    return 0;
}

#else   // TH_POOL_VER case.

//
// TRcrdRequest::getRequestString
// ==============================
//
// translate the type into the actual Command string
//#pragma const_seg("CONST_DATA")

const string &TRcrdRequest::getRequestString(ERequestType eType) const
{
    static const string zReqType[] = { "GET ", "HEAD ", "POST ", "CONNECT " };

    return zReqType[(int)eType%4];
}

//
// TRcrdRequest::execute
// =====================
//
// Execute a request.
int TRcrdRequest::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);

    string szStatus, szUri;
    CAsyncHTTPRequest* pConnection = NULL;
    int nConid;

    if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
    {
        GetTraceStream(VU.GetTrace()) << "TRcrdRequest: entering step " << VU.GetTRecordCallStep() << ".\r\n";
        ReleaseTraceStream();
    }

    switch(VU.GetTRecordCallStep())
    {
    case 0:
        {
			// Wait for the last request executed on that connection to complete.

            ctxt.set_phase_text(HTTP_REQUEST);

            nConid = intPrmVal(TParam::stConnid, ctxt);
            VU.SetCurrentConid(nConid);

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("request full"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdRequest: executing on connection " << VU.GetManagerNumber() << '/' << nConid << ".\r\n";
                ReleaseTraceStream();
            }

            // Make sure that the last request on the specified connection is over.
            VU.SetTRecordCallStep(1);   // Set before the test to avoid run-conditions.
            if(VU.SynchroniseHTTPClient(nConid) == FALSE)
            {
                return TOF_RES_ASYNC_STARTED;
            }
        }
        // No break since should continue if no need to synchronize.

		// Perform connection if required.
        // Cases are:
        //  1/ TCP-connecting directly. => TCP-connect to remote server.
        //  2/ SSL-connecting directly. => TCP-connect to remote server + SSL-connect.
        //  3/ TCP-connecting through a proxy. => TCP-connect to proxy (performed by the CONNECT TOF object) if not already done by TOF-connect or connection to proxy was broken.
        //  4/ SSL-connecting through a proxy. => TCP-connect to proxy (performed by the CONNECT TOF object) + HTTP-CONNECT to remote server + SSL-connect.

    case 1:
        {
            // Previous request on that connection is now over.
            // Eventually disconnect if previous request received a Connection-close message from server.

            nConid = VU.GetCurrentConid();

            // Get a (new) request slot to work with.
            pConnection = VU.ObtainRequestStruct(nConid);

            if(pConnection == NULL)
            {
                // All the request slots are already used, and was not able to allocate a new one. Fatal error.
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("connection errors"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdRequest: could not allocate new connection.\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("REQUEST ERROR: could not allocate new connection.", ERRMSG);

                return TOF_RES_FAILED;
            }

            pConnection->SetWatchStatsTo(VU.MustDumpStats());

            // In case of HTTP 1.0 and if the selected connection has already been used, then server should have 
            // disconnected. Can close the connection.
            if(pConnection->GetConnectionType() == CAsyncHTTPRequest::CNX_DISCONNECTS)
            {
                if(pConnection->GetUseProxy())
                {
                    pConnection->SetWillReconnectThroughProxy(TRUE);
                }

                VU.SetTRecordCallStep(2);   // Will enter the "case 2" section next time execute is called.
                if(VU.DisconnectClient(nConid) == TOF_RES_ASYNC_STARTED)    // Can't use CAsyncHTTPRequest::Disconnect().
                {
                    // That case occurs when using SSL protocole. 
                    // The framework will then first call:
                    //  1- CAsyncSecureSocket::Close
                    //  2- CAsyncSecureSocket::CompleteClose (when SSL close dialog is over)
                    //  3- CAsyncTCPSocket::Close
                    //  4- CAsyncHTTPRequest::OnEvent
                    return TOF_RES_ASYNC_STARTED;
                }
                else
                {
                    pConnection->SetConnectionId(nConid);
                }
            }
        }
        // No break since should continue if did not perform an async. disconnect.

    case 2:
        {
            // Execute the HTTP-CONNECTION request against the proxy if required (using the connection 
            // established by the CONNECT TOF object), or TCP-connect directly to the remote server.

            nConid = VU.GetCurrentConid();

            // Get a (new) request slot to work with.
            if(pConnection == NULL)
                pConnection = VU.ObtainRequestStruct(nConid);

            assert(pConnection != NULL);

            {
#ifndef DONTLOCKEVAL
                CLock guard (m_pFastLock);
#endif
                szUri = _uriP->Evaluate(ctxt);
            }

            // Check if that connection will be used for SSL requests LATER on
            // (for the moment, remain in standard TCP in case we need to perform a CONNECT request).
	        BOOL bSecureRequest;
	        if(find_in_string(szUri, "https://", FALSE, 0) != -1)
		        bSecureRequest = TRUE;
            else
                bSecureRequest = FALSE;

            // Parse the host:port string. In the HTTP/1.1 case, should find a 'Host' header field. Otherwise, parse the URI.
            string szRemoteServer, szPort;
            string szHeader; szHeader.reserve(512);

            BuildHeaderString(ctxt, szHeader);

            ExtractHostPort(ctxt, szHeader, szUri, szRemoteServer, szPort);

            if(szPort.empty())
            {
                if(bSecureRequest)
                    szPort = "443";
                else
                    szPort = "80";
            }

            // Store the values for later use.
            VU.SetHeaderText(szHeader);
            VU.SetRemoteServerName(szRemoteServer);
            VU.SetRemoteServerPort(szPort);
            VU.SetSecureModeUsed(bSecureRequest);
            VU.SetCurrentUri(szUri);

            // Check if the connection is established.
            if(pConnection->IsTCPConnected() == FALSE)
            {
			    if(VU.GetTrace() && VU.GetTrace()->IsTrace("request"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdRequest: Connecting " << nConid << " using connection slot number " << pConnection->GetRequestNumber() << ".\r\n";
                    ReleaseTraceStream();
                }

                pConnection->SetSecurityMode(CAsyncHTTPRequest::HTTP_NOT_SECURE);

                // Check if should first connect to a proxy server or directly to the remote server.
                if(pConnection->GetUseProxy() || pConnection->GetWillReconnectThroughProxy())
                {
                    // Connection to proxy, which was opened by a previous CONNECT TOF was lost. Reconnect.
                    if(pConnection->ReConnect() != 0)
                    {
                        // Note result cannot be TOF_RES_ASYNC_STARTED since won't connect to the proxy using SSL.
                        if(ctxt.GetVerbose(ERRMSG))
                            ctxt.log("REQUEST: ERROR re-connecting to the proxy.", ERRMSG);
                        return TOF_RES_FAILED;
                    }
                    pConnection->SetWillReconnectThroughProxy(FALSE);
                    pConnection->SetUseProxyTo(TRUE);
                }
                else
                {
                    // Is not yet connected to the web server or the connection was lost.
                    WORD wPort = (WORD)atoi(szPort.c_str());

                    // TCP-connect to the remote server, whether this is to later perform a secure request or not.
                    if(pConnection->Connect(szRemoteServer.c_str(), (WORD)wPort, (IocpEvent*) pConnection) != 0)
                    {
                        if(ctxt.GetVerbose(ERRMSG))
                            ctxt.log("REQUEST: ERROR connecting to " + szRemoteServer + ':' + szPort, ERRMSG);
                        return TOF_RES_FAILED;
                    }
                }
            }

            // In case of SSL proxy tunneling, and if not sent any requests yet, send a HTTP CONNECT request.
            if(bSecureRequest && pConnection->NeedConnectToProxy())
            {
                // Is connected to the proxy server since the 'Connect' TOF object. Should now HTTP-CONNECT to the remote server.

                // Build the CONNECT request.
                string szRequestText; szRequestText.reserve(255);

                string szURI = "CONNECT " + szRemoteServer + ':' + szPort + " HTTP/1.1";
                szRequestText = szURI + "\r\nUser-Agent: Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt)\r\nHost: "
                                + szRemoteServer + ':' + szPort + "\r\nContent-Length: 0\r\nPragma: no-cache\r\n\r\n";

                if(VU.GetTrace() && VU.GetTrace()->IsTrace("request"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdRequest: Executing (size is " << szRequestText.size() << "):\r\n" << szRequestText << ".\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(TRACEMSG))
                    ctxt.log("REQUEST: Connecting the proxy to the remote server to enable SSL session establishment.", TRACEMSG);

                pConnection->GetStats().SetUrl(szURI);

                pConnection->SetRequestText(szRequestText);

                DWORD dwRes = pConnection->InitiateCONNECT(NO_RESULT);
                if(dwRes != 0)
                {
                    if(ctxt.GetVerbose(ERRMSG))
                        ctxt.log("REQUEST: ERROR executing the CONNECT request through the proxy. Error is: " + int_to_string((LONG)dwRes), ERRMSG);
                    return TOF_RES_FAILED;
                }

                // Wait for the CONNECT request to complete.
                VU.SetTRecordCallStep(3);   // Will enter the "case 3" section next time execute is called.
                if(VU.SynchroniseHTTPClient(nConid) == FALSE)
                {
                    // Download on that connection is not over yet. Cannot continue.
                    return TOF_RES_ASYNC_STARTED;
                }
            }
        }
        // No break since should continue if did not start an HTTP-CONNECT request.

    case 3:
        {
            // In the case of https requests, execute the SSL connection sequence.

            nConid = VU.GetCurrentConid();

            if(pConnection == NULL)
            {
                // Get a (new) request slot to work with.
                pConnection = VU.ObtainRequestStruct(nConid);

                if(pConnection == NULL)
                {
                    // All the request slots are already used, and was not able to allocate a new one. Fatal error.
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("connection errors"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdRequest: could not allocate new connection.\r\n";
                        ReleaseTraceStream();
                    }

                    return TOF_RES_FAILED;
                }
            }
            // else moved directly from case 2 to case 3. Thus, already acquired a connection slot.

            if(szUri.empty())
            {
                szUri = VU.GetCurrentUri();
            }

	        BOOL bSecureRequest = VU.GetSecureModeUsed();
	        if(bSecureRequest)
            {
                pConnection->SetSecurityMode(CAsyncHTTPRequest::HTTP_SECURE);

                // Check if a certificate name is provided.
                string szCertificateName;
                {
#ifndef DONTLOCKEVAL
                    CLock guard (m_pFastLock);
#endif
            		szCertificateName = _certName->Evaluate(ctxt);
                }
                pConnection->SetCertificateName(szCertificateName);
            }
            else
            {
                pConnection->SetSecurityMode(CAsyncHTTPRequest::HTTP_NOT_SECURE);
            }

            if(bSecureRequest == TRUE)
            {
                // Perform a SSL connect.

                string szRemoteServer = VU.GetRemoteServerName();
                string szPort = VU.GetRemoteServerPort();

                assert(szRemoteServer.empty() == FALSE);
                assert(szPort.empty() == FALSE);

                // Perform SSL connection if required.
                if(pConnection->IsConnected() == FALSE)
                {
                    VU.SetTRecordCallStep(4);   // In case Connect returns TOF_RES_ASYNC_STARTED, will enter the "case 4" section 
                                                // next time execute is called.
                    DWORD dwConnRes = pConnection->Connect(szRemoteServer.c_str(), atoi(szPort.c_str()), (IocpEvent*) pConnection);
				    switch(dwConnRes)
				    {
				    case WSAEISCONN:
				    case TOF_RES_OK:
                        // Will move on to case 4.
					    break;

				    case TOF_RES_ASYNC_STARTED:
                        {
					        return TOF_RES_ASYNC_STARTED;
                        }

				    default:
					    {
						    // If error
						    ctxt.set_phase_text(HTTP_NONE);
						    // Log client error
                            if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                            {
                                GetTraceStream(VU.GetTrace()) << "ERROR: Connection to " << szRemoteServer << " on port " << szPort << " failed.\r\n";
                                ReleaseTraceStream();
                            }

                            if(ctxt.GetVerbose(ERRMSG))
    			    	    	ctxt.log("ERROR: Connection to " + szRemoteServer + " on port " + szPort + " failed.", ERRMSG);      

						    VU.ReleaseRequest(pConnection->GetRequestNumber());

    					    return TOF_RES_FAILED;
					    }
				    };
                }
            }
            // else don't need to connect in SSL.
        }
        // No break since should continue if we're performing a standard TCP connection.

	case 4:
        {
            // (SSL) Connection is over. Execute the request.

            nConid = VU.GetCurrentConid();

            // --- Get the variables from the tof.
			ERequestType    eType   = (ERequestType)intPrmVal(TParam::stType, ctxt);
            int		        nOptions= intPrmVal(TParam::stOptions, ctxt);

            string szBody;
            {
#ifndef DONTLOCKEVAL
                CLock guard (m_pFastLock);
#endif
                szBody = _bodyP->Evaluate(ctxt);
            }

            if(szUri.empty())
                szUri = VU.GetCurrentUri();

            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("REQUEST(" + int_to_string(nConid) + "): "	+ getRequestString(eType) + szUri, TRACEMSG);

	        // ------------------------------
            // Format headers

            if(pConnection == NULL)
            {
                // Get a (new) request slot to work with.
                pConnection = VU.ObtainRequestStruct(nConid);

                if(pConnection == NULL)
                {
                    // All the request slots are already used, and was not able to allocate a new one. Fatal error.
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("request errors"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdRequest: could not allocate new connection.\r\n";
                        ReleaseTraceStream();
                    }

                    return TOF_RES_FAILED;
                }
            }
            // else moved directly from case 2 to case 3. Thus, already acquired a connection slot.

	        string szHeader = VU.GetHeaderText();
            assert(szHeader.empty() == FALSE);

			string newBody;
            if(eType == HTTP_REQUEST_POST)
            {
				if (szBody.length())
				{
					Ascii2Binary(szBody, newBody);
				}

                // In case this is a POST, must make sure the optional parameter "Content-Length" is fine.
                DWORD dwStartPos = strinfind(szHeader.data(), szHeader.size()-1, "content-length", sizeof "content-length" -1);
                if(dwStartPos != ULONG_MAX)
                {
                    // Replace the size value.
                    dwStartPos = szHeader.find_first_not_of(": \t", dwStartPos+sizeof("content-length")-1);
                    DWORD dwEndPos = szHeader.find_first_of("\r\n \t", dwStartPos);

                    CHAR lenStr [20];
                    sprintf(lenStr, "%d", newBody.size());

                    szHeader.replace(dwStartPos, dwEndPos-dwStartPos, lenStr);
                }
                else
                {
                    // Add the right header field.
		            char buf[30];
		            sprintf(buf, "Content-Length: %d\r\n", newBody.size());
                    string::size_type CRLF_CRLFpos = szHeader.rfind("\r\n\r\n");
                    if(CRLF_CRLFpos != string::npos)
                        szHeader.insert(CRLF_CRLFpos + 2, buf);
                    else
		                szHeader.append(buf + string("\r\n"));
                }
            }
            else
                newBody = szBody;


            // ------------------------------
            // Perform the request.

            if(szUri.find("HTTP/1.0") != string::npos 
                && pConnection->GetSecurityMode() != CAsyncHTTPRequest::HTTP_SECURE
                && pConnection->GetUseProxy() == FALSE)
            {
                // Should expect to re-connect before executing the next request.
                pConnection->SetConnectionType(CAsyncHTTPRequest::CNX_DISCONNECTS);
            }
            else
            {
                // HTTP/1.1 as well as Secure mode usually uses Keep-alive.
                pConnection->SetConnectionType(CAsyncHTTPRequest::CNX_KEEP_ALIVE);
            }

            pConnection->GetStats().SetUrl(getRequestString(eType) + szUri);

            if(pConnection->GetSecurityMode() == CAsyncHTTPRequest::HTTP_SECURE || pConnection->GetUseProxy() == FALSE)
            {
                // Remove the http://server_name part from szUri if it exists.
                string::size_type nPos = szUri.find("http");
                if(nPos != string::npos)
                {
                    nPos += sizeof "http"-1;
                    if(szUri[nPos] == 's')
                        nPos++;
                    // Skip server_name.
                    nPos = szUri.find_first_of('/', nPos+3);    // Skip the '://' sequence.
                    if(nPos != string::npos)
                    {
                        szUri.erase(0, nPos);
                    }
                }
            }

            string szRequestText; szRequestText.reserve(512);
            szRequestText = getRequestString(eType) + szUri + "\r\n" + szHeader + newBody;
            pConnection->SetRequestText(szRequestText);

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("request"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdRequest: Executing (size is "
                    << szRequestText.size() << "):\r\n" << szRequestText << ".\r\n"
                    << ((_eResultsRequired==NO_RESULT)?
                        "Won't parse results.\r\n"
                        :(
                            (_eResultsRequired==HEADER_ONLY)?
                                "Will parse header only.\r\n"
                                :"Will parse the body.\r\n"
                         )
                       );
                ReleaseTraceStream();
            }

            pConnection->SetTimerName(strPrmVal(TParam::stTimer, ctxt));
	        if(!(pConnection->GetTimerName().empty()))
		        ctxt.StartTimer(pConnection->GetTimerName());

            DWORD dwReqRes;

            switch(eType)
            {
            case HTTP_REQUEST_GET:  // GET
                dwReqRes = pConnection->InitiateGET(_eResultsRequired);
                break;

            case HTTP_REQUEST_HEAD: // HEAD
                dwReqRes = pConnection->InitiateHEAD(_eResultsRequired);
                break;

            case HTTP_REQUEST_POST: // POST
                dwReqRes = pConnection->InitiatePOST(_eResultsRequired);
                break;

            default:
                // ERROR. Unsupported request type!?!
                assert(0);
                return TOF_RES_FAILED;
                break;
            };

            if(dwReqRes != 0)
            {
                if(ctxt.GetVerbose(TRACEMSG))
                {
                    LPVOID lpMsgBuf;
                    if(FormatMessage( 
                        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
                        FORMAT_MESSAGE_FROM_SYSTEM | 
                        FORMAT_MESSAGE_IGNORE_INSERTS,
                        NULL,
                        dwReqRes,
                        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                        (LPTSTR) &lpMsgBuf,
                        0,
                        NULL 
                    ) == 0)
                    {
                        char buff[30];
                        sprintf(buff, "%#lX", dwReqRes);

                        ctxt.log("ERROR: REQUEST(" + int_to_string(nConid) + "): " + getRequestString(eType) + szUri + " failed with error " + buff, TRACEMSG);
                    }
                    else
                    {
                        ctxt.log("ERROR: REQUEST(" + int_to_string(nConid) + "): " + getRequestString(eType) + szUri + " failed with error " + (LPCTSTR)lpMsgBuf, TRACEMSG);
                        LocalFree(lpMsgBuf);
                    }
                }

                return TOF_RES_FAILED;
            }

            // Eventually waits for the end of the request if the result status of the request is required before executing further.
            // Used if the error code of the request is required.
            szStatus    = strPrmVal(TParam::stRetSts, ctxt);

            if(_retCode || szStatus.empty() == FALSE)
            {
                // Synchronise on that request.
                VU.SetTRecordCallStep(5);   // If SynchroniseHTTPClient returns FALSE then will enter the "case 2" 
                                            // section next time execute is called.
                if(VU.SynchroniseHTTPClient(nConid) == FALSE)
                {
                    // Download on that connection is not over yet. Cannot continue.
                    return TOF_RES_ASYNC_STARTED;
                }
            }
            else
            {
                // No need to wait.
                VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.
                ctxt.set_phase_text(HTTP_NONE);

                return TOF_RES_OK;
            }
        }
//        break;    No break since may continue when information about the results is required.

    case 5:
        {
            // Request over. Parse the retcode and status values. Only when asked to.

            nConid = VU.GetCurrentConid();

            if(szStatus.empty())
                szStatus = strPrmVal(TParam::stRetSts, ctxt);

            // If some variables names are set, then request must compute the values.
            if(!(szStatus.empty() && _retCode == NULL))
            {
                if(pConnection == NULL)
                {
                    pConnection = VU.ObtainRequestStruct(nConid);

                    if(pConnection == NULL)
                    {
                        // All the request slots are already used, and was not able to allocate a new one. Fatal error.
                        if(VU.GetTrace() && VU.GetTrace()->IsTrace("request errors"))
                        {
                            GetTraceStream(VU.GetTrace()) << "TRcrdRequest: could not allocate new connection.\r\n";
                            ReleaseTraceStream();
                        }

                        return TOF_RES_FAILED;
                    }
                }

                CStatistics &stats = pConnection->GetStats();
	            if(_retCode)
                {
                    VU.set_variable(_retCode->Varnam(),
                        int_to_string((int)stats.GetHttpReplyCode()),
                        _retCode->Evaluate(ctxt));
                }

                if(szStatus.empty()   == FALSE)
                {
                    if(stats.GetHttpReplyCode() < 400)
                        VU.set_variable(szStatus, int_to_string(1));  // OK.
                    else
                        VU.set_variable(szStatus, int_to_string(0));  // Error.
                }
            }

            VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.
            ctxt.set_phase_text(HTTP_NONE);

            return TOF_RES_OK;
        }
        break;

    default:
        {
            nConid = VU.GetCurrentConid();

            ctxt.set_phase_text(HTTP_NONE);
            if(VU.GetTrace() && VU.GetTrace()->IsTrace("request errors"))
            {
                GetTraceStream(VU.GetTrace()) << "ERROR: entered default case in TRcrdRequest for connection " << nConid << ".\r\n";
                ReleaseTraceStream();
            }
            assert(0);

            return TOF_RES_FAILED;
        }
    };
}


void TRcrdRequest::BuildHeaderString(TContext& ctxt, string &szFormattedHeader) const
{
    list<string> headerList;
    list<string> headerWithList;
    list<string> headerWithoutList;

    {
#ifndef DONTLOCKEVAL
        CLock guard (m_pFastLock);
#endif
        _headerP->  EvaluateValues(ctxt, headerList);
        _withP->    EvaluateValues(ctxt, headerWithList);
	    _woutP->    EvaluateValues(ctxt, headerWithoutList);
    }

	string szWithHeader;
	string szWoutHeader;
    szWithHeader.reserve(255);
    szWoutHeader.reserve(255);

	list<string>::iterator j;
	for(j = headerList.begin(); j != headerList.end(); j++) 
	{
        string &str = *j;
        if(str[str.size()-1] == '\n')
            szFormattedHeader.append(str.data(), str.size()-1); // Remove the \n.
        else
            szFormattedHeader.append(str);
		szFormattedHeader.append("\r\n");
	}

	for(j = headerWithList.begin(); j != headerWithList.end(); j++) 
	{
        string &str = *j;
        if(str[str.size()-1] == '\n')
            szWithHeader.append(str.data(), str.size()-1); // Remove the \n.
        else
            szWithHeader.append(str);
		szWithHeader.append("\r\n");
	}

	for(j = headerWithoutList.begin(); j != headerWithoutList.end(); j++) 
	{
        string &str = *j;
        if(str[str.size()-1] == '\n')
            szWoutHeader.append(str.data(), str.size()-1); // Remove the \n.
        else
            szWoutHeader.append(str);
		szWoutHeader.append("\r\n");
	}

	string_lf2crlf(szFormattedHeader);
	string_lf2crlf(szWithHeader);
	string_lf2crlf(szWoutHeader);

    // Remove the "with" header field
	string::size_type offset = 0;
	string::size_type endLine = 0;
	string::size_type beginField = 0;
	string::size_type endField = 0;

	while ( (offset != string::npos) && ((endField = szWithHeader.find_first_of(" :", offset)) != string::npos) )
	{
		beginField = find_in_string(szFormattedHeader, szWithHeader.substr(offset, endField - offset), false, 0);
		endLine = szFormattedHeader.find("\r\n", beginField);
		if ( (beginField != string::npos) && (endLine != string::npos) )
			szFormattedHeader.erase(beginField, endLine - beginField + 2);
		offset = szWithHeader.find("\r\n", offset);
		offset += 2;
	}

	// Add the "with" header fields
    szFormattedHeader.append(szWithHeader);

	// And remove the "wout" ones
	offset = endLine = beginField = endField = 0;
	while ( (offset != string::npos) && ((endField = szWoutHeader.find_first_of(" \r\n", offset)) != string::npos) )
	{
		beginField = find_in_string(szFormattedHeader, szWoutHeader.substr(offset, endField - offset), false, 0);
		endLine = szFormattedHeader.find("\r\n", beginField);
		if ( (beginField != string::npos) && (endLine != string::npos) )
			szFormattedHeader.erase(beginField, endLine - beginField + 2);
		offset = szWoutHeader.find("\r\n", offset);
		offset += 2;
	}

	// Add an empty line
    szFormattedHeader.append("\r\n");
}

// Build the header string, then parse the server and port strings.
// In HTTP/1.1, should parse the 'Host' header field. Otherwise, should parse the URI string.
// szPort will be empty if no explicit communication port was found.
void TRcrdRequest::ExtractHostPort(TContext& ctxt, const string &szHeader, const string &szUri, string &szRemoteServer, string &szPort) const
{
    // Find the HTTP://... format.
    DWORD dwPos = szUri.find("://");
    assert(dwPos != string::npos);
    dwPos += 3;

    // Found the host name. Search for the port.
    DWORD dwCurs = dwPos;
    while(szUri[dwCurs] != ':' && szUri[dwCurs] != '\r' && szUri[dwCurs] != '\0' && szUri[dwCurs] != '/')
        dwCurs++;

    szRemoteServer.assign(&szUri[dwPos], dwCurs-dwPos);

    if(szUri[dwCurs] == ':')
    {
        // Read the communication port value.
        dwPos = ++dwCurs;
        while(szUri[dwCurs] != '/' && szUri[dwCurs] != '\r' && szUri[dwCurs] != '\0')
            dwCurs++;
        szPort.assign(&szUri[dwPos], dwCurs-dwPos);
    }
    else
    {
        // No explicit port. Return an empty string.
        szPort.erase();
    }
}

//
// Ascii2Binary
// ============
//
// Convert an ascii body to its binary value or generate a body with given size
//
int TRcrdRequest::Ascii2Binary(const string &inBuf, string &outBuf) const
{
	outBuf.reserve(inBuf.size());

    // if inBuf begin with "binary("
    if (begins_with(inBuf, "BINARY("))
    {
        // extract the body length
        int beginLen = strlen("binary(");
        int endLen = inBuf.find_last_not_of(')', beginLen);
        string szLen = inBuf.substr(beginLen, endLen - beginLen);
        int length = atoi(szLen.c_str());
        outBuf.assign(length, 'A');
        return 0;
    }

    // if inBuf does not begin with "0x", its not a binary body
    if (!begins_with(inBuf, "0X"))
    {
        outBuf = inBuf;
        return 0;
    }

    int bufLen = inBuf.size();

    char tmp[2];
    int hexcode;
    for (int i = 2; i < bufLen; i += 2)
    {
        tmp[0] = inBuf[i];
        tmp[1] = inBuf[i+1];
        sscanf(tmp, "%x", &hexcode);
        outBuf.append(1, (char)hexcode);
    }

    return 0;
}

#endif

//
// TRcrdDisconnect::prepare
// ========================
//
// perform any preparation before disconnecting
int TRcrdDisconnect::prepare(TScript* script, int objIndex, bool verbose) 
{
    return 0;
}

//
// TRcrdDisconnect::execute
// ========================
//
// execute the disconnect

#ifndef TH_POOL_VER

int TRcrdDisconnect::execute(TContext& ctxt)
{
    ctxt.set_phase_text(HTTP_DISCONNECT);

	// Get the variables from the tof
    unsigned nConid = intPrmVal(TParam::stConnid, ctxt);
	unsigned nCancel = intPrmVal(TParam::stOptions, ctxt);

    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("DISCONNECT(" + int_to_string(nConid) + ')', TRACEMSG);


	// ------------------------------
    // If there is no connection Id, disconnect all connections
	if (nConid == 0)
	{
		if (nCancel == 0)
		{
			if (CHTTPconn::receiveOnAllConnections(ctxt, 0, NULL) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
			{
				return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
			}
		}

		// Get connections map
		cons_map_t &conMap = ctxt.get_connections();

        cons_map_t::iterator j;

		// Disconnect all connections
		for(j = conMap.begin(); j != conMap.end(); j++) 
		{
			(static_cast<CHTTPconn*>((*j).second))->disconnect();
		}

        // Clear the map (destroy each element)
		while (!conMap.empty())
		{
			j = conMap.begin();
			ctxt.remCon((*j).first);
		}
		conMap.clear();

		ctxt.set_phase_text(HTTP_NONE);
		return 0;
	}
	// ------------------------------


	// ------------------------------
    // Get connection object
    CHTTPconn* pConnection = static_cast<CHTTPconn*>(ctxt.connection(nConid));
    if( pConnection == NULL )
	{
        if(ctxt.GetVerbose(ERRMSG))
            ctxt.log("Invalid connection id(" + int_to_string(nConid) + ") for disconnect", ERRMSG);
        ctxt.set_phase_text(HTTP_NONE);
        return 0; // Safe to carry on
    }
	// ------------------------------


	// ------------------------------
	// Get all remaining data on this connection
	if ( !(pConnection->isReceiveFinished()) && (nCancel == 0) )
	{
		if (CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
		{
			return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
		}
		// CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection);
	}
	// ------------------------------


	// ------------------------------
    // Disconnect and remove this connection object from map
    if( pConnection->disconnect() != 0 )
	{
        if(ctxt.GetVerbose(ERRMSG))
       		ctxt.log("Disconnect failed:" + pConnection->getLastError(), ERRMSG);
    }
	
    ctxt.remCon(nConid);
	// ------------------------------

    ctxt.set_phase_text(HTTP_NONE);

    return 0;
}

#else   // TH_POOL_VER case.

int TRcrdDisconnect::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);

    unsigned nConid = intPrmVal(TParam::stConnid, ctxt);

    if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
    {
        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: entering step " << VU.GetTRecordCallStep() << ".\r\n";
        ReleaseTraceStream();
    }

    switch(VU.GetTRecordCallStep())
    {
    case 0: // Must synchronize on the pending IOs.
        {
            ctxt.set_phase_text(HTTP_DISCONNECT);
	        unsigned nCancel = intPrmVal(TParam::stOptions, ctxt);

            if(ctxt.GetVerbose(TRACEMSG))
            {
                if(nCancel == 0)
                    ctxt.log("DISCONNECT(" + int_to_string(nConid) + ')', TRACEMSG);
                else
                    ctxt.log("DISCONNECT(" + int_to_string(nConid) + "), cancelling the current request(s)", TRACEMSG);
            }

            VU.SetTRecordCallStep(1);   // If SynchroniseHTTPClient returns FALSE then will enter the "case 1" 
                                        // section next time execute is called.
            if(nCancel == 0)
            {
                // Should not cancel the request(s).
                if(nConid == 0)
                {
                    // Synchronise on all clients.
                    if(VU.GetTrace())
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: Synchronising on all connections for " << VU.GetManagerNumber() << ".\r\n";
                        ReleaseTraceStream();
                    }

                    if(VU.SynchroniseAllHTTPClients() == FALSE)
                    {
                        return TOF_RES_ASYNC_STARTED;
                    }
                    else
                    {
                        // No pending requests.
                    }
                }
                else
                {
                    // Synchronise on that specific client.
                    if(VU.GetTrace())
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: Synchronising on " << VU.GetManagerNumber() << '/' << nConid << ".\r\n";
                        ReleaseTraceStream();
                    }

                    if(VU.SynchroniseHTTPClient(nConid) == FALSE)
                    {
                        return TOF_RES_ASYNC_STARTED;
                    }
                }
            }
            else
            {
                // Should finish with the pending IO but not receive all the data.
                if(nConid == 0)
                {
                    // Cancel all queries.
                    if(VU.GetTrace())
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: Cancelling all the requests on " << VU.GetManagerNumber() << ".\r\n";
                        ReleaseTraceStream();
                    }

                    VU.CancelAllQueries();

                    // Synchronise on all clients.
                    if(VU.SynchroniseAllHTTPClients() == FALSE)
                    {
                        return TOF_RES_ASYNC_STARTED;
                    }
                }
                else
                {
                    // Cancel that query.
                    if(VU.GetTrace())
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: Cancelling request " << VU.GetManagerNumber() << '/' << nConid << ".\r\n";
                        ReleaseTraceStream();
                    }

                    VU.CancelQuery(nConid);

                    // Synchronise on that request.
                    if(VU.SynchroniseHTTPClient(nConid) == FALSE)
                    {
                        return TOF_RES_ASYNC_STARTED;
                    }
                }
            }
        }
        // No break since should walk through case 1.

    case 1: // The client(s) is (are) over.
        {
            DWORD dwRes;

            VU.SetTRecordCallStep(2);   // In the case of secure connection, disconnect is a two steps process.
            if(nConid == 0)
            {
                // Disconnect all CAsyncHTTPRequest s.
                dwRes = VU.DisconnectAllClients();
            }
            else
            {
                // Disconnect that CAsyncHTTPRequest.
                dwRes = VU.DisconnectClient(nConid);
            }

            if(dwRes != 0)
            {
                if(dwRes == TOF_RES_ASYNC_STARTED)
                {
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("connection"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: Waiting for an async disconnect to complete.\r\n";
                        ReleaseTraceStream();
                    }
                    return TOF_RES_ASYNC_STARTED;
                }
                else
                {
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: could not disconnect connection " << nConid << ".\r\n";
                        ReleaseTraceStream();
                    }

                    if(ctxt.GetVerbose(ERRMSG))
                        ctxt.log("ERROR DISCONNECTING (" + int_to_string(nConid) + ')', ERRMSG);

                    return TOF_RES_OK;
                    // return TOF_RES_FAILED;   This is not considered as a relevant error.
                }
            }

            break;  // We're over.
        }

    case 2: // The client disconnection is over.
        {
            // Nothing to do.
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("DISCONNECTED (" + int_to_string(nConid) + ')', TRACEMSG);

            break;
        }

    default:
        assert(0);
        break;
    };

    ctxt.set_phase_text(HTTP_NONE);
    VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRcrdDisconnect::execute.

    return TOF_RES_OK;
}

#endif

//
// TRcrdHttpResponse::prepare
// ========================
//
// perform any preparation before reading the response
int TRcrdHttpResponse::prepare(TScript* script, int objIndex, bool verbose)
{
	const TParam* prmP = param(TParam::stData);
	_responseType = prmP->intVal();

	prmP = param(TParam::stConnid);
    string l_conid = prmP->strVal();

	_iConid = atoi(l_conid.c_str());
    for (string::size_type i=0; i<l_conid.size(); i++)
	{
		if (!isdigit(l_conid[i]))
		{
			_iConid = -1 ;
			break ;
		}
	}

    // Get variable string and expression string from 'name' parameters.
    prmP = param(TParam::stInto);
    string l_name = prmP->strVal();
	
    // Create variable reference object, for variable to be set.
    _varP = new TtVarRef(script);
	
    // Encode the variable reference; report an error if encode fails.
    if (!(_varP->Parse(l_name, 0)))
    {
		MSGERR("SET VARIABLE: " + l_name + " Failed during 'into' PREPARE with: " + _varP->Reason());
		return -1;
    }
	
    // Get pointer to TRecord for variable - as we need information about the variable
    // to process the expression string.
    TRecord* trecP = (*script).GetVardef(_varP->Varnam());
    if (trecP == NULL) // If not a defined variable
    {
		MSGERR("SET VARIABLE: " + l_name + " Failed during PREPARE with: Variable '" + _varP->Varnam() + "' not defined");
		return -1;
    }

    // Get variable options mask so that we can check the variable type and so
    // determine what type of expression we need (integer or character).
    prmP = trecP->param(TParam::stOptions);
    if (prmP == NULL) // If no 'options'  parameter (should always be present).
    {
		MSGERR("SET VARIABLE: " + l_name + " Failed during PREPARE with: Variable options missing");
		return -1;
    }

    int options = prmP->intVal();   // Variable options mask

    // Create an expression object (of the appropriate type)
    if (_exprP) delete _exprP;	// Delete any existing object first
	

    if (options & VTCHAR)
    {
		_exprP = new TtChrExpr(script);
    }
    else if (options & VTINT)
    {
        _exprP = new TtIntExpr(script);
    }
    else // Inrecognised variable type.
    {
		MSGERR("SET VARIABLE: " + l_name + " Failed during PREPARE with: Unrecognised variable type");
		return -1;
    }

    // Keep the variable size
    prmP = trecP->param(TParam::stSize);
    if(prmP == FALSE)
    {
        // TODO: ERR
        return -1;
    }
    _varSize = prmP->intVal();   // Variable length.

	prmP = param("identifier");
    if(prmP == NULL) 
		return -1;
    string l_ident = prmP->strVal();
    _identP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_identP->Parse(l_ident, 0)))
    {
		MSGERR("Connect : " + l_ident +
			   " Failed during 'identifier' PREPARE with: " + _identP->Reason());
		return -1;
    }

	if (_responseType == HEADER_VAR)
    {
#ifdef TH_POOL_VER
		unsigned int newIndex = objIndex;
		while (newIndex > 0)
		{
			newIndex--;
			TRecord * prevRecord = script->rec_at_index(newIndex);
			if (prevRecord && prevRecord->isType(TRcrdRequest::KEY))
			{
				TRcrdRequest *prevRequestRecord = (TRcrdRequest *)prevRecord;
				if ( (prevRequestRecord->_iConid == _iConid) || (prevRequestRecord->_iConid == 0) )
				{
                    // Eventually upgrade the store level to HEADER_ONLY.
                    if(prevRequestRecord->_eResultsRequired == NO_RESULT)
                        prevRequestRecord->_eResultsRequired = HEADER_ONLY;
					break;
				}
			}
		}
#endif  // TH_POOL_VER
		return 0 ;
    }

	// if the conid is -1, build a new map.
	if (_iConid == -1)
	{
		_varAddressMap[l_name] = _identP ;
		_identP = NULL ;
		return 0 ;
	}
	else
	{
		int newIndex = objIndex;
		while (newIndex > 0)
		{
			newIndex--;
			TRecord * prevRecord = script->rec_at_index(newIndex) ;
			if (prevRecord && prevRecord->isType(TRcrdRequest::KEY))
			{
				TRcrdRequest *prevRequestRecord = (TRcrdRequest *)prevRecord;
				if ( (prevRequestRecord->_iConid == _iConid) || (prevRequestRecord->_iConid == 0) )
				{
#ifdef TH_POOL_VER
                    // Upgrade the store level to ALL_DATA.
                    prevRequestRecord->_eResultsRequired = ALL_DATA;
#else
					prevRequestRecord->_bSaveDocument = true;
#endif
					break;
				}
			}
		}

		// Try to find a previous Response Record with the same conid 
		// and the same responseType
		while (objIndex > 0)
		{
			objIndex -- ;
			TRecord * prevRecord = script->rec_at_index(objIndex) ;
			// check that the previous record is also a TRcrdHttpResponse Record
			// with type==BODY
			if (prevRecord->isType(KEY))
			{
				TRcrdHttpResponse * prevResponseRecord = (TRcrdHttpResponse *)prevRecord ; 
				if (prevResponseRecord->_iConid == _iConid && 
					prevResponseRecord->_responseType == _responseType)
				{
					_addressQueryList = prevResponseRecord->_addressQueryList ;
					_varAddressMap = prevResponseRecord->_varAddressMap ;
					prevResponseRecord->_addressQueryList.clear() ;
					prevResponseRecord->_varAddressMap.clear() ;
					break ;
				}
			}
			else 
				break ;
		}
	}

	// If the address begins by " and finishes by " and 
	// there are no other " in the address, remove the "
	// else has to evaluate the address in execute later
	string address = l_ident ;
	if (address[0] == '"' && address.find("\"", 1) == address.size() - 1)
		address = address.substr(1, address.size() - 2);
	else
	{
		_varAddressMap[l_name] = _identP ;
		_identP = NULL ;
		return 0 ;
	}

	if (!address.empty())
	{
		AddressQuery * aq = new AddressQuery(l_name, address.c_str());
		// Normally, it should be a valid address, if not, will be evaluated in
		// the next
		if (aq->IsValidAddress())
		{
			_addressQueryList.push_back(aq) ;
			delete _identP ;
			_identP = NULL ;
		}
		else
		{
			delete aq ;
			_varAddressMap[l_name] = _identP ;
			_identP = NULL ;
			return 0 ;
		}
	}
	else
	{
		_varAddressMap[l_name] = _identP ;
		_identP = NULL ;
	}
/*
#ifdef TH_POOL_VER
    // TODO: Check that optimisation.
    const TParam* prm = param("data");
    if(prm == NULL)
        return -1;
    int nType = prm->intVal();
    prm = param("conid");
    if(prm == NULL)
        return -1;
    int nConid = prm->intVal();
    if(nType == HEADER_VAR)
    {
        // TODO: set "_eResultsRequired = HEADER_ONLY;" in the preceding TRcrdRequest on the nConid connection.
    }
#endif
*/
    return 0;
}



//
// TRcrdHttpResponse::execute
// ========================
//
// Wait for the response and parse the variables values.

#ifndef TH_POOL_VER

int TRcrdHttpResponse::execute(TContext& ctxt)
{
    ctxt.set_phase_text(HTTP_RESPONSE);
	int ret ;

	// if both map are empty, return

    int nConid      = intPrmVal(TParam::stConnid, ctxt);
    int nType       = intPrmVal(TParam::stData, ctxt);
    string szName ;
	string szIdent ;
	if (nType == HEADER_VAR)
	{
		szName   = strPrmVal(TParam::stInto, ctxt);
		szIdent  = "";
		{
			CTLockGuard<CTMutex> guard(m_mutex);
			szIdent =_identP->Evaluate(ctxt);
		}

        if(ctxt.GetVerbose(TRACEMSG))
            ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Name: "
                    + szName + " Identity: " + szIdent, TRACEMSG);
	}
	else if (_addressQueryList.empty() && _varAddressMap.empty())
		return 0;

    // ------------------------------
    // Get connection object
    CHTTPconn* pConnection = static_cast<CHTTPconn*>(ctxt.connection(nConid));
    if(pConnection == NULL)
    {
        // print statistics before exiting
        CHTTPconn::printAllStatistics(ctxt);

        ctxt.log("Invalid connection id(" + int_to_string(nConid) + ") for HttpResponse", ERRMSG);
        throw mdbParseException("Invalid connection id(" + int_to_string(nConid) + ") for HttpResponse");
        //return -1;
    }
    // ------------------------------

    // ------------------------------
    // Get header or all remaining data on this connection

    if ( !(pConnection->isReceiveFinished()) )
    {
        if (CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection, nType) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
        {
            return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
        }
    }

	// Header variable
	if (nType == HEADER_VAR)
	{
		int ret = 0;
		// If szIdent is empty, put the document in variable regards to its size
		if (szIdent == "")
			ret = ctxt.set_variable(szName, pConnection->getHttpResponse(HEADER_VAR, _varSize));
		//
		else
			ret = ctxt.set_variable(szName, pConnection->getHeaderResponse(szIdent));
	}
	// Body variables
	else
	{
		// ------------------------------
		//
		Parser::AddressValueList avl;
		TtChrExpr  * pIdent ;

		AddressQueryList tmpAddressQueryList = _addressQueryList;
		AddressQueryList newAddressQueryList;

		// Evaluate the variable addresses and add fill an AddressQueryList
	    for (variableAddressMap::iterator itr = _varAddressMap.begin(); itr != _varAddressMap.end(); itr ++)
	    {
		    pIdent = itr->second ;
		    if (pIdent)
		    {
			    string address ;
			    {
				    CTLockGuard<CTMutex> guard(m_mutex);
				    address = pIdent->Evaluate(ctxt);
			    }

			    // If the address is empty, get all the document.
			    if(address.empty())
			    {
                    if(ctxt.GetVerbose(TRACEMSG))
				        ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Name: " + itr->first + " Identity: " + address, TRACEMSG);

				    ret = ctxt.set_variable(itr->first, pConnection->getHttpResponse(BODY_VAR, _varSize));
				    if(ret == -1)
				    {
					    // print statistics before exiting
					    CHTTPconn::printAllStatistics(ctxt);

					    ctxt.log("Set variable failed(" + itr->first + ')', ERRMSG);
					    throw mdbParseException("Set variable failed(" + itr->first + ')');
				    }
				    continue ;
			    }

			    // put the addressQuery in the addressQueryList.
			    AddressQuery *newAddressQuery  = new AddressQuery(itr->first, address.c_str());
			    newAddressQueryList.push_back(newAddressQuery) ;
		    }
	    }

        // Merge the two lists.
        tmpAddressQueryList.insert(tmpAddressQueryList.end(), newAddressQueryList.begin(), newAddressQueryList.end());

		// Use the HTML parser to  get the values associated with addresses.
		if (pConnection->getBodyResponse(tmpAddressQueryList, avl))
		{
			// assciate each addressValue to each variable.
			for (Parser::AddressValueList::iterator avlItr = avl.begin(); avlItr != avl.end(); avlItr ++)
			{
				// get the variable name associated with the AddressQuery
				const AddressQuery * aq = avlItr->pAq_ ;
				string vName = aq->strVariable_ ;	

                if(ctxt.GetVerbose(TRACEMSG))
                    ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Name: "
							 + vName + " Identity: " + (aq->address_).ToString(), TRACEMSG);

				ret = ctxt.set_variable(vName, string((const char *)(avlItr->pcBegin_), avlItr->ulSize_));
				if(ret == -1)
				{
					// print statistics before exiting
					CHTTPconn::printAllStatistics(ctxt);

					ctxt.log("Set variable failed(" + vName + ")", ERRMSG);
					throw mdbParseException("Set variable failed(" + vName + ")");
				}
			}
		}

        // Delete the dynamicaly evaluated variables.
		for (AddressQueryList::iterator newitr = newAddressQueryList.begin(); newitr != newAddressQueryList.end(); newitr ++)
		{
			delete *newitr;
		}
	}


    if(ret == -1)
    {
        // print statistics before exiting
        CHTTPconn::printAllStatistics(ctxt);

        ctxt.log("Set variable failed(" + szName + ")", ERRMSG);
        throw mdbParseException("Set variable failed(" + szName + ")");
    }
    // ------------------------------


    ctxt.set_phase_text(HTTP_NONE);

    return 0;
}

#else   // TH_POOL_VER case.

template <class ACCESSOR_TYPE>
BOOL ParseBodyBuffers(CDataBuffHandler &messageBuffer, BOOL bContainContinueHeader, const AddressQueryList &aql, PhParser<ACCESSOR_TYPE>::AddressValueList &avl, unsigned long* pulErrorLine, unsigned long* pulErrorCol);

template <class ACCESSOR_TYPE>
BOOL StoreResults(TContext &ctxt, PhParser<ACCESSOR_TYPE>::AddressValueList &avl);

int TRcrdHttpResponse::execute(TContext& ctxt)
{
    int nType = intPrmVal(TParam::stData, ctxt);  // Location of the data: HEADER or BODY.

    if(nType == BODY_VAR && _addressQueryList.empty() && _varAddressMap.empty())
    {
        // Nothing to parse. Lload instructions were concatenated into a single one. Parsing will be performed in one of the next Lloads.
		return TOF_RES_OK;
    }

    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);

    int nConid;

    if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
    {
        GetTraceStream(VU.GetTrace()) << "TRcrdHttpResponse: entering step " << VU.GetTRecordCallStep() << ".\r\n";
        ReleaseTraceStream();
    }

    switch(VU.GetTRecordCallStep())
    {
    case 0: // Must synchronize on the pending IOs.
        {
            ctxt.set_phase_text(HTTP_RESPONSE);

            nConid = intPrmVal(TParam::stConnid, ctxt);
            VU.SetCurrentConid(nConid);

            // Synchronise on that request.
            VU.SetTRecordCallStep(1);   // In the case of secure connection, disconnect is a two steps process.
            if(VU.SynchroniseHTTPClient(nConid) == FALSE)
            {
                // Download on that connection is not over yet. Cannot continue.
                return TOF_RES_ASYNC_STARTED;
            }
        }
        // No break since should continue if no need to synchronize.

    case 1:
        {
            string szVarVal;
            nConid = VU.GetCurrentConid();

            CAsyncHTTPRequest* pConnection = VU.ObtainRequestStruct(nConid);
            if(pConnection == NULL)
            {
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdHttpResponse: Invalid connection id(" << nConid << ").\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("HTTPRESPONSE: Invalid connection id(" + int_to_string(nConid) + ')', ERRMSG);
                return TOF_RES_FAILED;
	        }

            CDataBuffHandler &messageBuff = pConnection->GetRequestResults();

            if(messageBuff.IsEmpty() == TRUE)
            {
                // ERROR: the data is not available (connection was broken inbetween ...).
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdHttpResponse: No data available for connection id(" << nConid << ").\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("HTTPRESPONSE: No data available for connection id(" + int_to_string(nConid) + ')', ERRMSG);
                return TOF_RES_FAILED;
            }

            BOOL bFoundContinueHeader = pConnection->FoundContinueHeader();

            if(nType == HEADER_VAR)
            {
		        string szName = strPrmVal(TParam::stInto, ctxt);
                string szIdent;
		        {
#ifndef DONTLOCKEVAL
                    CLock guard (m_pFastLock);
#endif
			        szIdent = _identP->Evaluate(ctxt);
		        }

                if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdHttpResponse: (" << nConid << "): Parsing header contents. Name: " << szName << " Identity: " << szIdent << ".\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(TRACEMSG))
                    ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Parsing header contents. Name: " + szName + " Identity: " + szIdent, TRACEMSG);

                if(szIdent.empty() == TRUE)
                {
                    // Return up to _dwVarLength characters from the header in szVarVal.
                    DumpHeader(messageBuff, szVarVal, _varSize, bFoundContinueHeader);
                }
                else
                {
                    // Parse header/cookie value matching expression szIdent.
                    ParseHeaderValue(messageBuff, szIdent, bFoundContinueHeader, szVarVal);
                }

                if(ctxt.GetVerbose(TRACEMSG))
                    ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Setting " + szName + " to \'" + szVarVal + '\'', TRACEMSG);

                // Store the header results.
				if(ctxt.set_variable(szName, szVarVal) == -1)
			    {
					// Error storing the results. Print statistics before exiting

                    string msg = "Set variable failed(" + szName + ')';

                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                    {
                        GetTraceStream(VU.GetTrace()) << msg << ".\r\n";
                        ReleaseTraceStream();
                    }

                    ctxt.log(msg, ERRMSG); throw mdbParseException(msg);
			    }
            }
            else    // nType == BODY_VAR.
            {
                // Handle multiple-address parsing at a time.
	            TtChrExpr *pIdent;

	            AddressQueryList tmpAddressQueryList = _addressQueryList;
	            AddressQueryList newAddressQueryList;

                try
                {
	                // Evaluate the variable addresses.
                    // Addresses that need parsing will be added to newAddressQueryList; 'empty' addresses will be evaluated there.
	                for (variableAddressMap::iterator itr = _varAddressMap.begin(); itr != _varAddressMap.end(); itr ++)
	                {
		                pIdent = itr->second ;
		                if (pIdent)
		                {
			                string address ;
			                {
#ifndef DONTLOCKEVAL
                                CLock guard (m_pFastLock);
#endif
                                address = pIdent->Evaluate(ctxt);
			                }

			                if(address.empty())
			                {
    			                // The address is empty, get all the document.
                                if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
                                {
                                    GetTraceStream(VU.GetTrace()) << "TRcrdHttpResponse: (" << nConid << "): Parsing header contents. Name: " << itr->first << " Identity: " << address << ").\r\n";
                                    ReleaseTraceStream();
                                }

                                if(ctxt.GetVerbose(TRACEMSG))
				                    ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Name: " + itr->first + " Identity: " + address, TRACEMSG);

                                DumpBody(messageBuff, szVarVal, _varSize, bFoundContinueHeader);
				                if(ctxt.set_variable(itr->first, szVarVal) == -1)
				                {
					                // Error storing the results. Print statistics before exiting

                                    string msg = "Set variable failed(" + itr->first + ')';

                                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                                    {
                                        GetTraceStream(VU.GetTrace()) << msg << ".\r\n";
                                        ReleaseTraceStream();
                                    }

                                    ctxt.log(msg, ERRMSG); throw mdbParseException(msg);
				                }
			                }
                            else
                            {
			                    // Need to use the parser. Put the addressQuery in the addressQueryList.
			                    AddressQuery *newAddressQuery  = new AddressQuery(itr->first, address.c_str());
			                    newAddressQueryList.push_back(newAddressQuery) ;
                            }
		                }
                        else
                        {
                            assert(0);
                        }
	                }

                    // Merge the static and variable address lists.
                    tmpAddressQueryList.insert(tmpAddressQueryList.end(), newAddressQueryList.begin(), newAddressQueryList.end());

                    BOOL bRes;
                    unsigned long ulErrorLine, ulErrorCol;

                    // Call the HTML parser on the address list.
                    if(pConnection->GetMessageLengthType() == CAsyncHTTPRequest::MSG_NON_IDENTITY)
                    {
                        // Data in chunk format.
	                    PhParser<CDataBuffListChunkAccessor>::AddressValueList avl;
                        bRes = ParseBodyBuffers<CDataBuffListChunkAccessor> (   pConnection->GetRequestResults(), 
                                                                                pConnection->DoesContainContinueHeader(), 
                                                                                tmpAddressQueryList,
                                                                                avl,
                                                                                &ulErrorLine,
                                                                                &ulErrorCol);
                        if(bRes == TRUE)
                        {
                            bRes = StoreResults<CDataBuffListChunkAccessor> (ctxt, avl);
                        }
                        else
                        {
                            if(ctxt.GetVerbose(ERRMSG))
                                ctxt.log("HTTPRESPONSE: HTML parser error for connection (" + int_to_string(nConid) + ") on line " + int_to_string((long)ulErrorLine) + ", column " + int_to_string((long)ulErrorCol), ERRMSG);
                            assert(0);
                        }
                    }
                    else
                    {
                        // Standard HTTP data.
	                    PhParser<CDataBuffListAccessor>::AddressValueList avl;
                        bRes = ParseBodyBuffers<CDataBuffListAccessor> (    pConnection->GetRequestResults(), 
                                                                            pConnection->DoesContainContinueHeader(), 
                                                                            tmpAddressQueryList, 
                                                                            avl,
                                                                            &ulErrorLine,
                                                                            &ulErrorCol);
                        if(bRes == TRUE)
                        {
                            bRes = StoreResults<CDataBuffListAccessor> (ctxt, avl);
                        }
                        else
                        {
                            if(ctxt.GetVerbose(ERRMSG))
                                ctxt.log("HTTPRESPONSE: HTML parser error for connection (" + int_to_string(nConid) + ") on line " + int_to_string((long)ulErrorLine) + ", column " + int_to_string((long)ulErrorCol), ERRMSG);
                            assert(0);
                        }
                    }
                }
                catch(mdbParseException &)
                {
                    // Perform cleanup: delete the dynamicaly evaluated variables.
	                for (AddressQueryList::iterator newitr = newAddressQueryList.begin(); newitr != newAddressQueryList.end(); newitr ++)
		                delete *newitr;

                    throw;
                }
            }   // End of BODY_VAR section.

            break;
        }

    default:
        {
            assert(0);
        }
    };

    ctxt.set_phase_text(HTTP_NONE);

	return TOF_RES_OK;
}

void SkipHeader(CDataBuffListAccessor &iter, BOOL bFoundContinueHeader)
{
    // First skip the CONTINUE header if exists.
    if(bFoundContinueHeader)
    {
        do
        {
            iter.SkipCRLF();
        }
        while(iter.IsEndOfBuffs() == FALSE && iter.IsCRLF() == FALSE);
        iter.SkipCRLF();
    }

    // Skip the real header.
    do
    {
        iter.SkipCRLF();
    }
    while(iter.IsEndOfBuffs() == FALSE && iter.IsCRLF() == FALSE);
    iter.SkipCRLF();
}

// Return up to dwVarLength characters from the header or body (depending on nDataArea) in szVarVal.
// WARNING: in the case of headers, returns only the real header (skip the CONTINUE header if applicable).
void TRcrdHttpResponse::DumpHeader(CDataBuffHandler &messageBuff, string &szVarVal, DWORD dwVarLength, BOOL bFoundContinueHeader) const
{
    CDataBuffListAccessor iter (&messageBuff.GetWholeBuffList());
    CDataBuffListAccessor iterEnd (iter);

    if(bFoundContinueHeader)
    {
        SkipHeader(iterEnd, FALSE);
    }

    // Compute header length.
    SkipHeader(iterEnd, FALSE);
    DWORD dwHeaderLen = iterEnd-iter;

    iter.BuildString(szVarVal, __min(dwVarLength, dwHeaderLen));
}

void TRcrdHttpResponse::DumpBody(CDataBuffHandler &messageBuff, string &szVarVal, DWORD dwVarLength, BOOL bFoundContinueHeader) const
{
    CDataBuffListAccessor iter (&messageBuff.GetWholeBuffList());

    // Skip the header.
    SkipHeader(iter, bFoundContinueHeader);

    // Build the result.
    iter.BuildString(szVarVal, dwVarLength);
}

template <class ACCESSOR_TYPE>
BOOL ParseBodyBuffers(CDataBuffHandler &messageBuffer, BOOL bContainContinueHeader, const AddressQueryList &aql, PhParser<ACCESSOR_TYPE>::AddressValueList &avl, unsigned long* pulErrorLine, unsigned long* pulErrorCol)
{
    // Move to the begining of the body section (use generic iterator).
    CDataBuffListAccessor iterSkipHeader(&messageBuffer.GetWholeBuffList());
    SkipHeader(iterSkipHeader, bContainContinueHeader);

    // Parse. Use either chunk or regular iterator.
    ACCESSOR_TYPE iterData;
    iterData.MoveIterTo(iterSkipHeader);

    PhLexer<ACCESSOR_TYPE>  lexer   (iterData);
    PhParser<ACCESSOR_TYPE> parser  (lexer, TRUE);

    return parser.Parse(aql, avl, pulErrorLine, pulErrorCol);
}


template <class ACCESSOR_TYPE>
BOOL StoreResults(TContext &ctxt, PhParser<ACCESSOR_TYPE>::AddressValueList &avl)
{
    string szVarVal;

	// Associate each addressValue to each variable.
	for(PhParser<ACCESSOR_TYPE>::AddressValueList::iterator avlItr = avl.begin(); avlItr != avl.end(); avlItr ++)
	{
        // Build the request result.
        ACCESSOR_TYPE &iterResultBeg = avlItr->pcBegin_;

        if(iterResultBeg != NULL)
        {
            if(iterResultBeg.BuildString(szVarVal, avlItr->ulSize_) == FALSE)
            {
                assert(0);
                return FALSE;
            }
        }

		// Get the variable name associated with the AddressQuery.
		const AddressQuery * aq = avlItr->pAq_ ;
		string vName = aq->strVariable_ ;	

        if(ctxt.GetVerbose(TRACEMSG))
        	ctxt.log("HTTPRESPONSE: Name: " + vName + " Identity: " + (aq->address_).ToString() + " Value:" + szVarVal, TRACEMSG);

        // Store the results.
		int ret = ctxt.set_variable(vName, szVarVal);
		if(ret == -1)
		{
			// Error storing the results. Print statistics before exiting.
            string msg = "Set variable failed(" + vName + ')';
			ctxt.log(msg, ERRMSG); throw mdbParseException(msg);
		}
	}

    return TRUE;
}

//
// Find a cookie value.
// This function expects there is only one cookie per "Set-Cookie" line in the header
// (ie: GET ...
//      Set-Cookie: name1 = value1
//      Set-Cookie: name2 = value2
// )
//
BOOL TRcrdHttpResponse::ParseCookieValue(CDataBuffHandler &messageBuff, const string &zName, BOOL bFoundContinueHeader, string &zCookieVal) const
{
    zCookieVal.erase();

    // Use regular expressions to find the cookie name.
    string zModifName = '^' + zName;
    string::size_type i = 0;
    while ( (i = zModifName.find_first_of('.', i)) != string::npos )
    {
        zModifName.insert(i, 1, '\\');
        i += 2;
    }
    i = 0;
    while ( (i = zModifName.find_first_of('*', i)) != string::npos )
    {
        zModifName.insert(i, 1, '.');
        i += 2;
    }
    i = 0;
    while ( (i = zModifName.find_first_of('?', i)) != string::npos )
    {
        zModifName.replace(i, 1, 1, '.');
        i += 2;
    }

    // Add '=' at end of cookie's name to differentiate cookies like "foo"
    // from those like "foofoo"
    // TODO: is that possible that we find tabs or spaces before '='?
    zModifName += '=';

    CRegExp regExp;
    regExp.RegComp(zModifName.c_str());

    CDataBuffListAccessor iter (&messageBuff.GetWholeBuffList());

    // First skip the CONTINUE header if exists.
    if(bFoundContinueHeader == TRUE)
    {
        SkipHeader(iter, FALSE);    // Set FALSE to skip only the first header.
    }

    while(iter.IsEndOfBuffs() == FALSE && iter.IsCRLF() == FALSE)
    {
        // Iterate through the header lines.
        // Skip white spaces (that might not be necessary, but we never know ...).
        if(iter.SkipWS() == FALSE) { assert(0); return FALSE; /*Found nothing but W.S.*/ }

        if(iter.MatchFirstWordToLower("set-cookie", sizeof "set-cookie" - 1) == TRUE)
	    {
            iter.Skip(sizeof "set-cookie" - 1);

		    // Skip the column.
            if(iter.SkipChar(':') == FALSE) { assert(0); return FALSE; }

		    // Skip the LWS.
            if(iter.SkipWS() == FALSE)      { assert(0); return FALSE; }

            // Build the cookie name string (+ the equal sign) and check if it matches.
            string zFoundCookieName;
            CDataBuffListAccessor base = iter;

            if(iter.SkipChar('=') == FALSE)                                                 { assert(0); return FALSE; }

            base.BuildString(zFoundCookieName, iter-base);

            BOOL bFound = (regExp.RegFind(zFoundCookieName.c_str()) != -1);

		    // Skip the LWS.
            if(iter.SkipWS() == FALSE)                                                      { assert(0); return FALSE; }

            // Copy the value.
            if(bFound)
            {
                // Return the found value.
                base = iter;
                if(iter.SkipTillAny(";\r\n", sizeof ";\r\n" - 1) == FALSE)                  { assert(0); return FALSE; }
                base.BuildString(zCookieVal, iter-base);
                zCookieVal = zFoundCookieName + zCookieVal;
                return TRUE;
            }
            else
            {
                // Skip to next header line.
                if(iter.SkipCRLF() == FALSE)                                                { assert(0); return FALSE; }
            }
	    }
        else
        {
            // Not a "Set-Cookie" line. Move to next line.
            iter.SkipCRLF();
        }
    }   // Parsing header lines.

#ifdef __DLLSA__

    // If this cookie-name is not available, show the others Set-Cookie fields (only for unique replay).
    logr->AddTrace(USERTRACE, ERRORLEVEL, ("Cannot find cookie " + zName + ", must be one of the following:").c_str());

    CDataBuffListAccessor iterDumpCookies(&messageBuff.GetWholeBuffList());

    // First skip the CONTINUE header if exists.
    if(bFoundContinueHeader == TRUE)
    {
        SkipHeader(iterDumpCookies, FALSE); // Set FALSE to skip only the first header.
    }

    while(iterDumpCookies.IsEndOfBuffs() == FALSE && iterDumpCookies.IsCRLF() == FALSE)
    {
        // Iterate through the header lines.
        // Skip white spaces (that might not be necessary, but we never know ...).
        if(iterDumpCookies.SkipWS() == FALSE) { assert(0); /*Found nothing but W.S.*/ }
        if(iterDumpCookies.MatchFirstWordToLower("set-cookie", sizeof "set-cookie" - 1) == TRUE)
	    {
            // Add end of line to msg.
            iterDumpCookies.SkipChar(':');
            iterDumpCookies.SkipWS();
            DWORD dwEOL;
            if(iterDumpCookies.FindAny("\r\n;,", 4, dwEOL) == FALSE)  { assert(0); break; }
            string zCookie;
            iterDumpCookies.BuildString(zCookie, dwEOL);
            logr->AddTrace(USERTRACE, ERRORLEVEL, zCookie.c_str());
        }

        iterDumpCookies.SkipCRLF();
    }
    logr->AddTrace(USERTRACE, ERRORLEVEL, "-- end of the available cookies list --");

#endif  __DLLSA__

    zCookieVal = zName + '=';

    return FALSE;  // Not found.
}

// Browse through the header searching for a header-field definition.
void TRcrdHttpResponse::ParseHeaderValue(CDataBuffHandler &messageBuff, const string &zIdent, BOOL bFoundContinueHeader, string &zHeaderValue) const
{
    DWORD dwPos = strinfind(zIdent.c_str(), zIdent.size() -1, "set-cookie", sizeof "set-cookie"-1);
    if(dwPos != ULONG_MAX)
    {
        // Requested to parse a cookie data from messageBuff.

        // Evaluate the cookie name.
        string::size_type beginName = zIdent.find_first_not_of(", ", (string::size_type) dwPos + strlen("Set-Cookie"));

		if(beginName == string::npos)
        {
            assert(0);
            zHeaderValue.erase();
			return ;
        }

		string::size_type endName = zIdent.find_first_of(" \r\n", beginName);

		if(endName == string::npos)
			endName = zIdent.size();

		string zCookieName = zIdent.substr(beginName, endName - beginName);

        // Parse the cookie value.
        if(ParseCookieValue(messageBuff, zCookieName, bFoundContinueHeader, zHeaderValue) == FALSE)
        {
            // TODO: error.
        }
    }
    else
    {
        // Parse header value matching expression zIdent.
        CDataBuffListAccessor iter (&messageBuff.GetWholeBuffList());

        // First skip the CONTINUE header if exists.
        if(bFoundContinueHeader == TRUE)
        {
            SkipHeader(iter, FALSE);    // Set FALSE to skip only the first header.
        }

        // Iterate through the header lines.
        while(iter.IsCRLF() == FALSE && iter.IsEndOfBuffs() == FALSE)   // Not the end of the header section.
        {
            // Skip white spaces (that might not be necessary, but we never know ...).
            if(iter.SkipWS() == FALSE)  { assert(0); break; /* TODO: Found nothing but W.S. Error !!!*/ }

            if(iter.MatchFirstWordToLower(zIdent.c_str(), zIdent.size() - 1) == TRUE)
	        {
                // Found the right header field. Return the associated value.
                if(iter.Skip(zIdent.size()) == FALSE)           { assert(0); break; }
                if(iter.SkipChar(':') == FALSE)                 { assert(0); break; }
                if(iter.SkipWS() == FALSE)                      { assert(0); break; }

                CDataBuffListAccessor tmpIter = iter;

                if(tmpIter.SkipCRLF() == FALSE)                 { assert(0); break; }

                if(iter.BuildString(zHeaderValue, tmpIter-iter-2) == FALSE) { assert(0); /* (Should not happen) */ }

                return;
            }
            else
            {
                // Move to next header-field.
                if(iter.SkipCRLF() == FALSE)                    { assert(0); break; }
            }
        }

        zHeaderValue.erase();    // TODO : not found.
    }
}

#endif


//
// TRcrdSynchronize::prepare
// ========================
//
// prepare
int TRcrdSynchronize::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prmP = param(TParam::stTimeOut);
    if (prmP == NULL) 
		return 0;
    int tgtvar = prmP->intVal();

	if (tgtvar == 0)
	    return 0;
    _labelIndex = PrepareGoto(this, TParam::stTmolab, script, true);
    return 0;
}


//
// TRcrdSynchronize::execute
// ========================
//
// read the response on all connections

#ifndef TH_POOL_VER

int TRcrdSynchronize::execute(TContext& ctxt)
{
    ctxt.set_phase_text(HTTP_SYNCHRONIZE);
    
    int nTimeout = intPrmVal(TParam::stTimeOut, ctxt);

    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("SYNCHRONIZE", TRACEMSG);

	CHTTPconn::enumReceiveCode retCode;

	if (nTimeout == 0)
		retCode = CHTTPconn::receiveOnAllConnections(ctxt, 0, NULL);
	else
		retCode = CHTTPconn::receiveOnAllConnections(ctxt, nTimeout, NULL, 0, true);
	
	if (retCode == CHTTPconn::HTTP_RECEIVE_EXIT_ON_LOCAL_TIMER)
	{
	    ctxt.set_phase_text(HTTP_NONE);
		return ExecuteGoto(this, ctxt, TParam::stTmolab, &_labelIndex, &m_mutex);
	}
	else
	if (retCode == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
	{
		return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
	}
	
    ctxt.set_phase_text(HTTP_NONE);

	return 0;
}

#else   // TH_POOL_VER case.

int TRcrdSynchronize::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);

    if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
    {
        GetTraceStream(VU.GetTrace()) << "TRcrdSynchronize: entering step " << VU.GetTRecordCallStep() << ".\r\n";
        ReleaseTraceStream();
    }

    switch(VU.GetTRecordCallStep())
    {
    case 0:
        {
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("SYNCHRONIZE", TRACEMSG);

            ctxt.set_phase_text(HTTP_SYNCHRONIZE);

            int nTimeout    = intPrmVal(TParam::stTimeOut, ctxt);
            int nGoto       = intPrmVal(TParam::stTmolab, ctxt);

            // Synchronise on that VU's requests.
            VU.SetTRecordCallStep(1);
            if(nTimeout == 0)
            {
                // Either value not specified or should wait indefinitely.
                if(VU.SynchroniseAllHTTPClients() == FALSE)
                {
                    // Downloads on that VU are not over yet. Must wait.
                    return TOF_RES_ASYNC_STARTED;
                }
            }
            else
            {
                // ERROR: do not care for time-outs with Thread pool version.
                assert(0);
            }
        }
        // No break since should move to case 1 if all the users are already synchronized.

    case 1:
        {
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("All requests synchronised.", TRACEMSG);

            ctxt.set_phase_text(HTTP_NONE);

            break;
        }

    default:
        {
            assert(0);
        }
    };

    VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.

	return TOF_RES_OK;
}

#endif

#ifndef TH_POOL_VER

//
// TRcrdWait::execute
// ========================
//
// receive some data from each connection during an amount of time 

int TRcrdWait::execute(TContext& ctxt)
{
	ctxt.set_phase_text(HTTP_WAIT);
    
	int nDelay = intPrmVal(TParam::stTDelay, ctxt);
	
    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("WAIT : Delay: " + int_to_string(nDelay), TRACEMSG);

	unsigned percent = ctxt.get_delay_percent();
	unsigned long realDelay = (nDelay*percent)/100 + (nDelay*percent)%100;

	if (CHTTPconn::receiveOnAllConnections(ctxt, realDelay, NULL) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
	{
		return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
	}

	ctxt.set_phase_text(HTTP_NONE);

	return 0;
}                                   

#else   // TH_POOL_VER case.

int TRcrdWait::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);

    switch(VU.GetTRecordCallStep())
    {
    case 0:
        {
            ctxt.set_phase_text(PHASE_WAIT);

            unsigned nDelay = intPrmVal(TParam::stTDelay, ctxt);

            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("WAIT : Delay: " + int_to_string(nDelay), TRACEMSG);

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("sync"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdWait: Initiating a wait operation for user " << VU.GetManagerNumber() << ". Duration is " << nDelay << ".\r\n";
                ReleaseTraceStream();
            }

            VU.SetTRecordCallStep(1);   // Will enter the "case 1" section next time execute is called.
            if(VU.StartEventTimer(nDelay) == FALSE)
            {
                assert(0);
                return TOF_RES_FAILED;
            }

            return TOF_RES_ASYNC_STARTED;  // Should let the framework call us again when the timer op. is completed.
        }

    case 1:
        {
            if(VU.GetTrace() && VU.GetTrace()->IsTrace("sync"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdWait: Waking up after a wait operation for user " << VU.GetManagerNumber() << ".\r\n";
                ReleaseTraceStream();
            }

            ctxt.set_phase_text(PHASE_NONE);

            VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.

            return TOF_RES_OK;   // Move to next TOF object.
        }

    default:
        VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.
        ctxt.set_phase_text(PHASE_NONE);
        assert(0);

        return TOF_RES_FAILED;
    };
}

#endif


#ifndef TH_POOL_VER

///////////////////////////////////////////////////////////////////////////////
// HELPER FUNCTIONS
// ================
///////////////////////////////////////////////////////////////////////////////
int server_specific_init(void)
{
    char* maxthr = getenv("TE_MAX_THREADS");
    int imaxthr;
    if(maxthr != NULL)
    {
		imaxthr = atoi(maxthr);
    }
    else
    {
		imaxthr = DEFAULT_THREAD_MAX;
    }

	CHTTPconn::statmtx = new CTMutex[(imaxthr/10) + 1];

	CHTTPconn::initHTMLParser();

    return 0;
}

// Perform any server specific one time cleanup
int server_specific_cleanup(void)
{
	if (CHTTPconn::statmtx)
		delete[] CHTTPconn::statmtx;

    return 0;
}

#else

// No equivalent in the thread pool model.

#endif


