/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_ASYNFILE_H_
#define _OPENSTA_ASYNFILE_H_

////////////////////////////////////////////////////////////////////////////////
// Asynchronous file definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "AsyncDevice.hpp"
#include "Iocp.hpp"

// This class implements an asynchronous file bound to an I/O completion port.
// THREAD UNSAFE. This should not be a problem, since an asynchronous device should
// always be accessed by a single thread at a time. The derivatives should be thread
// unsafe as well.
class CAsyncFile : public CAsyncDevice
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    CAsyncFile(CIOCP& iocp, CTrace* pTrace = NULL) : CAsyncDevice(iocp, pTrace)
    {
        m_hDevice = NULL;
    }
    virtual ~CAsyncFile()
    {
        if(m_hDevice != NULL)
            CloseHandle(m_hDevice);
    }

    //// --------------------------------------------------------------------------------
    //// User methods.
    //// --------------------------------------------------------------------------------

    // Open an existing file.
    // The file handle is then bound to the IOCP.
    // The dwDesiredAccess argument has the same semantics as its homonym
    // in the CreateFile() API.
    // Return value: 0 = success, other = GetLastError() failure code.
    virtual DWORD Open(LPCSTR zFilePath, DWORD dwDesiredAccess, IocpEvent *pEventTarget)
    {
/*        DWORD dwRes = 0;
        HANDLE hF = CreateFile(zFilePath, dwDesiredAccess, 0, NULL, OPEN_EXISTING,
            FILE_FLAG_OVERLAPPED|FILE_FLAG_RANDOM_ACCESS, NULL);
        if(hF == INVALID_HANDLE_VALUE)
            dwRes = GetLastError();
        else
        {
            // Bind the file handle to the completion port.
            dwRes = BindToIocp(hF, pEventTarget);
            if(dwRes == 0)
                m_hDevice = hF;
        }

        return dwRes;
*/
        return Open(zFilePath, dwDesiredAccess, OPEN_EXISTING, pEventTarget);
    }

//TODO: that one should be enough. The point is we have to define the pure virtual CAsyncDevice::Open.
    virtual DWORD Open(LPCSTR zFilePath, DWORD dwDesiredAccess, DWORD dwCreationDisposition, IocpEvent *pEventTarget)
    {
        DWORD dwRes = 0;
        HANDLE hF = CreateFile(zFilePath, dwDesiredAccess, 0, NULL, dwCreationDisposition,
            FILE_FLAG_OVERLAPPED|FILE_FLAG_RANDOM_ACCESS, NULL);
        if(hF == INVALID_HANDLE_VALUE)
            dwRes = GetLastError();
        else
        {
            // Bind the file handle to the completion port.
            dwRes = BindToIocp(hF, pEventTarget);
            if(dwRes == 0)
                m_hDevice = hF;
        }

        return dwRes;
    }

    // Disconnect and close the socket.
    // Return value: 0 = success, other = GetLastError() failure code.
    virtual DWORD Close()
    {
        DWORD dwRes = 0;
        if(m_hDevice != NULL)
        {
            if(!CloseHandle(m_hDevice))
                dwRes = GetLastError();
            m_hDevice = NULL;
        }

        return dwRes;
    }

    // Asynchronous file read operation.
    // Warning: this function reads from the beginning of the file.
    // Use the overloaded method to read from any position.
    // CAUTION: the buffer must remain valid until the call completes!
    // Return value: 0 = success, other = WSAGetLastError() failure code.
    virtual DWORD Read(LPSTR zBuffer, DWORD dwBufferLength)
    {
        return Read(zBuffer, dwBufferLength, 0);
    }

    // Asynchronous file read operation from an arbitrary offset.
    virtual DWORD Read(LPSTR zBuffer, DWORD dwBufferLength, DWORD dwOffset)
    {
        DWORD dwRes = 0;
        ZeroMemory(&m_overlapped, sizeof(OVERLAPPED));
        m_overlapped.Offset = dwOffset;
        m_iocp.IncrementOutstandingIoCount();
        if(!ReadFile(m_hDevice, zBuffer, dwBufferLength, NULL, &m_overlapped))
        {
            DWORD dwError = GetLastError();
            if(dwError != ERROR_IO_PENDING)
                dwRes = dwError;
        }

        return dwRes;
    }

    // Asynchronous file write operation.
    // Warning: this function writes from the beginning of the file.
    // Use the overloaded method to read from any position.
    // CAUTION: the buffer must remain valid until the call completes!
    // Return value: 0 = success, other = GetLastError() failure code.
    virtual DWORD Write(LPCSTR zBuffer, DWORD dwBufferLength = 0)
    {
        return Write(zBuffer, dwBufferLength, 0);
    }

    // Asynchronous file write operation to an arbitrary offset.
    virtual DWORD Write(LPCSTR zBuffer, DWORD dwBufferLength, DWORD dwOffset)
    {
        DWORD dwRes = 0;
        ZeroMemory(&m_overlapped, sizeof(OVERLAPPED));
        m_overlapped.Offset = dwOffset;
        m_iocp.IncrementOutstandingIoCount();
        if(!WriteFile(m_hDevice, zBuffer, dwBufferLength, NULL, &m_overlapped))
        {
            DWORD dwError = GetLastError();
            if(dwError != ERROR_IO_PENDING)
                dwRes = dwError;
        }

        return dwRes;
    }

    // Is the device open?
    virtual BOOL IsOpen()
    {
        return m_hDevice != NULL;
    }

private:
    // Forbidden.
    operator=(const CAsyncFile&);
};

#endif  // _OPENSTA_ASYNFILE_H_
