/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _ASYNCVIRTUALUSER_H_
#define _ASYNCVIRTUALUSER_H_

#include "TContext.h"
#include "ScriptCache.h"

#include "AsyncHTTPRequestManager.hpp"
#include "TimerQueue.hpp"

#ifndef CYRANO_TP_STANDALONE
#   include "TestExecuter.h"
#   include "CECSemaphore.h"
#endif

class CAsyncVirtualUserController;
class CAsyncFile;


///////////////////////////////////////////////////////////////////////////////
//
// Class provided for use instead of TContextData in the TContext m_ContextDataStack
// in order to allow script inclusion: stores the current status of a context
// before the call to another script, and restores the caller script's context
// when the called script is over.
//
///////////////////////////////////////////////////////////////////////////////

class CScriptContextData : public TContext::TContextData
{
public:
    CScriptContextData(DWORD dwParallelism);

public: // Public because data is public in TContextData.
    TaskScript *    m_pTaskScript;

    // For use by the script execution routines (TScript::run ...).
    DWORD   m_dwScriptLength;

    // For use by the asynchronous TRecord execute methods.
    DWORD   m_dwTRecordCallStep;    // Stores the position in the execution sequence between each call to
                                    // the TRecord::execute method. Must be reset to zero while moving to
                                    // the next TRecord.
private:
    //// Forbidden:
    CScriptContextData();
    CScriptContextData(const CScriptContextData&);
    operator=(const CScriptContextData&);
};

typedef enum  {
                EXEC_INIT,              // First step in processing.
                EXEC_RUNNING,           // Processing pure-CPU TOF objects.
                EXEC_WAIT_MUTEX,        // Waiting for an asynchronous mutex acquisition.
                EXEC_WAIT_SEMAPHORE,    // Waiting for an asynchronous semaphore acquisition.
                EXEC_WAIT_TIMER,        // Waiting for an asynchronous wait to finish.

                EXEC_HTTP_SYNC,         // Synchronise with an HTTP request before moving to next TOF.
                EXEC_ALL_HTTP_SYNC,     // Synchronise with all HTTP requests before moving to next TOF.

                EXEC_WAIT_DISCONNECTS,  // Wait for several asynchronous connections to disconnect before moving to next TOF.

                EXEC_OVER,              // The whole TOF list was successfully processed.
                EXEC_GOTO               // 
              } EVuExecPhase;


class CAsyncVirtualUser : public CAsyncHTTPRequestManager, public TContext, public IocpEvent, public IocpEventDispatcher
{
#define CURR_SCRIPT_CTXT_DATA ((CScriptContextData*)m_ContextDataStack.top())

public:
    CAsyncVirtualUser();
    ~CAsyncVirtualUser();

    virtual DWORD Init(CThreadPoolController& controller, CTimerQueue* pTimerQueue, 
                            DWORD dwManagerNumber, DWORD dwParallelism, CTrace* pTrace = NULL);

    static  BOOL    InitClass();
    static  void    UninitClass();

    void Initialise(const std::string &szTofFile
#ifndef CYRANO_TP_STANDALONE
                        , const std::string &szTestName, const ThreadDef &thrDef
#endif
                        );

    virtual void InitContextData()
    {
        CScriptContextData* pData = new CScriptContextData(m_dwParallelism);
        m_ContextDataStack.push(pData);
    }

    // Callbacks for use by CAsyncRequestManager derivatives.
    virtual void OnRequestFinished(DWORD dwRequestSlotNumber);      // Called when request is completely over. Mind the re-entering issues.
    virtual void OnRequestCancelled(DWORD dwRequestSlotNumber);
    virtual void OnRequestClosed(DWORD dwRequestNumber);

    // To process IOCP events.
    virtual void OnEvent      (DWORD /*dwNumberOfBytesTransferred*/, BOOL /*bIsIo*/, DWORD dwEventType);
    virtual void OnEventFailed(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType, DWORD dwError);

    ETofResult  ProcessTof();

    //// For use by the script executer.

    CAsyncHTTPRequest*  ObtainRequestStruct(DWORD dwConnectionId);    // Books a new request to execute the requests for dwConnectionId.
                                                                // If dwConnectionId has already been allocated, return the existing one.
    DWORD   DisconnectClient(DWORD dwConnectionId);
    DWORD   DisconnectAllClients();

    void    ReleaseRequest(DWORD dwConnectionId);
    void    ReleaseRequest(CAsyncHTTPRequest* pReq);

    // Asynchronous ops.
    BOOL    SynchroniseHTTPClient(DWORD dwClientId);    // Return TRUE if all the client is already over.
    BOOL    SynchroniseAllHTTPClients();                // Return TRUE if all the clients are already over.
    BOOL    StartEventTimer(DWORD dwNbMilliSeconds);

    // Mutexes and semaphores.
    void    SetMutexStartWaitTime(time_t startTime)         { m_mutexStartWaitTime = startTime; }
    time_t  GetMutexStartWaitTime()                 const   { return m_mutexStartWaitTime;      }

    void    SetSemaphoreStartWaitTime(time_t startTime)         { m_semaphoreStartWaitTime = startTime; }
    time_t  GetSemaphoreStartWaitTime()                 const   { return m_semaphoreStartWaitTime;      }
    BOOL    TryAcquireSemaphore(const std::string &s_name, unsigned l_scope);

    BOOL    MustDumpStats();

    void    CancelQuery(DWORD dwConnectionId);

    CTrace* GetTrace()  { return m_pTrace; }
#ifndef CYRANO_TP_STANDALONE
    void    GetVUInfo(ThreadInfo &tI)   const;
#endif

    int CallScript(const std::string& name);

    void    SetParamList(const TtValList &paramList)    { m_paramsInfo = paramList; }
    int     Entry(const TtValList &parameters);

    // TODO: This service should belong to TContext.
    CECVar *GetVariablePointer(const std::string &szVarName) { std::string szTempVarName = szVarName; return CURR_SCRIPT_CTXT_DATA->varlist.get_variable_ptr(szTempVarName); }

    //// For use by the script execution routines (TScript::run ...).

    void    SetScriptLength(DWORD dwScriptLength)           { CURR_SCRIPT_CTXT_DATA->m_dwScriptLength = dwScriptLength; }
    DWORD   GetScriptLength()                       const   { return CURR_SCRIPT_CTXT_DATA->m_dwScriptLength; }

    void    SetScriptPos(DWORD dwScriptPos);
    DWORD   GetScriptPos();
    const std::string &  GetCurrScriptName()    const   { return CURR_SCRIPT_CTXT_DATA->m_pTaskScript->GetName();   }

    // For use by the asynchronous TRecord execute methods so that they know what to do when called again.
    void    SetTRecordCallStep(DWORD dwCS)          { CURR_SCRIPT_CTXT_DATA->m_dwTRecordCallStep = dwCS;   }
    DWORD   GetTRecordCallStep()            const   { return CURR_SCRIPT_CTXT_DATA->m_dwTRecordCallStep;   }


    //// Data storing temporary info. for async. TOF objects, avoiding to compute the value several times
    //// if that object is processed in several steps.
    void                SetHeaderText(const std::string &szHeaderText)                      { m_szHeaderText = szHeaderText; }
    const std::string & GetHeaderText()                                             const   { return m_szHeaderText; }

    void                SetRemoteServerName(const std::string &szRemoteServerName)          { m_szRemoteServerName = szRemoteServerName; }
    const std::string & GetRemoteServerName()                                       const   { return m_szRemoteServerName; }

    void                SetRemoteServerPort(const std::string &szRemoteServerPort)          { m_szRemoteServerPort = szRemoteServerPort; }
    const std::string & GetRemoteServerPort()                                       const   { return m_szRemoteServerPort; }

    void                SetCurrentUri(const std::string &szCurrentUri)                      { m_szCurrentUri = szCurrentUri; }
    const std::string & GetCurrentUri()                                             const   { return m_szCurrentUri; }

    void                SetProxyServerName(const std::string &szProxyServerName)            { m_szProxyServerName = szProxyServerName; }
    const std::string & GetProxyServerName()                                        const   { return m_szProxyServerName; }

    void                SetProxyServerPort(DWORD dwProxyServerPort)         { m_dwProxyServerPort = dwProxyServerPort; }
    DWORD               GetProxyServerPort()                        const   { return m_dwProxyServerPort; }

    void    SetSecureModeUsed(BOOL bSecureMode)         { m_bSecureModeUsed = bSecureMode; }
    BOOL    GetSecureModeUsed()                 const   { return m_bSecureModeUsed; }

    void    SetCurrentConid(DWORD dwConid)              { m_dwCurrentConid = dwConid; }
    DWORD   GetCurrentConid()                   const   { return m_dwCurrentConid; }

    EVuExecPhase    GetStatus() const       { return m_execPhase; }
    ULONG           GetStartTime()  const   { return m_dwStartTime; }

#ifndef CYRANO_TP_STANDALONE
    DWORD   GetTestCaseResults(TestCaseStatus Status)
    {
	    if(Status == PASSED)
	        return m_dwTestCasePass;
	    else
	        return m_dwTestCaseFail;
    }
#endif


    enum { EVENT_TYPE_VIRTUAL_USER_START = 0x8000 };    // User-defined IOCP event, sent to wake the VUs.

    //// --------------------------------------------------------------------------------
    //// IocpEventDispatcher methods.
    //// --------------------------------------------------------------------------------
    IocpEvent* GetIocpEventTarget() {return this;}

    //// --------------------------------------------------------------------------------
    //// TContext virtual methods.
    //// --------------------------------------------------------------------------------
    virtual void CleanUp();
    virtual void SetNoteText(const std::string& msg);
    virtual std::string get_test_name();
#ifndef CYRANO_TP_STANDALONE
    virtual ThreadDef GetThreadDef();
#endif
    virtual void    StartScript(const std::string& fullscriptname);
    virtual void    EndScript();
    virtual int     ExitCommand(bool keepalive, const std::string& retvar);
    virtual int     TCterminate(int result, std::string desc);

    virtual void    OnFileIoOver()  { ProcessTof(); }
    virtual void    OnFileIoFailed(){ ProcessTof(); }


protected:

    //// --------------------------------------------------------------------------------
    //// Helpers.
    //// --------------------------------------------------------------------------------

    CAsyncRequest*      GetNotConnectedRequest();
    CAsyncHTTPRequest*  GetRequestFromConnId(DWORD dwConnectionId) const;

    virtual void OnAllRequestsFinished()  {/*Neutralize CAsyncRequestManager's one, which is based only on the requests instead of the TOF list.*/}
    virtual void OnAllRequestsCancelled() {/*Neutralize CAsyncRequestManager's one, which is based only on the requests instead of the TOF list.*/}

    // To process timer events (called by OnEvent).
    virtual void OnTimerFinished();
    virtual void OnTimerCancelled();

    // Called by OnEvent.
    virtual void OnStartProcessing();

    BOOL    CreateTask(const std::string &szFileName);
    void    ProcessEndOfRequest(DWORD dwRequestSlotNumber, BOOL bWasCancelled);

protected:

    //// --------------------------------------------------------------------------------
    //// Data.
    //// --------------------------------------------------------------------------------

    EVuExecPhase    m_execPhase;
    std::string     m_szMainTOFFile;    // Main calling script.

    //// Info. required by some sync conditions.

    // Info used to synchronise on a specific request.
    DWORD   m_dwWaitedConnId;

#ifndef CYRANO_TP_STANDALONE
    // Info used to synchronise on a mutex.

    // Info used to synchronise on a semaphore.
    CECSemaphore    *m_pCurrentSemaphore;
    std::string     m_szCurrentSemaphoreName;
#endif

    CTimerQueue     *m_pTimerQueue;         // Pointer to the timer queue defined in the VU Controller.

    DWORD   m_dwParallelism;    // Max. number of parallel requests.

    list<std::string>   m_Params;       // Current parameter list (passing by value only so stack not required).
    TtValList           m_paramsInfo;   // ParameterListInfo.

    time_t              m_mutexStartWaitTime;       // Used for the mutex lock time out.
    time_t              m_semaphoreStartWaitTime;   // Used for the semaphore try-acquire time out.

    //// Data storing info. for async. TOF objects, avoiding to compute the value several times
    //// if that object is processed in several steps.
    std::string m_szHeaderText;         // Computed header string.
    std::string m_szRemoteServerName;
    std::string m_szRemoteServerPort;
    std::string m_szCurrentUri;

    DWORD   m_dwCurrentConid;           // Current connection being processed. For use by TRcrdRequest.
    BOOL    m_bSecureModeUsed;

    // Connection info for use by TRcrdConnect while stepping from case 1 to case 2.
    std::string m_szProxyServerName;
    DWORD       m_dwProxyServerPort;


    DWORD       m_dwStartTime;          // Time in seconds since 1/1/1901.
    std::string m_noteText;
#ifndef CYRANO_TP_STANDALONE
    ThreadDef   m_ThreadDef;
#endif
    std::string m_szTestName;

    DWORD   m_dwTestCasePass;
    DWORD   m_dwTestCaseFail;

    DWORD       m_dwDisconnectToWait;   // Number of disconnections we have to wait for.

    static CFastLock *  s_pCacheSl;   // Allow only one VU to access the scripts cache at a time.

private:
    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------

    CAsyncVirtualUser& operator = (const CAsyncVirtualUser&);
};


inline void CAsyncVirtualUser::SetScriptPos(DWORD dwScriptPos)
{
    CURR_SCRIPT_CTXT_DATA->oi = dwScriptPos;
}

inline DWORD CAsyncVirtualUser::GetScriptPos()
{
    return CURR_SCRIPT_CTXT_DATA->oi;
}

#endif
