/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "stdafx.h"
#include "AsyncVirtualUserController.h"

#ifdef USE_PERF_DATA
#   include "PerfCounterKeys.hpp"
#endif

#ifdef __DLLSA__
#   include "TEMainStructure.hpp"
    extern CMainExecuterStructure *g_pTExecuter;
#endif


///////////////////////////////////////////////////////////////////////////////
//
// Class CAsyncVirtualUserController
//
///////////////////////////////////////////////////////////////////////////////

#ifdef USE_PERF_DATA
LONG *CAsyncVirtualUserController::m_spActiveUsers = NULL;
#endif

CAsyncVirtualUserController::CAsyncVirtualUserController(TScriptRule* pRule, CTrace *pTrace)
    : CThreadPoolController(pTrace), m_scriptCache(pRule), m_pTimerQueue(NULL), m_dwFailedUsersCount(0L), m_dwSuccessfullUsersCount(0L)
{}

CAsyncVirtualUserController::~CAsyncVirtualUserController()
{
    if(m_pTimerQueue)
        delete m_pTimerQueue;
}

DWORD CAsyncVirtualUserController::Init(DWORD dwThreadPoolSize, DWORD dwNumberOfConcurrentThreadsInIocp, DWORD dwInitialVUCount, DWORD dwVUCountGrowBy, DWORD dwBrowserParallelism)
{
#ifdef USE_PERF_DATA
    m_spActiveUsers = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_ACTIVEUSERS);
#endif
    DWORD res = CThreadPoolController::Init(dwThreadPoolSize, dwNumberOfConcurrentThreadsInIocp);

    if(res == 0)
    {
        m_pTimerQueue = new CTimerQueue(*m_pIocp, INFINITE, m_pTrace);

        if(m_pTimerQueue == NULL)
        {
            if(m_pTrace)
            {
                GetTraceStream(m_pTrace) << "Failed at initialising the virtual users controller." << "\r\n";
                ReleaseTraceStream();
            }
            return -1;
        }

        res = m_pTimerQueue->Init();
        if(res == 0)
            res = m_VirtualUsersList.Init(m_pIocp, m_pTimerQueue, this, dwInitialVUCount, dwVUCountGrowBy, dwBrowserParallelism, m_pTrace);
    }

    return res;
}

int CAsyncVirtualUserController::StartNewUsersGroup(DWORD dwNbUsers, const string &szTofFile
#ifndef CYRANO_TP_STANDALONE
                                                    , TestLevel_ptr pTL, const ThreadDef &thrdef
#endif
                                                    )
{
    // Alloc nbUsers clients.
    DWORD dwStartedUsers;

#ifndef CYRANO_TP_STANDALONE
    dwStartedUsers = m_VirtualUsersList.StartUsers(dwNbUsers, szTofFile, pTL, thrdef);
#else
    dwStartedUsers = m_VirtualUsersList.StartUsers(dwNbUsers, szTofFile);
#endif

    assert(dwStartedUsers == dwNbUsers);

#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
    InterlockedExchangeAdd(m_spActiveUsers, dwStartedUsers);
#   else
    *m_spActiveUsers += dwStartedUsers;
#   endif
#endif

    if(dwStartedUsers != dwNbUsers)
    {
        if(m_pTrace)
        {
            GetTraceStream(m_pTrace) << "Failed at starting users." << "\r\n";
            ReleaseTraceStream();
        }
        return -1;
    }
    else
        return 0;
}

void CAsyncVirtualUserController::CancelAllUsers()
{
    m_VirtualUsersList.CancelAllUsers();
}

void CAsyncVirtualUserController::OnAllRequestsFinished(DWORD dwVuId)
{
#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
    InterlockedDecrement(m_spActiveUsers);
#   else
    --*m_spActiveUsers;
#   endif
#endif

    DWORD dwRemainingUsers = m_VirtualUsersList.DropVirtualUser(dwVuId);

    if(dwRemainingUsers == 0)
    {
        // All VU have been processed. Bye bye everybody!
        if(m_pTrace && m_pTrace->IsTrace("full"))
        {
            GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::OnAllRequestFinished: user " << dwVuId << " terminated. Simulation is over.\r\n";
            ReleaseTraceStream();
        }
        Stop();
    }
    else
    {
        if(m_pTrace && m_pTrace->IsTrace("full"))
        {
            GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::OnAllRequestFinished: user " << dwVuId << " terminated. Remaining " << dwRemainingUsers << " users.\r\n";
            ReleaseTraceStream();
        }
    }
}

void CAsyncVirtualUserController::OnAllRequestsCancelled(DWORD dwVuId)
{
    OnAllRequestsFinished(dwVuId);
}

void CAsyncVirtualUserController::OnEvent(DWORD /*dwNumberOfBytesTransferred*/, BOOL bIsIo, DWORD dwEventType)
{
    if(bIsIo == TRUE)
    {
        // Should not receive any IOs there.
        assert(0);
    }
    else
    {
        switch(dwEventType)
        {
        case EVENT_TYPE_THREADPOOL_START:
            {
#ifdef __DLLSA__
                // Start the modelled user.
                if(StartNewUsersGroup(1, g_pTExecuter->GetTofFileNameAndPath()) == -1)
                {
                    if(m_pTrace && m_pTrace->IsTrace("errors"))
                    {
                        GetTraceStream(m_pTrace) << "Could not create the user.\r\n";
                        ReleaseTraceStream();
                    }
                }
#endif
                break;
            }

        case EVENT_TYPE_THREADPOOL_STOP_THREAD:
            {
                if(m_pTrace && m_pTrace->IsTrace("errors"))
                {
                    GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::OnEvent: ERROR: do not handle STOP THREAD events YET.\r\n";
                    ReleaseTraceStream();
                }
                break;
            }

        case EVENT_TYPE_TIMER:
            {
                if(m_pTrace && m_pTrace->IsTrace("errors"))
                {
                    GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::OnEvent: ERROR: do not handle timer events." << "\r\n";
                    ReleaseTraceStream();
                }
                break;
            }

        default:
            {
                if(m_pTrace && m_pTrace->IsTrace("errors"))
                {
                    GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::OnEvent: ERROR: unknown type." << "\r\n";
                    ReleaseTraceStream();
                }
                break;
            }
        };
    }
}

void CAsyncVirtualUserController::OnEventFailed(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType, DWORD dwError)
{
    if(m_pTrace && m_pTrace->IsTrace("errors"))
    {
        GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::OnEventFailed: ERROR." << "\r\n";
        ReleaseTraceStream();
    }
}


///////////////////////////////////////////////////////////////////////////////
//
// Class CVirtualUsersGroup
//
///////////////////////////////////////////////////////////////////////////////

CVirtualUsersGroup::CVirtualUsersGroup(CTrace* pTrace)
    : m_VirtualUsersArray(NULL), m_dwStartedUsersCount(0L), m_dwVirtualUsersCount(0L), m_pIocp(NULL), m_pTrace(pTrace)
{}

BOOL CVirtualUsersGroup::CreateUsers(DWORD dwNbUsers)
{
    BOOL bCouldCreate = FALSE;

    if(dwNbUsers != 0)
    {
        try
        {
            m_VirtualUsersArray = new CAsyncVirtualUser[dwNbUsers];
        }
        catch(...)
        {
            if(m_pTrace)
            {
                GetTraceStream(m_pTrace) << "CVirtualUsersGroup::CreateUsers: Error allocating the virtual users.\r\n";
                ReleaseTraceStream();
            }
        }

        if(m_VirtualUsersArray != NULL)
        {
            m_dwVirtualUsersCount = dwNbUsers;
            bCouldCreate = TRUE;
        }
        else
        {
            if(m_pTrace)
            {
                GetTraceStream(m_pTrace) << "CVirtualUsersGroup::CreateUsers: Error allocating the virtual users.\r\n";
                ReleaseTraceStream();
            }
            m_dwVirtualUsersCount = 0;
            bCouldCreate = FALSE;
        }
    }
    else
    {
        if(m_pTrace)
        {
            GetTraceStream(m_pTrace) << "CVirtualUsersGroup::CreateUsers: Cannot create a group with no virtual users.\r\n";
            ReleaseTraceStream();
        }
    }

    return bCouldCreate;
}

void CVirtualUsersGroup::CancelAllUsers()
{
    for(DWORD dw=0L ; dw<m_dwVirtualUsersCount ; dw++)
    {
        m_VirtualUsersArray[dw].CancelAllQueries();
    }
}

// Get all the users in that group ready for start.
BOOL CVirtualUsersGroup::InitUsers(CIOCP *pIocp, CAsyncVirtualUserController& controller, CTimerQueue* pTimerQueue, DWORD dwParallelism)
{
    static DWORD dwManagerID = 0;
    m_pIocp = pIocp;

    for(DWORD dw=0 ; dw<m_dwVirtualUsersCount ; dw++)
    {
        if(m_VirtualUsersArray[dw].Init(controller, pTimerQueue, dwManagerID, dwParallelism, m_pTrace) != 0)
        {
            if(m_pTrace)
            {
                GetTraceStream(m_pTrace) << "CVirtualUsersGroup::InitUsers: Error initialising virtual user number " << dw << ".\r\n";
                ReleaseTraceStream();
            }
            return FALSE;
        }

        dwManagerID++;
    }

    return TRUE;
}

// Start dwNbUsers from the current group. Return the number of users that could NOT be started.
DWORD CVirtualUsersGroup::StartMoreUsers(DWORD dwNbUsers, const string &szTofFile
#ifndef CYRANO_TP_STANDALONE
                                        , TestLevel_ptr pTL, const ThreadDef &thrdef
#endif
                                         )
{
    while(dwNbUsers > 0 && m_dwStartedUsersCount < m_dwVirtualUsersCount)
    {
#ifndef CYRANO_TP_STANDALONE
        m_VirtualUsersArray[m_dwStartedUsersCount].Initialise(szTofFile, pTL->TestName(), thrdef);
#else
        m_VirtualUsersArray[m_dwStartedUsersCount].Initialise(szTofFile);
#endif

        if(m_pIocp->PostCompletedIO(0, (IocpEventDispatcher*) &m_VirtualUsersArray[m_dwStartedUsersCount], (OVERLAPPED*)CAsyncVirtualUser::EVENT_TYPE_VIRTUAL_USER_START) != 0)
        {
            assert(0);
            if(m_pTrace && m_pTrace->IsTrace("full"))
            {
                GetTraceStream(m_pTrace) << "Start users ERROR." << "\r\n";
                ReleaseTraceStream();
            }
            break;
        }

        m_dwStartedUsersCount++;
        dwNbUsers--;
    }

    return dwNbUsers;
}

// Delete all the users from that group.
void CVirtualUsersGroup::DropAllUsers()
{
    if(m_VirtualUsersArray != NULL)
    {
        delete [] m_VirtualUsersArray;
        m_VirtualUsersArray = NULL;
        m_dwVirtualUsersCount = 0;
    }
}

// Append the pipe-separated list of all unique identifiers for that group's users (formerly Threads).
// Code based on TestExecuterImpl::GetThreadIdents().
void CVirtualUsersGroup::AppendThreadIdents(string &str) const
{
    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        str += m_VirtualUsersArray[dw].GetThreadID() + '|';
    }
}

DWORD CVirtualUsersGroup::GetMatchingThreadsCount(const string &sname) const
{
    DWORD dwThreadsCount = 0;

    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        if(m_VirtualUsersArray[dw].ScriptName() == sname)
            dwThreadsCount++;
    }

    return dwThreadsCount;
}

#ifndef CYRANO_TP_STANDALONE
DWORD CVirtualUsersGroup::GetTestCaseResults(TestCaseStatus Status) const
{
    DWORD dwRes = 0;

    // Ask all the groups to return their working threads UIDs to the end of 'str'.
    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        dwRes += m_VirtualUsersArray[dw].GetTestCaseResults(Status);
    }

    return dwRes;
}
#endif

// Return thread current scriptname based on unique ID.
BOOL CVirtualUsersGroup::GetVuScriptName(const char *ident, string &name) const
{
    // Browse all the threads in that group in search for the 'ident' thread and return its ScriptName.
    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        if(m_VirtualUsersArray[dw].GetThreadID() == ident)
        {
            name = m_VirtualUsersArray[dw].ScriptName();
            return TRUE;
        }
    }

    return FALSE;
}

// Return a thread's execution phase based on its unique ID.
BOOL CVirtualUsersGroup::GetVuStatus(const char *ident, EVuExecPhase &phase) const
{
    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        if(m_VirtualUsersArray[dw].GetThreadID() == ident)
        {
            phase = m_VirtualUsersArray[dw].GetStatus();
            return TRUE;
        }
    }

    return FALSE;
}

BOOL CVirtualUsersGroup::GetVuStartTime(const char *ident, ULONG &ulTime) const
{
    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        if(m_VirtualUsersArray[dw].GetThreadID() == ident)
        {
            ulTime = m_VirtualUsersArray[dw].GetStartTime();
            return TRUE;
        }
    }

    return FALSE;
}

#ifndef CYRANO_TP_STANDALONE
DWORD CVirtualUsersGroup::AppendThreadInformation(ThreadInfoList &TIList, DWORD dwStartPos) const
{
    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        m_VirtualUsersArray[dw].GetVUInfo(TIList[dwStartPos + dw]);
    }

    return m_dwStartedUsersCount;
}
#endif


// ////////////////////////////////////////////////////////////////////////////
//
// Class CVirtualUsersList
//
// ////////////////////////////////////////////////////////////////////////////

CVirtualUsersList::CVirtualUsersList()
    : m_pIocp(NULL), m_pTimerQueue(NULL), m_pTrace(NULL), m_pVUController(NULL), m_dwVirtualUsersCount(0L), 
    m_dwParallelism(0L), m_dwVUCountGrowBy(0L), m_currGroup(0)
{}

CVirtualUsersList::~CVirtualUsersList()
{
    DeleteGroups();
}

DWORD CVirtualUsersList::Init(CIOCP* pIocp, CTimerQueue* pTimerQueue, CAsyncVirtualUserController *pVUController, DWORD dwInitialVUCount, DWORD dwVUCountGrowBy, DWORD dwParallelism, CTrace *pTrace)
{
    m_pIocp             = pIocp;
    m_pTimerQueue       = pTimerQueue;
    m_pTrace            = pTrace;
    m_pVUController     = pVUController;
    m_dwParallelism     = dwParallelism;
    m_dwVUCountGrowBy   = dwVUCountGrowBy;

    if(dwInitialVUCount != 0)
    {
        // Pre-allocate users.
        if(CreateGroup(dwInitialVUCount) == FALSE)
        {
            if(m_pTrace)
            {
                GetTraceStream(m_pTrace) << "CVirtualUsersList::Init: error creating a new group with " << dwInitialVUCount << " users.\r\n";
                ReleaseTraceStream();
            }
   
            return (DWORD)-1;
        }
    }
    m_currGroup = 0L;

    return 0L;
}

// Protected method that creates a group of standing-by users.
BOOL CVirtualUsersList::CreateGroup(DWORD dwNbUsers)
{
    CVirtualUsersGroup *pNewGroup = new CVirtualUsersGroup(m_pTrace);

    if(pNewGroup != NULL)
    {
        if(pNewGroup->CreateUsers(dwNbUsers) == FALSE)
            return FALSE;
        if(pNewGroup->InitUsers(m_pIocp, *m_pVUController, m_pTimerQueue, m_dwParallelism) == FALSE)
            return FALSE;

        CLock guard (m_userGroupListFl);
        m_GroupsList.push_back(pNewGroup);
        m_currGroup = m_GroupsList.size() - 1;

        return TRUE;
    }
    else
    {
        if(m_pTrace)
        {
            GetTraceStream(m_pTrace) << "Failed creating a new group with " << dwNbUsers << " users.\r\n";
            ReleaseTraceStream();
        }
        return FALSE;
    }
}

// Return the number of users that were started actually.
DWORD CVirtualUsersList::StartUsers(    DWORD dwNbUsers, const string &szTofFile
#ifndef CYRANO_TP_STANDALONE
                                        , TestLevel_ptr pTL, const ThreadDef &thrdef
#endif
                                        )
{
    if(dwNbUsers == 0)
    {
        assert(0);
        return 0;
    }

    CLock guard (m_userGroupListFl);

    if(m_GroupsList.size() == 0)
    {
        // Must create the first VU group.
        if(CreateGroup(__max(dwNbUsers, m_dwVUCountGrowBy)) == FALSE)
            return 0;
        assert(m_GroupsList.size() != 0);
    }

#ifndef CYRANO_TP_STANDALONE
    DWORD dwRemaining = m_GroupsList[m_currGroup]->StartMoreUsers(dwNbUsers, szTofFile, pTL, thrdef);
#else
    DWORD dwRemaining = m_GroupsList[m_currGroup]->StartMoreUsers(dwNbUsers, szTofFile);
#endif
    if(dwRemaining != 0)
    {
        if(CreateGroup(__max(dwRemaining, m_dwVUCountGrowBy)) == FALSE)
        {
            InterlockedExchangeAdd((LONG*)&m_dwVirtualUsersCount, (LONG)(dwNbUsers-dwRemaining));
            return dwNbUsers-dwRemaining;
        }

#ifndef CYRANO_TP_STANDALONE
        dwRemaining = m_GroupsList[m_currGroup]->StartMoreUsers(dwRemaining, szTofFile, pTL, thrdef);
#else
        dwRemaining = m_GroupsList[m_currGroup]->StartMoreUsers(dwRemaining, szTofFile);
#endif
        assert(dwRemaining == 0);
    }

    InterlockedExchangeAdd((LONG*)&m_dwVirtualUsersCount, (LONG)dwNbUsers);

    return dwNbUsers;
}

void CVirtualUsersList::CancelAllUsers()
{
    CLock lock(m_userGroupListFl);

    for(vector<CVirtualUsersGroup*>::iterator i=m_GroupsList.begin() ; i<m_GroupsList.end() ; ++i)
    {
        (*i)->CancelAllUsers();
    }
}

DWORD CVirtualUsersList::DropVirtualUser(DWORD dwVuId)
{
    // No need to delete that user.
    return InterlockedDecrement((LONG*)&m_dwVirtualUsersCount);
}

void CVirtualUsersList::DeleteGroups()
{
    CLock lock(m_userGroupListFl);

    for(vector<CVirtualUsersGroup*>::iterator i=m_GroupsList.begin() ; i<m_GroupsList.end() ; ++i)
    {
        (*i)->DropAllUsers();
    }

    m_GroupsList.clear();
}

// Get pipe-separated list of all unique identifiers for users (formerly Threads) in this Test executer.
// Code based on TestExecuterImpl::GetThreadIdents().
void CVirtualUsersList::GetThreadIdents(string &str) const
{
    str.erase();
    str.reserve(1024);

    CLock guard (m_userGroupListFl);

    // Ask all the groups to append their working threads UIDs to the end of 'str'.
    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i<m_GroupsList.end() ; ++i)
    {
        (*i)->AppendThreadIdents(str);
    }

    str += '|';
}

DWORD CVirtualUsersList::GetMatchingThreadsCount(const string &sname) const
{
    DWORD dwThreadsCount = 0;

    CLock guard (m_userGroupListFl);

    // Ask all the groups to return their working threads UIDs to the end of 'str'.
    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i<m_GroupsList.end() ; ++i)
    {
        dwThreadsCount += (*i)->GetMatchingThreadsCount(sname);
    }

    return dwThreadsCount;
}

#ifndef CYRANO_TP_STANDALONE
DWORD CVirtualUsersList::GetTestCaseResults(TestCaseStatus Status) const
{
    DWORD dwRes = 0;

    CLock guard (m_userGroupListFl);

    // Ask all the groups to return their working threads UIDs to the end of 'str'.
    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i<m_GroupsList.end() ; ++i)
    {
        dwRes += (*i)->GetTestCaseResults(Status);
    }

    return dwRes;
}
#endif

// Return thread current scriptname based on unique ID.
string CVirtualUsersList::GetVuScriptName(const char *ident) const
{
    string scriptName;

    CLock guard (m_userGroupListFl);

    // Browse all the groups in search for the 'ident' thread and return its ScriptName.
    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i<m_GroupsList.end() ; ++i)
    {
        if((*i)->GetVuScriptName(ident, scriptName) == TRUE)
            return scriptName;
    }

    // Could not find that user.
    assert(0);
    return string();
}

// Browse the groups until we find 'ident' and return its status.
EVuExecPhase CVirtualUsersList::GetVuStatus(const char *ident) const
{
    EVuExecPhase execPhase;

    CLock guard (m_userGroupListFl);

    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i<m_GroupsList.end() ; ++i)
    {
        if((*i)->GetVuStatus(ident, execPhase) == TRUE)
            return execPhase;
    }

    // Could not find that user.
    assert(0);
    return EXEC_OVER;
}

// Browse all the groups in search for the 'ident' thread and return its ScriptName.
ULONG CVirtualUsersList::GetVuStartTime(const char *ident) const
{
    ULONG time;

    CLock guard (m_userGroupListFl);

    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i<m_GroupsList.end() ; ++i)
    {
        if((*i)->GetVuStartTime(ident, time) == TRUE)
            return time;
    }

    // Could not find that user.
    assert(0);
    return 0;
}

#ifndef CYRANO_TP_STANDALONE
// Pack info about all the virtual users.
DWORD CVirtualUsersList::GetThreadInformation(ThreadInfoList &TIList) const
{
    DWORD dwPosInList = 0;

    CLock guard (m_userGroupListFl);

    DWORD dwUsersCount = GetUsersCount();

    TIList.length(dwUsersCount);
    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i<m_GroupsList.end() ; ++i)
    {
        dwPosInList += (*i)->AppendThreadInformation(TIList, dwPosInList);
    }
    
    return dwUsersCount;
}
#endif
