/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_STATS_H_
#define _OPENSTA_STATS_H_

////////////////////////////////////////////////////////////////////////////////
// Statistics definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "Psapi.h"
#include <iosfwd>


class CTimeStats
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction/Destruction.
    //// --------------------------------------------------------------------------------
    CTimeStats()
    {
        ZeroMemory(&m_ftKernelTimeStart, sizeof(m_ftKernelTimeStart));
        ZeroMemory(&m_ftUserTimeStart,   sizeof(m_ftUserTimeStart));
    }

    //// --------------------------------------------------------------------------------
    //// Taking snapshots. Take the start and end snapshots. Then, you
    //// can retrieve the statistics between these with the Get* functions.
    //// Also, EndSnapshot() copies the end snapshot to the start snapshot,
    //// so a subsequent call to EndSnapshot() will compute statistics for
    //// times between the second and third snapshots. An so on...
    //// --------------------------------------------------------------------------------
    void StartSnapshot();
    void EndSnapshot();


    //// --------------------------------------------------------------------------------
    //// Retrieving statistics.
    //// --------------------------------------------------------------------------------

    // Global elapsed time between the snapshots (milliseconds).
    SYSTEMTIME GetGlobalDuration() const {return m_stGlobalDuration;}

    // Time spent in user mode for the process between the snapshots (milliseconds).
    SYSTEMTIME GetUserTime() const {return m_stUserTime;}

    // Time spent in kernel mode for the process between the snapshots (milliseconds).
    SYSTEMTIME GetKernelTime() const {return m_stKernelTime;}

    // Total time spent in user and kernel modes for the process between the snapshots (milliseconds).
    SYSTEMTIME GetTotalTime() const {return m_stTotalTime;}

    friend std::ostream& operator<<(std::ostream& os, const CTimeStats& ts);

private:
    //// --------------------------------------------------------------------------------
    //// Helpers.
    //// --------------------------------------------------------------------------------
    __int64 FileTimeToQuadWord(PFILETIME pFileTime)
    {
        __int64 qw;
        qw = pFileTime->dwHighDateTime;
        qw <<= 32;
        qw |= pFileTime->dwLowDateTime;
        return qw;
    }

    PFILETIME QuadWordToFileTime(__int64 qw, PFILETIME pFileTime)
    {
        pFileTime->dwHighDateTime = (DWORD)(qw >> 32);
        pFileTime->dwLowDateTime  = (DWORD)(qw & 0xFFFFFFFF);
        return pFileTime;
    }

    //// Temporary data corresponding to the first snapshot.
    FILETIME    m_ftKernelTimeStart;
    FILETIME    m_ftUserTimeStart;
    SYSTEMTIME  m_stTimerStart;

    //// Final data computed after the second snapshot.
    SYSTEMTIME  m_stGlobalDuration;
    SYSTEMTIME  m_stUserTime;
    SYSTEMTIME  m_stKernelTime;
    SYSTEMTIME  m_stTotalTime;
};


class CMemoryStats
{
public:
    CMemoryStats() {Snapshot();}

    //// --------------------------------------------------------------------------------
    //// Peak statistics.
    //// --------------------------------------------------------------------------------

    // Taking a snapshot. All Get* functions are then callable to
    // retrieve statistics.
    void Snapshot();

    // Working set size in bytes.
    DWORD GetWorkingSetSize() const {return m_pmc.WorkingSetSize;}

    // Peak working set size in bytes.
    DWORD GetPeakWorkingSetSize() const {return m_pmc.PeakWorkingSetSize;}

    // Peak page file usage in bytes.
    DWORD GetPeakPageFileUsage() const {return m_pmc.PeakPagefileUsage;}

    // Peak paged pool usage in bytes.
    DWORD GetPeakPagedPoolUsage() const {return m_pmc.QuotaPeakPagedPoolUsage;}

    // Peak nonpaged pool usage in bytes.
    DWORD GetPeakNonPagedPoolUsage() const {return m_pmc.QuotaPeakNonPagedPoolUsage;}

    friend std::ostream& operator<<(std::ostream& os, const CMemoryStats& ms);

private:
    PROCESS_MEMORY_COUNTERS m_pmc;
};

// Dump all the statistics to a stream.
std::ostream& operator<<(std::ostream& os, const CTimeStats& ts);
std::ostream& operator<<(std::ostream& os, const CMemoryStats& ms);


#endif  // _OPENSTA_STATS_H_
