/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_THREADPOOL_H_
#define _OPENSTA_THREADPOOL_H_

////////////////////////////////////////////////////////////////////////////////
// Thread pool definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "Iocp.hpp"
#include "Trace.hpp"

#ifdef USE_PERF_DATA
#   include "PerfCounterKeys.hpp"
#endif


// This class implements a thread pool scheduled by an I/O completion port.
// THREAD UNSAFE.
class CThreadPool
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    CThreadPool(CIOCP& iocp, DWORD dwSize, CTrace* pTrace = NULL);
    virtual ~CThreadPool();

    //// --------------------------------------------------------------------------------
    //// User methods.
    //// --------------------------------------------------------------------------------

    // Starts the pool. Creates the termination event (initially not set),
    // creates the threads (suspended), calls OnStart(), resumes the threads
    // if OnStart() returns 0, otherwise destroys the threads. Then, posts
    // pEventDisptacher (an IocpEventDispatcher interface pointer) to the IOCP's queue to
    // start the pool (if pEventDisptacher is not NULL).
    // Then, sits waiting for a termination event. When this event is set, all
    // threads are asynchronously notified to stop. This method then waits for
    // all threads to die, then calls OnExit(), deletes the termination event
    // and finally returns.
    DWORD Run(IocpEventDispatcher *pEventDispatcher = NULL, DWORD m_dwTimeoutMilliseconds = INFINITE);

    // Sets the termination event.
    DWORD Stop() {return (SetEvent(m_hTerminationEvent) ? 0 : GetLastError());}

    // Get the total number of threads initially created in the pool.
    DWORD GetPoolSize() const {return m_dwSize;}

    // Get the instant number of live threads in the pool. These are the
    // threads actually existing in the pool.
    DWORD GetLiveThreadCount() const
    {
#ifdef USE_PERF_DATA
        return *g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_THREADS);
#else
        return 0;
#endif
    }

    // Increment the number of live threads in the pool.
    DWORD IncrementLiveThreadCount()
    {
#ifdef USE_PERF_DATA
        return (DWORD)InterlockedIncrement(g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_THREADS));
#else
        return 0;
#endif
    }

    // Decrement the number of live threads in the pool.
    DWORD DecrementLiveThreadCount()
    {
#ifdef USE_PERF_DATA
        return (DWORD)InterlockedDecrement(g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_THREADS));
#else
        return 0;
#endif
    }

    // Clear the number of live threads in the pool.
    DWORD ClearLiveThreadCount()
    {
#ifdef USE_PERF_DATA
        return (DWORD)InterlockedExchange(g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_THREADS), 0);
#else
        return 0;
#endif
    }

    // Get the instant number of runnable threads in the pool. These are the
    // threads that are either in the IOCP's released thread list or paused
    // thread list, i. e., not in the IOCP's waiting thread queue.
    DWORD GetRunnableThreadCount() const
    {
#ifdef USE_PERF_DATA
        return *g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_ACTIVETHREADS);
#else
        return 0;
#endif
    }

    // Increment the number of runnable threads in the pool.
    DWORD IncrementRunnableThreadCount();

    // Decrement the number of runnable threads in the pool.
    DWORD DecrementRunnableThreadCount();

    // Get peak count of runnable threads in the pool.
    DWORD GetPeakRunnableThreadCount() const {return m_dwPeakRunnableThreads;}

    // A new thread is about to block on the IOCP.
    void ThreadWillBlockOnIocp();

    //// --------------------------------------------------------------------------------
    //// Properties.
    //// --------------------------------------------------------------------------------
    CIOCP&             m_iocp;      // Associated IOCP, on which the threads call GetQueuedCompletionStatus().

protected:

    ////////////////////// Thread Class //////////////////////
    // This class implements a thread of the pool.
    // THREAD UNSAFE.
    class CThread
    {
    public:
        //// --------------------------------------------------------------------------------
        //// Construction / Destruction.
        //// --------------------------------------------------------------------------------
        CThread(CThreadPool& threadPool, DWORD dwTimeoutMilliseconds = INFINITE, CTrace* pTrace = NULL);
        virtual ~CThread();

        //// --------------------------------------------------------------------------------
        //// User methods.
        //// --------------------------------------------------------------------------------

        // Creates and starts the thread with the ThreadFunction() as
        // main thread function. The parameter to ThreadFunction() is the this
        // pointer. Calls OnInit() after the thread is created.
        DWORD Create();

        // Close the thread's handle.
        void CloseHandle() {if(m_hThread) {::CloseHandle(m_hThread); m_hThread = NULL;}}

        // Simple thread manipulation.
        DWORD Resume();
        DWORD Suspend();
        operator HANDLE() const {return m_hThread;}

    protected:
        //// --------------------------------------------------------------------------------
        //// Callbacks.
        //// --------------------------------------------------------------------------------
        virtual void OnInit() {}
        virtual void OnExit() {}
        virtual void OnTimeout() {}
        virtual void OnError(DWORD /*dwError*/) {}

        // Thread function. Loops on a call to GetQueuedCompletionStatus().
        // When GetQueuedCompletionStatus() returns, forwards the completion status
        // via the IocpEventDispatcher interface.
        // If the thread times out while waiting for GetQueuedCompletionStatus(), then
        // OnTimeout() is called.
        // If an error arises, OnError() gets called.
        // If GetQueuedCompletionStatus() returns an event type of
        // IocpEvent::EVENT_TYPE_THREADPOOL_STOP_THREAD as the LPOVERLAPPED pointer,
        // the function exits.
        virtual DWORD OnRun();

        //// --------------------------------------------------------------------------------
        //// Helpers.
        //// --------------------------------------------------------------------------------

        // Actual thread function. Translates pContext into a pointer to the object,
        // then calls OnRun().
        static DWORD WINAPI ThreadFunction(LPVOID pContext);

        //// --------------------------------------------------------------------------------
        //// Data.
        //// --------------------------------------------------------------------------------
        HANDLE          m_hThread;                  // Thread's handle.
        DWORD           m_dwTimeoutMilliseconds;    // Timeout when waiting for the next completed I/O.
        CThreadPool&    m_threadPool;               // Back reference to the thread pool.
        CTrace         *m_pTrace;                   // The ostream trace support object.

    private:
        //// --------------------------------------------------------------------------------
        //// Forbidden.
        //// --------------------------------------------------------------------------------
        void operator=(const CThread&);
    };
    typedef CThread *PTHREAD;
    ////////////////////// END of Thread Class //////////////////////

public:
    //// --------------------------------------------------------------------------------
    //// Public callbacks.
    //// --------------------------------------------------------------------------------
    virtual DWORD OnExitThread(CThread& /*thread*/) {return 0;}

protected:
    //// --------------------------------------------------------------------------------
    //// Callbacks.
    //// --------------------------------------------------------------------------------
    virtual DWORD OnStart() {return 0;}
    virtual DWORD OnExit()  {return 0;}

    //// --------------------------------------------------------------------------------
    //// Helpers.
    //// --------------------------------------------------------------------------------

    // Stop the threads.
    virtual void StopThreads();

    // Deletes the thread objects, clears m_ppThreads, closes the termination event
    // handle and resets the state as it was before Run() was called, except for the
    // counters (number of runnable, live threads, peak number), which remains as they
    // were, and will be cleared by Run() if it is to be called again.
    virtual void Cleanup();

    //// --------------------------------------------------------------------------------
    //// Data.
    //// --------------------------------------------------------------------------------
    enum PoolState
    {
        STATE_STOPPED,
        STATE_STARTING,
        STATE_RUNNING,
        STATE_STOPPING
    };
    CThread         **m_ppThreads;              // Dynamically allocated array of threads.
    DWORD             m_dwSize;                 // Total number of threads initially created in the pool.
    DWORD             m_dwPeakRunnableThreads;  // Peak number of runnable threads in the pool.
    HANDLE            m_hTerminationEvent;      // Termination event.
    DWORD             m_dwState;                // The current state of the pool. See PoolState enumeration.
    DWORD             m_dwReadyThreads;         // Number of threads blocked on the IOCP. Valid during pool startup only.
    CTrace           *m_pTrace;                 // The ostream trace support object.

#ifdef USE_PERF_DATA
    LONG            * m_plActiveThreadsPerfCtr; // Pointer to performance counter for the active threads count.
#endif

private:
    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------
    void operator=(const CThreadPool&);
};

#endif  // _OPENSTA_THREADPOOL_H_
