/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

////////////////////////////////////////////////////////////////////////////////
// Timer queue implementation file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TimerQueue.hpp"
#include "Common.hpp"
#include "Fastlock.hpp"

#ifdef USE_PERF_DATA
#   include "PerfCounterKeys.hpp"
#endif

#include <process.h>


DWORD CTimerQueue::Init()
{
    DWORD dwRes = 0;

#ifdef USE_PERF_DATA
    m_pTimerQueueLengthPerfCount = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_TIMERQUEUELENGTH);
#endif

    // Create the queue's waitable synchronization (auto-reset) timer.
    m_hTimer = CreateWaitableTimer(NULL, FALSE, NULL);
    if(m_hTimer)
    {
        // Create the timer thread termination event.
        m_hTimerThread = BeginThreadEx(
            NULL,
            0,
            TimerThreadFunction,
            (void *)this,
            0,
            NULL
        );
        if(m_hTimerThread == 0)
        {
            dwRes = (DWORD)errno;
            CloseHandle(m_hTimer);
        }
    }
    else
        dwRes = GetLastError();

    return dwRes;
}

DWORD CTimerQueue::Push(IocpEventDispatcher *pEventDispatcher, DWORD dwDuration, BOOL bForce, DWORD dwExtraInfo)
{
    DWORD dwRes = 0;
    FILETIME   ftNow;
    unsigned __int64    qwNew;
    unsigned __int64    qwDuration;
    unsigned __int64    qwTop = 0;   // Top of queue before pushing the new value into the queue.

    if(m_bQueueOpen || bForce)
    {
        // Get "now" file time. It's better to do it before entering the
        // critical section, because the duration begins as close as possible
        // to the Push() call.
        GetSystemTimeAsFileTime(&ftNow);

        // Add the duration.
        qwNew = ftNow.dwHighDateTime;
        qwNew <<= 32;
        qwNew |= ftNow.dwLowDateTime;
        qwDuration = UInt32x32To64(dwDuration, 10000L);    // Convert duration from milliseconds to hundreds of nanoseconds.

        qwNew += qwDuration;

        DWORD dwSize;
        {
            CLock lock(m_sl);
            CTimerQueueEntry tqe(qwNew, pEventDispatcher, dwExtraInfo);
            dwSize = m_queue.size();
            if(dwSize > 0)
                qwTop = m_queue.top().m_qwTimestamp;
            m_queue.push(tqe);
            if(dwSize+1 > m_dwPeakSize)
                m_dwPeakSize = dwSize+1;
#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
            InterlockedExchange(m_pTimerQueueLengthPerfCount, (LONG)(dwSize+1));
#   else
            *m_pTimerQueueLengthPerfCount = (LONG)(dwSize+1);
#   endif
#endif
        }

        // If the queue was originally empty, or if the new timestamp is
        // less then the timestamp from the top of the queue, then
        // override the current timer setting with a new one.
        if(dwSize == 0 || qwNew < qwTop)
        {
            LARGE_INTEGER li;
            li.LowPart  = (DWORD)(qwNew & 0xFFFFFFFF);
            li.HighPart = (LONG) (qwNew >> 32);
            SetWaitableTimer(m_hTimer, &li, 0, NULL, NULL, FALSE);
        }
    }

    return dwRes;
}

void CTimerQueue::PopLoop()
{
    BOOL bContinue = TRUE;

    InterlockedExchange((LPLONG)&m_bQueueOpen, (LONG)TRUE);

    // Pop loop.
    while(bContinue)
    {
        DWORD dwReason = WaitForSingleObject(m_hTimer, m_dwTimeOutMilliseconds);

        if(dwReason == WAIT_OBJECT_0) // Wait OK
        {
            FILETIME   ftNow;
            __int64    qwNow;

            // Lock.
            CLock lock(m_sl);

            // Get "now" file time.
            GetSystemTimeAsFileTime(&ftNow);
            qwNow = ftNow.dwHighDateTime;
            qwNow <<= 32;
            qwNow |= ftNow.dwLowDateTime;

            // Pop from the queue all entries whose time has expired.
            BOOL bMore = TRUE;
            while(bMore && !m_queue.empty())
            {
                const CTimerQueueEntry tqe = m_queue.top();
                if(tqe.m_qwTimestamp <= qwNow)
                {
                    m_queue.pop();
#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
                    InterlockedDecrement(m_pTimerQueueLengthPerfCount);
#   else
                    m_pTimerQueueLengthPerfCount--;
#   endif
#endif
                    if(tqe.m_pEventDispatcher == NULL)
                    {
                        bContinue = FALSE;  // Must terminate the pop loop and exit the function.
                        break;              // Forget possible entries remaining in the queue.
                    }
                    else
                        m_iocp.PostCompletedIO(tqe.m_dwExtraInfo, tqe.m_pEventDispatcher, (LPOVERLAPPED)IocpEvent::EVENT_TYPE_TIMER);
                }
                else
                    bMore = FALSE;
            }

            if(bContinue && !m_queue.empty())
            {
                qwNow = m_queue.top().m_qwTimestamp;
                LARGE_INTEGER li;
                li.LowPart  = (DWORD)(qwNow & 0xFFFFFFFF);
                li.HighPart = (LONG) (qwNow >> 32);
                SetWaitableTimer(m_hTimer, &li, 0, NULL, NULL, FALSE);
            }
        } // else, timeout or error.
    }

    InterlockedExchange((LPLONG)&m_bQueueOpen, (LONG)FALSE);

    // Cancel any pending timer.
    CancelWaitableTimer(m_hTimer);
}

DWORD WINAPI CTimerQueue::TimerThreadFunction(LPVOID pContext)
{
    CTimerQueue *pTQ = (CTimerQueue*)pContext;

    if(pTQ->m_pTrace)
    {
        GetTraceStream(pTQ->m_pTrace)
            << "The timer thread is " << GetCurrentThreadId() << "\r\n";
        ReleaseTraceStream();
    }

    pTQ->PopLoop();

    if(pTQ->m_pTrace)
    {
        GetTraceStream(pTQ->m_pTrace)
            << "The timer thread is exiting..." << "\r\n";
        ReleaseTraceStream();
    }

    return 1;
}

CTimerQueue::~CTimerQueue()
{
    if(m_hTimerThread)
    {
        // Test if the timer thread is still running.
        if(WaitForSingleObject(m_hTimerThread, 0) == WAIT_TIMEOUT)
        {
            //// The timer thread is still running, so stop it.
            Push(NULL, 0, TRUE);
            WaitForSingleObject(m_hTimerThread, INFINITE);
        }

        CloseHandle(m_hTimerThread);
    }
    if(m_hTimer)
        CloseHandle(m_hTimer);
}
