/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
#ifndef __PHPARSE__
#define __PHPARSE__


#include "phlex.hpp"
#include "PString.hpp"
#include "HTMLTree.hpp"
#include <vector>
#include <list>
#include "Address.hpp"



// ---------------------------------------------------------------------------------------
// THREAD SAFETY: this class is NOT thread safe, but has no non-const static data members.
// =============  Therefore, thread safety is guaranteed only when an object of this class
//                is only accessed by it's creating thread.
// ---------------------------------------------------------------------------------------
// Usage:
// =====
//
// - The type T is the same as the type T of the lexical analyzer class.
//   It must have been defined to manage both token and nonterminal attributes.
//   Requirements for T:
//      - A default constructor.
//      - A copy constructor.
//      - An assignment operator.
//
// - To react to common parser events, override the following member functions as
//   specializations for your type T:
//      - OnAccept
//      - OnStopParser
//      - PerformSemanticAction (mandatory)
//
// - To add error recovery, override the following member functions as specializations for
//   your type T:
//      - OnError
//      - OnFatalError
//      - OnEnterErrorRecovery
//      - OnErrorRecovered
//
// - To produce a tracing version, define the macro PHTRACE.
// ---------------------------------------------------------------------------------------


template <typename U>
class PhParser
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    PhParser(PhLexer<U>& lexer, BOOL bStopWhenAllAddressesResolved = TRUE, std::ostream& os = std::cout)
        : lexer_(lexer), os_(os), bStopWhenAllAddressesResolved_(bStopWhenAllAddressesResolved),
          bProduceTree_(FALSE), pRoot_(NULL), pCurrentTag_(NULL),
          bHasHTMLTag_(FALSE), bHasHeadTag_(FALSE), bHasBodyTag_(FALSE), bHeadClosed_(FALSE),
          bInTable_(FALSE), bTableHeadOrFoot_(FALSE), bHasTbody_(FALSE), bDisable_(FALSE) {}


    enum AfterErrorAction
    {
        RAISE_FATAL_ERROR,      // Emit a fatal error.
        TRY_TO_RECOVER,         // Attempt to recover from error.
        STOP_PARSER             // Stop the parser & go to the accepting state (successful termination).
    };

    //// --------------------------------------------------------------------------------
    //// Semantic types
    //// --------------------------------------------------------------------------------
    struct AddressValue
    {
        AddressValue(const AddressQuery* pAq) : pAq_(pAq), pcBegin_(U()), ulSize_(0) {}

        // An address-query.
        const AddressQuery* pAq_;

        // The value at this address-query.
        U               pcBegin_;           // A pointer to the beginning of the value. NULL if not resolved.
        unsigned long   ulSize_;            // The number of characters of the value.
    };
    typedef std::list<AddressValue> AddressValueList;


    //// --------------------------------------------------------------------------------
    //// Parsing Functions
    //// --------------------------------------------------------------------------------

    // Parsing algorithm. Input: a list of HTML element addresses. Ouput: the resolved addresses.
    BOOL Parse(const AddressQueryList& aql, AddressValueList& avl, unsigned long* pulErrorLine = NULL, unsigned long* pulErrorCol = NULL, BOOL bProduceTree = FALSE);

    // Parsing algorithm. Ouput: an HTML tree.
    BOOL Parse(CTag*& pRoot, unsigned long* pulErrorLine = NULL, unsigned long* pulErrorCol = NULL);


    //// --------------------------------------------------------------------------------
    //// Accessors
    //// --------------------------------------------------------------------------------
    const PhLexer<U>& getLexer() const {return lexer_;}
    PhLexer<U>& getLexer() {return lexer_;}


//---------------------------------------------------------------------------

protected:

    enum ParsingPhase
    {
        PHASE_NORMAL,
        PHASE_ERROR
    };

    struct StackEntry
    {
        StackEntry() : ulSymbol_(0) {}
        StackEntry(unsigned long ulSymbol, const CSymbolAttributes<U>& sa = CSymbolAttributes<U>()) : ulSymbol_(ulSymbol), symbolAttributes_(sa) {}

        unsigned long ulSymbol_;
        CSymbolAttributes<U> symbolAttributes_;
    };

    //// --------------------------------------------------------------------------------
    //// Parsing helpers
    //// --------------------------------------------------------------------------------

    // Parsing table.
    long ActionGoto(unsigned long ulState, unsigned long ulSymbol);

    // Pop n entries from the stack.
    void Pop(unsigned long ulLen);


    //// --------------------------------------------------------------------------------
    //// Parsing events. To be overriden.
    //// --------------------------------------------------------------------------------

    // Called when the parser detects a syntax error.
    AfterErrorAction OnError(long lLine, long lCol, BOOL bDuringRecovery);

    // Called on a syntax error, when the parser enters the recovery phase.
    void OnEnterErrorRecovery(BOOL /*bDuringRecovery*/);

    // Called on an unrecovered lexical error.
    void OnFatalError(BOOL bDuringRecovery);

    // Called when recovered from a lexical error.
    void OnErrorRecovered();

    // Called after the parser reaches the ACCEPT state.
    void OnAccept() {}

    // Called when the parser is stopped by its caller
    void OnStopParser() {}

    // Called to execute a semantic action. Must be overriden.
    // Return TRUE to continue parsing, FALSE to stop.
    BOOL PerformSemanticAction(AddressValueList& avl, long lProduction,
        CSymbolAttributes<U>& leftSymbolAttributes, StackEntry* pProdAttr,
        ParsingPhase& ulPhase);


    //// --------------------------------------------------------------------------------
    //// Stack manipulation from within a semantic action.
    //// --------------------------------------------------------------------------------
    CSymbolAttributes<U>& GetATT(long l, StackEntry* pEntry) {return pEntry[l].symbolAttributes_;}
    const CSymbolAttributes<U>& GetATT(long l, StackEntry* pEntry) const {return pEntry[l].symbolAttributes_;}

    //// --------------------------------------------------------------------------------
    //// Semantic functions.
    //// --------------------------------------------------------------------------------
    void ProcessOpeningTag      (CSymbolAttributes<U>& tokenAttributes);
    void ProcessClosingTag      (CSymbolAttributes<U>& tokenAttributes);
    BOOL ProcessAttributeValue  (AddressValueList& avl, CSymbolAttributes<U>& attrValue);
    BOOL ProcessAttributeValue  (AddressValueList& avl, CSymbolAttributes<U>& attrName, CSymbolAttributes<U>& attrValue);
    BOOL ProcessText            (AddressValueList& avl, CSymbolAttributes<U>& text);
    BOOL ProcessScript          (AddressValueList& avl, CSymbolAttributes<U>& text);
    BOOL ProcessComment         (AddressValueList& avl, CSymbolAttributes<U>& comment);
    BOOL AddressCheck           (AddressValueList& avl, CSymbolAttributes<U>& valueAttributes);


    //// --------------------------------------------------------------------------------
    //// Parser Data
    //// --------------------------------------------------------------------------------
    typedef std::vector<StackEntry> StackType;
    std::ostream&   os_;            // The output stream
    StackType       stack_;         // Dynamic parser stack
    PhLexer<U>&     lexer_;         // Lexical Analyzer

    //// --------------------------------------------------------------------------------
    //// Constants.
    //// --------------------------------------------------------------------------------
    enum
    {
        BLANK       = -32768,   // Value of a parsing table entry that indicates a syntax error.
        ACCEPT      =  32767,   // "Accept" value of a parsing table entry.
        ERROR_TOKEN =      1,   // The error token. For error productions (recovery).
        END_OF_LIST =     -1    // Marker for end of list in AGPool array.
    };

    //// --------------------------------------------------------------------------------
    //// Semantic Data.
    //// --------------------------------------------------------------------------------
    BOOL          bProduceTree_;
    Address       currentAddress_;  // The current address.
    CTag*         pRoot_;
    CTag*         pCurrentTag_;
    BOOL          bStopWhenAllAddressesResolved_;
    BOOL          bHasHTMLTag_, bHasHeadTag_, bHasBodyTag_, bHeadClosed_;
    BOOL          bInTable_, bTableHeadOrFoot_, bHasTbody_;
    BOOL          bDisable_;        // Disable information collection when inside and unknown tag.

private:
    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------
    void operator=(const PhParser&);
};


//// --------------------------------------------------------------------------------
//// Semantic action macros.
//// --------------------------------------------------------------------------------

// Retrieving a symbol's attributes (right side of the production).
#define ATT(l) GetATT(l, pProdAttr)

// Exiting the error recovery mode, simulating successful error recovery.
#define ExitErrorRecovery() (ulPhase = PHASE_NORMAL, OnErrorRecovered(), bError = FALSE)


#include "phparseimpl.hpp"


#endif  // defined(__PHPARSE__)
