/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//==================================================================< Pragmas >
#ifdef  WIN32
#  pragma warning(disable : 4512 4710)
#endif  // WIN32


//=================================================================< Includes >
#ifndef __DEFINITION__
#  include "definition.h"
#endif  // __DEFINITION__

#ifndef __FUNCTOR__
#  include <functor.hpp>
#endif  // __FUNCTOR__

#ifndef __GATEWAY__
#  include "gateway.hpp"
#endif  // __GATEWAY__

#include <signal.h>

#include <iostream>
using std::cout;
using std::endl;


//======================================================< Definition of class >

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////// Globals Functions //
///////////////////////////////////////////////////////////////////////////////

void signalInterruptionFunction (int iSig) ;


///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// End of Globals Functions //
///////////////////////////////////////////////////////////////////////////////


//=====================================================================< Data >
CGMain*         GpProcessMain = NULL ;

CGTasksList     GnullTaskList ;
CGTask          GnullTask (NULL, GnullTask, GnullTaskList, CGTask::TASK_CONTROL, 0, false) ;


//====================================================================< Class >

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////// Class CGTask //
///////////////////////////////////////////////////////////////////////////////

//==================================================================< Members >
const char* CGTask::szTaskType_[] =
    {
        "CT", "ALT", "ACT", "GLT", "GCT"
    } ;

//============================================< Constructor(s) and destructor >
//-----------------------------------------------------------------------------
// CGTask::CGTask
//
// Constructor.
//
// Arguments : pInitInfo    : initialization informations pointer.
//             father       : task which create this new task.
//             tasksList    : tasks list to insert into.
//             eType        : type of the new task.
//             iTaskId      : task identifier.
//             bToBeCreated : task need to be created.
// Return    : none.
//-----------------------------------------------------------------------------
CGTask::CGTask (void* pInitInfo, CGTask& father, CGTasksList& tasksList, enumTaskType eType, int iTaskId, bool bToBeCreated /* = true */)
        : pInitInfo_ (pInitInfo), pData_ (NULL), pFather_ (&father), tasksList_ (tasksList), pThread_ (NULL), eType_ (eType), pStatus_ (NULL), bInitializationTerminated_ (false), bInitializationIsOk_ (false), bToBeReinitialized_ (false), iTaskId_ (iTaskId), iLevel_ (-2), name_ ()
{
    if (pFather_)
    {
        pFather_->childrenList_.push_back (this) ;
    }
    try
    {
        switch (eType_)
        {
        case TASK_CONTROL :
            if (bToBeCreated)
            {   // Error from bToBeCreated.
                throw CGTaskException ("CGTask: don't know how to create a new control task.", __FILE__, __LINE__) ;
            }
            else
            {
                pThread_ = new CTThread (CTCurrentThread::getThreadId () ) ;
                if (pThread_ == NULL)
                {   // Error from new.
                    throw CGTaskException ("CGTask: new failed.", __FILE__, __LINE__) ;
                }
            }
            break ;
        case TASK_ADMINISTRATION_LISTEN :
            if (bToBeCreated)
            {
                pThread_ = new CTThread (makeFunctorA1 (administrationListenTask, CGTask&, *this) ) ;
                if (pThread_ == NULL)
                {   // Error from new.
                    throw CGTaskException ("CGTask: new failed.", __FILE__, __LINE__) ;
                }
            }
            else
            {   // Error from bToBeCreated.
                throw CGTaskException ("CGTask: don't know how to add a current administration listen task.", __FILE__, __LINE__) ;
            }
            break ;
        case TASK_ADMINISTRATION_CLIENT :
            if (bToBeCreated)
            {
                pThread_ = new CTThread (makeFunctorA1 (administrationClientTask, CGTask&, *this) ) ;
                if (pThread_ == NULL)
                {   // Error from new.
                    throw CGTaskException ("CGTask: new failed.", __FILE__, __LINE__) ;
                }
            }
            else
            {   // Error from bToBeCreated.
                throw CGTaskException ("CGTask: don't know how to add a current administration client task.", __FILE__, __LINE__) ;
            }
            break ;
        case TASK_GATEWAY_LISTEN :
            if (bToBeCreated)
            {
                pThread_ = new CTThread (makeFunctorA1 (gatewayListenTask, CGTask&, *this) ) ;
                if (pThread_ == NULL)
                {   // Error from new.
                    throw CGTaskException ("CGTask: new failed.", __FILE__, __LINE__) ;
                }
            }
            else
            {   // Error from bToBeCreated.
                throw CGTaskException ("CGTask: don't know how to add a current gateway listen task.", __FILE__, __LINE__) ;
            }
            break ;
        case TASK_GATEWAY_CLIENT :
            if (bToBeCreated)
            {
                pThread_ = new CTThread (makeFunctorA1 (gatewayClientTask, CGTask&, *this) ) ;
                if (pThread_ == NULL)
                {   // Error from new.
                    throw CGTaskException ("CGTask: new failed.", __FILE__, __LINE__) ;
                }
            }
            else
            {   // Error from bToBeCreated.
                throw CGTaskException ("CGTask: don't know how to add a current gateway client task.", __FILE__, __LINE__) ;
            }
            break ;
        default :
            throw CGTaskException ("CGTask: eType_ is invalid.", __FILE__, __LINE__) ;
        }
    }
    catch (CGTaskException&)
    {
        if (pFather_)
        {
            for (CGTaskStlList::iterator itr = pFather_->childrenList_.begin(); itr != pFather_->childrenList_.end(); itr++)
            {
                if (*itr == this)
                {
                    pFather_->childrenList_.erase(itr) ;
                    break;
                }
            }
        }
        throw ;
    }
    if (pFather_)
    {
        iLevel_ = pFather_->iLevel_ + 1 ;
    }
    for (int i = 0 ; i < iLevel_ ; ++ i)
    {
        name_ += ">" ;
    }
    char    szBuf [6] ;
    sprintf (szBuf, "%hu", (unsigned short) iTaskId_) ;
    name_ += string (" ") + szTaskType_ [eType_] + " " + szBuf ;
} // CGTask::CGTask

CGTask::~CGTask ()
{
    if (pThread_)
    {
        delete pThread_ ;
        pThread_ = NULL ;
    }
} // CGTask::~CGTask

//=========================================================< Control methodes >
//-----------------------------------------------------------------------------
// CGTask::initializationTerminated
//
// Tell the task is initialized.
//
// Arguments : bInitializationIsOk : specify if the termination is ok or failed.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTask::initializationTerminated (bool bInitializationIsOk /* = true */)
{
    if (!bInitializationTerminated_)
    {   // Signal the condition from the list.
        bInitializationIsOk_ = bInitializationIsOk ;
        tasksList_.getConditionTaskInitialize ().signal () ;
        bInitializationTerminated_ = true ;
        pInitInfo_ = NULL ;
    }
} // CGTask::initializationTerminated

//-----------------------------------------------------------------------------
// CGTask::kill
//
// Kill the task.
//
// Arguments : bRemoveFromFather    : remove from father children list.
//             bRemoveFromTasksList : remove from taks list.
//             bChildrenAlso        : kill children also.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTask::kill (bool bRemoveFromFather /* = false */, bool bRemoveFromTasksList /* = false */, bool bChildrenAlso /* = false */)
{
    if ( (bRemoveFromFather) && (pFather_) )
    {   // Remove this task from the list of the father.
        for (CGTaskStlList::iterator itr = pFather_->childrenList_.begin(); itr != pFather_->childrenList_.end(); itr++)
        {
            if (*itr == this)
            {
                pFather_->childrenList_.erase(itr) ;
                break;
            }
        }
    }

    if (bRemoveFromTasksList)
    {   // Remove this task from the task list.
        for (CGTaskStlList::iterator itr = tasksList_.getTasksList ().begin(); itr != tasksList_.getTasksList ().end(); itr++)
        {
            if (*itr == this)
            {
                tasksList_.getTasksList ().erase(itr) ;
                break;
            }
        }
    }

    CGTask* pTask ;
    while (!childrenList_.empty () )
    {
        // Get the first task.
        pTask = childrenList_[0] ;
        childrenList_.erase(childrenList_.begin());

        if (bChildrenAlso)
        {   // Kill the task.
            pTask->kill (bRemoveFromFather, bRemoveFromTasksList, bChildrenAlso) ;

            // Delete the task.
            delete pTask ;
        }
        else
        {
            if ( (bRemoveFromFather) && (pFather_) )
            {   // Put the task in the father list.
                pFather_->childrenList_.push_back (pTask) ;
                pTask->pFather_ = pFather_ ;
            }
        }
    }

    if ( (pThread_) && (pThread_->getThreadId () != CTCurrentThread::getThreadId () ) )
    {   // Kill the thread.
        pThread_->kill ( (void*) -1) ;
        // Join the thread.
        CTCurrentThread::join (*pThread_, &pStatus_) ;
    }
} // CGTask::kill

//-----------------------------------------------------------------------------
// CGTask::join
//
// Join the task.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTask::join (void)
{
    if ( (pThread_) && (pThread_->getThreadId () != CTCurrentThread::getThreadId () ) )
    {   // Join the thread.
        CTCurrentThread::join (*pThread_, &pStatus_) ;
    }
} // CGTask::join

//-----------------------------------------------------------------------------
// CGTask::suspend
//
// Suspend the task.
//
// Arguments : bChildrenAlso : suspend children also.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTask::suspend (bool bChildrenAlso /* = false */)
{
    if (bChildrenAlso)
    {   // Suspend the children.
        for (CGTaskStlList::iterator itr = childrenList_.begin();itr != childrenList_.end(); itr++)
            (*itr)->suspend (bChildrenAlso) ;
    }

    if ( (pThread_) && (pThread_->getThreadId () != CTCurrentThread::getThreadId () ) )
    {   // Suspend the thread.
        pThread_->suspend () ;
    }
} // CGTask::suspend

//-----------------------------------------------------------------------------
// CGTask::resume
//
// Resume the task.
//
// Arguments : bChildrenAlso : resume children also.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTask::resume (bool bChildrenAlso /* = false */)
{
    if (bChildrenAlso)
    {   // Resume the children.
        for (CGTaskStlList::iterator itr = childrenList_.begin();itr != childrenList_.end(); itr++)
            (*itr)->resume  (bChildrenAlso) ;
    }

    if ( (pThread_) && (pThread_->getThreadId () != CTCurrentThread::getThreadId () ) )
    {   // Resume the thread.
        pThread_->resume () ;
    }
} // CGTask::resume

//-----------------------------------------------------------------------------
// CGTask::reinitialize
//
// Reinitialize the task.
//
// Arguments : bChildrenAlso : resume children also.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTask::reinitialize (bool bChildrenAlso /* = false */)
{
    if (bChildrenAlso)
    {   // Reinitialize the children.
        for (CGTaskStlList::iterator itr = childrenList_.begin();itr != childrenList_.end(); itr++)
            (*itr)->reinitialize (bChildrenAlso) ;
    }

    bToBeReinitialized_ = true ;
} // CGTask::reinitialize

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////// End of Class CGTask //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////// Class CGTasksList //
///////////////////////////////////////////////////////////////////////////////

//============================================< Constructor(s) and destructor >

//=========================================================< Control methodes >
//-----------------------------------------------------------------------------
// CGTasksList::addCurrentTask
//
// Add the current task in the list.
//
// Arguments : eType  : task type.
//             father : Parent task.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTasksList::addCurrentTask (CGTask::enumTaskType eType, CGTask& father)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    switch (eType)
    {
    case CGTask::TASK_CONTROL :
        ++ iNbTasks_ ;
        tasksList_.push_back (new CGTask (NULL, father, *this, eType, iNbTasks_, false) ) ;
        break ;
    case CGTask::TASK_ADMINISTRATION_LISTEN :
    case CGTask::TASK_ADMINISTRATION_CLIENT :
    case CGTask::TASK_GATEWAY_LISTEN :
    case CGTask::TASK_GATEWAY_CLIENT :
    default :
        throw CGTaskException ("addCurrentTask: eType is invalid.", __FILE__, __LINE__) ;
    }
} // CGTasksList::addCurrentTask

//-----------------------------------------------------------------------------
// CGTasksList::newTask
//
// Create a new task.
//
// Arguments : eType     : task type.
//             pInitInfo : initialization informations pointer.
//             father    : Parent task.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTasksList::newTask (CGTask::enumTaskType eType, void* pInitInfo, CGTask& father)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    switch (eType)
    {
    case CGTask::TASK_CONTROL :
        throw CGTaskException ("newTask: eType is invalid.", __FILE__, __LINE__) ;
        break ;
    case CGTask::TASK_ADMINISTRATION_LISTEN :
    case CGTask::TASK_ADMINISTRATION_CLIENT :
    case CGTask::TASK_GATEWAY_LISTEN :
    case CGTask::TASK_GATEWAY_CLIENT :
        {
            LockGuard guard (mutexTaskInitialize_) ;
            ++ iNbTasks_ ;
            CGTask* pTask = new CGTask (pInitInfo, father, *this, eType, iNbTasks_) ;
            tasksList_.push_back (pTask) ;
            conditionTaskInitialize_.wait () ;
            if (!pTask->initializationIsOk () )
            {
                for (CGTaskStlList::iterator itr =father.getChildrenList ().begin(); itr != father.getChildrenList ().end(); itr++)
                {
                    if (*itr == pTask)
                    {
                        father.getChildrenList ().erase(itr);
                        break ;
                    }
                }
                for (itr =tasksList_.begin(); itr != tasksList_.end(); itr++)
                {
                    if (*itr == pTask)
                    {
                        tasksList_.erase(itr);
                        break ;
                    }
                }

                throw CGTaskException ("newTask: new task's initialization failed.", __FILE__, __LINE__) ;
            }
        }
        break ;
    default :
        throw CGTaskException ("newTask: eType is invalid.", __FILE__, __LINE__) ;
    }
} // CGTasksList::newTask

//-----------------------------------------------------------------------------
// CGTasksList::killTask
//
// Kill a task and its children.
//
// Arguments : iIndex        : index of the task.
//             bChildrenAlso : kill children also.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTasksList::killTask (int iIndex, bool bChildrenAlso /* = false */)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    // Get the task from index.
    CGTask* pTask = tasksList_[iIndex] ;

    // Verify the task can be killed.
    if ( (iIndex == 0) && (pTask->getType () == CGTask::TASK_CONTROL) )
    {   // Error from iIndex.
        throw CGTaskException ("killTask: can't kill the control task.", __FILE__, __LINE__) ;
    }

    // Kill the task.
    pTask->kill (true, true, bChildrenAlso) ;

    // Delete the task.
    delete pTask ;
} // CGTasksList::killTask

//-----------------------------------------------------------------------------
// CGTasksList::killAll
//
// Kill all the tasks.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTasksList::killAll (void)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    CGTask* pTask ;

    // Free the list.
    for (CGTaskStlList::iterator itr = tasksList_.begin(); itr != tasksList_.end(); itr ++)
    {
        pTask = *itr ;
        // Kill the task.
        pTask->kill () ;
        delete pTask ;
    }
    tasksList_.clear();

} // CGTasksList::killAll

//-----------------------------------------------------------------------------
// CGTasksList::joinTask
//
// Join a task.
//
// Arguments : iIndex        : index of the task.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTasksList::joinTask (int iIndex)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    // Get the task from index.
    CGTask* pTask = tasksList_.at (iIndex) ;

    // Verify the task can be killed.
    if ( (iIndex == 0) && (pTask->getType () == CGTask::TASK_CONTROL) )
    {   // Error from iIndex.
        throw CGTaskException ("joinTask: can't join the control task.", __FILE__, __LINE__) ;
    }

    // Unlock the list.
    UnlockGuard unlock (monitor () ) ;

    // Kill the task.
    pTask->join () ;
} // CGTasksList::joinTask

//-----------------------------------------------------------------------------
// CGTasksList::joinAll
//
// Join all the tasks.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTasksList::joinAll (void)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    while (tasksList_.size() > 1)
    {
        // Get the first task.
        CGTask* pTask = tasksList_[1] ;

        // Unlock the list.
        UnlockGuard unlock (monitor () ) ;

        // Kill the task.
        pTask->join () ;
    }

    // Delete the last task.
    delete tasksList_.front () ;
} // CGTasksList::joinAll

//-----------------------------------------------------------------------------
// CGTasksList::suspendTask
//
// Suspend a task and its children.
//
// Arguments : iIndex        : index of the task.
//             bChildrenAlso : suspend children also.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTasksList::suspendTask (int iIndex, bool bChildrenAlso /* = false */)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    // Get the task from index.
    CGTask* pTask = tasksList_[iIndex] ;

    // Verify the task can be suspended.
    if ( (iIndex == 0) && (pTask->getType () == CGTask::TASK_CONTROL) )
    {   // Error from iIndex.
        throw CGTaskException ("suspendTask: can't suspend the control task.", __FILE__, __LINE__) ;
    }

    // Suspend the task.
    pTask->suspend (bChildrenAlso) ;
} // CGTasksList::suspendTask

//-----------------------------------------------------------------------------
// CGTasksList::suspendAll
//
// Suspend all the tasks.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTasksList::suspendAll (void)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    for (CGTaskStlList::iterator itr = tasksList_.begin(); itr != tasksList_.end(); itr ++)
    {   // Suspend the task.
        (*itr)->suspend () ;
    }
} // CGTasksList::suspendAll

//-----------------------------------------------------------------------------
// CGTasksList::resumeTask
//
// Resume a task and its children.
//
// Arguments : iIndex        : index of the task.
//             bChildrenAlso : resume children also.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTasksList::resumeTask (int iIndex, bool bChildrenAlso /* = false */)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    // Get the task from index.
    CGTask* pTask = tasksList_[iIndex] ;

    // Verify the task can be resumed.
    if ( (iIndex == 0) && (pTask->getType () == CGTask::TASK_CONTROL) )
    {   // Error from iIndex.
        throw CGTaskException ("resumeTask: can't resume the control task.", __FILE__, __LINE__) ;
    }

    // Resume the task.
    pTask->resume (bChildrenAlso) ;
} // CGTasksList::resumeTask

//-----------------------------------------------------------------------------
// CGTasksList::resumeAll
//
// Resume all the tasks.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTasksList::resumeAll (void)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    for (CGTaskStlList::iterator itr = tasksList_.begin(); itr != tasksList_.end(); itr ++)
    {   // Resume the task.
        (*itr)->resume () ;
    }
} // CGTasksList::resumeAll

//-----------------------------------------------------------------------------
// CGTasksList::reinitializeTask
//
// Reinitialize a task and its children.
//
// Arguments : iIndex        : index of the task.
//             bChildrenAlso : resume children also.
// Return    : none.
//-----------------------------------------------------------------------------
void CGTasksList::reinitializeTask (int iIndex, bool bChildrenAlso /* = false */)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    // Get the task from index.
    CGTask* pTask = tasksList_[iIndex] ;

    // Verify the task can be reinitialized.
    if ( (iIndex == 0) && (pTask->getType () == CGTask::TASK_CONTROL) )
    {   // Error from iIndex.
        throw CGTaskException ("reinitializeTask: can't reinitialize the control task.", __FILE__, __LINE__) ;
    }

    // Reinitialize the task.
    pTask->reinitialize (bChildrenAlso) ;
} // CGTasksList::reinitializeTask

//=============================================================< Get methodes >
//-----------------------------------------------------------------------------
// CGTasksList::getCurrentTask
//
// Get the current task.
//
// Arguments : none.
// Return    : currentTask : current task.
//-----------------------------------------------------------------------------
CGTask& CGTasksList::getCurrentTask (void)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;

    CGTask*                     pTask ;
    for (CGTaskStlList::iterator itr = tasksList_.begin(); itr != tasksList_.end(); itr ++)
    {   // Is the current task ?
        pTask = *itr;
        CTThread* pThread = pTask->getThread () ;
        if ( (pThread == NULL) || (pThread->getThreadId () == CTCurrentThread::getThreadId () ) )
        {
            break ;
        }
    }

    if (pTask == NULL)
    {
        // Put current task in the list.
        addCurrentTask (CGTask::TASK_CONTROL, GnullTask) ;

        // Get the current task from the list.
        for (CGTaskStlList::iterator itr = tasksList_.begin(); itr != tasksList_.end(); itr ++)
        {   // Is the current task ?
            pTask = *itr;
            CTThread* pThread = pTask->getThread () ;
            if ( (pThread == NULL) || (pThread->getThreadId () == CTCurrentThread::getThreadId () ) )
            {
                break ;
            }
        }

        if (pTask == NULL)
        {   // Error from this.
            throw CGTaskException ("getCurrentTask: can't get the current task.", __FILE__, __LINE__) ;
        }
    }

    // Return a current task reference.
    return *pTask ;
} // CGTasksList::getCurrentTask

//-----------------------------------------------------------------------------
// CGTasksList::index
//
// Search for a task and return its index.
//
// Arguments : testTask  : return if the task is the searched task.
//             pCompInfo : pointer to comparaison informations.
// Return    : iIndex    : index of the searched task.
//-----------------------------------------------------------------------------
int CGTasksList::index (fnTestTask testTask, void* pCompInfo)
{
    // Lock the list.
    LockGuard lock (monitor () ) ;
    bool bFound = false ;

    for (int iIndex=0; iIndex < tasksList_.size(); iIndex++)
    {
        if (testTask(tasksList_[iIndex], pCompInfo))
        {
            bFound = true ;
            break ;
        }
    }
    if (bFound)
        return iIndex ;
    // Return the searched task pointer.
    return string::npos ;
} // CGTasksList::index

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// End of Class CGTasksList //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////// Class CGMain //
///////////////////////////////////////////////////////////////////////////////

//==========================================================< Constantes data >
const char* CGMain::azOption_[CGMain::OPT_LAST] =
#ifdef  WIN32
    { "-console", "-service", "-install", "-uninstall", "-stop", "-pause", "-continue", "-help", "-adm_port", "-reuse_port", "-cli_cfg", "-log", "-trace", "-lib_arg"
    } ;
#endif  // WIN32
#ifdef  UNIX
{ "-console", "-daemon", "-install", "-uninstall", "-stop", "-pause", "-continue", "-help", "-adm_port", "-reuse_port", "-cli_cfg", "-log", "-trace", "-lib_arg"} ;
#endif  // UNIX


//============================================< Constructor(s) and destructor >
//-----------------------------------------------------------------------------
// CGMain::CGMain
//
// Constructor.
//
// Arguments : serviceDisplayName : Service name.
//             argc : Number of arguments.
//             argv : Arguments array.
// Return    : none.
//-----------------------------------------------------------------------------
CGMain::CGMain (const string& serviceDisplayName, int argc, char* argv[])
        : CTService (argv[0], serviceDisplayName), serviceDisplayName_ (serviceDisplayName), eCommand_ (SRV_START), bTraceMode_ (false), logFileName_ ("gateway.log"), adminPort_ ("3000"), clientConfig_ (""), conditionContinue_ (mutexContinue_), bWaitTasks_ (false), bReuseAddress_ (false)
{
    parse (argc, argv) ;
} // CGMain::CGMain

//-----------------------------------------------------------------------------
// CGMain::~CGMain
//
// Destructor.
//
// Arguments : serviceDisplayName : Service name.
//             argc : Number of arguments.
//             argv : Arguments array.
// Return    : none.
//-----------------------------------------------------------------------------
CGMain::~CGMain ()
{} // CGMain::~CGMain

//=========================================================< Control methodes >
//-----------------------------------------------------------------------------
// CGMain::parse
//
// Parse the command line arguments.
//
// Arguments : argc : Number of arguments.
//             argv : Arguments array.
// Return    : none.
//-----------------------------------------------------------------------------

void CGMain::parse (int argc, char* argv[])
{
    int         i ;
    string  arg ;
    bool        bLibraryCmdLineArg = false ;

    programName_ = argv[0] ;
    programName_.erase(0, programName_.find_last_of ('\\') + 1) ;
    programName_.erase (0, programName_.find_last_of ('/') + 1) ;
    for (i = 1 ; i < argc ; ++ i)
    {
        arg = argv[i] ;
        if (arg == azOption_[OPT_CONSOLE])
        {   // Run as a console.
            eCommand_ = SRV_CONSOLE ;
        }
        else if (arg == azOption_[OPT_SERVICE])
        {   // Run as a service.
            eCommand_ = SRV_SERVICE ;
        }
        else if (arg == azOption_[OPT_INSTALL])
        {   // Install the service.
            eCommand_ = SRV_INSTALL ;
        }
        else if (arg == azOption_[OPT_UNINSTALL])
        {   // Uninstall the service.
            eCommand_ = SRV_UNINSTALL ;
        }
        else if (arg == azOption_[OPT_STOP])
        {   // Stop the service.
            eCommand_ = SRV_STOP ;
        }
        else if (arg == azOption_[OPT_PAUSE])
        {   // Pause the service.
            eCommand_ = SRV_PAUSE ;
        }
        else if (arg == azOption_[OPT_CONTINUE])
        {   // Continue the service.
            eCommand_ = SRV_CONTINUE ;
        }
        else if (arg == azOption_[OPT_HELP])
        {   // Display help.
            eCommand_ = SRV_HELP ;
            break ;
        }
        else if (arg == azOption_[OPT_ADMINISTRATION_LISTEN_PORT])
        {   // Specify the administration listen port.
            if (argc > i + 1)
            {
                adminPort_ = argv[++ i] ;
            }
            else
            {   // Error from command line arguments.
                throw CTException (string ("parse: option ") + arg + " : not enough arguments.", __FILE__, __LINE__) ;
            }
        }
        else if (arg == azOption_[OPT_REUSE_PORT])
        {   // Specify that socket can be bound to an address that is already used.
            bReuseAddress_ = true ;
        }
        else if (arg == azOption_[OPT_CLIENT_CONFIG])
        {   // Specify the clients configuration (listen port, parse library, target address).
            if (argc > i + 1)
            {
                clientConfig_ = argv[++ i] ;
            }
            else
            {   // Error from command line arguments.
                throw CTException (string ("parse: option ") + arg + " : not enough arguments.", __FILE__, __LINE__) ;
            }
        }
        else if (arg == azOption_[OPT_LOGFILE])
        {   // Specify the log file for the gateway.
            if (argc > i + 1)
            {
                logFileName_ = argv[++ i] ;
            }
            else
            {   // Error from command line arguments.
                throw CTException (string ("parse: option ") + arg + " : not enough arguments.", __FILE__, __LINE__) ;
            }
        }
        else if (arg == azOption_[OPT_TRACE])
        {   // Activate the trace mode.
            bTraceMode_ = true ;
        }
        else if (arg == azOption_[OPT_LIBARG])
        {   // Activate the trace mode.
            bLibraryCmdLineArg = true ;
        }
        else
        {
            if (bLibraryCmdLineArg)
            {
                if (libraryCmdLineArg_.empty () )
                {
                    libraryCmdLineArg_ = arg ;
                }
                else
                {
                    libraryCmdLineArg_ += " " + arg ;
                }
            }
            else
            {   // Error from command line arguments.
                throw CTException (string ("parse: option ") + arg + " : bad argument.", __FILE__, __LINE__) ;
            }
        }
    }
} // CGMain::parse

//-----------------------------------------------------------------------------
// CGMain::go
//
// Execute the service command.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGMain::go (void)
{
    logFile_.open (logFileName_.c_str()) ;
    if (logFile_.is_open () )
    {
        logFileList_.push_back (logFile_) ;
    }
    if (getStartMode () == START_CONSOLE)
    {
        logFileList_.push_back (cout) ;
    }
    switch (eCommand_)
    {
    case SRV_START :
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Running in default mode (as a " << getStartModeLabel () << ")..." << endl ;
        }
        start () ;
        break ;
    case SRV_CONSOLE :
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Running in console mode..." << endl ;
        }
        start (START_CONSOLE) ;
        break ;
    case SRV_SERVICE :
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
#ifdef  WIN32
            logFileList_ << "Try to start the service..." << endl ;
#endif  // WIN32
#ifdef  UNIX
            logFileList_ << "Running as a daemon..." << endl ;
#endif  // UNIX
        }

        start (START_SERVICE) ;
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "ok." << endl ;
        }
        break ;
    case SRV_INSTALL :
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Installing the service..." << endl ;
        }
        install () ;
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "ok." << endl ;
#ifdef  UNIX
            logFileList_ << "Just type " << programName_ << " " << azOption_[OPT_SERVICE] << endl ;
#endif  // UNIX
        }

        break ;
    case SRV_UNINSTALL :
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Uninstalling the service..." << endl ;
        }
        uninstall () ;
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "ok." << endl ;
        }
        break ;
    case SRV_STOP :
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Stop the service..." << endl ;
        }
        control (SRVCTRL_STOP) ;
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "ok." << endl ;
        }
        break ;
    case SRV_PAUSE :
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Suspend the service..." << endl ;
        }
        control (SRVCTRL_PAUSE) ;
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "ok." << endl ;
        }
        break ;
    case SRV_CONTINUE :
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Continue the service..." << endl ;
        }
        control (SRVCTRL_CONTINUE) ;
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "ok." << endl ;
        }
        break ;
    case SRV_HELP :
        help () ;
        break ;
    default :
        // Error from eCommand_.
        throw CTException ("go: eCommand_ is invalid.", __FILE__, __LINE__) ;
    }
} // CGMain::go

//-----------------------------------------------------------------------------
// CGMain::help
//
// Display help on the standard output.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGMain::help (void)
{
    if (eCommand_ == SRV_HELP)
    {
        CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;

        logFileList_ << "Usage : " << programName_ << " " << azOption_[OPT_HELP] << endl ;
        logFileList_ << "          " << azOption_[OPT_HELP] << ": this help." << endl ;
        logFileList_ << "        " << programName_ << " " << azOption_[OPT_INSTALL] << "|" << azOption_[OPT_UNINSTALL] << endl ;
        logFileList_ << "          " << azOption_[OPT_INSTALL] << ": install the service." << endl ;
        logFileList_ << "          " << azOption_[OPT_UNINSTALL] << ": uninstall the service." << endl ;
        logFileList_ << "        " << programName_ << " [" << azOption_[OPT_CONSOLE] << "|" << azOption_[OPT_SERVICE] << "] [" << azOption_[OPT_ADMINISTRATION_LISTEN_PORT] << " <port>] [" << azOption_[OPT_REUSE_PORT] << " ] [" << azOption_[OPT_CLIENT_CONFIG] << " <config>]" << endl ;
        logFileList_ << "        " << string (' ', programName_.length () ) << " [" << azOption_[OPT_LOGFILE] << " <logfile>] [" << azOption_[OPT_TRACE] << "] [" << azOption_[OPT_LIBARG] << " library command line]" << endl ;
        logFileList_ << "          " << azOption_[OPT_CONSOLE] << ": run as a console." << endl ;
#ifdef  WIN32
        logFileList_ << "          " << azOption_[OPT_SERVICE] << ": run as a service." << endl ;
#endif  // WIN32
#ifdef  UNIX
        logFileList_ << "          " << azOption_[OPT_SERVICE] << ": run as a daemon." << endl ;
#endif  // UNIX
        logFileList_ << "          " << azOption_[OPT_ADMINISTRATION_LISTEN_PORT] << " <port>: specify the administration listen port." << endl ;
        logFileList_ << "            <port> must be a decimal or hexadecimal number." << endl ;
        logFileList_ << "          " << azOption_[OPT_REUSE_PORT] << ": force to use the specified administration listen port." << endl ;
        logFileList_ << "          " << azOption_[OPT_CLIENT_CONFIG] << " <config>: specify the client configuration." << endl ;
        logFileList_ << "            <config> must be like <port[;parse_library];[default:]target>" << endl ;
        logFileList_ << "              <port> is the client listen port." << endl ;
        logFileList_ << "                     it must be a decimal or hexadecimal number." << endl ;
        logFileList_ << "              <parse_library> is the name of the packet parsing library." << endl ;
        logFileList_ << "              <target> is the target tcp address to redirect clients to." << endl ;
        logFileList_ << "                       it must be like <[default:]host name|IP address:port number>." << endl ;
        logFileList_ << "                       default indicate the client can't specify the TCP address server to connect to." << endl ;
        logFileList_ << "          " << azOption_[OPT_LOGFILE] << " <logfile>: specify the log file." << endl ;
        logFileList_ << "            <logfile> log file name." << endl ;
        logFileList_ << "          " << azOption_[OPT_TRACE] << ": activate the trace mode." << endl ;
        logFileList_ << "          " << azOption_[OPT_LIBARG] << ": next arguments are specifics to the libraries." << endl ;
        logFileList_ << "        " << programName_ << " " << azOption_[OPT_STOP] << "|" << azOption_[OPT_PAUSE] << "|" << azOption_[OPT_CONTINUE] << endl ;
        logFileList_ << "          " << azOption_[OPT_STOP] << ": stop the service." << endl ;
        logFileList_ << "          " << azOption_[OPT_PAUSE] << ": suspend the service." << endl ;
        logFileList_ << "          " << azOption_[OPT_CONTINUE] << ": continue the service." << endl ;
    }
    else
    {   // Error from eCommand_.
        throw CTException ("help: eCommand_ is invalid.", __FILE__, __LINE__) ;
    }
} // CGMain::help

//-----------------------------------------------------------------------------
// CGMain::signalInterrupt
//
// Interruption from user.
//
// Arguments : iSig : Signal number (always SIGINT).
// Return    : none.
//-----------------------------------------------------------------------------
void CGMain::signalInterrupt (int iSig)
{
    switch (iSig)
    {
    case SIGINT :   // User interuption.
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Aborting the gateway..." << endl ;
        }
        bWaitTasks_ = false ;
        // Say to abort.
        conditionContinue_.signal () ;
        break ;
    default :   // Error from iSig.
        throw CTException ("signalInterrupt: iSig is invalid.", __FILE__, __LINE__) ;
    }
} // CGMain::signalInterrupt

//-----------------------------------------------------------------------------
// CGMain::abort
//
// Interruption from program.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGMain::abort (void)
{
    {
        CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
        logFileList_ << "Aborting the gateway..." << endl ;
    }
    // Say to abort.
    conditionContinue_.signal () ;
} // CGMain::abort

//============================================================< User methodes >
//-----------------------------------------------------------------------------
// CGMain::init
//
// Initialization service function.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGMain::init (void)
{
    {
        CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;

        logFileList_    << endl ;

        logFileList_                                    << endl
        << "Initialize the gateway..."  << endl ;
    }

    // Save the this pointer.
    GpProcessMain = this ;

    // Set the signal handler for SIG_INT.
    if (::signal (SIGINT, signalInterruptionFunction) == SIG_ERR)
    {   // Error from ::signal ().
        throw CTException (string ("init: ::signal failed: ") + strerror (errno), __FILE__, __LINE__, errno) ;
    }

    // Lock the mutex continue.
    mutexContinue_.lock () ;

    // Put current task in the list.
    tasksList_.addCurrentTask (CGTask::TASK_CONTROL, GnullTask) ;

    try
    {
        // Create the administration listen task.
        tasksList_.newTask (CGTask::TASK_ADMINISTRATION_LISTEN, (void*) &adminPort_, tasksList_.getCurrentTask () ) ;

        if (!clientConfig_.empty () )
        {   // Create the gateway listen task.
            bool            bDefault = false ;
            string      target ;
            port_t          tListenPort = 0 ;
            string      library ;

            if (clientConfig_.find_first_of (';') == string::npos)
            {   // Get only the listen port.
                sscanf (clientConfig_.c_str (), "%hu", &tListenPort) ;
            }
            else
            {   // Get the listen port plus the library name and/or the target address.
                // Get only the listen port.
                string  port = clientConfig_ .substr(0, clientConfig_.find_first_of (';') ) ;
                sscanf (port.c_str (), "%hu", &tListenPort) ;

                // Get the library name and/or the target address.
                string  next = clientConfig_.substr(clientConfig_.find_first_of (';') + 1, clientConfig_.size () - clientConfig_.find_first_of (';') - 1) ;
                if (next.find_first_of (';') == string::npos)
                {   // Get the library name or the target address.
                    if (next.find_first_of (':') == string::npos)
                    {   // Get the library name.
                        library = next ;
                    }
                    else
                    {   // Get the target address.
                        target = next ;
                        string strTemp = target.substr (0, target.find_first_of (':') ) ;
                        if (stricmp(strTemp.c_str(), "default") == 0)
                            bDefault = true;
                        if (bDefault)
                        {
                            target.erase (0, target.find_first_of (':') + 1) ;
                        }
                    }
                }
                else
                {   // Get the library name and the target address.
                    library = next.substr (0, next.find_first_of (';') ) ;
                    target = next.substr (next.find_first_of (';') + 1, next.length () - next.find_first_of (';') - 1) ;
                    string strTemp = target.substr(0, target.find_first_of (':') ) ;
                    if (stricmp(strTemp.c_str(), "default") == 0)
                        bDefault = true;
                    if (bDefault)
                    {
                        target.erase (0, target.find_first_of (':') + 1) ;
                    }
                }
            }

            // Create the gateway listen task.
            CGGatewayInfo   gatewayInfo (false, target, tListenPort, library, !bDefault) ;
            tasksList_.newTask (CGTask::TASK_GATEWAY_LISTEN, (void*) &gatewayInfo, tasksList_.getCurrentTask () ) ;
        }
    }
    catch (const CGTaskException& )
    {
        throw;
    }
} // CGMain::init

//-----------------------------------------------------------------------------
// CGMain::main
//
// Main service function.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGMain::main (void)
{
    {
        CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
        logFileList_ << "Gateway is running..." << endl ;
    }

    // Wait for condition continue.
    conditionContinue_.wait () ;
} // CGMain::main

//-----------------------------------------------------------------------------
// CGMain::exit
//
// Exit service function.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGMain::exit (void)
{
    {
        CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
        logFileList_ << "Exit the gateway..." << endl ;
    }

    // Unlock the mutex continue.
    mutexContinue_.unlock () ;

    if (bWaitTasks_)
    {
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Wait all clients..." << endl ;
        }
        // Join all the tasks.
        tasksList_.joinAll () ;
    }
    else
    {
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Kill all clients..." << endl ;
        }
        // Kill all the tasks.
        tasksList_.killAll () ;
    }

    // Unset the signal handler for SIG_INT.
    if (::signal (SIGINT, SIG_DFL) == SIG_ERR)
    {   // Error from ::signal ().
        throw CTException (string ("exit: ::signal failed: ") + strerror (errno), __FILE__, __LINE__, errno) ;
    }

    // Clear the global pointer.
    GpProcessMain = NULL ;
} // CGMain::exit

//-----------------------------------------------------------------------------
// CGMain::controlNotify
//
// Main service function.
//
// Arguments : eServiceControl : action asked by the service program manager.
// Return    : none.
//-----------------------------------------------------------------------------
void CGMain::controlNotify (enumServiceControl eServiceControl)
{
    switch (eServiceControl)
    {
    case SRVCTRL_STOP :
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Stopping the gateway..." << endl ;
        }
        conditionContinue_.signal () ;
        break ;
    case SRVCTRL_PAUSE :
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Supending the gateway..." << endl ;
        }
        // Suspend all the tasks.
        tasksList_.suspendAll () ;
        break ;
    case SRVCTRL_CONTINUE :
        {
            CTLockGuard<CTMutex>    guard (logFileList_.getMutex () ) ;
            logFileList_ << "Resuming the gateway..." << endl ;
        }
        // Resume all the tasks.
        tasksList_.resumeAll () ;
        break ;
    }
} // CGMain::controlNotify

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////// End of Class CGMain //
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////// Globals Functions //
///////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// signalInterruptionFunction
//
// Signal interruption function.
//
// Arguments : iSig : Signal number.
// Return    : none.
//-----------------------------------------------------------------------------
void signalInterruptionFunction (int iSig)
{
    try
    {
        if (GpProcessMain)
        {
            GpProcessMain->signalInterrupt (iSig) ;
        }
        else
        {   // Error from GpProcessMain.
            throw CTException ("signalInterruptionFunction: GpProcessMain is NULL.", __FILE__, __LINE__) ;
        }
    }
    catch (CTException& e)
    {
        cerr << e << endl ;
    }
} // signalInterruptionFunction

//-----------------------------------------------------------------------------
// main
//
// Main function.
//
// Arguments : argc    : Number of arguments.
//             argv    : Arguments array.
// Return    : iStatus : Status exit code.
//-----------------------------------------------------------------------------
int main (int argc, char* argv[], char* [])
{
    int iStatus ;

    try
    {
        CGMain  service ("CYRANO Gateway", argc, argv) ;
        service.go () ;
        iStatus = 0 ;
    }
    catch (CTException& e)
    {
        cerr << e << endl ;
        iStatus = (int) e.getNumber () ;
    }

    return iStatus ;
} // main

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// End of Globals Functions //
///////////////////////////////////////////////////////////////////////////////
