/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/


// IECache.h: interface for the CIECache class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_IECACHE_H__C291E92D_03B9_11D4_9C91_00A0C9CFA893__INCLUDED_)
#define AFX_IECACHE_H__C291E92D_03B9_11D4_9C91_00A0C9CFA893__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#pragma warning (disable: 4786)
#include <map>
#include <string>
#include <list>

// ===================================================================================================
// Capture file format:
// -------------------
//
// <LastConnId><BodyBase><ClientHeaderBase><ServerHeaderBase><Index><Body><ClientHeader><ServerHeader>
//
// where:
// <LastConnId> = last connection id used in this capture. 4 bytes integer.
// <BodyBase>         = offset of the beginning of the bodies part within the file.
// <ClientHeaderBase> = offset of the beginning of the client headers part within the file.
// <ServerHeaderBase> = offset of the beginning of the server headers part within the file.
// <Index> = For each couple (URL, connection id), contains the offsets to the body and client
//           and server headers of the HTTP query. Format: list of records of the
//           form: <URL\0><Connid><BodyOffset><ClientHeaderOffset><ServerHeaderOffset>
// <URL\0> = null-terminated ASCII string representing a canonical HTTP URL.
// <Connid> = connection id of the HTTP query
// <BodyOffset> = offset of the HTTP query's body (HTML file ou resource file: .gif, .jpg...)
//                from <BodyBase>.
// <ClientHeaderOffset> = offset of the HTTP query's client headers from <ClientHeaderBase>.
// <ServerHeaderOffset> = offset of the HTTP query's server headers from <ServerHeaderBase>.
// <Body> = Contains the bodies of HTTP requests. Format: list of records of the form:
//          <BodySize><QueryBody>
// <BodySize> = size of the query's body in bytes. 4 bytes integer.
// <QueryBody> = the actual body of the query.
// <ClientHeader> = Contains the client headers of HTTP requests. Format: list of records of the form:
//          <ClientHeaderSize><QueryClientHeader>
// <ClientHeaderSize> = size of the query's client headers string in bytes. 4 bytes integer.
// <QueryClientHeader> = the actual client headers string.
// <ServerHeader> = Contains the server headers of HTTP requests. Format: list of records of the form:
//          <ServerHeaderSize><QueryServerHeader>
// <ServerHeaderSize> = size of the query's server headers string in bytes. 4 bytes integer.
// <QueryServerHeader> = the actual server headers string.
//
// This format can be viewed as the following diagram:
//
//                      --------------
//                      | LastConnId |
//                      --------------
//               +------| BodyBase   |
//               |      --------------------
//               | +----| ClientHeaderBase |
//               | |    --------------------
//               | | +--| ServerHeaderBase |
//               | | |  -------------------------------------------------------------------------------------------------------
// Index:        | | |  | URL                            | \0 | Connid | BodyOffset | ClientHeaderOffset | ServerHeaderOffset |
//               | | |  -------------------------------------------------------------------------------------------------------
//               | | |     .
//               | | |     .
//               | | |     .
//               | | |  ------
//               | | |  | \0 |
//               | | |  ----------------------------------------------
// Body:         +----->| BodySize | QueryBody                       |
//                 | |  ----------------------------------------------
//                 | |      .
//                 | |      .
//                 | |      .
//                 | |  ----------------------------------------------
// Client Header:  +--->| ClientHeaderSize | QueryClientHeader       |
//                   |  ----------------------------------------------
//                   |      .
//                   |      .
//                   |      .
//                   |  ----------------------------------------------
// Server Header:    +->| ServerHeaderSize | QueryServerHeader       |
//                      ----------------------------------------------
//                          .
//                          .
//                          .
//
// ===================================================================================================



struct URLKey
{
    CString m_strURL;
    DWORD   m_dwConnId;
    DWORD   m_dwRId;

    URLKey()
    {}

    URLKey(const CString& strURL, DWORD dwConnId, DWORD dwRid)
            : m_strURL(strURL), m_dwConnId(dwConnId), m_dwRId(dwRid)
    {}

    URLKey(const URLKey& uk)
            : m_strURL(uk.m_strURL), m_dwConnId(uk.m_dwConnId), m_dwRId(uk.m_dwRId)
    {}

    ~URLKey()
    {}

    void operator=(const URLKey& uk)
    {
        m_strURL   = uk.m_strURL;
        m_dwConnId = uk.m_dwConnId;
        m_dwRId    = uk.m_dwRId;
    }
    bool operator<(const URLKey& uk) const
    {
        bool bRes = false;

        if(m_strURL < uk.m_strURL)
            bRes = true;
        else if(m_strURL == uk.m_strURL)
        {
            if(m_dwConnId < uk.m_dwConnId)
                bRes = true;
            else if(m_dwConnId == uk.m_dwConnId)
                bRes = m_dwRId < uk.m_dwRId;
        }

        return bRes;
    }
};

struct URLData
{
    LPBYTE  m_pbBody;
    LPCSTR  m_pcClient;
    LPCSTR  m_pcServer;

    URLData()
    {}

    URLData(LPBYTE pbBody, LPCSTR pcClient, LPCSTR pcServer)
            : m_pbBody(pbBody), m_pcClient(pcClient), m_pcServer(pcServer)
    {}

    URLData(const URLData& ud)
            : m_pbBody(ud.m_pbBody), m_pcClient(ud.m_pcClient), m_pcServer(ud.m_pcServer)
    {}

    ~URLData()
    {}

    void operator=(const URLData& ud)
    {
        m_pbBody   = ud.m_pbBody;
        m_pcClient = ud.m_pcClient;
        m_pcServer = ud.m_pcServer;
    }
};

class CIECache
{
public:
    //// Construction / Destruction.

    // Constructor.
    // zCaptureFile: optional path of the file captured by the CYRANO HTTP Gateway.
    CIECache(LPCTSTR zCaptureFile = NULL);

    // Destructor.
    ~CIECache();

    //// User methods.

    // Initialization. Must be called before calling CacheURL().
    // This method initializes the object and loads the capture file,
    // if one was specified.
    // Return: TRUE if success, FALSE if failure.
    BOOL Init();

    // Store a valid URL into the Wininet cache.
    // strOrigURL: the URL.
    // dwConnId: Tof connection id. Ignored if pData is not NULL.
    // dwRId: Tof request id (relative to a connection id). Ignored if pData is not NULL.
    //        If pData is NULL and dwRid is not zero, then the URL's data is extracted
    //        from the capture file. If pData is NULL and dwRId is zero, then the method
    //        tries to find index entries with dwRId = 1 and dwRId = 2. If only the
    //        entry with dwRId = 1 exists, then it is used. Otherwise, if the two
    //        exist, there is an ambiguity and the function returns FALSE.
    // pData: the data at the URL. If NULL, the method attempts to extract the data
    //        using the capture file.
    // dwDataLength: the number of bytes pointed to by pData.
    // Return: TRUE if success, FALSE if failure.
    BOOL CacheURL(const CString& strOrigURL, DWORD dwConnId = 0, DWORD dwRId = 0, const BYTE* pData = NULL, DWORD dwDataLength = 0);

	//Add primary and all associated secondary urls to ie cache
	BOOL CacheAllForPrimaryURL(const CString& strURL,DWORD dwConnId=0,DWORD dwRId=0);
    
    // Returns the last connection id used in the capture file.
    DWORD GetLastConnId() const;

    // Retrieves a body for a given tuple (URL, connection id, request id).
    // Returns NULL if no capture file was specified at object construction, or if
    // no entry in the capture file matches the tuple, or if you specified 0 as dwRId
    // and multiple entries match (URL, connection id).
    // The pdwDataLength output argument gives the data's length in bytes.
    LPBYTE GetBody(const CString& strOrigURL, DWORD dwConnId, LPDWORD pdwDataLength, DWORD dwRId = 0) const;

    // Retrieves a client header for a given tuple (URL, connection id, request id).
    // Returns NULL if no capture file was specified at object construction, or if
    // no entry in the capture file matches the tuple, or if you specified 0 as dwRId
    // and multiple entries match (URL, connection id).
    // WARNING: the data pointed to by the return value is NOT null-terminated: you
    // must read the pdwDataLength output argument after the call to get the data's
    // length in bytes.
    LPCSTR GetClientHeader(const CString& strOrigURL, DWORD dwConnId, LPDWORD pdwDataLength, DWORD dwRId = 0) const;

    // Retrieves a server header for a given tuple (URL, connection id, request id).
    // Returns NULL if no capture file was specified at object construction, or if
    // no entry in the capture file matches the tuple, or if you specified 0 as dwRId
    // and multiple entries match (URL, connection id).
    // WARNING: the data pointed to by the return value is NOT null-terminated: you
    // must read the pdwDataLength output argument after the call to get the data's
    // length in bytes.
    LPCSTR GetServerHeader(const CString& strOrigURL, DWORD dwConnId, LPDWORD pdwDataLength, DWORD dwRId = 0) const;

    // Builds a list of URLs matching a given pattern.
    // The pattern is a string of the form:
    // abc*def*ghi
    // where the star (*) represents a (possibly empty) string of characters.
    // dwConnId filters the results.
    void QueryURL(const CString& strQuery, DWORD dwConnId, std::list<URLKey>& outList) const;

    // Stores all URLs from the capture file into the Wininet cache.
    // This assumes that you specified a capture file. This method is mainly for
    // purposes. The URLs are taken (and cached) in the order defined by URLKey::operator<().
    BOOL CacheAllURLsFromCaptureFile();

	// This method merges the capture file referenced by zCaptureFile2 into the
    // capture file referenced by zCaptureFile1. If the merge is successful, the
    // capture file referenced by zCaptureFile2 is deleted.
    // WARNING: To perform this operation, neither file must be loaded by a CIECache object.
    static BOOL MergeCaptureFile(LPCTSTR zCaptureFile1, LPCTSTR zCaptureFile2);

    // Builds a list of HTML URLs matching a given pattern.
    // The pattern is a string of the form:
    // abc*def*ghi
    // where the star (*) represents a (possibly empty) string of characters.
    // dwConnId filters the results.
    void QueryHTMLURL(const CString& strQuery, DWORD dwConnId, std::list<URLKey>& outList) const;

    BOOL IsHTMLURL(const CString& strOrigURL, DWORD dwConnId, DWORD dwRId) const;
    BOOL IsPrimaryURL(const CString& strOrigURL, DWORD dwConnId, DWORD dwRId) const;

private:

    typedef std::map<URLKey, URLData> Index;

    //// Helpers.

    // Returns TRUE if the key pointed to by it matches the pattern defined by
    // strPatterns.
    typedef std::list<std::string> stringList;
    BOOL Match(const stringList& strPatterns, DWORD dwConnId,
               const Index::const_iterator it, BOOL bLeadingStar,
               BOOL bTrailingStar) const;

    //// Data.
    CString m_strCaptureFile;
    LPBYTE  m_pbCaptureData;
    LPBYTE  m_pbBody;
    LPCSTR  m_pcClientHeader;
    LPCSTR  m_pcServerHeader;
    LPBYTE  m_pbIndex;
    Index   m_index;
};

#endif // !defined(AFX_IECACHE_H__C291E92D_03B9_11D4_9C91_00A0C9CFA893__INCLUDED_)
