/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     Exec
// @module  Exec Class | This module contains the implementation of
//          the Exec class.  This class allows the retrieval of statistics
//          via a command line.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include "OtaDefs.h"
#include <OtaExec.h>
#include <OSTAORB/OSTAORB.h>
#include "TPLog.h"

///////////////////////////////////////////////////////////////////////////////
#include <string>

#ifndef WIN32
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/wait.h>
#endif

///////////////////////////////////////////////////////////////////////////////
// Constants
//const char* OtaExec::CLASS_NAME = "OtaExec";
#define TIMEOUT_MESSAGE "*!@TIMEOUT@!*"

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor initialises the member variables
// @parm    Pointer to CORBA helper functionality.
// @parm    Name of the object.
OtaExec::OtaExec()
{
#ifdef WIN32
    m_hProcess = NULL;
    m_hStdinWr = NULL;
    m_hStdoutRd = NULL;
    m_hStdoutWr = NULL;
#else
    m_pid = 0;
    m_childstdin = 0;
    m_childstdout = 0;
#endif

    // Error information
    const char* where = "OtaExec::OtaExec";

}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor
OtaExec::~OtaExec()
{
    // Kill command process
    if (ProcessExists())
    {
        KillCmd();
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method executes the passed string as a command line.
// @rdesc   The result of the query.
// @parm    String representing the requested command line.
bool OtaExec::RunSimpleQuery( const string query, long *result )
{
    // In order to issue an OS command (e.g. to run a shell script), we need
    // to create a process to execute the command.  Having created the process,
    // there are two possible scenarios:
    // 1) the command runs once, writing the result to its stdout, then the process exits.
    //    In this case we will create another process next time.
    // 2) the shell script (or equivalent) does its job, but then waits for a newline
    //    to be read from its stdin - this is written by ota_ora, and indicates that
    //    the next "poll" is to be performed.  This process continues until the process
    //    receives EOF on its stdin, indicating that it should exit.

    // If we do not have a process already, or the process has exited, create one
    if (!ProcessExists())
    {
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"Creating command process");
        if (!ExecCommand(query))
        {
            OtaAddTrace(DBGSTD,ERRORLEVEL,"Failed to execute command: " + query);
            return false;
        }
    }

    // Write newline to process's stdin
    OtaAddTrace(DBGSTD,DEBUGLEVEL,"Triggering command process");
    TriggerCmd();

    // Read line from process's stdout
    char *res = GetCmdOutput();

    // Convert result to integer
    if (res == NULL)
    {
        return false;
    }
    else
    {
        long      n1;
        int       n2;
        int       rv;

        rv = sscanf(res, "%ld%n", &n1, &n2);
        if ((rv != 1)
                || (res[n2] != '\0'))
        {
            OtaAddTrace(CYR_OTA,ERRORLEVEL,"No valid integer result from command: " + query);
            return false;
        }
        else
        {
            *result = n1;
            return true;
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Detailed queries not implemented for Exec
// @parm    The detailed query string.
bool OtaExec::RunDetailedQuery( const string query, string *result )
{
    return false;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Detailed queries not implemented for Exec
// @rdesc   The header string.
// @parm    The query that will produce the headers.
string OtaExec::GetDetailedHeaders( const string query )
{
    return "";
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Check for process existence
bool OtaExec::ProcessExists()
{
#ifdef WIN32
    // If process handle is NULL, process doesn't exist!
    if (m_hProcess == NULL)
    {
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"Process handle = NULL");
        return false;
    }

    // Check for process exit */
    if (WaitForSingleObject(m_hProcess, 0) == WAIT_OBJECT_0)
    {
        CloseHandle(m_hProcess);
        m_hProcess = NULL;
        CloseHandle(m_hStdinWr);
        m_hStdinWr = NULL;
        CloseHandle(m_hStdoutRd);
        m_hStdoutRd = NULL;
        CloseHandle(m_hStdoutWr);
        m_hStdoutWr = NULL;
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"Process has exited");
        return false;
    }
    else
    {
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"Process exists");
        return true;
    }

#else   //UNIX
    pid_t rv = waitpid(m_pid, NULL, WNOHANG);
    if (rv == -1)
    {
        if (errno == ECHILD)
        {
            // pid does not exist
            return false;
        }
        else if (errno == EINTR)
        {
            return true;
        }
        else
        {
            OtaAddTrace(DBGSTD,ERRORLEVEL,"waitpid failed: " + string(strerror(errno)));
            return true;
        }
    }
    else if (rv == 0)
    {
        // child is running, but status is not available
        return true;
    }
    else if (rv == m_pid)
    {
        // in zombie state
        return true;
    }
    else
    {
        return true;
    }

#endif
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Execute the command
bool OtaExec::ExecCommand(const string query)
{
#ifdef WIN32
    HANDLE hOutputReadTmp;
    HANDLE hOutputRead;
    HANDLE hOutputWrite;
    HANDLE hInputWriteTmp;
    HANDLE hInputRead;
    HANDLE hInputWrite;
    HANDLE hErrorWrite;
    SECURITY_ATTRIBUTES sa;
    PROCESS_INFORMATION pi;
    STARTUPINFO si;

    // Set up the security attributes struct.
    sa.nLength= sizeof(SECURITY_ATTRIBUTES);
    sa.lpSecurityDescriptor = NULL;
    sa.bInheritHandle = TRUE;

    // Create the child output pipe.
    if (!CreatePipe(&hOutputReadTmp,&hOutputWrite,&sa,0))
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"CreatePipe() failed: " + LastErrorText());
        return false;
    }

    // Create a duplicate of the output write handle for the std error
    // write handle. This is necessary in case the child application
    // closes one of its std output handles.
    if (!DuplicateHandle(GetCurrentProcess(),hOutputWrite,
                         GetCurrentProcess(),&hErrorWrite,0,
                         TRUE,DUPLICATE_SAME_ACCESS))
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"DuplicateHandle() failed: " + LastErrorText());
        return false;
    }
    if (!CreatePipe(&hInputRead,&hInputWriteTmp,&sa,0))
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"CreatePipe() failed: " + LastErrorText());
        return false;
    }

    // Create new output read handle and the input write handles. Set
    // the Properties to FALSE. Otherwise, the child inherits the
    // properties and, as a result, non-closeable handles to the pipes
    // are created.
    if (!DuplicateHandle(GetCurrentProcess(),hOutputReadTmp,
                         GetCurrentProcess(),
                         &hOutputRead, // Address of new handle.
                         0,FALSE, // Make it uninheritable.
                         DUPLICATE_SAME_ACCESS))
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"DuplicateHandle() failed: " + LastErrorText());
        return false;
    }
    if (!DuplicateHandle(GetCurrentProcess(),hInputWriteTmp,
                         GetCurrentProcess(),
                         &hInputWrite, // Address of new handle.
                         0,FALSE, // Make it uninheritable.
                         DUPLICATE_SAME_ACCESS))
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"DuplicateHandle() failed: " + LastErrorText());
        return false;
    }

    // Close inheritable copies of the handles we do not want to be
    // inherited.
    if (!CloseHandle(hOutputReadTmp))
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"CloseHandle() failed: " + LastErrorText());
    }
    if (!CloseHandle(hInputWriteTmp))
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"CloseHandle() failed: " + LastErrorText());
    }

    // Set up the startup info struct.
    ZeroMemory(&si,sizeof(STARTUPINFO));
    si.cb = sizeof(STARTUPINFO);
    si.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
    si.hStdOutput = hOutputWrite;
    si.hStdInput  = hInputRead;
    si.hStdError  = hErrorWrite;
    si.wShowWindow = SW_HIDE;

    // Create the process to execute the command
    OtaAddTrace(DBGSTD,DEBUGLEVEL,"Creating process");
    if (!CreateProcess(NULL,
                       (char*)query.data(), // command line
                       NULL,                // process security attributes
                       NULL,                // primary thread security attributes
                       TRUE,                // handles are inherited
                       CREATE_NEW_CONSOLE,  // creation flags
                       NULL,                // use parent's environment
                       NULL,                // Current directory
                       &si,                 // STARTUPINFO pointer
                       &pi))                // receives PROCESS_INFORMATION
    {
        OtaAddTrace(CYR_OTA,ERRORLEVEL,"Failed to execute command:" + query
                    + " : " + LastErrorText());
        return false;
    }

    // Save process handle
    m_hProcess = pi.hProcess;

    // Close thread handle
    if (!CloseHandle(pi.hThread))
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"CloseHandle() failed: " + LastErrorText());
    }

    // Close pipe handles (do not continue to modify the parent).
    // We need to make sure that no handles to the write end of the
    // output pipe are maintained in this process or else the pipe will
    // not close when the child process exits and the ReadFile will hang.
    if (!CloseHandle(hInputRead ))
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"CloseHandle() failed: " + LastErrorText());
    }
    if (!CloseHandle(hErrorWrite))
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"CloseHandle() failed: " + LastErrorText());
    }

    // Save handles for reading from and writing to the created process
    m_hStdinWr = hInputWrite;
    m_hStdoutWr = hOutputWrite;
    m_hStdoutRd = hOutputRead;

    return true;

#else // UNIX
    int       fdsparent_to_child[2];
    int       fdschild_to_parent[2];
    int       rv;

    if (pipe(fdsparent_to_child) < 0)
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"pipe(fdsparent_to_child) failed: " + string(strerror(errno)));
        return false;
    }

    if (pipe(fdschild_to_parent) < 0)
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"pipe(fdschild_to_parent) failed: " + string(strerror(errno)));
        return false;
    }

    rv = fork();
    if (rv == -1)
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"fork failed: " + string(strerror(errno)));
        return false;
    }
    else if (rv != 0)
    {
        // parent
        m_pid = rv;
        m_childstdin = fdsparent_to_child[1];
        m_childstdout = fdschild_to_parent[0];
    }
    else
    {
        // child
        close(0);       // close stdin
        close(1);       // close stdout

        dup2(fdsparent_to_child[0], 0);     // new stdin is pipe from parent to child
        dup2(fdschild_to_parent[1], 1);     // new stdout

        close(fdsparent_to_child[0]);
        close(fdschild_to_parent[1]);

        {
            // split text into words to be passed to exec
            int n;
            char *p, *p1;
#define MAXARGS 20
            char *args[ MAXARGS ];
            int argn;

            p1 = (char*) malloc( strlen(query.data()) + 1 );
            if ( p1 == NULL )
            {
                // remember we have already forked
                exit( 1 );
            }

            p = p1;
            strcpy( p, query.data() );

            argn = 0;
            for(;;)
            {
                // lose leading white space
                n = strspn( p, "\t " );
                p += n;
                args[ argn ] = p;

                // find end of word
                n = strcspn( p, "\t " );
                if ( n == 0 )
                {
                    args[ argn+1 ] = NULL;
                    break;
                }

                p += n;
                *p = '\0';
                p++;

                if ( argn++ >= MAXARGS )
                {
                    OtaAddTrace(DBGSTD,ERRORLEVEL,"too many args");
                    free( p1 );
                    // remember we have already forked
                    exit( 1 );
                }
            }

            execvp(args[ 0 ], args);

            // error if reached here
            OtaAddTrace(DBGSTD,ERRORLEVEL,"execvp failed: " + string(strerror(errno)));

            free( p1 );

            // remember we have already forked
            exit( 1 );
        }
    }

    return true;
#endif
}

void OtaExec::TriggerCmd()
{
#ifdef WIN32
    DWORD     dwWritten;

    // Write newline to child's stdin
    if (!WriteFile(m_hStdinWr, "\n", (DWORD) 1, &dwWritten, NULL))
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"WriteFile() failed: " + LastErrorText());
    }
#else
    int       rv;
    const char nl = '\n';

    rv = write(m_childstdin, &nl, 1);
    if (rv != 1)
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"write() failed: " + string(strerror(errno)));
    }
#endif
}

char* OtaExec::GetCmdOutput()
{
    static char chBuf[256];
    int       rv;

#ifdef WIN32
    BOOL    bSuccess;
    DWORD   dwRead;

    // Read line from child's stdout
    bSuccess = ReadFile(m_hStdoutRd, chBuf, sizeof(chBuf) - 1, &dwRead, NULL)
               && (dwRead > 0);


    if (!bSuccess)
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"Read from child process failed: " + LastErrorText());
        return NULL;
    }
    rv = dwRead;


#else
    rv = read(m_childstdout, chBuf, sizeof(chBuf) - 1);
    if (rv < 0)
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"Read from child process failed: " + string(strerror(errno)));
        return NULL;
    }
#endif
    chBuf[rv] = '\0';

    // drop new line/carriage returns at end
    while ( (rv > 0) && ((chBuf[rv - 1] == '\n') || (chBuf[rv - 1] == '\r')))
    {
        chBuf[rv - 1] = '\0';
        rv--;
    }
    return chBuf;
}

#ifdef WIN32
void OtaExec::CmdTimeout()
{
    OtaAddTrace(DBGSTD,DEBUGLEVEL,"In CmdTimeout" );

    // Get access to this thread function

    DWORD   dwWritten;

    // Wait 5 seconds for event to be signalled
    OtaAddTrace(DBGSTD,DEBUGLEVEL,"WaitForSingleObject" );
    if (WaitForSingleObject (m_hEvent, 5000) != WAIT_OBJECT_0)
    {
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"Timed out" );
        // Write timeout message to child's stdout
        if (!WriteFile(m_hStdoutWr, TIMEOUT_MESSAGE, strlen(TIMEOUT_MESSAGE), &dwWritten, NULL))
        {
            OtaAddTrace(DBGSTD,ERRORLEVEL,"Write to pipe failed: " + LastErrorText());
        }
    }
    OtaAddTrace(DBGSTD,DEBUGLEVEL,"Exiting CmdTimeout" );
    ExitThread (0);
}
#endif

void OtaExec::KillCmd()
{
#ifdef WIN32
    OtaAddTrace(DBGSTD,DEBUGLEVEL,"KillCmd");
    // Close handles to stdin and stdout
    if (m_hStdinWr != NULL)
        CloseHandle(m_hStdinWr);
    if (m_hStdoutRd != NULL)
        CloseHandle(m_hStdoutRd);
    if (m_hStdoutWr != NULL)
        CloseHandle(m_hStdoutWr);

#if 0
    // Generate CTRL/C signal to kill off command process
    OtaAddTrace(DBGSTD,DEBUGLEVEL,"GenerateConsoleCtrlEvent");
    GenerateConsoleCtrlEvent(CTRL_C_EVENT, 0);

    if (m_hProcess != NULL)
    {
        // Wait 2 seconds for process to terminate
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"WaitForSingleObject");
        if (WaitForSingleObject(m_hProcess, 2000) != WAIT_OBJECT_0)
        {
            // Otherwise kill it off
            OtaAddTrace(DBGSTD,DEBUGLEVEL,"TerminateProcess");
            TerminateProcess(m_hProcess, 0);
        }

        // Close process handle
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"CloseHandle");
        CloseHandle(m_hProcess);
    }
#endif
#else //UNIX
    // Close standard input
    int rv;
    rv = close(m_childstdin);

    if (rv != 0)
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"close( childstdin ) failed: " + string(strerror(errno)));
    }

    rv = close(m_childstdout);
    if (rv != 0)
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"close( childstdout ) failed: " + string(strerror(errno)));
    }

    sleep(1);

    // Kill processes
    rv = kill(m_pid, SIGINT);

    if (rv == 0)
    {
        // OK
    }

    else if (rv == -1 && errno == ESRCH)
    {
        // process has already gone away
    }

    else
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"kill failed: " + string(strerror(errno)));
    }

    sleep(1);

    // Wait for processes
    rv = waitpid(m_pid, NULL, WNOHANG);

    if (rv == m_pid)
    {
        // OK
    }

    else if (rv == -1 && errno == ECHILD)
    {
        // OK
    }

    else
    {
        OtaAddTrace(DBGSTD,ERRORLEVEL,"kill failed: " + string(strerror(errno)));
    }
#endif
}

#ifdef WIN32
string OtaExec::LastErrorText()
{
    static char    buffer[1024];

    DWORD   dwError = GetLastError();   // Get last error
    LPVOID  lpMessageBuffer = NULL;     // Pointer to error text buffer

    if (FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                      NULL,
                      dwError,
                      LANG_NEUTRAL,
                      (LPSTR) & lpMessageBuffer,
                      0,
                      NULL))
    {
        sprintf(buffer, "%s", lpMessageBuffer);
    }
    else
    {
        sprintf(buffer, "Unknown error %d", dwError);
    }

    if (lpMessageBuffer != NULL)
        LocalFree((HLOCAL) lpMessageBuffer);

    return buffer;
}
#endif
